/*!  \file arwen.cpp
 *   \brief Main driver for ARWEN hydrodynamics code.
 *   
 *   ARWEN stands for Astrophysical Research With Enhanced Numerics.
 *   ARWEN is intended to be a simple-to-use, yet powerful, astrophysical
 *   hydrodynamics code.
 *   
 *   This version of ARWEN includes MPI parallelization - the global grid can
 *   be subdivided into local grid "tiles" which can be advanced on each 
 *   processor. 
 *
 *   At present ARWEN solves the inviscid equations of hydrodynamics on
 *   a uniform grid, using either unsplit PLMDE or split PPMLR. The former is 
 *   not as robust as PPMLR (and is limited to smaller timesteps due to 
 *   stricter wave speed and courant number constraints), but is more accurate
 *   and also has the advantage of being able to easily include MHD. 
 *
 *   \warning Any problem-specific ".h" files which use an 
 *      "#ifdef MAIN #else" struture MUST also be "#included" in arwen.cpp 
 *      where "MAIN" is defined (otherwise the code may compile successfully but
 *      will fail to link).
 *
 *   \notes Compilation: make
 *          Run:         mpiexec -np <no. processors> ./arwen
 * 
 *   \author Julian Pittard (Original version 25.05.11)
 *   \version 0.1-development (Undómiel):
 *   \date Last modified: 18.11.11 (JMP)
 */

#define MAIN

//  Header files:
#include <cstdlib>
#include <limits.h>
#include <iomanip>
#include <iostream>
#include <string>
#include "defs.h"
#include "constants.h"
#include "gas.h"
#include "global.h"
#include "globalMPI.h"
#include "grid.h"
#include "gridlet.h"
#include "pdcell.h"
#include "star.h"
//#include <gperftools/profiler.h>

using namespace std;

// Function declarations
extern void computeLoop();
extern void initiate(int argc, char *argv[]);
extern void prin1D();
extern void prin2D();
extern void prin2Dfits();
extern void prin3Dfits();
extern void quit();
extern void setup();

int main(int argc, char *argv[]){
  // Locals

  // Initialize MPI
  if (niProc == 1 && njProc == 1 && nkProc == 1) mpi = false;
  else mpi = true;

  if (debugMPI) cout << "mpi = " << mpi << "\n";

  initiate(argc, argv);
  if (debugMPI) cout << "procRank = " << procRank << "\n";
  if (debugMPI) cout << "mpi  = " << mpi << "\n";

  // Setup the initial conditions and write to disk
  setup();
  if (ncycle == 0){
    if (nd == 1) prin1D();
    else if (nd == 2){
      //prin2D();
      prin2Dfits();
      ++nfile2d;
    }
    else{
      //prin2D();
      if (!mpi){
	prin2Dfits();
        ++nfile2d;
      }
      prin3Dfits();
      ++nfile3d;
    }
  }
  
  // Evolve...
  //ProfilerStart("arwen.prof");
  computeLoop();
  //ProfilerStop();
  
  // Write to disk the final solution
  //prin1D();
  
  quit();
}




