/*!  \file computeLoop.cpp
 *   \brief Main computational loop.
 * 
 *   \author Julian Pittard (Original version 01.06.11)
 *   \version 0.1-development (Undómiel):
 *   \date Last modified: 01.06.11 (JMP)
 */

#undef MAIN

//  Header files:
#include <cmath>              //needed to check for infinities and NaNs
#include <iomanip> 
#include <iostream>           //  The C++ iostream.
#include <math.h>
#include "defs.h"
#include "constants.h"
#include "gas.h" 
#include "global.h"
#include "grid.h"
#include "physconst.h"


// Function declarations
void computeTimestep();
double ridt();
extern void determineGlobalTimestep();
extern double netCoolingRate(const double rho, const double temp);
extern double pdtmp(double pdcell);
extern void prin1D();
extern void prin2D();
extern void prin2Dfits();
extern void prin3Dfits();
extern void prinShkRadiusAndEnergyAndMassFractions();
extern void quit();
extern void timing(bool init);
extern void updateSplit();
extern void updateUnsplit();

using namespace std;


/*!  \brief Main computational loop.
 * 
 *   Marches the solution forward in time until the desired number of steps
 *   or the target time is achieved. During this period files are output
 *   as desired and the safe timestep is computed. The actual solution update
 *   is computed in update(). Problem-specific code after each update is
 *   handled in endofcomputeloop().
 *
 *   \author Julian Pittard (Original version 01.06.11)
 *   \version 0.1-development (Undómiel):
 *   \date Last modified: 24.08.11 (JMP)
 */
void computeLoop(){

  // Locals
  bool loop = true;
  double smalltime = 1.0e-6;
  
  // Number of steps per cycle..
#ifdef UNSPLIT
  double ndt = 1.0;
#else
  double ndt = double(nd);
#endif
  
  // Begin timing of compute loop
  timing(true);
  sc = ncycle;
  
  while (loop){

    //Increase counters
    ncycle++;
    ncycp1d++;
    ncycp2d++;
    ncycp3d++;

    // Adjust timestep if needed (nd timesteps are performed per cycle)
    if (t + ndt*dt > endtime){       //set dt to land on endtime
      dt = (endtime - t)/ndt;
    }
    if ( (timep1d + ndt*dt - tprin1d)/tprin1d > -smalltime ){ // set dt to land on tprin
      dt = (tprin1d - timep1d)/ndt;
      ncycp1d = nprin1d;  //changing the counters ensures file output
    }
    if ( (timep2d + ndt*dt - tprin2d)/tprin2d > -smalltime ){ 
      dt = (tprin2d - timep2d)/ndt;
      ncycp2d = nprin2d;
    }
    if ( (timep3d + ndt*dt - tprin3d)/tprin3d > -smalltime ){ 
      dt = (tprin3d - timep3d)/ndt;
      ncycp3d = nprin3d;
    }
    
    // Update the fluid variables
#ifdef UNSPLIT
    updateUnsplit();
#else
    updateSplit();
#endif

    // Output marching to terminal
    if (ncycle%ncycshow == 0){
      if (procRank == 0) cout << "ncycle = " << ncycle << "; time = " << t << "; dt = " << dt << "\n";
    }
    
    // Dump files
    if ((ncycp1d >= nprin1d) || (timep1d >= tprin1d )){
      //cout << "ncycp1d = " << ncycp1d << "; nprin1d = " << nprin1d << "; timep1d = " << timep1d << "; tprin1d = " << tprin1d << "\n";
      //quit();
      timep1d = 0.;
      ncycp1d = 0;
      prin1D();
      // Calculate shock radius and energy fractions and write out
      prinShkRadiusAndEnergyAndMassFractions();
      // Update tprin1d (useful if expanding grid)
      tprin1d *= 1.5;
    }
    if ((ncycp2d >= nprin2d) || (timep2d >= tprin2d )){
      timep2d = 0.;
      ncycp2d = 0;
      //prin2D();
      prin2Dfits();
      ++nfile2d;
    }
    if ((ncycp3d >= nprin3d) || (timep3d >= tprin3d )){
      timep3d = 0.;
      ncycp3d = 0;
      // prin3D();
      prin3Dfits();
      ++nfile3d;
    }
    
    // Barrier is necessary here to prevent some processes exiting before
    // all the files are written

    //barrier();

    // Calculate shock radius and write out
    //fshk_radius = findForwardShock();
    //prinShkRadius();
    
    // Update dt
    computeTimestep();
      
    // Check if the loop is finished

    if ((ncycle >= ncycend)||(t >= endtime)){
      if (procRank==0) cout << "Target reached, total number of cycles = " << ncycle << "\n";
      loop = false;
    }

    //if (ncycle > 4580) nprin1d = 1;
  } // do while

  // Final data dump
  prin1D();
  // Calculate shock radius and energy fractions and write out
  prinShkRadiusAndEnergyAndMassFractions();

  // Output the elapsed time of the compute loop
  timing(false);

  return;
}



void computeTimestep(){

  double dtx,dt3;

#ifdef UNSPLIT
  // Additional timestep controls necessary starting the UNSPLIT method with large gradients
  if (problem == "SNR"){
    if      (ncycle < 10) cn = 0.000025;
    else if (ncycle < 20) cn = 0.00025;
    else if (ncycle < 30) cn = 0.0025;
    else if (ncycle < 40) cn = 0.025;
    else                  cn = 0.25;
    //if ((nd == 1) && (ncycle > 50)) cn = 0.4;
  }
  else if (problem == "WBB"){
    if      (ncycle < 50) cn = 0.000025;
    else if (ncycle < 200) cn = 0.00025;
    else if (ncycle < 500) cn = 0.0025;
    else if (ncycle < 800) cn = 0.025;
    else                   cn = 0.25;
    if ((nd == 1) && (ncycle > 1300)) cn = 0.4;
  }
#endif

  /*
    // Additional time step control if using certain wind injection mechanisms
    if      (ncycle < 50) cn = 0.000025;
    else if (ncycle < 200) cn = 0.00025;
    else if (ncycle < 500) cn = 0.0025;
    else if (ncycle < 800) cn = 0.025;
    else                   cn = 0.25;
    if ((nd == 1) && (ncycle > 1300)) cn = 0.4;
  */
  
  // global time constraint for given courant number

  dtx = cn / ridt();

  //std::cout << "dtx = " << dtx << "; dt = " << dt << "\n";
  
  //if (ncycle > 0){
    // limiting constraint on rate of increase of dt
    //dt3 = 1.1 * dt; // this causes problems if dt is adjusted for FITS file dumps etc. (but commenting this out prevents the timestep from increasing)
    dt3 = dtx; 
  //}
  //else dt3 = dtx;
  
  // use smallest required timestep                                        
  dt = min( dt3, dtx );

  // if timestep becomes too small, stop the program!

  if (t/dt > 1.0e10){
    if (procRank == 0){
      cout << "Timestep has become too small: dt = " << dt << "\n";
      cout << "                             time = " << t << "\n";
      cout << "                           ncycle = " << ncycle << "\n"; 
    }
    quit();
  }

  //if (ncycle == 0) dt = min(dt,1.0e7);
  
  // Determine the global timestep if more than one processor
  if (mpi) determineGlobalTimestep();
  
  return;

}



/*!  \brief Compute the safe timestep.
 *
 *   The limiting timestep in the lagrangian code can be much larger than that
 *   in an Eulerian code. In a lagrangian code, dt=cn*max(dx/ux,dy/uy,del/cs)
 *   where del=min(dx,dy), while in an eulerian code dt=cn*max(dx/(ux+cs)).
 *   The VH-1 guide recommends a value of cn=0.6 or less for stability.
 *   svel = cs/del. Use R*d(theta) if y geometry is angular.
 * 
 *   \author Julian Pittard (Original version 02.06.11)
 *   \version 0.1-development (Undómiel):
 *   \date Last modified: 09.08.11 (JMP)
 */
double ridt(){

  int j = 0;
  int k = 0;
  double xvel,yvel,zvel,cs,width,widthy,widthz;
  double ri_dt = 0.0;
      
  cout << setprecision(4);
  //cout << "ridt start: procRank = " << procRank << "\n";

  if (nd == 1){
    for (int i = lg.irs; i <= lg.ire; ++i){
      cs = sqrt(gam*lg.P0[iqe][k][j][i]/lg.P0[iqd][k][j][i]);
      xvel = fabs(lg.P0[iqu0][k][j][i]);
#ifdef UNSPLIT
      ri_dt = max((xvel+cs)/lg.zdx[i],ri_dt);
#else
      ri_dt = max(max(xvel,cs)/lg.zdx[i],ri_dt);      
#endif
      /*
      // Cooling limit on timestep
      double rho = lg.P0[iqd][k][j][i];
      double pre = lg.P0[iqe][k][j][i];
      double pdcell = pre/rho;
      double temp = pdtmp(pdcell);	  
      double net_cooling_rate = netCoolingRate(rho,temp); // erg/s/cm^3 (positive if cooling, negative if heating)
	    	  
      //double Eint = pre/(tmpg(temp)-1.0); // use temperature-dependent gamma
      double Eint = pre/gamm; 
      double t_cool = Eint/net_cooling_rate;
      ri_dt = max(fabs(1.0/t_cool),ri_dt);
      */

      //cout << "ridt: i = " << i << "; cs = " << cs << "; xvel = " << xvel << "; ri_dt = " << ri_dt << "\n";
    }
    //quit();
  }
  else if (nd == 2){
    for (int j = lg.jrs; j <= lg.jre; ++j){
      for (int i = lg.irs; i <= lg.ire; ++i){
        widthy = lg.zdy[j];
        if (gg.ngeomy > 2) widthy = widthy*lg.zxc[i];
        cs = sqrt(gam*lg.P0[iqe][k][j][i]/lg.P0[iqd][k][j][i]);
	xvel = fabs(lg.P0[iqu0][k][j][i]);
	yvel = fabs(lg.P0[iqu0+1][k][j][i]);
#ifdef UNSPLIT
	ri_dt = max((xvel+cs)/lg.zdx[i],max((yvel+cs)/widthy,ri_dt));
#else
	width = min(lg.zdx[i],widthy);
	ri_dt = max(xvel/lg.zdx[i],max(yvel/widthy,max(cs/width,ri_dt)));
#endif
      }
    }
  }
  else{ //nd == 3
    for (int k = lg.krs; k <= lg.kre; ++k){
      for (int j = lg.jrs; j <= lg.jre; ++j){
        for (int i = lg.irs; i <= lg.ire; ++i){
	  widthy = lg.zdy[j];
	  widthz = lg.zdz[k];
  	  if(gg.ngeomy > 2) widthy = widthy*lg.zxc[i];
	  if(gg.ngeomz > 2) widthz = widthz*lg.zxc[i];
	  if(gg.ngeomz == 5) widthz = widthz*sin(lg.zyc[j]);
          cs = sqrt(gam*lg.P0[iqe][k][j][i]/lg.P0[iqd][k][j][i]);
	  xvel = fabs(lg.P0[iqu0][k][j][i]);
	  yvel = fabs(lg.P0[iqu0+1][k][j][i]);
	  zvel = fabs(lg.P0[iqu0+2][k][j][i]);
#ifdef UNSPLIT
	  ri_dt = max((xvel+cs)/lg.zdx[i],max((yvel+cs)/widthy,max((zvel+cs)/widthz,ri_dt)));
#else
	  width = min(lg.zdx[i],min(widthy,widthz));
	  ri_dt = max(xvel/lg.zdx[i],max(yvel/widthy,max(cs/width,ri_dt)));
#endif
	} 
      }
    }
  }

  return ri_dt;

}


