/*!  \file  dust.cpp
 *   \brief Contains functions needed to compute various dust physics.
 *
 *   \author Julian Pittard (Original version 17.11.17)
 *   \version 0.1-development (Undómiel):
 *   \date Last modified: 16.06.11 (JMP)
 */

#undef MAIN

//  Header files:
#include <cstdlib>
#include <cmath>
#include <iomanip> 
#include <iostream>
#include <vector>
#include "defs.h"
#include "array.h"
#include "constants.h"
#include "funcs.h"
#include "gas.h"
#include "global.h"
#include "grid.h"
#include "physconst.h"
#include "star.h"

using namespace std;

// File scope variables

// Function declarations
void dustEvolution();
void dustSputterDS79();
extern double findForwardShock();
extern void quit();



// Comment out the rest of the file if DUST is not defined
#ifdef DUST

// Evolve the dust (e.g. due to sputtering).
void dustEvolution(){

  dustSputterDS79();
  return;
}



// Sputter the dust using the Draine & Salpeter (1979) prescription.
// For T > 1e6 K, the dust lifetime tau_d = 1e6 (a/n) yr (a in microns).
// where a is the grain radius and n the gas nucleon number density.
// See p156-157 of "WBB" book for further details. Sputtered atoms
// contribute to the gas density (and if are co-moving with the gas
// the gas maintains its bulk velocity and pressure).
// JMP 22/11/17 - Correctly working
void dustSputterDS79(){

  const double dens_g = grainBulkDensity;            // (g/cm^3)
  const double a2 = grainRadius*grainRadius;
  const double massD = (4.0/3.0)*pi*pow(grainRadius,3)*dens_g; // grain mass (g)
  
  double rho,nH,ntot,z,rhod,nD,rhod_dot,drhod,rhodnew,rhonew,pre,temp;

  double avgm = avgmassamb;
  double const_1 = avgm/boltzman;

  for (int k = lg.krs; k <= lg.kre; k++){
    for (int j = lg.jrs; j <= lg.jre; j++){
      for (int i = lg.irs; i <= lg.ire; i++){
        //cout << "i = " << i << "; j = " << j << "; k = " << k << "\n";
	rho = lg.P0[iqd][k][j][i];
	nH = rho*(10.0/14.0)/massh;           // solar with 10 H per 1 He, avg nucleon mass mu_nu = 14/11. 
        ntot = 1.1*nH;                        // total nucleon number density
	z = lg.P0[iqal0][k][j][i];            // dust mass fraction 
	rhod = rho*z;                         // dust mass density (g/cm^3)
        nD = rhod/massD;                      // grain number density (cm^-3)
	pre = lg.P0[iqe][k][j][i];
	temp = const_1*pre/rho;               // K
	
	if (temp > 1.0e6 && z > 0.0){
	  rhod_dot = 1.33e-17*dens_g*a2*ntot*nD; // dust destruction rate (g cm^-3 s^-1)
	  drhod = rhod_dot*dt;
	  rhodnew = max(0.0,rhod - drhod);    // new dust density
	  rhonew = rho + drhod;               // new gas  density
	  //if ((ncycle > 10000)&&(z > 1.0e-3)){
	  //  cout << "rho = " << rho << "; ntot  =" << ntot << "; rhod = " << rhod << "; nD = " << nD << "; drhod = " << drhod << "\n";
	  //  quit();
	  //}
	  lg.P0[iqd][k][j][i] = rhonew;
	  lg.P0[iqal0][k][j][i] = rhodnew/rhonew; 
	}
      }
    }
  }
  
  return;
}

#endif



