/*!  \file grid.h
 *   \brief This is a repository for the grid variables and constants.
 *
 *   \param nd (const int) defines the dimensionality of the grid.
 *   \param imax (const int) is the number of zones in 1st direction.
 *   \param jmax (const int) is the number of zones in 2nd direction.
 *            If nd = 1, jmax MUST equal 1.
 *   \param kmax (const int) is the number of zones in 3rd direction.
 *            If nd < 3, kmax MUST equal 1.
 *   \param smax (const int) is the total number of advected scalars (min=1)
 *   \param cmax (const int) is the number of advected colours (min=1)
 *   \param nghost (const int) is the number of ghost cells to use
 *   \param cn (const double) is the courant number
 *   \param geom (string) specifies the grid geometry (XYZ or ZRP)
 *   \param ngeomx (int) stores the geometry of the 1st grid dimension
 *   \param ngeomy (int) stores the geometry of the 2nd grid dimension
 *   \param ngeomz (int) stores the geometry of the 3rd grid dimension
 *   \param nleftx (int) stores the left grid boundary type of the 1st dimension
 *   \param nlefty (int) stores the left grid boundary type of the 2nd dimension
 *   \param nleftz (int) stores the left grid boundary type of the 3rd dimension
 *   \param nrightx (int) stores the right grid boundary type of the 1st dimension
 *   \param nrighty (int) stores the right grid boundary type of the 2nd dimension
 *   \param nrightz (int) stores the right grid boundary type of the 3rd dimension
 *   \param xmin (double) is the left edge of the 1st grid dimension
 *   \param ymin (double) is the left edge of the 2nd grid dimension
 *   \param zmin (double) is the left edge of the 3rd grid dimension
 *   \param xmax (double) is the right edge of the 1st grid dimension
 *   \param ymax (double) is the right edge of the 2nd grid dimension
 *   \param zmax (double) is the right edge of the 3rd grid dimension
 *   \param P0 (<double,4>) stores the cell averaged primitives
 *   \param Pstar (<double,4>) stores the cell averaged primitives at the end of the predictor step
 *   \param zxa (<double>) stores the left edge coord of cells in the 1st dimension
 *   \param zya (<double>) stores the left edge coord of cells in the 2nd dimension
 *   \param zza (<double>) stores the left edge coord of cells in the 3rd dimension
 *   \param zxc (<double>) stores the coord of cell centres in the 1st dimension
 *   \param zyc (<double>) stores the coord of cell centres in the 2nd dimension
 *   \param zzc (<double>) stores the coord of cell centres in the 3rd dimension
 *   \param zxg (<double>) stores the coord of cell centre of mass in the 1st dimension
 *   \param zyg (<double>) stores the coord of cell centre of mass in the 2nd dimension
 *   \param zzg (<double>) stores the coord of cell centre of mass in the 3rd dimension
 *   \param zdx (<double>) stores the width of cells in the 1st dimension
 *   \param zdy (<double>) stores the width of cells in the 2nd dimension
 *   \param zdz (<double>) stores the width of cells in the 3rd dimension
 * 
 *   \author Julian Pittard (Original version 31.05.11)
 *   \version 0.1-development (Undómiel):
 *   \date Last modified: 18.08.11 (JMP)
 */

#include <cmath>
#include <string>      

using namespace std;

// Constants related to the grid. Others are defined in constants.h.
#define NDIM 1                ///< Number of dimensions
#define IMAXX 1200             ///< Number of cells in i-direction of global grid
#define JMAXX 1             ///< Number of cells in j-direction of global grid
#define KMAXX 1             ///< Number of cells in k-direction of global grid
#define MMAXX 1200             ///< Maximum number of cells in any direction - max(IMAXX,JMAXX,KMAXX)
#define NIPRO 1               ///< Number of processors in i-direction
#define NJPRO 1               ///< Number of processors in j-direction
#define NKPRO 1               ///< Number of processors in k-direction
#define MAXX  1200             ///< Maximum number of cells in any direction on a LOCAL grid

static const int niProc = NIPRO;  ///< Number of processors in i-direction
static const int njProc = NJPRO;  ///< Number of processors in j-direction
static const int nkProc = NKPRO;  ///< Number of processors in k-direction

static const int smax = 2;    ///< total number of advected scalars (min=1)
static const int cmax = 2;    ///< number of advected colours (min=1)

#if NDIM == 1
static const int nd   = 1;           ///< dimensionality of grid
#elif NDIM == 2
static const int nd   = 2;           ///< dimensionality of grid
#else
static const int nd   = 3;           ///< dimensionality of grid
#endif

static const int iqmax = 2+nd+smax;  ///< total number of hydro variables per cell (hydro: density, pressure, nd velocity, smax)
#ifdef UNSPLIT
static const int nghost = 2;         ///< number of ghost zones at each boundary, surrounding the grid
#else
static const int nghost = 3;         ///< number of ghost zones at each boundary, surrounding the grid
#endif

const int iqd = 0;  ///< density
const int iqu0 = 1; ///< velocity0/mtm0 (in 2D the velocity components are iqu0 and iqu0+1)
const int iqe = iqu0+nd; ///< pressure/energy
const int iqal0 = iqe+1; ///< advected scalar 0 (with smax=2, the scalar components are iqal0 and iqal0+1)


// Struct for global grid
struct ggrid{
  static const int imax = IMAXX;         ///< number of zones in 1st direction
  static const int jmax = JMAXX;         ///< number of zones in 2nd direction
  static const int kmax = KMAXX;         ///< number of zones in 3rd direction
  static const int mmax = MMAXX;         ///< maximum number of zones in any direction
  static const int ncells = imax*jmax*kmax;///< the number of cells on the grid
  string geom;                           ///< specify geometry (XYZ, ZRP or RTP)
  int ngeomx, ngeomy, ngeomz;            ///< geometry flags
  int nleftx, nlefty, nleftz, nrightx, nrighty, nrightz; ///< boundary flags
  double xmin;   ///< min coord of 1st axis on global grid (real cells ONLY)
  double xmax;   ///< max coord of 1st axis on global grid (real cells ONLY)
  double ymin;   ///< min coord of 2nd axis on global grid (real cells ONLY)
  double ymax;   ///< max coord of 2nd axis on global grid (real cells ONLY)
  double zmin;   ///< min coord of 3rd axis on global grid (real cells ONLY)
  double zmax;   ///< max coord of 3rd axis on global grid (real cells ONLY)
  double dotflo,uotflo,votflo,wotflo,potflo,dinflo,uinflo,vinflo,winflo,pinflo,einflo;
  double asotflo[smax],asinflo[smax];
  double xmax0;  ///< initial value of xmax
};


// Struct for grid local to each processor
struct lgrid{
  static const int imax = IMAXX/niProc;  ///< number of real zones in 1st direction
  static const int jmax = JMAXX/njProc;  ///< number of real zones in 2nd direction
  static const int kmax = KMAXX/nkProc;  ///< number of real zones in 3rd direction
  static const int mmax = MAXX;            ///< maximum number of zones in any direction

  static const int imaxd = imax + 2*nghost; ///< total number of zones in i-direction (including ghosts)
#if NDIM > 1
  static const int jmaxd = jmax + 2*nghost; ///< total number of zones in j-direction
#else
  static const int jmaxd = 1;
#endif
#if NDIM == 3
  static const int kmaxd = kmax + 2*nghost; ///< total number of zones in k-direction
#else
  static const int kmaxd = 1;
#endif

  static const int maxsweep = mmax + 2*nghost; // the MAX of the dimensions (including ghosts)

  int nmin,nmax;    // indices of first and last grid cells in sweeps

  // Local grid extent. global grid extent is defined in structure ggrid.
  double xmin;   ///< min coord of 1st axis on local grid (real cells ONLY). 
  double xmax;   ///< max coord of 1st axis on local grid (real cells ONLY)
  double ymin;   ///< min coord of 2nd axis on local grid (real cells ONLY)
  double ymax;   ///< max coord of 2nd axis on local grid (real cells ONLY)
  double zmin;   ///< min coord of 3rd axis on local grid (real cells ONLY)
  double zmax;   ///< max coord of 3rd axis on local grid (real cells ONLY)

  // Cell averaged primitive variables
  // Memory is dynamically assigned to create 4D arrays of dimensions [iqmax][kmaxd][jmaxd][imaxd]
  double**** P0;
#ifdef UNSPLIT
  double**** Pstar;
#endif

  // Cooling array used in radiate functions, of dimensions [kmaxd][jmaxd][imaxd]
  double*** dei;

#ifdef MASSLOAD
  // Array used to store the reservoir of gas still available for mass-loading,
  // of dimensions [kmaxd][jmaxd][imaxd]
  double*** rhoML;
#endif
  
  double zxa[imaxd+1]; // left edges of cells in x-direction
  double zxc[imaxd+1]; // centre 
  double zxg[imaxd+1]; // centre of mass
  double zdx[imaxd+1]; // width
  double zya[jmaxd+1]; // left edges of cells in y-direction
  double zyc[jmaxd+1]; // centre 
  double zyg[jmaxd+1]; // centre of mass
  double zdy[jmaxd+1]; // width
  double zza[kmaxd+1]; // left edges of cells in z-direction
  double zzc[kmaxd+1]; // centre 
  double zzg[kmaxd+1]; // centre of mass
  double zdz[kmaxd+1]; // width

  double Pbc[iqmax][3][2]; // for storing fixed boundary conditions: (iq,ijk,lr)
  
  int irs, ire, jrs, jre, krs, kre; // inclusive range of real cells on the grid

  static const int paramax=9; ///< total number of parabola's needed
  double zparax[paramax][maxsweep]; // parabolic coeffs in x-direction
  double zparay[paramax][maxsweep]; // parabolic coeffs in y-direction
  double zparaz[paramax][maxsweep]; // parabolic coeffs in z-direction
  double dr0;           // initial grid spacing in x-direction
  double radius;        // radius out to current column of angular zones
  double theta;
  double stheta;
};


// Variables
#ifdef MAIN
lgrid lg; // grid local to each processor
ggrid gg; // global grid
#else
extern lgrid lg;
extern ggrid gg;
#endif

