/*!  \file heatCool.cpp
 *   \brief Contains general heating and cooling functions.
 *
 *   \author Julian Pittard (Original version 20.10.17)
 *   \version 0.1-development (Undómiel):
 *   \date Last modified: 08.08.11 (JMP)
 */


#undef MAIN

//  Header files:
#include <cmath>
#include <fstream>
#include <iomanip>
#include <iostream>
#include <vector>
#include "defs.h"
#include "constants.h"
#include "funcs.h"
#include "gas.h"
#include "global.h"     
#include "globalMPI.h"
#include "grid.h"        
#include "physconst.h"
#include "star.h"

// Function declarations
void adjustPressureDueToCooling();
void calculateEquilibriumTemperatureCurve(string coolCurveFile);
double calculateTemperature(double dens, double pre, double y);
double netCoolingRate(const double rho, const double temp);
void radiateExact();
void radiateHeatCool();
void restrictCool();
void setupHeatingAndCooling(string coolCurveFile);
double (*func)(double, double); // function pointer for the function to root-find
double zbrent(double (*func)(double,double), double x1, double x2, double tol, double rho);

extern double pdtmp(double pdcell);
extern void quit();
extern double tmpmu(double temp);

using namespace std;

// File-scope constants and variables
const bool restrictUnresolvedCooling = true;

#ifdef DUST  
  const int ncool = 2;          // number of cooling curves
#else
  const int ncool = 1;          // number of cooling curves
#endif

// Structure to store cooling curve data
struct coolingCurve{
  int ntmax;
  string coolCurveFile;
  vector<double> logt,lambdac,te,loglambda;
  double t_min,t_max,logtmin,logtmax,dlogt;
};

coolingCurve cc[ncool];


/*!  \brief Calculate the gas temperature, given its density, pressure
 *   and ionization fraction.
 * 
 *   \author Julian Pittard (Original version 20.10.17)
 *   \version 1.0-stable (Evenstar):
 *   \date Last modified: 20.10.17 (JMP)
 */
/*
double calculateTemperature(double dens, double pre, double y){
  double x = 1.0 - y;                         // neutral fraction
  double nH = dens/muH;                       // cm^-3 (total H number density)
  //double nHI = x*nH;                          // cm^-3 (no. density of neutral H)
  double nHII = y*nH;                         // cm^-3 (no. density of ionized H)
  double ne = 1.1*nHII;                       // cm^-3 (no. density of electons, assuming 10% He by number and that He is singly ionized whenever H is - see Mackey et al. 2015, A&A, 573, A10)
                
  return pre/((nH+ne)*boltzman);       // K (see Rijkhorst et al. 2006)
}
*/


/*!  \brief Perform an EXACT calculation of radiative cooling (no heating).
 *
 *   For a given cooling curve, calculate the temporal evolution function (TEF)
 *   so that an exact integration scheme for the cooling can be used (see
 *   Townsend 2009, ApJS, 181, 391). 
 *   Based on exactcooling_good.f in other_programs/ExactCooling.
 *
 *   This is called from operatorsplit(), and is activated once per step, after
 *   the hydro calculation is complete. This is completely separate from the
 *   calcsourceterms() function, and should help to improve the stability of
 *   the code. Hence we update U.
 *
 *   Various possibilities for bugs exist. Current limitations include:
 *    1) The mininum (floor) temperature must not be below the minimum
 *       temperature in the cooling curve.
 *    2) The temperature scales of each cooling curve must be identical.
 *    3) At least 2 cooling curves are expected (these can be identical
 *       if required).
 *    4) The cooling curves must have a uniform logarithmic spacing in 
 *       temperature (because of the guess for the index kk).
 * 
 *   \author Julian Pittard (Original version 13.09.11)
 *   \version 1.0-stable (Evenstar):
 *   \date Last modified: 13.09.11 (JMP)
 */
void radiateExact(){

  const int ncmax = 2;          // maximum number of cooling curves
  const int ntmax = 121;        // maximum number of temperatures in a cooling curve
  //string coolcurve[ncmax];         // cooling curve filenames
  
  bool restrictUnresolvedCooling = true;
  int ncool = 2;                   // actual number of cooling curves (CWB,WBB,SNR = 2)

  //coolcurve[0] = "solar0.005-100keV";
  //coolcurve[1] = "solar0.005-100keV";
  
  int cc,kk,indx;
  double dens,pre,col,avgm,const_1,temp,mu;
  double tmpold,tmpnew,dtemp,fac2,pnew;
  double logtemp,invy,y_new,y_temp;
  double lambda_init[ncmax],powr,t_cool,temp_new_town;

  // The following variables need their values "preserved" between calls
  // (or alternatively they could be made global or file-scope)
  static bool first = true;
  static double t_min,t_max,t_ref,logtmin,logtmax,dlogt;
  static double te[ntmax];                // cooling curve temperatures     (assumed identical between curves)
  static double logt[ntmax];              // cooling curve log temperatures (assumed identical between curves)
  static double lambda[ncmax][ntmax];     // cooling curve lambda's
  static double loglambda[ncmax][ntmax];  // cooling curve log lambda's
  static double y[ncmax][ntmax];          // cooling curve y's
  static double alpha[ncmax][ntmax];      // cooling curve alpha's
  static double lambda_ref[ncmax];        // cooling curve reference lambda's
  static double fac[ncmax];               // 

  // If first call to this subroutine we need to read data files from disk.

  if (first){

    // Read in cooling curve(s). It is assumed that the temperature binning is 
    // uniform in log space i.e. dlog(T) is a constant 

    for (int nc = 0; nc < ncool; nc++){
      //ifstream file(coolcurve[nc].c_str()); //C++ method (coolcurve must be a string type)
      //ifstream file("solar0.005-100keV");
      ifstream file("cooling_curve_solar_lambdaJMP_fromWang2014.txt"); 
      if (!file) {
        cerr << "radiate_exact: failed to open file\n";
        quit();
      } 
      //cout << "Reading in cooling curve " << coolcurve[nc] << "\n"; 
      for (int n = 0; n < ntmax; n++){
        //cout << "n = " << n << "\n";
        file >> logt[n] >> lambda[nc][n];
        te[n] = pow(10,logt[n]);
        loglambda[nc][n] = log10(lambda[nc][n]);
        //cout << n << "\t" << te[n] << "\t" << lambda[nc][n] << "\n";
      }
      file.close();
      if (tmin < te[0]){
        cout << "Minimum temperature is below that in cooling curve\n"; // " << nc << ": " << coolcurve[nc] << ".\n";
        //cout << " tmin = " << tmin << " (K); te[0] = " << te[0] << " (K). Aborting!\n";
        //quit();
      }
    } // loop over reading ncool cooling curves

    t_min = te[0];       // K
    t_max = te[ntmax-1]; // K
    logtmax = logt[ntmax-1];
    logtmin = logt[0];
    dlogt = (logtmax-logtmin)/float(ntmax-1);

    // Construct alpha(k) (lambda(k) = lambda(n), t(k) = t(n))
      
    for (int n=0; n < ncool; n++){
      for (int k=0; k < ntmax-1; k++){
        alpha[n][k] = log10(lambda[n][k+1]/lambda[n][k])/log10(te[k+1]/te[k]);
      }

      // Construct the temporal evolution function (TEF)
      // T_ref is an arbitrary reference temperature, and we will set T_ref = T_max
      
      t_ref = t_max;
      y[n][ntmax-1] = 0.0; // zero time to cool from T_ref to T_ref
      lambda_ref[n] = lambda[n][ntmax-1]; // value of lambda at reference temperature
      fac[n] = lambda_ref[n]/t_ref;

      // Construct the TEF using Eq A6 in Townsend (2009)

      for (int k=ntmax-2; k >= 0; k--){
        if (alpha[n][k] != 1.0){
          y[n][k] = y[n][k+1] - (1.0/(1.0-alpha[n][k]))*fac[n]*(te[k]/lambda[n][k])* 
	    (1.0 - pow((te[k]/te[k+1]),(alpha[n][k]-1.0)));
        }    
        else{
          y[n][k] = y[n][k+1] - fac[n]*(te[k]/lambda[n][k])*log(te[k]/te[k+1]);
        }
      }

      // Write out the TEF
      //cout << "TEF, coolcurve n = " << n << "\n";
      for (int m = 0; m < ntmax; m++){
	//cout << logt[m] << "\t" << lambda[n][m] << "\t" << y[n][m] << "\t" << 1.0/y[n][m] << "\n";
      }
      //quit();
    } // loop over number of cooling curves
         
    first = false;
  }


  // Now loop over cells, calculating the exact cooling rate

  for (int k = lg.krs; k <= lg.kre; k++){
    for (int j = lg.jrs; j <= lg.jre; j++){
      for (int i = lg.irs; i <= lg.ire; i++){
        //cout << "i = " << i << "; j = " << j << "; k = " << k << "\n";
        lg.dei[k][j][i] = 0.0;      // initialize
	dens = lg.P0[iqd][k][j][i];     // g/cm^3
        pre  = lg.P0[iqe][k][j][i];
        col  = lg.P0[iqal0][k][j][i];   // (=1.0 for solar, 0.0 for WC)
        //if (col > 0.5){
	  //avgm = stavgm[0][0];
	  avgm = avgmass;
          cc = 0;                // use coolcurve1 for ALL cooling
	//} 
        //else{
        //  avgm = stavgm[1][0];      //(see "Multiple Winds" book p151)
	//  cc = 1;                // use coolcurve2 for ALL cooling
        //}
        const_1 = avgm/boltzman;
        temp = const_1*pre/dens;   // K
	mu = avgm/massh;
	tmpold = temp;

	//cout << "temp = " << temp << "\n";
	
        //temp = 2.0e4;           //use only for code testing

        // Exit (with zero cooling) if temp <= te[0]. We use te[0] instead of tmin, 
        // because if tmin<te[0] it won't trigger if t>tmin but <te[0] and the exact 
        // integration will fail.
        if (temp > 1.1*te[0]){

 	  // Calculate y_temp = y(T). This is the value of y at the desired starting 
	  // temperature. Use Eq. A5 to do this. First find nearest temp and therefore 
	  // the value of the index kk. Also calculate lambda_init
     
  	  logtemp = log10(temp);
	  kk = (logtemp-logtmin)/dlogt;
          //if ((i == 48)&&(j == 2)){
          //  cout << "temp = " << temp << "; kk = " << kk << "\n";
	  //}
	
	  if (kk == ntmax-1){
	    y_temp = 0.0;
	    lambda_init[cc] = lambda[cc][ntmax-1];
          }
          else{
	    if (alpha[cc][kk] != 1.0){
	      y_temp = y[cc][kk] + (1.0/(1.0-alpha[cc][kk]))*fac[cc]*(te[kk]/lambda[cc][kk])*(1.0 - pow((te[kk]/temp),(alpha[cc][kk]-1.0)));  //Eq A6
              //if ((i == 49)&&(j == 2)){
              //  cout << "kk = " << kk << "; cc = " << cc << "; y[cc][kk] = " << y[cc][kk] << "; alpha[cc][kk] = " << alpha[cc][kk] << "; te[kk] = " << te[kk] << "; fac[cc] = " << fac[cc] << "; lambda[cc][kk] = " << lambda[cc][kk] << "\n";
	      //}
            }
            else{
	      y_temp = y[cc][kk] + fac[cc]*(te[kk]/lambda[cc][kk])*log(te[kk]/temp);
            }
	    lambda_init[cc] = lambda[cc][kk]*pow((temp/te[kk]),alpha[cc][kk]);
          }

          //if ((i == 49)&&(j == 2)){
          //  cout << "Starting temp T = " << temp << "\n";
          //  cout << "            Y(T)= " << y_temp << "\n";
          //  cout << "       lambda(T)= " << lambda_init[cc] << "\n";
	    //  quit();
          //}

	  // Calculate the single point cooling time. See p188 of "Computing Notes 2017".
          t_cool = 1.0/((gam-1.0)*(dens*mu/(massh*boltzman))*lambda_init[cc]/temp); // This assumes that Edot = (rho/mH)^2 lambda(T)
          //if (rank == 0) cout << "t_cool = " << t_cool << " (s)\n";

	  // Calculate the new temperature given a timestep dt

	  // Townsend's exact integration scheme (Eq 26)
	  // y_new is the value of the square bracket on the RHS
	  y_new = y_temp + (temp/t_ref)*(lambda_ref[cc]/lambda_init[cc])*(dt/t_cool);

	  // The new temperature is now given using Eq A7 to obtain 
	  // Y^-1(y_new) = temp_new (see Eq. 26)
	  // We need to construct the inverse of the TEF (i.e. Y^-1 - see Eq. A7)
	  // First obtain the indx of y_new. This means that y_new lies 
	  // between y(kk) and y(kk+1)

          indx = ntmax-2;
          bool tempmin = false;
          while ((y_new > y[cc][indx]) && (indx > 0)){
	    indx--;
	  }
	  if (indx == 0) tempmin = true;
	  if (tempmin == false){
            kk = indx;
            //if ((i == 49)&&(j == 2)){
            //  cout << "y_new = " << y_new << "\n";
            //  cout << "kk = " << kk << "\n";
            //  cout << "y(k)   = " << y[indx][cc] << "\n";
            //  cout << "y(k+1) = " << y[indx+1][cc] << "\n";
            //}

	    // Now calculate the inverse of the TEF
	    fac2 = lambda[cc][kk]/(te[kk]*fac[cc]);
            if (alpha[cc][kk] != 1.0){
              powr = 1.0/(1.0-alpha[cc][kk]);
              invy = te[kk]*pow((1.0-(1.0-alpha[cc][kk])*fac2*(y_new - y[cc][kk])),powr);
            }
            else invy = te[kk]*exp(-fac2*(y_new-y[cc][kk])); 
            temp_new_town = invy;
	  }
          else temp_new_town = te[0];
        
	  // Store change in energy, dei = delta(temp) * density

  	  tmpnew = max(te[0],temp_new_town); // use te[0] instead of tmin
	  tmpnew = min(tmpnew,tmax);
          lg.dei[k][j][i]= (tmpold - tmpnew)*dens;

	
          //if ((i == 49)&&(j == 2)){
          //  cout << "ijk = " << i << " " << j << " " << k << "; tmpold = " << tmpold << "; temp_new_town = " << temp_new_town << "; tmin = " << tmin << "\n";
          //}
          //quit();

	} // if T > 1.1*te[0]
	else{
	  // Set pressure to tmin
          lg.P0[iqe][k][j][i] = tmin*dens/const_1;
	}

      } // i loop
    }   // j loop
  }     // k loop


  // Restrict cooling at unresolved interfaces
  if (restrictUnresolvedCooling) restrictCool();

  // Adjust pressure due to cooling
  adjustPressureDueToCooling();

  return;
}


/*!  \brief Restrict the cooling rate at unresolved interfaces between hot 
 *          diffuse gas and cold dense gas.
 *
 *   Replace deltaE with minimum of neighboring deltaEs at the interface.
 *
 *   \author Julian Pittard (Original version 13.09.11)
 *   \version 1.0-stable (Evenstar):
 *   \date Last modified: 13.09.11 (JMP)
 */
void restrictCool(){

  int itp,ibt,jtp,jbt,ktp,kbt;
  double scrch[lgrid::imaxd], dis[lgrid::imaxd];
  double drhox[lgrid::imaxd], drhoy[lgrid::imaxd], drhoz[lgrid::imaxd];

  for (int k = lg.krs; k <= lg.kre; k++){
    ktp = min(k+1,lg.kre);
    kbt = max(k-1,lg.krs);
    for (int j = lg.jrs; j <= lg.jre; j++){

      // Locate cloud interfaces => dis = 1, otherwise, dis = -1

      jtp = min(j+1,lg.jre);
      jbt = max(j-1,lg.jrs);
      for (int i = lg.irs+1; i < lg.ire; i++){
	scrch[i] = min(lg.P0[iqd][k][j][i+1],lg.P0[iqd][k][j][i-1]);
	drhox[i] = lg.P0[iqd][k][j][i+1] - lg.P0[iqd][k][j][i-1];
        drhox[i] = copysign(drhox[i],(lg.P0[iqe][k][j][i-1]/lg.P0[iqd][k][j][i-1]
		     - lg.P0[iqe][k][j][i+1]/lg.P0[iqd][k][j][i+1]))/scrch[i] - 2;
	if (nd > 1){
  	  scrch[i] = min(lg.P0[iqd][k][jtp][i],lg.P0[iqd][k][jbt][i]);
	  drhoy[i] = lg.P0[iqd][k][jtp][i] - lg.P0[iqd][k][jbt][i];
          drhoy[i] = copysign(drhoy[i],(lg.P0[iqe][k][jbt][i]/lg.P0[iqd][k][jbt][i]
		     - lg.P0[iqe][k][jtp][i]/lg.P0[iqd][k][jtp][i]))/scrch[i] - 2;
        }
	if (nd == 3){
  	  scrch[i] = min(lg.P0[iqd][ktp][j][i],lg.P0[iqd][kbt][j][i]);
	  drhoz[i] = lg.P0[iqd][ktp][j][i] - lg.P0[iqd][kbt][j][i];
          drhoz[i] = copysign(drhoz[i],(lg.P0[iqe][kbt][j][i]/lg.P0[iqd][kbt][j][i]
		     - lg.P0[iqe][ktp][j][i]/lg.P0[iqd][ktp][j][i]))/scrch[i] - 2;
        }
	if      (nd == 1) dis[i] = drhox[i];
	else if (nd == 2) dis[i] = max(drhox[i],drhoy[i]);
	else              dis[i] = max(drhox[i],max(drhoy[i],drhoz[i]));
      }
      dis[lg.irs] = -1.0;
      dis[lg.ire] = -1.0;

      for (int i = lg.irs; i <= lg.ire; i++){
	itp = min(i+1,lg.ire);
	ibt = max(i-1,lg.irs);
	if      (nd == 1) scrch[i] = min(lg.dei[k][j][ibt],lg.dei[k][j][itp]);
	else if (nd == 2) scrch[i] = min(lg.dei[k][j][ibt],min(lg.dei[k][j][itp],
                                     min(lg.dei[k][jtp][i],lg.dei[k][jbt][i])));
        else              scrch[i] = min(lg.dei[k][j][ibt],min(lg.dei[k][j][itp],
			             min(lg.dei[k][jtp][i],min(lg.dei[k][jbt][i],
                                     min(lg.dei[ktp][j][i],lg.dei[kbt][j][i])))));
	if (dis[i] > 0.0) lg.dei[k][j][i] = scrch[i];
      }

    } // j loop
  } // k loop
  
  return;
}


/*!  \brief Adjust the pressure due to cooling.
 *
 *   Uses dei(). dei is positive if cooling.
 *
 *   \author Julian Pittard (Original version 13.09.11)
 *   \version 1.0-stable (Evenstar):
 *   \date Last modified: 13.09.11 (JMP)
 */
void adjustPressureDueToCooling(){
  double dens,pre,col,pdcell,tmpold,muold,dtemp,tmpnew,pnew;
  
  for (int k = lg.krs; k <= lg.kre; k++){
    for (int j = lg.jrs; j <= lg.jre; j++){
      for (int i = lg.irs; i <= lg.ire; i++){
        
	dens = lg.P0[iqd][k][j][i];     // g/cm^3
        pre  = lg.P0[iqe][k][j][i];
        col  = lg.P0[iqal0][k][j][i];   // (=1.0 for solar, 0.0 for WC)

	// Calculate temperature from ratio of p/rho using a temperature-dependent avg particle mass  
	pdcell = pre/dens;
	tmpold = pdtmp(pdcell);
     
	dtemp = lg.dei[k][j][i]/dens;
	tmpnew = max((tmpold-dtemp),tmin);
	tmpnew = min(tmpnew,tmax);
	pnew = pre*tmpnew/tmpold; // this scaling is the best way to handle the change
	
        lg.P0[iqe][k][j][i] = pnew;         
        //if (j == 2){
	//if (tmpnew > 1.1*tmpold){
	//  cout << "j = " << j << "; i = " << i << "; tmpold = " << tmpold << "; tmpnew = " << tmpnew << "; preold = " << pre << "; prenew = " << pnew << "\n"; //lg.P0[iqe][k][j][i] << "\n";
	//  cout << "dens = " << dens << "; avgm = " << avgm << "; const_1 = " << const_1 << "; dtemp = " << dtemp << "\n";
	//  quit();
	//}

	//if (i == lg.ire){// && ncycle < 3){
	//  cout << "[AdjustPressureDueToCooling] tmpold = " << tmpold << " (K); tmpnew = " << tmpnew << " (K)\n";
	//  cout << "j = " << j << "; i = " << i << "; tmpold = " << tmpold << "; tmpnew = " << tmpnew << "; preold = " << pre << "; prenew = " << pnew << "\n"; //lg.P0[iqe][k][j][i] << "\n";
	//  cout << "dens = " << dens << "; dtemp = " << dtemp << "\n";
  	//  quit();
	//}

      } // i loop
    }   // j loop
  }     // k loop
  return;
}


  
/*!  \brief Heat and cool the gas.
 * 
 *   The cooling rate is given by (rho/mh)^2 * lambda.
 *   See numerical_cooling.cpp in other_programs/Cooling/lowTcool.
 *
 *   \author Julian Pittard (Original version 20.10.17)
 *   \version 1.0-stable (Evenstar):
 *   \date Last modified: 20.10.17 (JMP)
 */
void radiateHeatCool(){
      
  // Loop over cells, calculating the heating/cooling rate.
  // Adjust the gas pressure (temperature) as appropriate. 
  for (int k = lg.krs; k <= lg.kre; k++){
    for (int j = lg.jrs; j <= lg.jre; j++){
      for (int i = lg.irs; i <= lg.ire; i++){

	double dens = lg.P0[iqd][k][j][i];       // g/cm^3
        double pre  = lg.P0[iqe][k][j][i];
 
	// Calculate temperature from the ratio of p/rho using a temperature-dependent avg particle mass  
	double pdcell = pre/dens;
	double temp = pdtmp(pdcell);
	double mu = tmpmu(temp);  // calculate mu from temp (assumes solar abundance)
	double avgm = mu*massh;
	  
	double tempold = temp;
	double logtemp = log10(temp);
	double rhomh = dens/massh;

	//cout << "dt = " << dt << "\n";
	//dt = 1.0e14; // set dt to a high value (only for testing)
	
	double dtint = 0.0;
	bool firstLoop = true;
	bool firstLoopCooling = true;
	double temp_prev,net_cooling_rate_prev;
	double tempnew;
	
        while (dtint < dt){

	  //double dens = 2.0e-21; double temp = 30.0;
	  
	  double net_cooling_rate = netCoolingRate(dens,temp); // erg/s/cm^3 (positive if cooling, negative if heating)
	    	  
          //double Eint = pre/(tmpg(temp)-1.0); // use temperature-dependent gamma
          double Eint = pre/gamm; 
	  double t_cool = Eint/net_cooling_rate;
	  //std::cout << "t_cool = " << t_cool << "\n"; quit();
	  
	  double dtcool = 0.2*abs(t_cool); // safe timestep (20% change in temperature)
          if ((dtint + dtcool) > dt) dtcool = dt - dtint;
	  dtint += dtcool;
	  
	  // Calculate new temperature, and update pressure
	  double dEint = -net_cooling_rate*dtcool;
	  tempnew = temp*(Eint + dEint)/Eint;	
	  tempnew = max(tmin,tempnew);
	  tempnew = min(tempnew,tmax);

	  //if (i == lg.ire){
	  //  cout << "ncycle = " << ncycle << "; Temp = " << temp << " (K); Net cooling rate (negative means net heating) = " << net_cooling_rate; // << "; Heating rate = " << rhomh*GammaHeat <<
	  //  cout << "; tempnew = " << tempnew << " (K)\n";
	    //quit();
	  //}
	  //quit();
	  
	  if (firstLoop){
	    if (net_cooling_rate < 0.0) firstLoopCooling = false;
	    else                        firstLoopCooling = true;
	    firstLoop = false;
	  }
	  else{
	    if ((firstLoopCooling && net_cooling_rate < 0.0) || (!firstLoopCooling && net_cooling_rate > 0.0)){
	      // originally cooling/heating but now heating/cooling, so jump out.
	      // By definition there will be at least 2 subcycles.
	      // Calculate the new temperature through interpolation (this isn't perfect, but is good enough)
	      double grad = (temp - temp_prev)/(net_cooling_rate - net_cooling_rate_prev);
	      double dtotal_cool = (0.0 - net_cooling_rate_prev);	      
	      tempnew = temp_prev + dtotal_cool*grad;
	      //if (i == lg.ire){
	      //  std::cout << "temp_prev = " << temp_prev << "; net_cooling_rate_prev = " << net_cooling_rate_prev << "\n";
	      //  std::cout << "temp      = " << temp      << "; net_cooling_rate      = " << net_cooling_rate << "\n";
	      //  std::cout << "grad = " << grad << "; dtotal_cool = " << dtotal_cool << "; tempnew = " << tempnew << "\n";
	      //}
	      dtint = 1.1*dt; // force jump out
	    }
	  }

  	  // Record previous temp and total_cool value
	  temp_prev = temp;
	  net_cooling_rate_prev = net_cooling_rate;
	  
	  // Swap for next cycle
          temp = tempnew;
  	  logtemp = log10(temp);

	} // end of cooling sub-cycling
	//if (ncycle == 6) quit();
	
	// Calculate change in energy, dei = delta(temp) * density
	// dei is positive if cooling.
        lg.dei[k][j][i] = (tempold - tempnew)*dens;	
      }
    }
  }

  // Restrict cooling at unresolved interfaces
  if (restrictUnresolvedCooling) restrictCool();

  // Adjust pressure due to cooling
  adjustPressureDueToCooling();

  return;
}



/*!  \brief Setup function for the heating and cooling.
 *
 *   \author Julian Pittard (Original version 20.10.17)
 *   \version 1.0-stable (Evenstar):
 *   \date Last modified: 20.10.17 (JMP)
 */
void setupHeatingAndCooling(string coolCurveFile){

  // Specify the cooling curves and the number of temperature bins.
  // The first cooling curve should extend to the same or lower temperature
  // than the second.
  // The cooling_KI02_4.0_CLOUDY_7.6_MEKAL_correct_low.txt gas cooling rate
  // is Edot = lambda(T)*(rho/mH)^2. See p140 of "Computing Notes 2017" book. 
  // It has the correct lambda for log T > 4.0.
  // It extends from logT = 0.0-9.0, using 181 bins.

  cc[0].coolCurveFile = coolCurveFile;
  cc[0].ntmax = 181;
    
  // Read in each cooling curve. It is assumed that the temperature binning is 
  // uniform in log space i.e. dlog(T) is a constant
  for (int nc = 0; nc < ncool; ++nc){
    ifstream file(cc[nc].coolCurveFile.c_str()); 
    if (!file) {
      cerr << "radiateHeatCool: failed to open file " << cc[nc].coolCurveFile << "\n";
      quit();
    } 
    for (int n = 0; n < cc[nc].ntmax; ++n){
      double logt,lambda;
      file >> logt >> lambda;
      cc[nc].logt.push_back(logt);
      cc[nc].lambdac.push_back(lambda);
      cc[nc].te.push_back(pow(10,logt));
      cc[nc].loglambda.push_back(log10(lambda));
      //cout << "nc = " << nc << "; n = " << n << "\t" << cc[nc].te[n] << "\t" << cc[nc].lambdac[n] << "\n";
    }
    file.close();
    
    cc[nc].t_min   = cc[nc].te[0];             // K
    cc[nc].t_max   = cc[nc].te[cc[nc].ntmax-1]; // K
    cc[nc].logtmax = cc[nc].logt[cc[nc].ntmax-1];
    cc[nc].logtmin = cc[nc].logt[0];
    cc[nc].dlogt   = (cc[nc].logtmax-cc[nc].logtmin)/float(cc[nc].ntmax-1);
  }

  // Check that tmin is not below the minimum temperature in the first
  // cooling curve    
  if (tmin < cc[0].te[0]){
    cout << "Minimum temperature is below that in cooling curve\n"; 
    cout << " tmin = " << tmin << " (K); te[0] = " << cc[0].te[0] << " (K). Aborting!\n";
    quit();
  }
               
  if (procRank == 0){
    cout << "Finished setupHeatingAndCooling! ntmax = " << cc[0].ntmax << "\n";
    cout << "Cooling curve te[0]     = " << cc[0].te[0] << " (K); lambda(T) = " << cc[0].lambdac[0] << "\n";
    cout << "Cooling curve te[ntmax] = " << cc[0].te[cc[0].ntmax-1] << " (K); lambda(T) = " << cc[0].lambdac[cc[0].ntmax-1] << "\n";
  }
  return;
}



/*!  \brief Returns the equilibrium temperature, given a density.
 *
 *   \author Julian Pittard (Original version 20.10.17)
 *   \version 1.0-stable (Evenstar):
 *   \date Last modified: 20.10.17 (JMP)
 */
double equilibriumTemperature(double rho){
  // Set values which bracket the root

  const int nc = 0;
  func = netCoolingRate;
  double tol = 1.0e-3;
  double x1 = cc[nc].te[0];
  double x2 = cc[nc].te[cc[nc].ntmax-1];

  double root = zbrent(func,x1,x2,tol,rho);
  //if (procRank == 0) std::cout << "The netCoolingRate function has a root at Temp = " << root << "\n";
  //quit();
  return root;
}



/*!  \brief Calculate the equilibrium temperature as a function of density, for a given 
 *          heating rate and cooling curve.
 *
 *   \author Julian Pittard (Original version 20.10.17)
 *   \version 1.0-stable (Evenstar):
 *   \date Last modified: 20.10.17 (JMP)
 */
void calculateEquilibriumTemperatureCurve(string coolCurveFile){

  const int nc = 0;
  func = netCoolingRate;
  double tol = 1.0e-3;
  
  // Set temperature values which bracket the root
  double x1 = cc[nc].te[0];
  double x2 = cc[nc].te[cc[nc].ntmax-1];

  // Set density range
  const int nval = 81;
  double rhomin = 1.0e-26;
  double rhomax = 1.0e-18;
  double logrhomin = log10(rhomin);
  double logrhomax = log10(rhomax);
  double dlogrho = (logrhomax - logrhomin)/double(nval-1);

  string outfile = "equilibriumTemperatureCurve.dat";

  ofstream file(outfile.c_str()); //C++ method (outfile must be a string type)
  if (!file) {
    cerr << "Failed to open output file: " << outfile;
    quit();
  } 
  file << "# GammaHeat = " << GammaHeat << "; coolingcurve = " << coolCurveFile << "\n";
  file << "#   rho            Teq             mu           P/k = rho*Teq/(mu*massh)\n"; 

  // Loop over density
  for (int n = 0; n < nval; ++n){
    double rho = pow(10,logrhomin + double(n)*dlogrho);
    double Teq = zbrent(func,x1,x2,tol,rho);
    // Now calculate the temperature-dependent average particle mass and the pressure
    double mu = tmpmu(Teq);
    double poverk = rho*Teq/(mu*massh); 
    file << scientific << setprecision(4);
    file << rho << "\t" << Teq << "\t" << mu << "\t" << poverk << "\n";
    //if (procRank == 0) std::cout << "The netCoolingRate function has a root at Temp = " << root << "\n";
    //quit();
  }
  file.close();
  return;
}


/*!  \brief Given a density and a temperature, return the net cooling rate.
 *
 *   The returned value is positive if the gas is cooling, and vice-versa.
 *
 *   \author Julian Pittard (Original version 20.10.17)
 *   \version 1.0-stable (Evenstar):
 *   \date Last modified: 20.10.17 (JMP)
 */
double netCoolingRate(const double rho, const double temp){
  
  const int nc = 0; // assume that there is only one cooling curve to consider
  double lambda_cool = 0.0;
	  
  double logtemp = log10(temp);
  double rhomh = rho/massh;

  // Cooling is only calculated if the temperature is
  // within the temperature range of the cooling curve.
  if (temp >= cc[nc].te[0] && temp <= cc[nc].te[cc[nc].ntmax-1]){     
    int kk = int((logtemp-cc[nc].logtmin)/cc[nc].dlogt);

    if (kk == cc[nc].ntmax-1)
	lambda_cool = cc[nc].lambdac[cc[nc].ntmax-1];
    else{
      // temperature lies between logt[kk] and logt[kk+1]
      double grad = (cc[nc].loglambda[kk+1] - cc[nc].loglambda[kk])/cc[nc].dlogt; 
      lambda_cool = pow(10,cc[nc].loglambda[kk] + grad*(logtemp - cc[nc].logt[kk]));
    }
  }

  // Calculate cooling and heating rates
  double cooling_rate = rhomh*rhomh*lambda_cool; // erg/s/cm^3
  double heating_rate = rhomh*GammaHeat;               

  double net_cooling_rate = cooling_rate - heating_rate;
  return net_cooling_rate;
}


double zbrent(double (*func)(double,double), double x1, double x2, double tol, double rho){
  const int itmax = 100;
  const double eps = 1.0e-8;
  double d,e,p,q,r;

  double a = x1;
  double b = x2;
  double fa = func(rho,a);
  double fb = func(rho,b);
  if ((fa > 0.0 && fb > 0.0) || (fa < 0.0 && fb < 0.0)){
    std::cout << "Root must be bracketed for zbrent!\n";
    std::cout << "a = " << a << "; fa = " << fa << "\n";
    std::cout << "b = " << b << "; fb = " << fb << "\n";
    quit();
  }
  double c = b;
  double fc = fb;
  for (int iter = 0; iter < itmax; ++iter){
    if ((fb > 0.0 && fc > 0.0) || (fb < 0.0 && fc < 0.0)){
      // Rename a, b and c, and adjust bounding interval d
      c = a;
      fc = fa;
      d = b - a;
      e = d;
    }
    if (fabs(fc) < fabs(fb)){
      a = b;
      b = c;
      c = a;
      fa = fb;
      fb = fc;
      fc = fa;
    }
    // Convergence check
    double tol1 = 2.0*eps*fabs(b) + 0.5*tol;
    double xm = 0.5*(c - b);
    if (fabs(xm) <= tol1 || fb == 0.0){
      return b;
    }
    if (fabs(e) >= tol1 && fabs(fa) > fabs(fb)){
      // Attempt inverse quadratic interpolation
      double s = fb/fa;
      if (a == c){
	p = 2.0*xm*s;
	q = 1.0 - s;
      }
      else{
	q = fa/fc;
	r = fb/fc;
	p = s*(2.0*xm*q*(q-r) - (b-a)*(r-1.0));
	q = (q-1.0)*(r-1.0)*(s-1.0);
      }
      // Check whether in bounds
      if (p > 0.0) q = -q;
      p = fabs(p);
      if (2.0*p < fmin(3.0*xm*q - fabs(tol1*q),fabs(e*q))){
	// Accept interpolation
	e = d;
	d = p/q;
      }
      else{
	d = xm;
	e = d;
      }
    }
    else{
      // Bounds decreasing too slowly. Use bisection.
      d = xm;
      e = d;
    }
    // Move last best guess to a
    a = b;
    fa = fb;
    // Evaluate new trial root
    if (fabs(d) > tol1){
      b += d;
    }
    else{
      b += std::copysign(tol1,xm);
    }
    fb = func(rho,b);
  }
  std::cout << "Zbrent exceeds maximum iterations!\n";
  return b; // current best guess
}



