/*!  \file hydroGeneral.cpp
 *   \brief Contains general hydro functions.
 *
 *   For example, fillGhostCells.
 *
 *   \author Julian Pittard (Original version 01.08.11)
 *   \version 0.1-development (Undómiel):
 *   \date Last modified: 08.08.11 (JMP)
 */


#undef MAIN

//  Header files:
#include <cmath>
#include <fstream>
#include <iomanip>
#include <iostream>
#include "defs.h"
#include "constants.h"
#include "funcs.h"
#include "gas.h"
#include "global.h"     
#include "globalMPI.h"
#include "grid.h"        
#include "physconst.h"

// Function declarations
void fillGhostCells(double**** P);
void operatorSplit();
double pminFixUp(double avgmass, double tamb, double r, double p);
void riemann(int direc, double pL[], double pR[], double flux[]);
void riemannNL(double dl, double ul, double pl, double cl, double dr, double ur, double pr, double cr, double rs[3]);
double wave(double p, double p0);
extern void barrier();
extern void dustEvolution();
extern void mapWind(bool init);
extern void massLoad();
extern void quit();
extern void radiateExact();
extern void radiateHeatCool();

using namespace std;


/*!  \brief Perform operator split activities (e.g. cooling, non-hydro evolution of scalars etc.)
 *
 *   \author Julian Pittard (Original version 09.08.11)
 *   \version 0.1-development (UndÃ³miel):
 *   \date Last modified: 09.08.11 (JMP)
 */
void operatorSplit(){

#ifdef DUST
  dustEvolution();
#endif

  if (problem == "WBB"){

#ifdef MASSLOAD
    massLoad();
#endif    
#ifdef DUST    
    radiateHeatCool();
#else
    radiateHeatCool();
    //radiateExact();
#endif
    mapWind(false); 

  }
  else if (problem == "SNR"){

#ifdef MASSLOAD
    massLoad();
#endif    
#ifdef DUST    
    radiateHeatCool();
#else
    radiateExact();
#endif

  }
  else if (problem == "CWB"){
    radiateExact();
    mapWind(false);
  }
  
  return;
}


/*!  \brief Fill the ghost cells on the grid with their necessary values.
 *
 *   This function acts on the Grid, not a gridlet. The grid is passed in 
 *   as P. This is either P0 (for the split update, or during the 
 *   predictor step for the unsplit update) or Pstar (during the corrector 
 *   step of the unsplit update). The boundaries are filled in the order XYZ 
 *   in order to fill the corner cells properly.
 *
 *   Each edge of a Grid represents either:
 *     - (1) a physical boundary at the edge of the global grid; in which case BCs are
 *           specified by an integer flag set by the user in setup().
 *     - (2) the boundary between local grid domains resulting from decomposition of a larger
 *           global grid using MPI; in which case BCs are handled by MPI calls
 *
 *   For case (1) -- PHYSICAL BOUNDARIES
 *     The values of the integer flags (nleftx, etc.) are:
 *       - 1 = reflecting; 2 = outflow; 3 = fixed, 4 = periodic; 5 = outflow ONLY
 *
 *   For case (2) -- MPI BOUNDARIES
 *     We do the parallel synchronization by having every local grid domain:
 *      - 1) Post non-blocking receives for data from both L and R Grids
 *      - 2) Construct the appropriate MPI_TYPE_VECTORs for the array sub-blocks we
 *           wish to send. Send data to the Grids on both L and R.
 *      - 3) Check for receives (data is automatically received in the correct
 *           ghost cells of the local grid domain).
 *     Steps 1-3 above are cycled through in order of XYZ. We cannot start the Y
 *     communication until after the X communication has finished.
 *     If the Grid is at the edge of the Domain, we set BCs as in case (1).
 *
 *   This subroutine follows a similar approach as mhd_bvals.c in the Athena hydrodynamics code.
 * 
 *   Notes on MPI calls:
 *       int MPI_Irecv(void *buf, int count, MPI_Datatype datatype, int source,
 *             int tag, MPI_Comm comm, MPI_Request *request)
 *
 *       where: buf      = initial address of receive buffer (choice) 
 *              count    = number of elements in receive buffer (integer) 
 *              datatype = datatype of each receive buffer element (handle) 
 *              source   = procRank of source (integer) 
 *              tag      = message tag (integer) 
 *              comm     = communicator (handle) 
 *              request  = communication request (handle) - this is an output parameter which indicates
 *                                                          whether the receive has completed
 *
 *       reqsL[] is a 1D array with 2 elements - it simply holds two values which allow
 *        checking of whether both the send and recieve communication on the L has occured. 
 *       reqsR[] performs the same task for communication on the R.
 *       The MPI_Waitall() routine is used to check this.
 *
 *   \author Julian Pittard (Original version 28.03.12)
 *   \version 1.0-stable (Evenstar):
 *   \date Last modified: 19.04.12 (JMP)
 */
void fillGhostCells(double**** P){

  bool mpiCommLeft = false;
  bool mpiCommRight = false;
  int ierr;

  // reqsL[] is a 1D array with 2*iqmax elements - it holds two values for each iq which allow
  // checking of whether both the send and recieve communication on the L has occured. 
  // reqsR[] performs the same task for communication on the R.
  MPI_Request reqsL[2*iqmax]; // there is 1 send and 1 receive request for communication from/to the L per iq
  MPI_Request reqsR[2*iqmax]; // there is 1 send and 1 receive request for communication from/to the R per iq
  MPI_Status stats[2*iqmax];
  
  //*********************
  // Left X ghost cells
  //*********************

  // Determine whether local domain is on global left boundary with more than
  // one processor and periodic BC's, or if the local domain is not on the
  // global left boundary
  if (((niProc > 1)&&(ipro == 0)&&(gg.nleftx == 3)) || (ipro != 0)){                               
    mpiCommLeft = true;
    for (int iq = 0; iq < iqmax; ++iq){
      // Receive/send data from/to processor to the left X
      // Receive data from L
      ierr = MPI_Irecv(&(P[iq][0][0][0]),1,MPI_SUBI,liProID,dataLtoR_tag+2*iq,
                  MPI_COMM_WORLD, &reqsL[2*iq]);
      // Send data to L 
      ierr = MPI_Isend(&(P[iq][0][0][lg.irs]),1,MPI_SUBI,liProID,dataRtoL_tag+2*iq,
                  MPI_COMM_WORLD, &reqsL[2*iq+1]);
    }
  }
  else{ // (ipro == 0) - local domain is on global left X boundary  
    if (gg.nleftx == 0){ // reflecting
      for (int k = lg.krs; k <= lg.kre; ++k){
        for (int j = lg.jrs; j <= lg.jre; ++j){
          for (int i = 0; i < nghost; ++i){
            for (int n = 0; n < iqmax; ++n){
	      P[n][k][j][i] = P[n][k][j][2*nghost-1-i];
  	    }
	    P[iqu0][k][j][i] *= -1.0; // reverse x-velocity
	  }
        }
      }
    }
    else if (gg.nleftx == 1){ // outflow
      for (int k = lg.krs; k <= lg.kre; ++k){
        for (int j = lg.jrs; j <= lg.jre; ++j){
          for (int i = 0; i < nghost; ++i){
            for (int n = 0; n < iqmax; ++n){
    	      P[n][k][j][i] = P[n][k][j][lg.irs];
  	    }
	  }
        }
      }
    }
    else if (gg.nleftx == 2){ // fixed
       for (int k = lg.krs; k <= lg.kre; ++k){
        for (int j = lg.jrs; j <= lg.jre; ++j){
          for (int i = 0; i < nghost; ++i){
            for (int n = 0; n < iqmax; ++n){
	      P[n][k][j][i] = lg.Pbc[n][0][0];
	    }
	  }
        }
      }
    }
    else if (gg.nleftx == 3){ // periodic
      for (int k = lg.krs; k <= lg.kre; ++k){
        for (int j = lg.jrs; j <= lg.jre; ++j){
          for (int i = 0; i < nghost; ++i){
            for (int n = 0; n < iqmax; ++n){
	      P[n][k][j][i] = P[n][k][j][lg.ire+1-nghost+i];
	    }
	  }
        }
      }
    }
    else if (gg.nleftx == 4){ // outflow only
      for (int k = lg.krs; k <= lg.kre; ++k){
        for (int j = lg.jrs; j <= lg.jre; ++j){
          for (int i = 0; i < nghost; ++i){
            for (int n = 0; n < iqmax; ++n){
	      P[n][k][j][i] = P[n][k][j][lg.irs];
	    }
	    if (P[iqu0][k][j][i] > 0.0) P[iqu0][k][j][i] *= -1.0;
	  }
        }
      }
    }
  }
  
  //*********************
  // Right X ghost cells
  //*********************
  // Determine whether local domain is on global right X boundary with
  // more than one processor and periodic BC's, or if it is not on the
  // global right X boundary
  if (((ipro == niProc - 1)&&(niProc > 1)&&(gg.nrightx == 3))|| 
     ((niProc > 1)&&(ipro != niProc-1))){                    
    mpiCommRight = true;
    for (int iq = 0; iq < iqmax; ++iq){
      // Receive/send data from/to processor to the right X      
      // Receive from R
      ierr = MPI_Irecv(&(P[iq][0][0][lg.ire+1]),1,MPI_SUBI,riProID,dataRtoL_tag+2*iq,
                  MPI_COMM_WORLD, &reqsR[2*iq]);
      // Send data to R
      ierr = MPI_Isend(&(P[iq][0][0][lg.ire-nghost+1]),1,MPI_SUBI,riProID,dataLtoR_tag+2*iq,
                  MPI_COMM_WORLD, &reqsR[2*iq+1]);
    }
  }
  else{
    if (gg.nrightx == 0){ // reflecting
      for (int k = lg.krs; k <= lg.kre; ++k){
        for (int j = lg.jrs; j <= lg.jre; ++j){
          for (int i = 0; i < nghost; ++i){
            for (int n = 0; n < iqmax; ++n){
	      P[n][k][j][lg.ire+1+i] = P[n][k][j][lg.ire-i];
	    }
	    P[iqu0][k][j][lg.ire+1+i] *= -1.0; // reverse x-velocity
  	  }
        }
      }
    }
    else if (gg.nrightx == 1){ // outflow
      for (int k = lg.krs; k <= lg.kre; ++k){
        for (int j = lg.jrs; j <= lg.jre; ++j){
          for (int i = 0; i < nghost; ++i){
            for (int n = 0; n < iqmax; ++n){
	      P[n][k][j][lg.ire+1+i] = P[n][k][j][lg.ire];
	    }
	  }
        }
      }
    }
    else if (gg.nrightx == 2){ // fixed
      for (int k = lg.krs; k <= lg.kre; ++k){
        for (int j = lg.jrs; j <= lg.jre; ++j){
          for (int i = 0; i < nghost; ++i){
            for (int n = 0; n < iqmax; ++n){
	      P[n][k][j][lg.ire+1+i] = lg.Pbc[n][0][1];
	    }
	  }
        }
      }
    }
    else if (gg.nrightx == 3){ // periodic
      for (int k = lg.krs; k <= lg.kre; ++k){
        for (int j = lg.jrs; j <= lg.jre; ++j){
          for (int i = 0; i < nghost; ++i){
            for (int n = 0; n < iqmax; ++n){
	      P[n][k][j][lg.ire+1+i] = P[n][k][j][lg.irs+i];
	    }
	  }
        }
      }
    }
    else if (gg.nrightx == 4){ // outflow only
      for (int k = lg.krs; k <= lg.kre; ++k){
        for (int j = lg.jrs; j <= lg.jre; ++j){
          for (int i = 0; i < nghost; ++i){
            for (int n = 0; n < iqmax; ++n){
	      P[n][k][j][lg.ire+1+i] = P[n][k][j][lg.ire];
	    }
	    if (P[iqu0][k][j][lg.ire+1+i] > 0.0) P[iqu0][k][j][lg.ire+1+i] *= -1.0;
	  }
        }
      }
    }
  }

  // Check send and receives have completed
  if (mpiCommLeft){
    // wait for all communication to the L to complete
    ierr = MPI_Waitall(2*iqmax, reqsL, stats);
  }
  if (mpiCommRight){
    // wait for all communication to the R to complete
    ierr = MPI_Waitall(2*iqmax, reqsR, stats);
  }
  //if (procRank == 0) cout << "Finished exchanging all data in i-direction!\n";

      
  if (nd > 1){
    
    //*********************
    // Left Y ghost cells
    //*********************
    
    mpiCommLeft = false;
    mpiCommRight = false;
    // Determine whether the local domain is on the global left Y boundary,
    // there are more than 1 processors along Y, and it is periodic BC's,
    // or if it is not on the global left Y boundary
    if (((njProc > 1)&&(jpro == 0)&&(gg.nlefty == 3)) || (jpro != 0)){                               
      mpiCommLeft = true;
      for (int iq = 0; iq < iqmax; ++iq){
        // Receive from L
        ierr = MPI_Irecv(&(P[iq][0][0][0]),1,MPI_SUBJ,ljProID,dataLtoR_tag+2*iq,
                  MPI_COMM_WORLD, &reqsL[2*iq]);
        // Send data to L 
        ierr = MPI_Isend(&(P[iq][0][lg.jrs][0]),1,MPI_SUBJ,ljProID,dataRtoL_tag+2*iq,
                  MPI_COMM_WORLD, &reqsL[2*iq+1]);
      }
    }
    else{
      if (gg.nlefty == 0){ // reflecting
        for (int k = lg.krs; k <= lg.kre; ++k){
          for (int i = 0; i < lgrid::imaxd; ++i){
            for (int j = 0; j < nghost; ++j){
              for (int n = 0; n < iqmax; ++n){
	        P[n][k][j][i] = P[n][k][2*nghost-1-j][i];
	      }
	      P[iqu0+1][k][j][i] *= -1.0; // reverse y-velocity
	    }
          }
        }
      }
      else if (gg.nlefty == 1){ // outflow
        for (int k = lg.krs; k <= lg.kre; ++k){
          for (int i = 0; i < lgrid::imaxd; ++i){
            for (int j = 0; j < nghost; ++j){
              for (int n = 0; n < iqmax; ++n){
	        P[n][k][j][i] = P[n][k][lg.jrs][i];
	      }
	    }
          }
        }
      }
      else if (gg.nlefty == 2){ // fixed
        for (int k = lg.krs; k <= lg.kre; ++k){
          for (int i = 0; i < lgrid::imaxd; ++i){
            for (int j = 0; j < nghost; ++j){
              for (int n = 0; n < iqmax; ++n){
	        P[n][k][j][i] = lg.Pbc[n][1][0];
	      }
	    }
          }
        }
      }
      else if (gg.nlefty == 3){ // periodic
        for (int k = lg.krs; k <= lg.kre; ++k){
          for (int i = 0; i < lgrid::imaxd; ++i){
            for (int j = 0; j < nghost; ++j){
              for (int n = 0; n < iqmax; ++n){
	        P[n][k][j][i] = P[n][k][lg.jre+1-nghost+j][i];
	      }
	    }
          }
        }
      }
      else if (gg.nlefty == 4){ // outflow only
        for (int k = lg.krs; k <= lg.kre; ++k){
          for (int i = 0; i < lgrid::imaxd; ++i){
            for (int j = 0; j < nghost; ++j){
              for (int n = 0; n < iqmax; ++n){
	        P[n][k][j][i] = P[n][k][lg.jrs][i];
	      }
	      if (P[iqu0+1][k][j][i] > 0.0) P[iqu0+1][k][j][i] *= -1.0;
	    }
          }
        }
      }
    }

    //*********************
    // Right Y ghost cells
    //*********************
    // Determine whether the local domain is on the global right Y boundary,
    // there is more than 1 processor along y, and periodic BC's, or if it is
    // not on the global right Y boundary
    if (((njProc > 1)&&(jpro == njProc - 1)&&(gg.nrighty == 3))|| 
       ((njProc > 1)&&(jpro != njProc-1))){                    
      mpiCommRight = true;
      for (int iq = 0; iq < iqmax; ++iq){
        // Receive from R
        ierr = MPI_Irecv(&(P[iq][0][lg.jre+1][0]),1,MPI_SUBJ,rjProID,dataRtoL_tag+2*iq,
                  MPI_COMM_WORLD, &reqsR[2*iq]);
        // Send data to R
        ierr = MPI_Isend(&(P[iq][0][lg.jre-nghost+1][0]),1,MPI_SUBJ,rjProID,dataLtoR_tag+2*iq,
                  MPI_COMM_WORLD, &reqsR[2*iq+1]);
      }
    }
    else{
      if (gg.nrighty == 0){ // reflecting
        for (int k = lg.krs; k <= lg.kre; ++k){
          for (int i = 0; i < lgrid::imaxd; ++i){
            for (int j = 0; j < nghost; ++j){
              for (int n = 0; n < iqmax; ++n){
	        P[n][k][lg.jre+1+j][i] = P[n][k][lg.jre-j][i];
	      }
	      P[iqu0+1][k][lg.jre+1+j][i] *= -1.0; // reverse y-velocity
	    }
          }
        }
      }
      else if (gg.nrighty == 1){ // outflow
        for (int k = lg.krs; k <= lg.kre; ++k){
          for (int i = 0; i < lgrid::imaxd; ++i){
            for (int j = 0; j < nghost; ++j){
              for (int n = 0; n < iqmax; ++n){
	        P[n][k][lg.jre+1+j][i] = P[n][k][lg.jre][i];
	      }
	    }
          }
        }
      }
      else if (gg.nrighty == 2){ // fixed
        for (int k = lg.krs; k <= lg.kre; ++k){
          for (int i = 0; i < lgrid::imaxd; ++i){
            for (int j = 0; j < nghost; ++j){
              for (int n = 0; n < iqmax; ++n){
	        P[n][k][lg.jre+1+j][i] = lg.Pbc[n][1][1];
	      }
	    }
          }
        }
      }
      else if (gg.nrighty == 3){ // periodic
        for (int k = lg.krs; k <= lg.kre; ++k){
          for (int i = 0; i < lgrid::imaxd; ++i){
            for (int j = 0; j < nghost; ++j){
              for (int n = 0; n < iqmax; ++n){
	        P[n][k][lg.jre+1+j][i] = P[n][k][lg.jrs+j][i];
	      }
	    }
          }
        }
      }
      else if (gg.nrighty == 4){ // outflow only
        for (int k = lg.krs; k <= lg.kre; ++k){
          for (int i = 0; i < lgrid::imaxd; ++i){
            for (int j = 0; j < nghost; ++j){
              for (int n = 0; n < iqmax; ++n){
	        P[n][k][lg.jre+1+j][i] = P[n][k][lg.jre][i];
	      }
	      if (P[iqu0+1][k][lg.jre+1+j][i] > 0.0) P[iqu0+1][k][lg.jre+1+j][i] *= -1.0;
	    }
          }
        }
      }
    }
    
    // Check if send and receives have completed
    if (mpiCommLeft){
      // wait for all communication to the L to complete
      ierr = MPI_Waitall(2*iqmax, reqsL, stats);
    }
    if (mpiCommRight){
      // wait for all communication to the R to complete
      ierr = MPI_Waitall(2*iqmax, reqsR, stats);
    }
    //if (procRank == 0) cout << "Finished exchanging all data in j-direction!\n";
    
  } // nd > 1

  if (nd == 3){
    
    //*********************
    // Left Z ghost cells
    //*********************

    mpiCommLeft = false;
    mpiCommRight = false;
    // Determine whether the local domain is on the global left Z boundary,
    // there are more than 1 processors along z, and periodic BC's, or if the
    // local domain is not on global left Z boundary
    if (((nkProc > 1)&&(kpro == 0)&&(gg.nleftz == 3))|| (kpro != 0)){
      mpiCommLeft = true;
      for (int iq = 0; iq < iqmax; ++iq){
        // Receive from L
        ierr = MPI_Irecv(&(P[iq][0][0][0]),1,MPI_SUBK,lkProID,dataLtoR_tag+2*iq,
                  MPI_COMM_WORLD, &reqsL[2*iq]);
        // Send data to L 
        ierr = MPI_Isend(&(P[iq][lg.krs][0][0]),1,MPI_SUBK,lkProID,dataRtoL_tag+2*iq,
                  MPI_COMM_WORLD, &reqsL[2*iq+1]);
      }
    }
    else{
      if (gg.nleftz == 0){ // reflecting
        for (int j = 0; j < lgrid::jmaxd; ++j){
          for (int i = 0; i < lgrid::imaxd; ++i){
            for (int k = 0; k < nghost; ++k){
              for (int n = 0; n < iqmax; ++n){
	        P[n][k][j][i] = P[n][2*nghost-1-k][j][i];
	      }
	      P[iqu0+2][k][j][i] *= -1.0; // reverse z-velocity
	    }
          }
        }
      }
      else if (gg.nleftz == 1){ // outflow
        for (int j = 0; j < lgrid::jmaxd; ++j){
          for (int i = 0; i < lgrid::imaxd; ++i){
            for (int k = 0; k < nghost; ++k){
              for (int n = 0; n < iqmax; ++n){
	        P[n][k][j][i] = P[n][lg.krs][j][i];
	      }
	    }
          }
        }
      }
      else if (gg.nleftz == 2){ // fixed
        for (int j = 0; j < lgrid::jmaxd; ++j){
          for (int i = 0; i < lgrid::imaxd; ++i){
            for (int k = 0; k < nghost; ++k){
              for (int n = 0; n < iqmax; ++n){
	        P[n][k][j][i] = lg.Pbc[n][2][0];
	      }
	    }
          }
        }
      }
      else if (gg.nleftz == 3){ // periodic
        for (int j = 0; j < lgrid::jmaxd; ++j){
          for (int i = 0; i < lgrid::imaxd; ++i){
            for (int k = 0; k < nghost; ++k){
              for (int n = 0; n < iqmax; ++n){
	        P[n][k][j][i] = P[n][lg.kre+1-nghost+k][j][i];
	      }
	    }
          }
        }
      }
      else if (gg.nleftz == 4){ // outflow only
        for (int j = 0; j < lgrid::jmaxd; ++j){
          for (int i = 0; i < lgrid::imaxd; ++i){
            for (int k = 0; k < nghost; ++k){
              for (int n = 0; n < iqmax; ++n){
	        P[n][k][j][i] = P[n][lg.krs][j][i];
	      }
	      if (P[iqu0+2][k][j][i] > 0.0) P[iqu0+2][k][j][i] *= -1.0;
	    }
          }
        }
      }
    }

    //*********************
    // Right Z ghost cells
    //*********************

    // Determine whether the local domain is on the global right Z boundary,
    // there are more than 1 processors along z, and reflecting BC's, or
    // if it is not on global right Z boundary
    if (((nkProc > 1)&&(kpro == nkProc - 1)&&(gg.nrightz == 3))||
       ((nkProc > 1)&&(kpro != nkProc-1))){                    
      mpiCommRight = true;
      for (int iq = 0; iq < iqmax; ++iq){
        // Receive from R
        ierr = MPI_Irecv(&(P[iq][lg.kre+1][0][0]),1,MPI_SUBK,rkProID,dataRtoL_tag+2*iq,
                  MPI_COMM_WORLD, &reqsR[2*iq]);
        // Send data to R
        ierr = MPI_Isend(&(P[iq][lg.kre-nghost+1][0][0]),1,MPI_SUBK,rkProID,dataLtoR_tag+2*iq,
                  MPI_COMM_WORLD, &reqsR[2*iq+1]);
      }
    }
    else{      
      // Rightz ghost cells
      if (gg.nrightz == 0){ // reflecting
        for (int j = 0; j < lgrid::jmaxd; ++j){
          for (int i = 0; i < lgrid::imaxd; ++i){
            for (int k = 0; k < nghost; ++k){
              for (int n = 0; n < iqmax; ++n){
	        P[n][lg.kre+1+k][j][i] = P[n][lg.kre-k][j][i];
	      }
	      P[iqu0+2][lg.kre+1+k][j][i] *= -1.0; // reverse z-velocity
	    }
          }
        }
      }
      else if (gg.nrightz == 1){ // outflow
        for (int j = 0; j < lgrid::jmaxd; ++j){
          for (int i = 0; i < lgrid::imaxd; ++i){
            for (int k = 0; k < nghost; ++k){
              for (int n = 0; n < iqmax; ++n){
	        P[n][lg.kre+1+k][j][i] = P[n][lg.kre][j][i];
	      }
	    }
          }
        }
      }
      else if (gg.nrightz == 2){ // fixed
        for (int j = 0; j < lgrid::jmaxd; ++j){
          for (int i = 0; i < lgrid::imaxd; ++i){
            for (int k = 0; k < nghost; ++k){
              for (int n = 0; n < iqmax; ++n){
	        P[n][lg.kre+1+k][j][i] = lg.Pbc[n][2][1];
	      }
	    }
          }
        }
      }
      else if (gg.nrightz == 3){ // periodic
        for (int j = 0; j < lgrid::jmaxd; ++j){
          for (int i = 0; i < lgrid::imaxd; ++i){
            for (int k = 0; k < nghost; ++k){
              for (int n = 0; n < iqmax; ++n){
	        P[n][lg.kre+1+k][j][i] = P[n][lg.krs+k][j][i];
	      }
    	    }
          }
        }
      }
      else if (gg.nrightz == 4){ // outflow only
        for (int j = 0; j < lgrid::jmaxd; ++j){
          for (int i = 0; i < lgrid::imaxd; ++i){
            for (int k = 0; k < nghost; ++k){
              for (int n = 0; n < iqmax; ++n){
	        P[n][lg.kre+1+k][j][i] = P[n][lg.kre][j][i];
	      }
	      if (P[iqu0+2][lg.kre+1+k][j][i] > 0.0) P[iqu0+2][lg.kre+1+k][j][i] *= -1.0;
	    }
          }
        }
      }
    }
    
    // Check send and receives have completed
    if (mpiCommLeft){
      // wait for all communication to the L to complete
      ierr = MPI_Waitall(2*iqmax, reqsL, stats);
    }
    if (mpiCommRight){
      // wait for all communication to the R to complete
      ierr = MPI_Waitall(2*iqmax, reqsR, stats);
    }
    //if (procRank == 0) cout << "Finished exchanging all data in k-direction!\n";
   
  } // nd == 3
  
  return;
}


/*!  \brief Given left and right states (pL,pR), calculate the time-averaged flux.
 *
 *   Uses Sam's method in mg_g (i.e. in the hflux routine).
 *
 *   Note that direc=0 for the 1st axis (i.e. "X" in "XYZ" geometry), direc=1 for the 2nd axis, etc.
 *   This means that direc is different from the value passed to hflux in mg.
 *
 *   \todo Write a fixup routine more like those in mg_g.
 *
 *   \author Julian Pittard (Original version 18.03.13)
 *   \version 1.0-stable (Evenstar):
 *   \date Last modified: 21.03.13 (JMP)
 */
void riemann(int direc, double pL[iqmax], double pR[iqmax], double flux[iqmax]){

  double linear = 0.1;
  bool nlrieman = false;
  bool cwrieman = false;
  double rs[3];
  double vel[3];
  double cl,cr,ul,dl,pl,ur,dr,pr,cel,cer,p,u,d;
  double mflux,ud,ev,cvis,vis,tv;
    
  cl = 1.0; // defaults
  cr = 1.0;
    
  // Zero fluxes (safety)
  for (int n = 0; n < iqmax; ++n)
    flux[n] = 0.0;

  // Check that the values in pL and pR are sensible using problem-specific checks.
  pL[iqd] = max(pL[iqd],dfloor);
  pR[iqd] = max(pR[iqd],dfloor);
  pL[iqe] = pminFixUp(avgmassamb,tamb,pL[iqd],pL[iqe]);
  pR[iqe] = pminFixUp(avgmassamb,tamb,pR[iqd],pR[iqe]);

  int iqu = iqu0 + direc;

  if (cwrieman){
    // Use the Collea & Woodward (1984) 2-shock riemann solver
    //riemannCW84(direc,pL,pR,d,u,p,vel)
    cout << "Not converted yet!\n";
    quit();
  }
  else{    
    // Use the riemann solver in mg
    // Left state
    ul = pL[iqu]; // normal velocity
    dl = pL[iqd];
    pl = pL[iqe];

    // Right state
    ur = pR[iqu]; // normal velocity
    dr = pR[iqd];
    pr = pR[iqe];

    // Calculate sound speeds
    cel = sqrt(gam*pl/dl);
    cl = dl*cel;
    cer = sqrt(gam*pr/dr);
    cr = dr*cer;

    // Linear Riemann solver
    p = (cr*pl + cl*pr - cr*cl*(ur - ul))/(cr + cl);
    u = (cr*ur + cl*ul - (pr - pl))/(cr + cl);
    if ((abs(p - pl) > linear*pl) || (abs(p - pr) > linear*pr)){
      // Use non-linear Riemann solver
      nlrieman = true;
      rs[0] = 1.0; // dummy argument (d)
      rs[1] = u;
      rs[2] = p;
      riemannNL(dl, ul, pl, cl, dr, ur, pr, cr, rs);
      d = rs[0];
      u = rs[1];
      p = rs[2];
    }
     
    if (u >= 0.0){
      if (ul - cel > 0.0){
	d = dl;
        u = ul;
	p = pl;
      }
      else{
	d = cl*dl/(cl + (u - ul)*dl);
      }
      for (int n = 0; n < nd; ++n){
        vel[n] = pL[iqu0+n];
      }
    }
    else{
      if (ur + cer <= 0.0){
	d = dr;
	u = ur;
	p = pr;
      }
      else{
	d = cr*dr/(cr - (u - ur)*dr);
      }
      for (int n = 0; n < nd; ++n){
        vel[n] = pR[iqu0+n];
      }
    }
    vel[direc] = u;
  }
  
  // Make sure that the time-averaged fluid variables at the cell interface
  // (d, u and p) have sensible values. The pressure in particular is problem-specific
  // and so we load the interface values into a primitive variable for checking
  // by the routine pminFixUp.
  d = fmax(d,dfloor);
  p = pminFixUp(avgmassamb,tamb,d,p);

  // Mass flux and energy flux
  mflux = d*u;
  flux[iqd] = mflux;
  flux[iqe] = p;
  // Add velocity bits into momentum fluxes and energy
  for (int n = 0; n < nd; ++n){
    ud = vel[n];
    flux[iqu0+n] = mflux*ud;
    flux[iqe] += 0.5*d*ud*ud;
  }
  // Calculate internal energy
  ev = p/gamm;
  // Add this into energy flux, add pressure into normal momentum
  flux[iqe] = u*(flux[iqe] + ev);
  flux[iqu] += p;
  // Scalars
  if (u > 0.0){
    for (int n = 0; n < smax; ++n){
      flux[iqal0+n] = mflux*pL[iqal0+n];
    }
  }
  else{
    for (int n = 0; n < smax; ++n){
      flux[iqal0+n] = mflux*pR[iqal0+n];
    }
  }

  // Artificial dissipation
  cvis = fmax(cl,cr);
  vis = avisp*cvis;
  for (int n = 0; n < nd; ++n){
    tv = vis*(pL[iqu0+n] - pR[iqu0+n]);
    flux[iqu0+n] += tv;
    flux[iqe] += tv*vel[n];
  }
  flux[iqe] += avise*cvis*(pL[iqe]/pL[iqd] - pR[iqe]/pR[iqd]);
 
  //cout << "pL(dup) = " << pL[iqd] << " " << pL[iqu0] << " " << pL[iqe] << " " << "; pR(dup) = " << " " << pR[iqd] << " " << pR[iqu0] << " " << pR[iqe] << " " << "; flux(dup) = " << " " << flux[iqd] << " " << flux[iqu0] << " " << flux[iqe] << "\n";
  //quit();

  return;

}


/*!  \brief Non-linear Riemann solver.
 *
 *   Calculates the resolved state which has dl,ul,pl on the left and dr,ur,pr
 *   on the right.
 *
 *   \author Julian Pittard (Original version 18.03.13)
 *   \version 1.0-stable (Evenstar):
 *   \date Last modified: 21.03.13 (JMP)
 */
void riemannNL(double dl, double ul, double pl, double cl, double dr, double ur,
  double pr, double cr, double rs[3]){

  int nit = 0;
  double wl, wr, c, ci, p0;
  const double tol=1.0e-4;
  bool doit = true;
  double d = rs[0];
  double u = rs[1];
  double p = rs[2];
  
  if ((p < pl) && (p < pr)){
    // Non-iterative solution for two rarefactions
    ci = cr/(dr*(double)pow((double)pr, (double)g4));
    p = (0.5*g1*(ul - ur) + cr/dr + cl/dl)/(ci + cl/
      (dl*(double)pow((double)pl, (double)g4)));
    if (p < pfloor)
      p = pfloor;
    else
      p = (double)pow((double)p, (double)g5);
  }
  else{
    // Iterative solution for shock/rarefaction
    p0 = p;
    wl = -cl*wave(p,pl);
    wr = cr*wave(p,pr);
    // Iteration
    while (doit){
      p = (wr*pl - wl*pr + wr*wl*(ur - ul))/(wr - wl);
      if ((mabs(p - p0) > tol*p) && (p > 0.0)){
	p0 = p;
	wl = -cl*wave(p,pl);
	wr = cr*wave(p,pr);
	nit++;
	if (nit > 50){
	  cout << "convergence failure in riemann solver" << '\n';
	  // Use linear guess
	  p = (cr*pl + cl*pr - cr*cl*(ur - ul))/(cr + cl);
	  p = max(p, pfloor);
	  wl = -cl;
	  wr = cr;
	  doit = false;
	}
	else if (nit > 20)
	  cout << "slow convergence in riemann solver nit " <<
	    nit << '\n';
      }
      else
	doit = false;
    }
    // Calculate resolved density and velocity
    u = (wr*ur - wl*ul - pr + pl)/(wr -wl);
    // Check for position of contact
    if (u > 0.0){
      // Contact is on right of interface
      d = wl*dl/(wl - (u -ul)*dl);
      // Check velocity of left waves
      if (p < pl){
	// Left wave is rarefaction
	c = sqrt(gam*p/d);
	if ((u - c) > 0.0){
	  // Tail is on right of interface
	  cl /= dl;
	  if ((ul - cl) < 0.0){
	    // Head is on left of interface -- spans interface 
	    u = g8*(ul +g7*cl);
	    p = pl*pow((double)(u/cl), (double)g5);
	    d = gam*p/(u*u);
	  }
	  else{
	    // Rarefaction does not span interface
	    d = dl;
	    u = ul;
	    p = pl;
	  }
	}
      }
      else{
	// Left wave is shock
	if ((wl/dl+ul) > 0.0){
	  // On right of interface
	  d = dl;
	  u = ul;
	  p = pl;
	}
      }
    }
    else{
      //     Contact is on left of interface 
      d = wr*dr/(wr - (u - ur)*dr);
      // Check velocity of right waves
      if (p < pr){
	// Right wave is rarefaction 
	c = sqrt(gam*p/d);
	if ((u + c) < 0.0){
	  // Tail is on left of interface
	  cr /= dr;
	  if ((ur + cr) > 0.0){
	    // Head is on right of interface -- spans interface
	    u = g8*(ur - g7*cr);
	    p = pr*pow((double)(-u/cr), (double)g5);
	    d = gam*p/(u*u);
	  }
	  else{
	    // Rarefaction does not span interface
	    d = dr;
	    u = ur;
	    p = pr;
	  }
	}
      }
      else{
	// Right wave is shock 
	if ((wr/dr+ur) < 0.0){
	  // On left of interface 
	  d = dr;
	  u = ur;
	  p = pr;
	}
      }
    }
  }

  rs[0] = d;
  rs[1] = u;
  rs[2] = p;
  
  return;

}


/*!  \brief Calculate the nonlinear wave speed. 
 *
 *   This function calculates the wave speed for a wave connecting
 *   states with pressures p0 and p ahead of and behind respectively.
 *   From Sam Falle's wave function.
 *
 *   \author Julian Pittard (Original version 21.03.13)
 *   \version 1.0-stable (Evenstar):
 *   \date Last modified: 22.03.13 (JMP)
 */
double wave(double p, double p0){

  double x, w;
  x = p/p0;
  if (mabs(x - 1.0) < 1.0e-03)
    // Use linear expression
    w = 1.0 + 0.5*g3*(x - 1.0);
  else{
    // Use non-linear expression
    if (x >= 1.0)
      // Shock
      w = sqrt(1.0 + g3*(x - 1.0));
    else
      // Rarefaction
      w = g4*(1.0 - x)/(1.0 - pow((double)x, (double)g4));
  }

  return w;

}


/*!  \brief Calculate the minimum pressure, given the density. 
 *
 *   This is problem specific.
 *
 *   \author Julian Pittard (Original version 21.03.13)
 *   \version 1.0-stable (Evenstar):
 *   \date Last modified: 22.03.13 (JMP)
 */
double pminFixUp(double avgmass, double tamb, double r, double p){
  double const1 = boltzman*tamb/avgmass;
  double pmin;
    
  if (isnan(p) || isinf(p)) p = pfloor;
  if (isnan(r) || isinf(r)) r = dfloor;
        
  if ((problem == "WBB") || (problem == "SNR") || (problem == "CWB")){
    pmin = max(r*const1,pfloor);
    p = max(p,pmin);
  }
  else{
    p = max(p,pfloor);
  }
  return p;
}


