/*!  \file hydroSplit.cpp
 *   \brief Perform the directionally split hydro update.
 *
 *   \author Julian Pittard (Original version 01.08.11)
 *   \version 0.1-development (Undómiel):
 *   \date Last modified: 08.08.11 (JMP)
 */


#undef MAIN

//  Header files:
#include <fstream>
#include <iomanip>
#include <iostream>
#include <cmath>
#include "defs.h"
#include "constants.h"
#include "gas.h"
#include "global.h"     
#include "grid.h"        
#include "physconst.h"
#include "star.h"

// Function declarations
void evolve(int n1, int n2, int ngeom);
void fictForces(double xf[], double uf[], double vf[], double wf[], double fict[]);
void flatten();
void pminFixUpArray(int nmin, int nmax);
void ppm(int n1, int n2, int nleft, int nright, int ngeom, double para[][lgrid::maxsweep] );
void remap( int nleft, int nright, int ngeom);
void riemannCW84(int lmin, int lmax);
void states(int n1, int n2);
void sweepx();
void sweepy();
void sweepz();
void updateSplit();
void volume ( int ngeom);
//extern void fillGhostCells(double P[][lgrid::kmaxd][lgrid::jmaxd][lgrid::imaxd]);
extern void fillGhostCells(double**** P);
extern void localGridSetup();
extern void operatorSplit();
extern void parabola(int nmin, int nmax, double para[][lgrid::maxsweep], double a[], double deltaa[], double a6[], double al[], double flat[], double jump[], int iflat, int iskip, int icd);
extern void paraset(int nmin, int nmax, double zpara[][lgrid::maxsweep], double dx[], double xa[]);
extern double tmpmu(double temp);
extern void quit(); 
extern void radiateExact();
extern void restrictCool(double dei[lgrid::kmaxd][lgrid::jmaxd][lgrid::imaxd]);

using namespace std;

// File scope variables
  string sweep;
  double r[lgrid::maxsweep];         // 1D vector of extracted density
  double p[lgrid::maxsweep];         //                        pressure
  double u[lgrid::maxsweep];         //                        x-velocity
  double v[lgrid::maxsweep];         //                        y-velocity
  double w[lgrid::maxsweep];         //                        z-velocity
  double e[lgrid::maxsweep];         // 1D vector of internal energy density
  double as[smax][lgrid::maxsweep]; // 2D Matrix of extracted advected scalars

  double act[lgrid::maxsweep];       // 1D vector of actual forces     

  double xa [lgrid::maxsweep];       // 1D vector of grid coords
  double dx [lgrid::maxsweep];       //                   cell widths
  double xa0[lgrid::maxsweep];       // 1D vector of grid coords
  double dx0[lgrid::maxsweep];       //                   cell widths
  double dvol[lgrid::maxsweep];      // 1D vector of grid cell volumes
  double dvol0[lgrid::maxsweep];     //                   

  int shck[lgrid::maxsweep];
  double dr[lgrid::maxsweep], du[lgrid::maxsweep], dp[lgrid::maxsweep],
          dv  [lgrid::maxsweep], dw  [lgrid::maxsweep], 
          r6  [lgrid::maxsweep], u6  [lgrid::maxsweep], p6  [lgrid::maxsweep],
          v6  [lgrid::maxsweep], w6  [lgrid::maxsweep],  
          rl  [lgrid::maxsweep], ul  [lgrid::maxsweep], pl  [lgrid::maxsweep], 
          vl  [lgrid::maxsweep], wl  [lgrid::maxsweep], 
          rrgh[lgrid::maxsweep], urgh[lgrid::maxsweep], prgh[lgrid::maxsweep],
          vrgh[lgrid::maxsweep], wrgh[lgrid::maxsweep], 
          rlft[lgrid::maxsweep], ulft[lgrid::maxsweep], plft[lgrid::maxsweep], 
          vlft[lgrid::maxsweep], wlft[lgrid::maxsweep], 
          umid[lgrid::maxsweep], pmid[lgrid::maxsweep], rmid[lgrid::maxsweep],
          vmid[lgrid::maxsweep], wmid[lgrid::maxsweep], 
          flat[lgrid::maxsweep], jump[lgrid::maxsweep], 
    das_1d[lgrid::maxsweep], as6_1d[lgrid::maxsweep], asl_1d[lgrid::maxsweep]; 
  double das[smax][lgrid::maxsweep],as6[smax][lgrid::maxsweep],asl[smax][lgrid::maxsweep],
    asrgh[smax][lgrid::maxsweep],aslft[smax][lgrid::maxsweep],asmid[smax][lgrid::maxsweep];


/*!  \brief Advance the solution using a directionally split method.
 *
 *   1D hydro sweeps are performed in alternating directions.
 * 
 *   \warning Assume forces are constant over 2 or 3 hydro steps 
 *
 *   \author Julian Pittard (Original version 02.06.11)
 *   \version 0.1-development (Undómiel):
 *   \date Last modified: 09.08.11 (JMP)
 */
void updateSplit(){

  // The number of same-direction sweeps per cycle is equal to nd.
  // The timestep must be unchanged during this.
  
  // 1D sweeps
  fillGhostCells(lg.P0);
              sweepx();
  if (nd > 1) sweepy();
  if (nd > 2) sweepz();

  operatorSplit();

  t       += dt;
  timep1d += dt;
  timep2d += dt;
  timep3d += dt;

  // For 2D/3D calcs continue 1D sweeps different order...
  if (nd > 1){
    fillGhostCells(lg.P0);
    // 1D sweeps different order
    if (nd > 1) sweepy();
    if (nd > 2) sweepz();
                sweepx();

    operatorSplit();

    t       += dt;
    timep1d += dt;
    timep2d += dt;
    timep3d += dt;
  }
  
  // For 3D calcs continue 1D sweeps different order...
  if (nd == 3){
    fillGhostCells(lg.P0);
    // 1D sweeps different order
    if (nd > 2) sweepz();
                sweepx();
    if (nd > 1) sweepy();

    operatorSplit();

    t       += dt;
    timep1d += dt;
    timep2d += dt;
    timep3d += dt;
  }

  return;
}


/*!  \brief Function to perform the hydro update
 *
 * Compute volume elements, flatten (smooth)
 * the flow near shocks, calculate the input states to the Riemann solver,
 * obtain the zone face fluxes, and evolve the flow. Remap to a 
 * fixed or evolving grid if necessary
 *
 */
void hydroUpdate(int n1, int n2, int nleft, int nright, int ngeom, int ntot, double para[][lgrid::maxsweep] ){

  // Compute cell volumes
  volume ( ngeom);

  // Flatten (smooth) the flow near shocks 
  flatten();

  // Calculate the input states to the Riemann solver
  ppm( n1, n2, nleft, nright, ngeom, para );

  // Call the Riemann solver to obtain the zone face averages. For the
  // lagrangian solver we only require umid and pmid, while for the Eulerian
  // solver we require in addition rmid, vmid, wmid, and asmid

  riemannCW84( lg.nmin-1, lg.nmax+2);      // OK for Lagrangian and Eulerian

  // Do lagrangian update using umid and pmid, or Eulerian update using
  // umid, rmid, pmid, vmid, wmid and asmid. 
  evolve( n1, n2, ngeom);

  // Remap to a fixed or expanding grid if necessary
  //if (lagrangian)
  remap( nleft, nright, ngeom);

  return;
}


/*!  \brief Performs 1D sweeps in the x-direction.
 *
 *   For an expanding grid simply calculate a new xa0 and dx0 for the expansion.
 *   The parabolic coefficients in zparax should be calculated  
 *   from the original grid before the expansion, and updated
 *   each time step. However, if we are scaling a uniform grid, 
 *   the coeffs. remain constant, in which case there is no need
 *   to recompute...
 *
 *   \author Julian Pittard (Original version 02.06.11)
 *   \version 0.1-development (Undómiel):
 *   \date Last modified: 01.08.11 (JMP)
 */
void sweepx(){
 
  //Locals
  static bool first = true;
  double u2,v2,w2;

  sweep = "x";

  // Set global ranges for x-sweep
  lg.nmin = lg.irs;                    // first real cell
  lg.nmax = lg.ire;                    // last real cell

  const bool expand = true;
  const double tpower = 0.6;
  bool expanding = false;
  
  // Expand the grid for each i. dr0 is the initial value and is not updated!
  // NOTE: This only works for single processor runs where lg and gg are identical....
  double rmax_ini = gg.xmax0;
  double edot = 0.5*stars[0].mdot*stars[0].vinf*stars[0].vinf;

  // Only calculate the start time for the grid expansion once (since remapRadius changes once the expansion starts)
  static double tstart;
  if (first){
    //tstart = 0.8*pow(rmax_ini,5.0/3.0)*1.57*pow(ramb/edot,1.0/3.0); // based on WBB adiabatic expansion
    tstart = (rmax_ini - remapRadius)/stars[0].vinf; // based on freely expanding wind
    //cout << "tstart = " << tstart << "\n"; quit();
    first = false;
  }
  
  if (expand && t > tstart) expanding = true;
  if (expanding){
    double rmax = rmax_ini*pow(t/tstart,0.6); // new size of grid (R_FS \propto t^0.6, R_RS \propto t^0.4)
    gg.xmax = rmax;
    lg.xmax = rmax;
    double dr = gg.xmax/gg.imax;
    remapRadius = float(remapi)*dr;
    //cout << setprecision(9) << "dr = " << dr << "; dr0 = " << lg.dr0 << "\n"; quit();
    for (int i = 0; i < lgrid::imaxd; i++){
      int n = i - nghost;
      dx0[i] = dr;
      xa0[i] = float(n)*dr;
      //cout << "i = " << i << "; nghost = " << nghost << "; dr = " << dr << "; xa0 = " << xa0[i] << "\n";
      // xa0(n) = xa0(n) + 0.4*sin(float(ncycle)/1.0)*dr; ! wiggle the grid
    }
    //quit();
  }
  
  // Loop over each row...
  for (int k = lg.krs; k <= lg.kre; k++){
    for (int j = lg.jrs; j <= lg.jre; j++){
      // Put state variables into 1D arrays, including ghost zones.
      // Also extract the required component of the effective force(s)
      // (gravity plus radiative continuum) and the radiative line force(s). 
      for (int i = 0; i < lgrid::imaxd; i++){
	r  [i] = lg.P0[iqd][k][j][i];
	p  [i] = lg.P0[iqe][k][j][i];
	u  [i] = lg.P0[iqu0][k][j][i];
#if NDIM > 1
	v  [i] = lg.P0[iqu0+1][k][j][i];
#else
	v  [i] = 0.0;
#endif
#if NDIM == 3
	w  [i] = lg.P0[iqu0+2][k][j][i];
#else
	w  [i] = 0.0;
#endif
        for (int m = 0; m < smax; m++){
  	  as[m][i] = lg.P0[iqal0+m][k][j][i];  
        }
        if (!expanding){
          xa0[i] = lg.zxa[i];
          dx0[i] = lg.zdx[i];
	}
        xa [i] = lg.zxa[i];
        dx [i] = lg.zdx[i];
        p  [i] = max(pfloor,p[i]);
	u2 = u[i]*u[i];
	v2 = v[i]*v[i];
	w2 = w[i]*w[i]; 
	e  [i] = p[i]/(r[i]*gamm)+0.5*(u2 + v2 + w2);
	/*
        if (force){
          if (rank == 0) cout << "Not coded for force in sweepx yet\n";
          quit();
          //act[i] = fx[k][j][i];
        }
        else
	*/
        act[i] = 0.0;
        //printf("%3d %15.6e %15.6e %15.6e\n",n,p[i],r[i],u[i]);
      }

      // Perform the hydro update
      hydroUpdate(j, k, gg.nleftx, gg.nrightx, gg.ngeomx, lgrid::maxsweep, lg.zparax);

      // Put updated values in real cells back into 3D arrays, dropping ghost zones

      for (int i = lg.irs; i <= lg.ire; i++){
	lg.P0[iqd][k][j][i] = r[i];
	lg.P0[iqe][k][j][i] = p[i];
	lg.P0[iqu0][k][j][i] = u[i];
#if NDIM > 1	
	lg.P0[iqu0+1][k][j][i] = v[i];
#endif
#if NDIM == 3	
	lg.P0[iqu0+2][k][j][i] = w[i];
#endif
        for (int m = 0; m < smax; m++){
  	  lg.P0[iqal0+m][k][j][i] = as[m][i];
        }
      }

    }  // j loop
  }    // k loop

  
  // Store the new coords and compute new parabolic coeffs
  if (expanding){
    //localGridSetup();  
    for (int i = 0; i < lgrid::imaxd; i++){
      lg.zxa[i] = xa0[i];
      lg.zdx[i] = dx0[i];
      lg.zxc[i] = lg.zxa[i] + 0.5*lg.zdx[i];
      double x1 = lg.zxa[i];
      double x2 = lg.zxa[i] + lg.zdx[i];
      lg.zxg[i] = 0.75*(x2*x2*x2 + x1*x2*(x2 + x1) + x1*x1*x1)/(x2*x2 + x1*x2 + x1*x1); // RTP spherically symmetric

    }
    paraset(nghost,lgrid::imax-1+nghost,lg.zparax,lg.zdx,lg.zxa);
  }
  
  
  return;
}


/*!  \brief Performs 1D sweeps in the y-direction.
 *
 *   For an expanding grid simply calculate a new xa0 and dx0 for the expansion.
 *   The parabolic coefficients in zparay should be calculated  
 *   from the original grid before the expansion, and updated
 *   each time step. However, if we are scaling a uniform grid, 
 *   the coeffs. remain constant, in which case there is no need
 *   to recompute...
 *
 *   \author Julian Pittard (Original version 02.06.11)
 *   \version 0.1-development (Undómiel):
 *   \date Last modified: 02.06.11 (JMP)
 */
void sweepy(){

  sweep = "y";
  
  // Set global ranges for y-sweep
  lg.nmin = lg.jrs;                    // first real cell
  lg.nmax = lg.jre;                    // last real cell

  // Expand the grid for each j. dr0 and dsep0 are the initial values and are 
  // not updated!
  /*
  if (expand){
    dr = dr0*pow((dsep/dsep0),expande);
    for (j=0; j<jmax; j++){
      n = j + nghost;
      dx0[n] = dr;
      xa0[n] = float(j-jpivot)*dx0[n];
      // xa0[n] = xa0[n] + 0.4*sin(float(ncycle)/1.0)*dr; ! wiggle the grid
    }
  }
  */
  
  // Loop over each row...
  for (int k = lg.krs; k <= lg.kre; k++){
    for (int i = lg.irs; i <= lg.ire; i++){
      lg.radius = lg.zxc[i];

      // Put state variables into 1D arrays, including ghost zones.
      // Also extract the required component of the effective force(s)
      // (gravity plus radiative continuum) and the radiative line force(s). 
      for (int j = 0; j < lgrid::jmaxd; j++){
	r  [j] = lg.P0[iqd][k][j][i];
	p  [j] = lg.P0[iqe][k][j][i];
	u  [j] = lg.P0[iqu0+1][k][j][i];
#if NDIM == 3
	v  [j] = lg.P0[iqu0+2][k][j][i];
#else
	v  [j] = 0.0;
#endif
	w  [j] = lg.P0[iqu0][k][j][i];
        for (int m = 0; m < smax; m++){
  	  as[m][j] = lg.P0[iqal0+m][k][j][i];  
        }
        //if (!expand){
          xa0[j] = lg.zya[j];
          dx0[j] = lg.zdy[j];
	//}
        xa [j] = lg.zya[j];
        dx [j] = lg.zdy[j];
        p  [j] = max(pfloor,p[j]);
	e  [j] = p[j]/(r[j]*gamm)+0.5*(u[j]*u[j]+v[j]*v[j]+w[j]*w[j]);

	/*
        if (force){
          if (rank == 0) cout << "Not coded for force in sweepy yet\n";
          quit();
          //act[j] = fy[k][j][i];
        }
        else 
	*/
	act[j] = 0.0;
      }

      // Perform the hydro update

      hydroUpdate(i, k, gg.nlefty, gg.nrighty, gg.ngeomy, lgrid::maxsweep, lg.zparay);

      // Put updated values back into 3D arrays, dropping ghost zones

      for (int j = lg.jrs; j <= lg.jre; j++){
	lg.P0[iqd][k][j][i] = r[j];
	lg.P0[iqe][k][j][i] = p[j];
	lg.P0[iqu0+1][k][j][i] = u[j];
#if NDIM == 3	
	lg.P0[iqu0+2][k][j][i] = v[j];
#endif
	lg.P0[iqu0][k][j][i] = w[j];
        for (int m = 0; m < smax; m++){
  	  lg.P0[iqal0+m][k][j][i] = as[m][j];
        }
      }

    }  // i loop
  }    // k loop

  // Store the new coords
  /*
  if (expand){
    for (j=0;j < jmax; j++){
      n = j+nghost;
      zya[j] = xa0[n];
      zdy[j] = dx0[n];
    }
    ymin = zya[0];
    ymax = zya[jmax-1] + zdy[jmax-1];
    yming = ymin;
    ymaxg = ymax;
    para_setup();  // compute new parabolic coeffs
  }
  */
  
  return;
}


/*!  \brief Performs 1D sweeps in the z-direction.
 *
 *   For an expanding grid simply calculate a new xa0 and dx0 for the expansion.
 *   The parabolic coefficients in zparaz should be calculated  
 *   from the original grid before the expansion, and updated
 *   each time step. However, if we are scaling a uniform grid, 
 *   the coeffs. remain constant, in which case there is no need
 *   to recompute...
 *
 *   \author Julian Pittard (Original version 02.06.11)
 *   \version 0.1-development (Undómiel):
 *   \date Last modified: 04.08.11 (JMP)
 */
void sweepz(){

  sweep = "z";
  
  // Set global ranges for z-sweep
  lg.nmin = lg.krs;                    // first real cell
  lg.nmax = lg.kre;                    // last real cell

  // Expand the grid for each k. dr0 and dsep0 are the initial values and are 
  // not updated!
  /*
  if (expand){
    dr = dr0*pow((dsep/dsep0),expande);
    for (k=0; k<kmax; k++){
      n = k + nghost;
      dx0[n] = dr;
      xa0[n] = float(k-kpivot)*dx0[n];
      // xa0[n] = xa0[n] + 0.4*sin(float(ncycle)/1.0)*dr; ! wiggle the grid
    }
  }
  */
  
  // Loop over each row...

  for (int j = lg.jrs; j <= lg.jre; j++){
    for (int i = lg.irs; i <= lg.ire; i++){
      lg.radius = lg.zxc[i];
      lg.theta  = lg.zyc[j];
      lg.stheta = sin(lg.theta);
      lg.radius = lg.radius * lg.stheta;

      // Put state variables into 1D arrays, including ghost zones.
      // Also extract the required component of the effective force(s)
      // (gravity plus radiative continuum) and the radiative line force(s). 

      for (int k = 0; k < lgrid::kmaxd; k++){
	r  [k] = lg.P0[iqd][k][j][i];
	p  [k] = lg.P0[iqe][k][j][i];
	u  [k] = lg.P0[iqu0+2][k][j][i];
	v  [k] = lg.P0[iqu0][k][j][i];
	w  [k] = lg.P0[iqu0+1][k][j][i];
        for (int m = 0; m < smax; m++){
  	  as[m][k] = lg.P0[iqal0+m][k][j][i];  
        }
        //if (!expand){
          xa0[k] = lg.zza[k];
          dx0[k] = lg.zdz[k];
	//}
        xa [k] = lg.zza[k];
        dx [k] = lg.zdz[k];
        p  [k] = max(pfloor,p[k]);
	e  [k] = p[k]/(r[k]*gamm)+0.5*(u[k]*u[k]+v[k]*v[k]+w[k]*w[k]);

	/*
        if (force){
          if (rank == 0) cout << "Not coded for force in sweepz yet\n";
          quit();
          //act[k] = fz[k][j][i];
        }
        else */
	act[k] = 0.0;
      }

      // Perform the hydro update

      hydroUpdate(i, j, gg.nleftz, gg.nrightz, gg.ngeomz, lgrid::maxsweep, lg.zparaz);

      // Put updated values back into 3D arrays, dropping ghost zones

      for (int k = lg.krs; k <= lg.kre; k++){
	lg.P0[iqd][k][j][i] = r[k];
	lg.P0[iqe][k][j][i] = p[k];
	lg.P0[iqu0+2][k][j][i] = u[k];
	lg.P0[iqu0][k][j][i] = v[k];
	lg.P0[iqu0+1][k][j][i] = w[k];
        for (int m = 0; m < smax; m++){
  	  lg.P0[iqal0+m][k][j][i] = as[m][k];
        }
      }

    }  // i loop
  }    // j loop

  /*
  // Store the new coords
  if (expand){
    for (k=0;k < kmax; k++){
      n = k+nghost;
      zza[k] = xa0[n];
      zdz[k] = dx0[n];
    }
    zmin = zza[0];
    zmax = zza[kmax-1] + zdz[kmax-1];
    zming = zmin;
    zmaxg = zmax;
    para_setup(); // compute new parabolic coeffs
  }
  */
  
  return;
}


/*!  \brief Updates the fluid variables using the time averaged fluxes from
 *          the Riemann solver.
 *
 *   For the lagrangian solver, umid and pmid are used to update the 
 *   velocity, density, and total energy. For the Eulerian solver,
 *   rmid, vmid, wmid, and asmid are also used.
 *   Physical zones are from lg.nmin to lg.nmax.  Zone boundary numbers run from
 *   lg.nmin to lg.nmax+1.
 *
 *   smallp is used to prevent negative temperatures - the value is
 *   problem-specific. ewnd*rho(n) also sets a temperature floor, for
 *   instance preventing adiabatic cooling - this may not be desirable
 *   for some problems. 
 *
 *   Various error checking occurs: e.g. that there are no NaNs or INFs in
 *   the updated fluid variables, that advected scalars have sensible values,
 *   etc.
 *
 *   04.08.11 (JMP) - Replaced 1D Matrices with built-in arrays
 *
 *   \author Julian Pittard (Original version 05.06.11)
 *   \version 0.1-development (Undómiel):
 *   \date Last modified: 04.08.11 (JMP)
 */
void evolve(int n1, int n2, int ngeom){

  // Variables for PPMDE scheme 
  //double pmin, scrch1, artd, arte, csp, hdt, dtdvol, dtdx, aux1, smallp;
  //double rold[lgrid::maxsweep], rinv[lgrid::maxsweep], 
  //    rhoflx[lgrid::maxsweep], cflx[lgrid::maxsweep], uflx[lgrid::maxsweep], vflx[lgrid::maxsweep], 
  //  wflx[lgrid::maxsweep], eflx[lgrid::maxsweep], dr[lgrid::maxsweep];
  //double asflx[smax][lgrid::maxsweep];
  //double as_ex[smax], as_exn1[smax], as_exn2[smax], as_flx[smax];
  //double avis[lgrid::maxsweep]; // artifical viscosity

  double dtheta;
  double fict0[lgrid::maxsweep], fict1[lgrid::maxsweep],
    amid[lgrid::maxsweep], uold[lgrid::maxsweep], dm[lgrid::maxsweep], dtbdm[lgrid::maxsweep],
    xa1[lgrid::maxsweep], xa2[lgrid::maxsweep], xa3[lgrid::maxsweep], dvol1[lgrid::maxsweep], upmid[lgrid::maxsweep];

  const double third = 1./3.;

  //bool errorcheck = true;
  //bool errorcheck = false;
  //const double rndoff = 1.0e-09;

  //if (lagrangian){
    //grid position evolution
    for (int n = lg.nmin-2; n <= lg.nmax+2; n++){
      dm   [n] = r[n] * dvol[n];
      dtbdm[n] = dt / dm[n];
      xa1  [n] = xa[n];
      dvol1[n] = dvol[n];                          // store original zone vol
      xa   [n] = xa[n] + dt * umid[n] / lg.radius;
      upmid[n] = umid[n] * pmid[n];
    }

    xa1[lg.nmin-3] = xa[lg.nmin-3];
    xa1[lg.nmax+3] = xa[lg.nmax+3];
    for (int n = lg.nmin-2; n <= lg.nmax+2; n++){
      xa2[n]   = xa1[n] + 0.5*dx[n];
      dx [n]   = xa[n+1] - xa[n];    
      xa3[n]   = xa [n] + 0.5*dx[n];
    }

    // Calculate forces using coordinates at t. Note that fictitious
    //   forces at t+dt depend on updated velocity, but we ignore this
    fictForces(xa2,u,v,w,fict0); // at t    (xa1/xa2 is the original grid)
    fictForces(xa3,u,v,w,fict1); // at t+dt (xa/xa3  is the evolved  grid)
    
    // Calculate dvolume and average area based on geometry of sweep
    // Maybe loop here can be from lg.nmin-1 to lg.nmax+2?
    if (ngeom == 0){
      for (int n = lg.nmin-2; n <= lg.nmax+2; n++){
        dvol[n] = dx[n];
        amid[n] = 1.;
      }
    }
    else if (ngeom == 1){
      for (int n = lg.nmin-2; n <= lg.nmax+2; n++){
        dvol[n] = dx[n]*(xa[n]+.5*dx[n]);
        amid[n] = .5*(xa[n] + xa1[n]); // average area of interface 
                                       // between t and t+dt
      }
    }
    else if (ngeom == 2){
      for (int n = lg.nmin-2; n <= lg.nmax+2; n++){
        dvol[n] = dx[n]*(xa[n]*(xa[n]+dx[n])+dx[n]*dx[n]*third);
        amid[n] = (xa[n]-xa1[n])*(third*(xa[n]-xa1[n])+xa1[n])+xa1[n]*xa1[n];
      }
    }
    else if (ngeom == 3){
      for (int n = lg.nmin-2; n <= lg.nmax+2; n++){
        dvol[n] = dx[n]*lg.radius;
        amid[n] = 1.;
      }
    }
    else if (ngeom == 4){
      for (int n = lg.nmin-2; n <= lg.nmax+2; n++){
        dvol[n] = (cos(xa[n])-cos(xa[n+1]))*lg.radius;
        dtheta  = xa[n] - xa1[n];
        if (dtheta == 0.0) amid[n] = sin(xa[n]);
        else amid[n] = (cos(xa1[n])-cos(xa[n]))/dtheta;
      }
    }
    else if (ngeom == 5){
      for (int n = lg.nmin-2; n <= lg.nmax+2; n++){
        dvol[n] = dx[n]*lg.radius;
        amid[n] = 1.;
      }
    }
    else{
      if (procRank == 0) cout << "Geometry" << ngeom << " not implemented\n";
      quit();
    }

    // Evolve for the lagrangian code. For the density evolution, all we  
    // have to do is watch the change in the geometry.
    // velocity evolution due to pressure acceleration and forces, both 
    // fictious and external.
    // Total energy evolution (do not include fictious forces)
    // The fictious forces do not change the total energy, but rather 
    // redistribute the momentum between the different coordinate directions.
    // Note we rely on smallp to prevent negative temperatures. ewnd*rho[n]
    // will reheat material which may not be wanted if we have cooling on.
    // Because this is a lagrangian code, advected scalars are unchanged, and 
    // are not updated here.

    for (int n = lg.nmin-1; n <= lg.nmax+1; n++){
      r[n] *= ( dvol1[n] / dvol[n] );
      r[n] = max(r[n],dfloor);
      uold [n] = u[n];
      u[n] -= dtbdm[n] * (pmid[n+1]-pmid[n])
                         * 0.5 * (amid[n+1]+amid[n])
	   + 0.5*dt*(act[n]+fict0[n]+act[n]+fict1[n]);
      e[n] -= dtbdm[n]*
                 (amid[n+1]*upmid[n+1] - amid[n]*upmid[n])
	   + .5*dt*(uold[n]*act[n] + u[n]*act[n]);
      p[n]  = r[n]*gamm*(e[n]-.5*(u[n]*u[n] + v[n]*v[n] + w[n]*w[n]));

      //smallp = max((r[n]/avgmassamb)*boltzman*tamb,pfloor);
      //p[n] = max(p[n],smallp);
      //cout << "evolve:" << n << " " << r[n] << " " << u[n] << " " << p[n] << " " << xa[n] << " " << dx[n] << " " << dvol[n] << " " << dvol1[n] << " " << e[n] << "\n";
    }

    pminFixUpArray(lg.nmin,lg.nmax);
    //quit();
    /*
  }
  else{

// Evolve for the Eulerian code
  } //Eulerian
    */
    
  return;
}


/*!  \brief Check the minimum pressure value. This is problem dependent.
 *
 *   \author Julian Pittard (Original version 05.06.11)
 *   \version 0.1-development (Undómiel):
 *   \date Last modified: 10.08.11 (JMP)
 */
void pminFixUpArray(int nmin, int nmax){

  double pmin;
  //double const1 = boltzman*tamb/avgmassamb;
  double mu = tmpmu(tmin);
  double avgmass = mu*massh;
  double const1 = boltzman*tmin/avgmass;
  
  if (problem == "WBB" || problem == "SNR"){
    for (int n = nmin-1; n <= nmax+1; ++n){
      pmin = max(r[n]*const1,pfloor);
      p[n] = max(p[n],pmin);
    }
  }
  else{
    for (int n = nmin-1; n <= nmax+1; ++n){
      p[n] = max(p[n],pfloor);
    }

  }  
  return;
}


/*!  \brief Solve the Riemann shock tube problem for the left and right 
 *    input states
 *
 *   Uses the Colella & Woodard (1984) iterative, approximate 2-shock solver.
 *   The Newton iteration procedure described in van Leer (1979).
 *
 *   Because of the way the code is constructed, some of the ghost cell values 
 *   will be NaNs, but this is OK as long as this information is far enough 
 *   removed to not propagate into the real grid cells.
 *
 *   Note that the Euleian solver needs the information in w_lft and w_rgh.
 *  
 *   \author Julian Pittard (Original version 05.06.11)
 *   \version 0.1-development (Undómiel):
 *   \date Last modified: 10.08.11 (JMP)
 */
void riemannCW84(int lmin, int lmax){

  //Locals
  double gamfac1, gamfac2;
  double tol,small_p;
  double umidl, umidr, pmold, p_mid;
  double r_r, p_r, u_r, r_l, p_l, u_l;
  double w_r, c_r, z_r, w_l, c_l, z_l;
  double v_r, v_l;
  //double rs,ps,us,vs,ws,w_s,c_s,sig,c_mid,ls,lmid,zeta;
  //double w_lft[lgrid::maxsweep], w_rgh[lgrid::maxsweep], c_lft[lgrid::maxsweep], 
  //  c_rgh[lgrid::maxsweep], z_lft[lgrid::maxsweep], z_rgh[lgrid::maxsweep];
  //double ass[smax];

  //double gamma = gam;
  gamfac1 = 0.5*(gam+1.0)/gam;
  gamfac2 = gam + 1.0;
  tol     = 1.0e-3;
  small_p = 1.0e-20;

//
// Input variables are:
//
//   lmin   = zone number of first physical zone
//   lmax   = zone number of first ghost zone on right (lmax=lg.nmax+1)
//   gam    = equation of state gamma
//   r_r     = density state on the right side of the boundary
//   r_l     = density state on the left side of the boundary
//   p_r     = pressure state on the right side of the boundary
//   p_l     = pressure state on the left side of the boundary
//   u_r     = velocity state on the right side of the boundary
//   u_l     = velocity state on the left side of the boundary
//  and for Eulerian part only,
//   v_r     = transverse velocity state on the right side of the boundary
//   v_l     = transverse velocity state on the left side of the boundary
//   w_r     = transverse velocity state on the right side of the boundary
//   w_l     = transverse velocity state on the left side of the boundary
//   asr(m) = (m)th advected scalar on the right side of the boundary
//   asl(m) = (m)th advected scalar on the left side of the boundary
//
// Output variables are:
// 
//   umid   = time averaged velocity at the zone interface
//   pmid   = time averaged pressure at the zone interface
//   rmid   = time averaged density at the zone interface
//   vmid   = time averaged transverse velocity at the zone interface
//   wmid   = time averaged transverse velocity at the zone interface
//   asmid  = time averaged (m)th advected scalar at the zone interface
//
//


  for (int l = lmin; l <= lmax; l++){
    p_l = plft[l];
    u_l = ulft[l];
    r_l = rlft[l];
    p_r = prgh[l];
    u_r = urgh[l];
    r_r = rrgh[l];

// Obtain first guess for Pmid by assuming W_lft, W_rgh = C_lft, C_rgh
//
    c_l = sqrt(gam*p_l*r_l);
    c_r = sqrt(gam*p_r*r_r);
    // See Eq. 72 in Fryxell et al. (2000, ApJSS, 131, 273) for the next two lines
    p_mid = p_r - p_l - c_r*(u_r-u_l);     
    p_mid = p_l + p_mid * c_l/(c_l+c_r);      
    p_mid = max(small_p,p_mid);

    v_l = 1.0/r_l; // specific volume (saves divisions in the iterator)
    v_r = 1.0/r_r;
//
// Iterate up to 8 times using Newton's method to converge on correct Pmid
//     -use previous guess for pmid to get wavespeeds: w_l, w_r
//     -find the slope in the u-P plane for each state: z_l, z_r
//     -use the wavespeeds and pmid to guess umid on each side: umidl, umidr
//     -project tangents from (pmid,umidl) and (pmid,umidr) to get new pmid
//     -make sure pmid does not fall below floor value for pressure
//

    for (int n = 1; n <= 8; n++){
      pmold = p_mid;
      w_l  = 1.0 + gamfac1*(p_mid - p_l) / p_l;   //Eq 77
      w_r  = 1.0 + gamfac1*(p_mid - p_r) / p_r;
      w_l  = c_l * sqrt(w_l);      
      w_r  = c_r * sqrt(w_r);   
      //w_lft[l] = w_l; // Save - these are needed in the Eulerian part 
      //w_rgh[l] = w_r;
      z_l  = 4.0 * w_l * w_l * v_l;      
      z_r  = 4.0 * w_r * w_r * v_r;   
      z_l  = -z_l * w_l/(z_l - gamfac2*(p_mid - p_l));      
      z_r  =  z_r * w_r/(z_r - gamfac2*(p_mid - p_r));     
      umidl = u_l - (p_mid - p_l) / w_l;      
      umidr = u_r + (p_mid - p_r) / w_r;   
      p_mid  = p_mid + (umidr - umidl)
                * (z_l  * z_r ) / (z_r  - z_l );   
      p_mid  = max(small_p,p_mid);
      if (fabs(p_mid-pmold)/p_mid < tol) break;
    }
    // 
    // Calculate umid by averaging umidl, umidr based on new pmid
    //
    umidl = u_l - (p_mid - p_l) / w_l;      
    umidr = u_r + (p_mid - p_r) / w_r;   
    umid[l]  = 0.5*(umidl + umidr);
    pmid[l] = p_mid;
    //cout << "riemann: " << l << " " << r_l << " " << p_l << " " << u_l << " " << r_r << " " << p_r << " " << u_r << " " << umid[l] << " " << pmid[l] << "\n";
  }

// At this point the Lagrangian solver returns umid and pmid. For the
// Eulerian solver, some extra computation needs to be done (see Flash 
// paper p 293)


  return;
}


/*!  \brief Look for signs of strong shocks and contact discontinuities.
 *
 *   Computes values for flat, jump and shck. Global numerical diffusion can
 *   be added by increasing the minimum value of the flattening parameter from
 *   0.0 to say 0.2 (see Sutherland 2010, Ap&SS, 327, 173, section 6.1.7).
 *   However, tests have shown that this can seriously mess up flow aligned with
 *   the grid, while it does not do as good a job at stopping the carbuncle
 *   instability in 2D axisymmetric sims of CWBs as the code in the Fortran version
 *   of vh1xe.
 *
 *   \var flat (double) returns a value between 0.0 (smooth flow) 
 *        and 1.0 (strong shock) using a simplified method of C&W: eqn. 
 *        A.1 and A.2.
 *
 *   \var jump (double) returns a measure of the pressure jump,
 *    which is needed in order to determine if a CD exists and therefore
 *    whether to apply steepening in the Eulerian solver.
 *
 *   10.08.11 (JMP) - reduced loops from +/- 4 to +/- 2, and sweep[n] to +/- 3.
 *
 *   \author Julian Pittard (Original version 05.06.11)
 *   \version 0.1-development (Undómiel):
 *   \date Last modified: 10.08.11 (JMP)
 */
void flatten(){

  //Locals
  int n;
  double steep[lgrid::maxsweep];
  double omega1, omega2, epsilon, delp1, delp2, shock;
  double temp1, temp2;

  // Parameters as suggested in C&W

  omega1 = 0.75;
  omega2 = 5.0;
  //if (lagrangian) omega2  = 5.0;
  //else omega2 = 10.0;
  epsilon = 0.33;

  // Look for presence of a shock using pressure gradient and sign of
  // velocity jump:  shock = 1 if there is a shock in the zone, else shock = 0
  // Compute steepness parameter based on steepness of pressure jump IF 
  // there is a shock.

  for (n = lg.nmin-2; n <= lg.nmax+2; n++){
    //shck[n] = 0;                                   // initialize
    delp1 = p[n+1] - p[n-1];
    delp2 = p[n+2] - p[n-2];
    if(fabs(delp2) < smallvalue) delp2 = smallvalue;
    jump[n] = fabs(delp1)/min(p[n+1],p[n-1]);
    shock = jump[n]-epsilon;                       // Eq 46 in Fryxell etal
    shock = max(0.0,shock);
    if(shock > 0.0) shock = 1.0;
    if(u[n-1] < u[n+1]) shock = 0.0;               // Eq 47 in Fryxell etal
    //if(shock == 1.0) shck[n] = 1;
    temp1 = ( delp1 / delp2 - omega1 ) * omega2;
    steep[n] = shock * max( 0., min(1.0,temp1) );  // Eq 45 including Eqs 46
                                                   // and 47 
  }

  // Set phony boundary conditions for the steepness parameter

  steep[lg.nmin-3] = steep[lg.nmin-2];
  steep[lg.nmax+3] = steep[lg.nmax+2];

  // Set flatening coefficient based on the steepness in neighboring zones
  for (n = lg.nmin-2; n <= lg.nmax+2; n++){
    //if (lagrangian){
      temp2   = max( steep[n-1], max(steep[n], steep[n+1]) );
      flat[n] = max( 0.0, min( 0.5, temp2 ) );
      //flat[n] = max( 1.0, min( 0.5, temp2 ) ); // global diffusion can be added by increasing the
                                               // minimum value of the flattening parameter
      /*}
    else{
      if (p[n+1] < p[n-1]){             // Eq 48 in Fryxell
        flat[n] = max(steep[n],steep[n+1]);
      }
      else flat[n] = max(steep[n],steep[n-1]);
      //flat[n] = max(1.0,flat[n]);   // adjust the numerical value to add/increase numerical diffusion
      }*/
  }

  return;
}


/*!  \brief Calculate volume of cells.
 *
 *   \author Julian Pittard (Original version 05.06.11)
 *   \version 0.1-development (Undómiel):
 *   \date Last modified: 05.06.11 (JMP)
 */
void volume(int ngeom){

  int l;
  const float third = 0.333333333;

  // Calculate dvolume and average area based on geometry of sweep

  if (ngeom == 0){
    lg.radius = 1.0;
    for (l = lg.nmin-2; l <= lg.nmax+2; l++){
      dvol [l] = dx [l];
      dvol0[l] = dx0[l];
    }
  }
  else if (ngeom == 1){
    lg.radius = 1.0;
    for (l = lg.nmin-2; l <= lg.nmax+2; l++){
      dvol [l] = dx [l]*(xa [l]+0.5*dx [l]);
      dvol0[l] = dx0[l]*(xa0[l]+0.5*dx0[l]);
    }
  }
  else if (ngeom == 2){
    lg.radius = 1.0;
    for (l = lg.nmin-2; l <= lg.nmax+2; l++){
      dvol [l]=dx [l]*(xa [l]*(xa [l]+dx [l])+dx [l]*dx [l]*third);
      dvol0[l]=dx0[l]*(xa0[l]*(xa0[l]+dx0[l])+dx0[l]*dx0[l]*third);
    }
  }
  else if (ngeom == 3){
    for (l = lg.nmin-2; l <= lg.nmax+2; l++){
      dvol [l] = dx [l]*lg.radius;
      dvol0[l] = dx0[l]*lg.radius;
    }
  }
  else if (ngeom == 4){
    for (l = lg.nmin-2; l <= lg.nmax+2; l++){
      dvol [l] = (cos(xa [l])-cos(xa [l+1]))*lg.radius;
      dvol0[l] = (cos(xa0[l])-cos(xa0[l+1]))*lg.radius;
    }
  }
  else if (ngeom == 5){
    for (l = lg.nmin-2; l <= lg.nmax+2; l++){
      dvol [l] = dx [l] * lg.radius;
      dvol0[l] = dx0[l] * lg.radius;
    }
  }

  return;
}


/*!  \brief Function which calculates fictitious (ie. coriolis and 
 *          centrifugal) forces. 
 *
 *   \var fict[] (double) returns with the fictitious forces.
 *
 *   Actual forces (eg. gravitational and radiative) should be calculated 
 *   elsewhere and extracted into the array act(n) in sweepx, 
 *   sweepy and sweepz.
 *
 *   \author Julian Pittard (Original version 05.06.11)
 *   \version 0.1-development (Undómiel):
 *   \date Last modified: 04.08.11 (JMP)
 */
void fictForces(double xf[], double uf[], double vf[], double wf[], double fict[]){

  int n;
  float xf0;

  // Initialize the fictious array

  for (n = 0; n < lgrid::maxsweep; n++){
    fict[n] = 0.0;
  }

  // Perform the force calculations
 
  if (sweep == "x"){
    if ((gg.ngeomx == 0) && ((gg.ngeomy == 0)||(gg.ngeomy == 1))){

      // Cartesian or cyclindrical radial
      // Do nothing, since already initialized to zero...
    }
    else if ((gg.ngeomx == 1) && (gg.ngeomy == 3)){

      // Cylindrical polar

      for (n = 0; n < lgrid::maxsweep; n++){
        fict[n] = vf[n]*vf[n]/xf[n];
      }
    }
    else if (gg.ngeomx == 2){

      // Spherical

      for (n = 0; n < lgrid::maxsweep; n++){
        if (xf[n] == 0) fict[n] = 0.;
        else fict[n] = (wf[n]*wf[n]+vf[n]*vf[n])/xf[n];
      }
    }
  }


  if (sweep == "y"){
     if ((gg.ngeomx == 0) && ((gg.ngeomy == 0)||(gg.ngeomy == 1))){

      // Cartesian or cyclindrical radial (do nothing)

    }
    else if ((gg.ngeomx == 1) && (gg.ngeomy == 3)){

      // Cylindrical polar (2D)

      for (n = 0; n < lgrid::maxsweep; n++){
        fict[n] = -uf[n]*wf[n]/lg.radius;
      }
    }
    else if ((gg.ngeomy == 3) && (gg.ngeomx == 2)){

      // Spherical equator (2D)

      for (n = 0; n < lgrid::maxsweep; n++){
        fict[n] = -uf[n]*wf[n]/lg.radius;
      }
    }
    else if ((gg.ngeomy == 4) && (gg.ngeomx == 2)){

      // Spherical

      for (n = 0; n < lgrid::maxsweep; n++){
        if (xf[n] == 0.0) xf0 = smallvalue;
        else xf0 = xf[n];
        fict[n] = -uf[n]*wf[n]/lg.radius
	  +vf[n]*vf[n]/lg.radius * cos(xf[n])/sin(xf0);
        if (xf[n] == 0.0) fict[n] = 0.0;
      }
    }
  }

  if (sweep == "z"){
    if (gg.ngeomz == 0){

      // Cartesian (do nothing)
      
    }
    else if (gg.ngeomz == 3){

      // Cylindrical

      for (n = 0; n < lgrid::maxsweep; n++){
        fict[n] = -uf[n]*wf[n]/lg.radius;
      }
    }
    else if (gg.ngeomz == 5){

      // Spherical

      for (n = 0; n < lgrid::maxsweep; n++){
        fict[n] = -uf[n]*wf[n]/lg.radius
	  -uf[n]*vf[n]/lg.radius * cos(lg.theta)/lg.stheta;
      }
    }
  }

  return;
}


/*!  \brief Performs the 1D lagrangian or eulerian hydrodynamics evolution.
 *
 *   It uses the 1D arrays of the fluid variables extracted from the sweep
 *   functions. For the lagrangian evolution:
 *   \li obtain parabolic interpolations of rho, u, P
 *   \li compute input states from these interpolations for Riemann problem
 *   \li call the Riemann solver to find the time averages umid, pmid
 *   \li evolve the finite difference equations to get updated
 *      values of rho, u, and E
 *
 *   For the eulerian evolution:
 *   \li obtain parabolic interpolations of rho, u, P, v, w, and the advected
 *     advected scalars
 *   \li compute input states from these interpolations for Riemann problem
 *
 *   n1 and n2 are indices of the 1D array sweep extracted from the 3D solution
 *
 *   04.08.11 (JMP) - Modifies for use with built-in arrays.
 *
 *   \author Julian Pittard (Original version 05.06.11)
 *   \version 0.1-development (Undómiel):
 *   \date Last modified: 10.08.11 (JMP)
 */
void ppm(int n1, int n2, int nleft, int nright, int ngeom, double para[][lgrid::maxsweep] ){

  //Locals
  int icd;

  // Interpolate parabolae for fluid variables. For the Eulerian solver
  // we obtain parabolae for the transverse velocities and the advected
  // scalars in addition to p,r, and u

  //if (lagrangian == false){
  //  icd = 1;  // cd steepening applied to eulerian version 
  //}
  //else icd = 0;
  icd = 0;
  
  parabola(lg.nmin-2,lg.nmax+2,para,r,dr,r6,rl,flat,jump,1,0,icd);
  parabola(lg.nmin-2,lg.nmax+2,para,p,dp,p6,pl,flat,jump,1,0,0);
  parabola(lg.nmin-2,lg.nmax+2,para,u,du,u6,ul,flat,jump,1,0,0);

  /*
  //if (lagrangian == false){
    parabola(lg.nmin-2,lg.nmax+2,para,v,dv,v6,vl,flat,jump,1,0,0);
    parabola(lg.nmin-2,lg.nmax+2,para,w,dw,w6,wl,flat,jump,1,0,0);
    //pass 1D vector for each advected scalar
    for (m = 0; m < smax; m++){
      //pass contents of row [m] extracted from a 2D array into parabola
      parabola(lg.nmin-2,lg.nmax+2,para,as[m],das_1d,as6_1d,asl_1d,flat,jump,1,0,1);
      //copy contents of doubles back into the mth row of a 2D double
      copy(das_1d,das_1d + lgrid::maxsweep,das[m]); //not sure this works!
      copy(as6_1d,as6_1d + lgrid::maxsweep,as6[m]);
      copy(asl_1d,asl_1d + lgrid::maxsweep,asl[m]);
    }
  //}
  */
  
  // Integrate parabolae to get input states for Riemann problem
  states( n1, n2 );

  return;
}


/*!  \brief Computes the left and right input states for the Riemann solver.
 *
 *   Takes the values of rho, u, and P at the left hand
 *   side of the zone, the change accross the zone, and the parabolic 
 *   coefficients, p6, u6, and rho6, and computes the left and right states
 *   (integrated over the characteristics) for each variable for input 
 *   to the Riemann solver.
 *
 *   \author Julian Pittard (Original version 02.06.11)
 *   \version 0.1-development (Undómiel):
 *   \date Last modified: 04.08.11 (JMP)
 */
void states(int n1, int n2){
//
// Input variables are:
//    pl    =  pressure at left edge of zone
//    dp    =  pressure slope
//    p6    =  pressure parabola coeffecient
//    ul    =  velocity at left edge of zone
//    du    =  velocity slope
//    u6    =  velocity parabola coeffecient
//    rl    =  density at left edge of zone
//    dr    =  density slope
//    r6    =  density parabola coeffecient
//
// Output variables are:
//    plft =  average pressure in the + wave state (left of interface)
//    prgh =  average pressure in the - wave state (right of interface)
//    ulft =  average velocity in the + wave state (left of interface)
//    urgh =  average velocity in the - wave state (right of interface)
//    rlft =  average density in the  + wave state (left of interface)
//    rrgh =  average density in the  - wave state (right of interface)
//

  // Locals
  bool check_error;
  //double betap, beta0, betam, tbetap, tbeta0, tbetam,
  //    plftp, plft0, plftm, ulftp, ulft0, ulftm, rlftp, rlft0, rlftm,
  //    prghp, prgh0, prghm, urghp, urgh0, urghm, rrghp, rrgh0, rrghm,
  //    vlft0, wlft0, vrgh0, wrgh0, svell, svelr, al, ar,
  //    chilp, chil0, chilm, chirp, chir0, chirm, dtdx;
  //double aslft0[smax], asrgh0[smax];

  double fothd, hdt;
  double Cdtdx[lgrid::maxsweep], fCdtdx[lgrid::maxsweep];
  double fict[lgrid::maxsweep];

  //check_error = true;
  check_error = false;

  fothd = 4./3.;
  hdt   = 0.5*dt;
  
  // Include gravitational and ficticious forces

  fictForces(xa,u,v,w,fict);

  // If the solver is Lagrangian, calculate states

  //if (lagrangian){
    // Calculate the domain of dependence along space coordinate,
    // C*dt, by multiplying the Eulerian sound speed in 
    // each zone by dt.  Divide by two to save flops later.
    //
    // If angular coordinate, then divide out by lg.radius to get physical 
    // dimensions

    for (int n = lg.nmin-2; n <= lg.nmax+2; n++){
      Cdtdx [n] = sqrt(gam*p[n]/r[n])/(dx[n]*lg.radius);
      svel      = max(svel,Cdtdx[n]);
      Cdtdx [n] = Cdtdx[n]*hdt;                            
      fCdtdx[n] = 1. - fothd*Cdtdx[n];
    }

    // Obtain averages of rho, u, and P over the domain (+/-)Cdt
    //                    lft is the + wave on the left  side of the boundary
    //                    rgh is the - wave on the right side of the boundary


    // This loop needs to be from nmin-3 (if from lg.nmin-2 it creates NaNs on the
    // boundary). Actually lg.nmin-2 is OK for 1D SOD.

    for (int n = lg.nmin-3; n <= lg.nmax+2; n++){
      int k = n+1;
      // Time-averaged left state at right boundary of cell i
      plft[k] = pl[n]+dp[n]-Cdtdx[n]*(dp[n]-fCdtdx[n]*p6[n]); // Eq 56 Fryxell
      ulft[k] = ul[n]+du[n]-Cdtdx[n]*(du[n]-fCdtdx[n]*u6[n]);
      rlft[k] = rl[n]+dr[n]-Cdtdx[n]*(dr[n]-fCdtdx[n]*r6[n]);
      plft[k] = max(smallvalue,plft[k]);
      rlft[k] = max(smallvalue,rlft[k]);
      ulft[k] = ulft[k] + hdt*(act[k]+fict[k]);

      // Time-averaged right state at left boundary of cell i
      prgh[n] = pl[n] + Cdtdx[n]*(dp[n]+fCdtdx[n]*p6[n]);     //Eq 54
      urgh[n] = ul[n] + Cdtdx[n]*(du[n]+fCdtdx[n]*u6[n]);
      rrgh[n] = rl[n] + Cdtdx[n]*(dr[n]+fCdtdx[n]*r6[n]);
      prgh[n] = max(smallvalue,prgh[n]);
      rrgh[n] = max(smallvalue,rrgh[n]);
      urgh[n] = urgh[n] + hdt*(act[n]+fict[n]);
      //if (check_error) checkstates(n,plft[k],rlft[k],ulft[k],prgh[n],rrgh[n],urgh[n]);
    }

  /*}
  else{ 

// Calculate states for Eulerian solver
// Obtain initial guesses. For the positive and negative characteristics,
// only the pressure, normal velocity, and density need to be computed.
// For the zero characteristic, the transverse velocities and mass fractions
// must also be included.

  for (int n = lg.nmin-1; n <= lg.nmax+1; n++){
    int k = n+1;
    svel = sqrt(gam*p[n]/r[n]);// this is correct - Colella & Woodward use
                               // C = sqrt(gam p rho), where C^2 = (rho svel)^2
    dtdx = dt/(dx[n]*lg.radius);  // lg.radius is not equal to 1.0 if the coord
                               // is annular

    // Domain of dependence for left state

    betap = max((u[n]+svel)*dtdx,0.0); //
    beta0 = max((u[n]     )*dtdx,0.0); // - Eq 59 
    betam = max((u[n]-svel)*dtdx,0.0); //                
    tbetap = 1.0 - (2.0/3.0)*betap;
    tbeta0 = 1.0 - (2.0/3.0)*beta0;        
    tbetam = 1.0 - (2.0/3.0)*betam;        

    // Time-averaged state on left side of the right interface of cell n 
    plftp = pl[n]+dp[n] - 0.5*betap*(dp[n]-p6[n]*tbetap);// Eq 58
    plft0 = pl[n]+dp[n] - 0.5*beta0*(dp[n]-p6[n]*tbeta0);
    plftm = pl[n]+dp[n] - 0.5*betam*(dp[n]-p6[n]*tbetam);
    ulftp = ul[n]+du[n] - 0.5*betap*(du[n]-u6[n]*tbetap);
    ulft0 = ul[n]+du[n] - 0.5*beta0*(du[n]-u6[n]*tbeta0);
    ulftm = ul[n]+du[n] - 0.5*betam*(du[n]-u6[n]*tbetam);
    rlftp = rl[n]+dr[n] - 0.5*betap*(dr[n]-r6[n]*tbetap);
    rlft0 = rl[n]+dr[n] - 0.5*beta0*(dr[n]-r6[n]*tbeta0);
    rlftm = rl[n]+dr[n] - 0.5*betam*(dr[n]-r6[n]*tbetam);
    vlft0 = vl[n]+dv[n] - 0.5*beta0*(dv[n]-v6[n]*tbeta0);
    wlft0 = wl[n]+dw[n] - 0.5*beta0*(dw[n]-w6[n]*tbeta0);
    for (int m = 0; m < smax; m++){   // advected scalars
      aslft0[m] = asl[m][n]+das[m][n] - 0.5*beta0*(das[m][n]-as6[m][n]*tbeta0);
    }

    // Domain of dependence for right state

    betap = max(-(u[n]+svel)*dtdx,0.0); //
    beta0 = max(-(u[n]     )*dtdx,0.0); // - Eq 61 
    betam = max(-(u[n]-svel)*dtdx,0.0); //                 
    tbetap = 1.0 - (2.0/3.0)*betap;        
    tbeta0 = 1.0 - (2.0/3.0)*beta0;        
    tbetam = 1.0 - (2.0/3.0)*betam;        

    // Time-averaged state on right side of the left interface of cell n 
    prghp = pl[n] + 0.5*betap*(dp[n]+p6[n]*tbetap);      //Eq 60
    prgh0 = pl[n] + 0.5*beta0*(dp[n]+p6[n]*tbeta0);        
    prghm = pl[n] + 0.5*betam*(dp[n]+p6[n]*tbetam);        
    urghp = ul[n] + 0.5*betap*(du[n]+u6[n]*tbetap);        
    urgh0 = ul[n] + 0.5*beta0*(du[n]+u6[n]*tbeta0);        
    urghm = ul[n] + 0.5*betam*(du[n]+u6[n]*tbetam);        
    rrghp = rl[n] + 0.5*betap*(dr[n]+r6[n]*tbetap);        
    rrgh0 = rl[n] + 0.5*beta0*(dr[n]+r6[n]*tbeta0);        
    rrghm = rl[n] + 0.5*betam*(dr[n]+r6[n]*tbetam);        
    vrgh0 = vl[n] + 0.5*beta0*(dv[n]+v6[n]*tbeta0);        
    wrgh0 = wl[n] + 0.5*beta0*(dw[n]+w6[n]*tbeta0);        
    for (int m = 0; m < smax; m++){   // advected scalars
      asrgh0[m] = asl[m][n] + 0.5*beta0*(das[m][n]+as6[m][n]*tbeta0);
    }

    // Correct initial guesses. Compute the 6 chi quantities (Eqs 62 and 63)
    // I've checked this against the code in FLASH - everything seems OK

    svell = sqrt(gam*plftp/rlftp); // related to clft in FLASH
    svelr = sqrt(gam*prghm/rrghm);
    al = rlftp*svell;              // rho*c
    ar = rrghm*svelr;
    chilp = 0.0;
    chilm =  (0.5/al)*(ulftp - ulftm - (plftp-plftm)/al);
    chirp = -(0.5/ar)*(urghm - urghp + (prghm-prghp)/ar);
    chirm = 0.0;
    chil0 = (plftp-plft0)/(al*al) + 1.0/rlftp - 1.0/rlft0;
    chir0 = (prghm-prgh0)/(ar*ar) + 1.0/rrghm - 1.0/rrgh0;
    if (u[n]      <= 0.0) chil0 = 0.0;  // lambda_i_0
    if (u[n]-svel <= 0.0) chilm = 0.0;  // lambda_i_-
    if (u[n]+svel >= 0.0) chirp = 0.0;  // lambda_i+1_+
    if (u[n]      >= 0.0) chir0 = 0.0;  // lambda_i+1_0

    // Now use the chi quantities to correct the initial guess for the left
    // and right states (Eqs 65-67). The left and right states for v,w,col
    // are simply the average of these quantities under the zero characteristic

    plft[k] = plftp + al*al*(chilp + chilm);
    ulft[k] = ulftp + al*(chilp - chilm);
    rlft[k] = 1.0/(1.0/rlftp - chilp - chil0 - chilm);
    plft[k] = max(pfloor,plft[k]);
    rlft[k] = max(dfloor,rlft[k]);
    ulft[k] = ulft[k] + hdt*(act[k]+fict[k]);
    vlft[k] = vlft0; // transverse velocity
    wlft[k] = wlft0; // transverse velocity
    for (int m = 0; m < smax; m++){  // advected scalars
      aslft[m][k] = aslft0[m]; 
    }

    prgh[n] = prghm + ar*ar*(chirp + chirm);
    urgh[n] = urghm + ar*(chirp - chirm);
    rrgh[n] = 1.0/(1.0/rrghm - chirp - chir0 - chirm);
    prgh[n] = max(pfloor,prgh[n]);
    rrgh[n] = max(dfloor,rrgh[n]);
    urgh[n] = urgh[n] + hdt*(act[n]+fict[n]);
    vrgh[n] = vrgh0;
    wrgh[n] = wrgh0;
    for (int m = 0; m < smax; m++){  // advected scalars
      asrgh[m][n] = asrgh0[m]; 
    }
  }
  } // Lagrangian or Eulerian
  */
    
  return;
}


/*!  \brief Remap from the updated lagrangian grid to the fixed 
 *    Eulerian grid.
 *
 *   The mass, momentum, and energy are remapped using the 
 *   piecewise parabolic functions. No flattening is used on remap 
 *   (pass dummy array: dm to parabola.f)
 *
 *   04.08.11 - replaced Matrices with built-in arrays
 *
 *   \author Julian Pittard (Original version 07.06.11)
 *   \version 0.1-development (Undómiel):
 *   \date Last modified: 04.08.11 (JMP)
 */
void remap(int nleft, int nright, int ngeom){

  // Locals
  double rr, ur, vr, wr, pr, er;
  double deltx, fractn, fractn2;

  double du[lgrid::maxsweep], ul[lgrid::maxsweep], u6[lgrid::maxsweep],
    dv[lgrid::maxsweep], vl[lgrid::maxsweep], v6[lgrid::maxsweep],
    dw[lgrid::maxsweep], wl[lgrid::maxsweep], w6[lgrid::maxsweep],
    de[lgrid::maxsweep], el[lgrid::maxsweep], e6[lgrid::maxsweep],
    dp[lgrid::maxsweep], pl[lgrid::maxsweep], p6[lgrid::maxsweep],
    dr[lgrid::maxsweep], rl[lgrid::maxsweep], r6[lgrid::maxsweep];

  double dm[lgrid::maxsweep], dm0[lgrid::maxsweep], delta[lgrid::maxsweep],
    fluxr[lgrid::maxsweep], fluxu[lgrid::maxsweep],
    fluxe[lgrid::maxsweep], fluxv[lgrid::maxsweep], fluxw[lgrid::maxsweep];

  double asr[smax];
  
  //double as_ex[smax];
  //double das_1d[lgrid::maxsweep], as6_1d[lgrid::maxsweep], asl_1d[lgrid::maxsweep]; 

  double das[smax][lgrid::maxsweep],asl[smax][lgrid::maxsweep],as6[smax][lgrid::maxsweep],
    fluxas[smax][lgrid::maxsweep];

  double para[lgrid::paramax][lgrid::maxsweep];

  const double third  = 1. / 3.;
  //const double twothd = 2. * third;
  const double fourthd= 4. * third;

  // Generate interpolation functions, saving da, al for constructing left 
  // and right total energy states. (dm is passed as dummy array)

  paraset(lg.nmin,lg.nmax,para,dx,xa);
  //cout << scientific << setprecision(4);
  //for (n = lg.nmin; n <= lg.nmax; n++){
    //cout << n << "\t" << para[1][n] << "\t" << para[2][n] << "\t" << para[3][n] << "\t" << para[4][n] << "\t" << para[5][n] << "\n";
    //cout << n << "\t" << r(n) << "\t" << dr(n) << "\t" << r6(n) << "\t" << rl(n) << "\n";
  //}
  parabola(lg.nmin,lg.nmax,para,r,dr,r6,rl,dm,dm,0,0,0);
  parabola(lg.nmin,lg.nmax,para,u,du,u6,ul,dm,dm,0,0,0);
  parabola(lg.nmin,lg.nmax,para,v,dv,v6,vl,dm,dm,0,0,0);
  parabola(lg.nmin,lg.nmax,para,w,dw,w6,wl,dm,dm,0,0,0);
  parabola(lg.nmin,lg.nmax,para,p,dp,p6,pl,dm,dm,0,0,0);
  //quit();

  for (int m=0;m<smax;m++){
    //pass contents of row [m] extracted from a 2D array into parabola
    parabola(lg.nmin,lg.nmax,para,as[m],das[m],as6[m],asl[m],dm,dm,0,0,0);
  }

  // Use the profiles for density, pressure, and velocities to 
  // calculate consistent values of the left and right values of total energy

  for (int n = lg.nmin; n <= lg.nmax; n++){
    rr = rl[n] + dr[n];
    ur = ul[n] + du[n];
    pr = pl[n] + dp[n];
    vr = vl[n] + dv[n];
    wr = wl[n] + dw[n];
    for (int m=0;m<smax;m++){
      asr[m] = asl[m][n] + das[m][n];
    }
    el[n] = pl[n]/(rl[n]*gamm)+.5*(ul[n]*ul[n] + vl[n]*vl[n] + wl[n]*wl[n]);
    er    = pr   /(rr   *gamm)+.5*(ur*ur   +vr*vr   +wr*wr);
    de[n] = er-el[n];
    //cout << n << "\t" << el[n] << "\t" << er << "\t" << de[n]  << "\n";
  }
  //quit();

  // Apply boundary conditions for flow onto the grid

  if ((nleft == 0) || (nleft == 1) || (nleft == 4)){
    rl[lg.nmin-1] = rl[lg.nmin];
    dr[lg.nmin-1] = 0.;
    r6[lg.nmin-1] = 0.;
    vl[lg.nmin-1] = vl[lg.nmin];
    dv[lg.nmin-1] = 0.;
    v6[lg.nmin-1] = 0.;
    wl[lg.nmin-1] = wl[lg.nmin];
    dw[lg.nmin-1] = 0.;
    w6[lg.nmin-1] = 0.;
    if (nleft == 4){
      if (ul[lg.nmin] > 0.0){
	ul[lg.nmin-1] = -ul[lg.nmin]; // opposite to enforce zero flux
      } 
      else{
	ul[lg.nmin-1] = ul[lg.nmin];
      }
    }
    else ul[lg.nmin-1] = ul[lg.nmin];
    du[lg.nmin-1] = 0.;
    u6[lg.nmin-1] = 0.;
    el[lg.nmin-1] = el[lg.nmin];
    de[lg.nmin-1] = 0.;
    e6[lg.nmin-1] = 0.;
    for (int m=0;m<smax;m++){
      asl[m][lg.nmin-1] = asl[m][lg.nmin];
      das[m][lg.nmin-1] = 0.;
      as6[m][lg.nmin-1] = 0.;
    }
  }
  else if(nleft == 2){
    rl[lg.nmin-1] = gg.dinflo;
    dr[lg.nmin-1] = 0.0;
    r6[lg.nmin-1] = 0.0;
    vl[lg.nmin-1] = gg.vinflo;
    dv[lg.nmin-1] = 0.0;
    v6[lg.nmin-1] = 0.0;
    wl[lg.nmin-1] = gg.winflo;
    dw[lg.nmin-1] = 0.0;
    w6[lg.nmin-1] = 0.0;
    ul[lg.nmin-1] = gg.uinflo;
    du[lg.nmin-1] = 0.0;
    u6[lg.nmin-1] = 0.0;
    el[lg.nmin-1] = gg.einflo;
    de[lg.nmin-1] = 0.0;
    e6[lg.nmin-1] = 0.0;
    for (int m=0;m<smax;m++){
      asl[m][lg.nmin-1] = gg.asinflo[m];
      das[m][lg.nmin-1] = 0.0;
      as6[m][lg.nmin-1] = 0.0;
    }
  }
  else if(nleft == 3){
    rl[lg.nmin-1] = rl[lg.nmax];
    dr[lg.nmin-1] = dr[lg.nmax];
    r6[lg.nmin-1] = r6[lg.nmax];
    vl[lg.nmin-1] = vl[lg.nmax];
    dv[lg.nmin-1] = dv[lg.nmax];
    v6[lg.nmin-1] = v6[lg.nmax];
    wl[lg.nmin-1] = wl[lg.nmax];
    dw[lg.nmin-1] = dw[lg.nmax];
    w6[lg.nmin-1] = w6[lg.nmax];
    ul[lg.nmin-1] = ul[lg.nmax];
    du[lg.nmin-1] = du[lg.nmax];
    u6[lg.nmin-1] = u6[lg.nmax];
    el[lg.nmin-1] = el[lg.nmax];
    de[lg.nmin-1] = de[lg.nmax];
    e6[lg.nmin-1] = e6[lg.nmax];
    for (int m=0;m<smax;m++){
      asl[m][lg.nmin-1] = asl[m][lg.nmax];
      das[m][lg.nmin-1] = das[m][lg.nmax];
      as6[m][lg.nmin-1] = as6[m][lg.nmax];
    }
  }
  else if(nleft == 5){
    rl[lg.nmin-1] = gg.dinflo;
    dr[lg.nmin-1] = 0.0;
    r6[lg.nmin-1] = 0.0;
    vl[lg.nmin-1] = gg.winflo;
    dv[lg.nmin-1] = 0.0;
    v6[lg.nmin-1] = 0.0;
    wl[lg.nmin-1] = gg.uinflo;
    dw[lg.nmin-1] = 0.0;
    w6[lg.nmin-1] = 0.0;
    ul[lg.nmin-1] = gg.vinflo;
    du[lg.nmin-1] = 0.0;
    u6[lg.nmin-1] = 0.0;
    el[lg.nmin-1] = gg.einflo;
    de[lg.nmin-1] = 0.0;
    e6[lg.nmin-1] = 0.0;
    for (int m=0;m<smax;m++){
      asl[m][lg.nmin-1] = gg.asinflo[m];
      das[m][lg.nmin-1] = 0.0;
      as6[m][lg.nmin-1] = 0.0;
    }
  }
  if ((nright == 0) || (nright == 1) || (nright == 4)){
    rl[lg.nmax+1] = rl[lg.nmax];
    dr[lg.nmax+1] = 0.;
    r6[lg.nmax+1] = 0.;
    vl[lg.nmax+1] = vl[lg.nmax];
    dv[lg.nmax+1] = 0.;
    v6[lg.nmax+1] = 0.;
    wl[lg.nmax+1] = wl[lg.nmax];
    dw[lg.nmax+1] = 0.;
    w6[lg.nmax+1] = 0.;
    if (nright == 4){
      if (ul[lg.nmax] < 0.0){
	ul[lg.nmax+1] = -ul[lg.nmax]; // opposite to enforce zero flux
      }
      else{
	ul[lg.nmax+1] = ul[lg.nmax];
      }
    }
    else ul[lg.nmax+1] = ul[lg.nmax];
    du[lg.nmax+1] = 0.;
    u6[lg.nmax+1] = 0.;
    el[lg.nmax+1] = el[lg.nmax];
    de[lg.nmax+1] = 0.;
    e6[lg.nmax+1] = 0.;
    for (int m=0;m<smax;m++){
      asl[m][lg.nmax+1] = asl[m][lg.nmax];
      das[m][lg.nmax+1] = 0.;
      as6[m][lg.nmax+1] = 0.;
    }
  }
  else if(nright == 2){
    rl[lg.nmax+1] = gg.dinflo;
    dr[lg.nmax+1] = 0.0;
    r6[lg.nmax+1] = 0.0;
    vl[lg.nmax+1] = gg.vinflo;
    dv[lg.nmax+1] = 0.0;
    v6[lg.nmax+1] = 0.0;
    wl[lg.nmax+1] = gg.winflo;
    dw[lg.nmax+1] = 0.0;
    w6[lg.nmax+1] = 0.0;
    ul[lg.nmax+1] = gg.uinflo;
    du[lg.nmax+1] = 0.0;
    u6[lg.nmax+1] = 0.0;
    el[lg.nmax+1] = gg.einflo;
    de[lg.nmax+1] = 0.0;
    e6[lg.nmax+1] = 0.0;
    for (int m=0;m<smax;m++){
      asl[m][lg.nmax+1] = gg.asinflo[m];
      das[m][lg.nmax+1] = 0.0;
      as6[m][lg.nmax+1] = 0.0;
    }
  }
  else if (nright == 3){
    rl[lg.nmax+1] = rl[lg.nmin];
    dr[lg.nmax+1] = dr[lg.nmin];
    r6[lg.nmax+1] = r6[lg.nmin];
    vl[lg.nmax+1] = vl[lg.nmin];
    dv[lg.nmax+1] = dv[lg.nmin];
    v6[lg.nmax+1] = v6[lg.nmin];
    wl[lg.nmax+1] = wl[lg.nmin];
    dw[lg.nmax+1] = dw[lg.nmin];
    w6[lg.nmax+1] = w6[lg.nmin];
    ul[lg.nmax+1] = ul[lg.nmin];
    du[lg.nmax+1] = du[lg.nmin];
    u6[lg.nmax+1] = u6[lg.nmin];
    el[lg.nmax+1] = el[lg.nmin];
    de[lg.nmax+1] = de[lg.nmin];
    e6[lg.nmax+1] = e6[lg.nmin];
    for (int m=0;m<smax;m++){
      asl[m][lg.nmax+1] = asl[m][lg.nmin];
      das[m][lg.nmax+1] = das[m][lg.nmin];
      as6[m][lg.nmax+1] = as6[m][lg.nmin];
    }
  }
  else if(nright == 5){
    rl[lg.nmax+1] = gg.dinflo;
    dr[lg.nmax+1] = 0.;
    r6[lg.nmax+1] = 0.;
    vl[lg.nmax+1] = gg.winflo;
    dv[lg.nmax+1] = 0.;
    v6[lg.nmax+1] = 0.;
    wl[lg.nmax+1] = gg.uinflo;
    dw[lg.nmax+1] = 0.;
    w6[lg.nmax+1] = 0.;
    ul[lg.nmax+1] = gg.vinflo;
    du[lg.nmax+1] = 0.;
    u6[lg.nmax+1] = 0.;
    el[lg.nmax+1] = gg.einflo;
    de[lg.nmax+1] = 0.;
    e6[lg.nmax+1] = 0.;
    for (int m=0;m<smax;m++){
      asl[m][lg.nmax+1] = gg.asinflo[m];
      das[m][lg.nmax+1] = 0.0;
      as6[m][lg.nmax+1] = 0.0;
    }
  }

  // Use the constructed left and right states to get parabolas for total energy

  parabola(lg.nmin,lg.nmax,para,e,de,e6,el,dm,dm,0,1,0);
  //for (n = lg.nmin; n <= lg.nmax; n++){
  //  cout << n << "\t" << e[n] << "\t" << de[n] << "\t" << e6[n] << "\t" << el[n] << "\n";
  //}
  //quit();

  // Calculate the volume of the overlapping subshells (delta)

  if (ngeom == 0){
    for (int n = lg.nmin; n <= lg.nmax+1; n++){
      delta[n] = xa[n] - xa0[n];
    }
  }
  else if (ngeom == 1){
    for (int n = lg.nmin; n <= lg.nmax+1; n++){
      delta[n] = xa[n] - xa0[n];
      delta[n] = delta[n]*(xa0[n]+.5*delta[n]);
    }
  }
  else if (ngeom == 2){
    for (int n = lg.nmin; n <= lg.nmax+1; n++){
      delta[n] = xa[n] - xa0[n];
      delta[n] = delta[n]*(xa0[n]*(xa0[n]
		 + delta[n]) + delta[n]*delta[n]*third);
    }
  }
  else if (ngeom == 3){
    for (int n = lg.nmin; n <= lg.nmax+1; n++){
      delta[n] = (xa[n] - xa0[n]) * lg.radius;
    }
  }
  else if (ngeom == 4){
    for (int n = lg.nmin; n <= lg.nmax+1; n++){
      delta[n] = (cos(xa0[n]) - cos(xa[n])) * lg.radius;
    }
  }
  else if (ngeom == 5){
    for (int n = lg.nmin; n <= lg.nmax+1; n++){
      delta[n] = (xa[n] - xa0[n]) * lg.radius;
    }
  }

  // Calculate the total mass (fluxr), momentum (fluxu), and energy (fluxe)
  // in the subshell created by the overlap of the Lagrangian and Eulerian 
  // grids. If the zone face has moved to the left (deltx > 0), use the 
  // integral from the left side of zone n (fluxrr).  If the zone face has 
  // moved to the right (deltx < 0), use the integral from the right side 
  // of zone k=n-1 (fluxrl).

  for (int n = lg.nmin; n <= lg.nmax+1; n++){
    deltx = xa[n] - xa0[n];

    if(deltx >= 0.0){
      int k = n - 1;
      fractn  = 0.5*deltx/dx[k];
      fractn2 = 1.0 - fourthd*fractn;
      fluxr[n] = delta[n]*(rl[k]+dr[k]-fractn*(dr[k]-fractn2*r6[k]));
      fluxu[n]=(ul[k]+du[k]-fractn*(du[k]-fractn2*u6[k]))*fluxr[n];
      fluxv[n]=(vl[k]+dv[k]-fractn*(dv[k]-fractn2*v6[k]))*fluxr[n];
      fluxw[n]=(wl[k]+dw[k]-fractn*(dw[k]-fractn2*w6[k]))*fluxr[n];
      fluxe[n]=(el[k]+de[k]-fractn*(de[k]-fractn2*e6[k]))*fluxr[n];
      for (int m = 0; m < smax; m++){
	fluxas[m][n]=(asl[m][k]+das[m][k] - 
		     fractn*(das[m][k]-fractn2*as6[m][k]))*fluxr[n];
      }
    }
    else{
      fractn   = 0.5*deltx/dx[n];
      fractn2  = 1.0 + fourthd*fractn;
      fluxr[n] = delta[n]*(rl[n] - fractn*(dr[n]+fractn2*r6[n]));
      fluxu[n]=(ul[n] - fractn*(du[n]+fractn2*u6[n]))*fluxr[n];
      fluxv[n]=(vl[n] - fractn*(dv[n]+fractn2*v6[n]))*fluxr[n];
      fluxw[n]=(wl[n] - fractn*(dw[n]+fractn2*w6[n]))*fluxr[n];
      fluxe[n]=(el[n] - fractn*(de[n]+fractn2*e6[n]))*fluxr[n];
      for (int m = 0; m < smax; m++){
        fluxas[m][n]=(asl[m][n] - fractn*(das[m][n] + 
					fractn2*as6[m][n]))*fluxr[n];
      }
    }
  }

  // Advect mass, momentum, and energy by moving the subshell quantities 
  // into the appropriate Eulerian zone. ewnd*rho[n]
  // will reheat material which may not be wanted if we have cooling on.
  // Advect colour variable in the same fashion as rho - do colour first so that
  // dmu and dmu0 calculated from rho can be used for advection of u etc.

  for (int n = lg.nmin; n <= lg.nmax; n++){
    dm [n] = r[n] * dvol[n];
    dm0[n] = (dm[n] + fluxr[n] - fluxr[n+1]);
    r  [n] = dm0[n]/dvol0[n];
    r  [n] = max(dfloor,r[n]);
    dm0[n] = 1./(r[n]*dvol0[n]);
    for (int m = 0; m < smax; m++){
      as[m][n] = (as[m][n]*dm[n] + fluxas[m][n]-fluxas[m][n+1])*dm0[n];
    }
    
    u  [n] = (u[n]*dm[n] + fluxu[n]-fluxu[n+1])*dm0[n];
    v  [n] = (v[n]*dm[n] + fluxv[n]-fluxv[n+1])*dm0[n];
    w  [n] = (w[n]*dm[n] + fluxw[n]-fluxw[n+1])*dm0[n];
    e  [n] = (e[n]*dm[n] + fluxe[n]-fluxe[n+1])*dm0[n];
    p  [n] = gamm*r[n]*(e[n]-0.5*(u[n]*u[n] + v[n]*v[n] + w[n]*w[n]));

    //smallp = max((r[n]/avgmassamb)*boltzman*tamb,pfloor);
    //p[n] = max(p[n],smallp);

    // Set sensible limits for the advected scalars
    //as_check(as_ex);
    for (int m=0; m < smax; m++){
      as[m][n] = max(0.0,min(1.0,as[m][n]));
    }
      
  }

  pminFixUpArray(lg.nmin,lg.nmax);

  return;
}


