/*!  \file hydroUnsplit.cpp
 *   \brief Update the solution using a directionally unsplit method. 
 *
 *   See ~/work/stellarcluster/SSCwind_withIndividualSNe/ARWENv1.0
 *   for a more sophisticated version of this code, which deals
 *   with the grid being split across multiple processors by MPI. 
 *
 *   \author Julian Pittard (Original version 09.08.11)
 *   \version 0.1-development (Undómiel):
 *   \date Last modified: 09.08.11 (JMP)
 */

#undef MAIN

//  Header files:
#include <cmath>
#include <fstream>
#include <iostream>
#include "defs.h"
#include "constants.h"
#include "funcs.h"
#include "gas.h"
#include "global.h"
#include "grid.h"
#include "gridlet.h"
#include "physconst.h"

// Function declarations
void calculateAreas(int i, int j, int k);
void calculateFluxes(int i, int j, int k);
void calculateP(int i, int j, int k);
void calculatePstar(int i, int j, int k);
void calculateSourceTerms(int i, int j, int k);
void calculateVolumes(int i, int j, int k);
void constructGridlet(int i, int j, int k, double**** P);
void consToPrim(double c[], double p[]);
void primToCons(double p[], double c[]);
void setupGridletIndices();
void unsplit();
void updateUnsplit();
extern void barrier();
extern void fillGhostCells(double**** P);
extern void operatorSplit();
extern double pminFixUp(double avgmass, double tamb, double r, double p);
extern void quit();
extern void riemann(int direc, double pL[], double pR[], double flux[]);

using namespace std;

// Global arrays with file scope
double pL[iqmax];
double pR[iqmax];


/*!  \brief Update the fluid variables by advancing the solution.
 *
 *   A directionally unsplit update is performed on specific regions 
 *   (gridlets) on the grid.
 * 
 *   \author Julian Pittard (Original version 02.06.11)
 *   \version 0.1-development (Undómiel):
 *   \date Last modified: 09.08.11 (JMP)
 */
void updateUnsplit(){

#ifdef UNSPLIT
  setupGridletIndices();
  
  // Alternate operator split activities to achieve 2nd order in time
  if (ncycle%2 == 0){
    unsplit();
    operatorSplit();
  }
  else{
    operatorSplit();
    unsplit();
  }
  
  t       += dt;
  timep1d += dt;
  timep2d += dt;
  timep3d += dt;
#endif
  
  return;
}


#ifdef UNSPLIT
/*!  \brief Set the values of the gridlet indices, which are only
 *   visible to the functions in this file.
 *
 *   \author Julian Pittard (Original version 09.08.11)
 *   \version 0.1-development (Undmiel):
 *   \date Last modified: 09.08.11 (JMP)
 */
void setupGridletIndices(){
  gl.irs = nghost;
  gl.ire = nghost+ig;
  if (nd > 1){
    gl.jrs = nghost;
    gl.jre = nghost+jg;
  }
  else{
    gl.jrs = 0;
    gl.jre = 1;
  }
  if (nd == 3){
    gl.krs = nghost;
    gl.kre = nghost+kg;
  }
  else{
    gl.krs = 0;
    gl.kre = 1;
  }

  if (ncycle == 0){
    cout << "lg.irs jrs krs = " << lg.irs << " " << lg.jrs << " " << lg.krs << "\n";
    cout << "lg.ire jre kre = " << lg.ire << " " << lg.jre << " " << lg.kre << "\n";

    cout << " gridlet irs jrs krs = " << gl.irs << " " << gl.jrs << " " << gl.krs << "\n";
    cout << " gridlet ire jre kre = " << gl.ire << " " << gl.jre << " " << gl.kre << "\n";
  }
  //quit();
  return;
}


/*!  \brief Advance the hydro solution using a directionally unsplit scheme.
 *
 *   \author Julian Pittard (Original version 28.03.12)
 *   \version 1.0-stable (Evenstar):
 *   \date Last modified: 19.04.12 (JMP)
 */
void unsplit(){

  //cout << "unsplit: dt = " << dt << "\n";

  fillGhostCells(lg.P0);

  // Predictor step: calculates Pstar at t^(n+1)
  // Loop over gridlets
  for (int k = 0; k < nkg; ++k){
    for (int j = 0; j < njg; ++j){
      for (int i = 0; i < nig; ++i){
	constructGridlet(i,j,k,lg.P0); // the gridlet p is equal to the grid P0
	calculateVolumes(i,j,k);
	calculateAreas(i,j,k);
	calculateSourceTerms(i,j,k);
	calculateFluxes(i,j,k);
	calculatePstar(i,j,k);
      }
    }
  }
  
  fillGhostCells(lg.Pstar);

  // Corrector step: 
  // Loop over gridlets
  for (int k = 0; k < nkg; ++k){
    for (int j = 0; j < njg; ++j){
      for (int i = 0; i < nig; ++i){
	constructGridlet(i,j,k,lg.Pstar); // the gridlet p is equal to the grid Pstar
	calculateVolumes(i,j,k);
	calculateAreas(i,j,k);
	calculateSourceTerms(i,j,k);
	calculateFluxes(i,j,k);
	calculateP(i,j,k);
      }
    }
  }
  
  return;
}


/*!  \brief Construct gridlet.
 *
 *   Pgrid is passed in. This is either P0 (during the predictor step) or Pstar
 *   (during the corrector step).
 *
 *   The gridlet indices igg, jgg and kgg are passed in.
 *
 *   \author Julian Pittard (Original version 08.03.13)
 *   \version 1.0-stable (Evenstar):
 *   \date Last modified: 05.09.13 (JMP)
 */
void constructGridlet(int igg, int jgg, int kgg, double**** Pgrid){

  // Loop over all (real and boundary) gridlet cells, extracting
  // data from real and boundary cells on the local grid.
  // igg = gridlet index
  // ig = number of real cells in gridlet
  // ii = total number of cells in gridlet
  // il = index of cell on grid (which stores ghost and real cells)

  int kl,jl,il;
    
  for (int n = 0; n < iqmax; ++n){
    for (int k = 0; k < kk; ++k){
      kl = kgg*kg + k;
      for (int j = 0; j < jj; ++j){
	jl = jgg*jg + j;
	for (int i = 0; i < ii; ++i){
	  il = igg*ig + i;
	  gl.p[n][k][j][i] = Pgrid[n][kl][jl][il];
	}
      }
    }
  }

  //for (int i = 0; i < ii; ++i){
  //  cout << "Gridlet: i = " << i << "; p[iqe] = " << p[iqe][0][2][i] << "\n";
  //}
  //quit();
  
  return;
}


/*!  \brief Calculate the volume of cells on the gridlet.
 *
 *   The gridlet indices igg, jgg and kgg are passed in.
 *
 *   \author Julian Pittard (Original version 08.03.13)
 *   \version 1.0-stable (Evenstar):
 *   \date Last modified: 05.09.13 (JMP)
 */
void calculateVolumes(int igg, int jgg, int kgg){

  int gi,gj,gk;
  double r1,r2;

  // Default values
  double dx = 1.0;
  double dy = 1.0;
  double dz = 1.0;
  int j = 0;
  int k = 0;
    
  if (gg.geom == "XYZ"){
    if (nd == 1){
      for (int i = 0; i < ii; ++i){
	gi = igg*ig + i; // index of cell on grid
        dx = lg.zdx[gi];
	gl.vol[k][j][i] = dx;
      }
    }
    else if (nd == 1){
      for (int j = 0; j < jj; ++j){
	gj = jgg*jg + j; // index of cell on grid
	dy = lg.zdy[gj];
	for (int i = 0; i < ii; ++i){
	  gi = igg*ig + i; // index of cell on grid
	  dx = lg.zdx[gi];
	  gl.vol[k][j][i] = dx*dy;
	}
      }
    }
    else{
      for (int k = 0; k < kk; ++k){
        gk = kgg*kg + k; // index of cell on grid
        dz = lg.zdz[gk];
        for (int j = 0; j < jj; ++j){
	  gj = jgg*jg + j; // index of cell on grid
	  dy = lg.zdy[gj];
	  for (int i = 0; i < ii; ++i){
	    gi = igg*ig + i; // index of cell on grid
	    dx = lg.zdx[gi];
	    gl.vol[k][j][i] = dx*dy*dz;
	  }
        }
      }
    }
  }
  else if (gg.geom == "ZRP"){
    if (nd == 1){
      for (int i = 0; i < ii; ++i){
	gi = igg*ig + i; // index of cell on grid
        dx = lg.zdx[gi];
	gl.vol[k][j][i] = dx;
      }
    }
    else if (nd == 2){
      for (int j = 0; j < jj; ++j){
	gj = jgg*jg + j;   // index of cell on grid
	r1 = lg.zya[gj];   // inner radius of cell
	r2 = lg.zya[gj+1]; // outer radius
	for (int i = 0; i < ii; ++i){
	  gi = igg*ig + i; // index of cell on grid
	  dx = lg.zdx[gi];
	  gl.vol[k][j][i] = mabs(r2*r2 - r1*r1)*dx*pi;
	}
      }
    }
    else{
      cout << "calculateVolumes: check this!\n";
      quit();
      for (int k = 0; k < kk; ++k){
        gk = kgg*kg + k; // index of cell on grid
        dz = lg.zdz[gk];
        for (int j = 0; j < jj; ++j){
  	  gj = jgg*jg + j;   // index of cell on grid
	  r1 = lg.zya[gj];   // inner radius of cell
	  r2 = lg.zya[gj+1]; // outer radius
	  for (int i = 0; i < ii; ++i){
	    gi = igg*ig + i; // index of cell on grid
	    dx = lg.zdx[gi];
	    gl.vol[k][j][i] = mabs(r2*r2 - r1*r1)*dx*dz; // replace dz with dtheta?
	  }
	}
      }
    }
  }
  else if (gg.geom == "RTP"){
    // dy and dz should be fractions of pi if 2D/3D, and =1.0 if 1D
    if (nd == 1){
      for (int i = 0; i < ii; ++i){
	gi = igg*ig + i; // index of cell on grid
        r1 = lg.zxa[gi];    // inner radius of cell
	r2 = lg.zxa[gi+1];  // outer radius
	gl.vol[k][j][i] = 4.0*pi*mabs(r2*r2*r2 - r1*r1*r1)*dy*dz/3.0;
	//cout << "calcVol: i = " << i << "; vol = " << vol[k][j][i] << "\n"; 
      }
    }
    else if (nd == 2){
      for (int j = 0; j < jj; ++j){
	gj = jgg*jg + j; // index of cell on grid
	dy = lg.zdy[gj];
	for (int i = 0; i < ii; ++i){
	  gi = igg*ig + i; // index of cell on grid
	  r1 = lg.zxa[gi];    // inner radius of cell
	  r2 = lg.zxa[gi+1];  // outer radius
	  gl.vol[k][j][i] = 4.0*pi*mabs(r2*r2*r2 - r1*r1*r1)*dy*dz/3.0; // check this!
	}
      }
    }
    else{
      for (int k = 0; k < kk; ++k){
        gk = kgg*kg + k; // index of cell on grid
        dz = lg.zdz[gk];
        for (int j = 0; j < jj; ++j){
	  gj = jgg*jg + j; // index of cell on grid
	  dy = lg.zdy[gj];
	  for (int i = 0; i < ii; ++i){
	    gi = igg*ig + i; // index of cell on grid
	    r1 = lg.zxa[gi];    // inner radius of cell
	    r2 = lg.zxa[gi+1];  // outer radius
	    gl.vol[k][j][i] = 4.0*pi*mabs(r2*r2*r2 - r1*r1*r1)*dy*dz/3.0; // check this!
	  }
	}
      }
    }
  }

  //for (int k = 0; k < kk; ++k){
  //  for (int j = 0; j < jj; ++j){
  //    for (int i = 0; i < ii; ++i){
  //	cout << i << "\t" << j << "\t" << k << "\t" << gl.vol[k][j][i] << "\n";
  //    }
  //  }
  //}
  //for (int i = 0; i < 15; ++i){
  //   cout << i << "\t" << gl.vol[0][0][i] << "\n";
  //}
  //quit();
  return;
}


/*!  \brief Calculate the area of the faces of the cells on the gridlet.
 *
 *   The gridlet indices igg, jgg and kgg are passed in.
 *
 *   \author Julian Pittard (Original version 08.03.13)
 *   \version 1.0-stable (Evenstar):
 *   \date Last modified: 05.09.13 (JMP)
 */
void calculateAreas(int igg, int jgg, int kgg){

  int gi,gj,gk;
  double r1,r2;

  // Default values
  double dx = 1.0;
  double dy = 1.0;
  double dz = 1.0;
  int i = 0;
  int j = 0;
  int k = 0;
    
  if (gg.geom == "XYZ"){
    for (int k = 0; k < kk; ++k){
      gk = kgg*kg + k; // index of cell on grid
      dz = lg.zdz[gk];
      for (int j = 0; j < jj; ++j){
	gj = jgg*jg + j; // index of cell on grid
	dy = lg.zdy[gj];
	for (int i = 0; i < ii; ++i){
	  gi = igg*ig + i; // index of cell on grid
	  dx = lg.zdx[gi];
	  gl.area[0][k][j][i] = dy*dz; // face normal along x
	  gl.area[1][k][j][i] = dx*dz; // face normal along y
	  gl.area[2][k][j][i] = dx*dy; // face normal along z
	}
      }
    }
  }
  else if (gg.geom == "ZRP"){
    if (nd == 1){
      for (int i = 0; i < ii; ++i){
	gi = igg*ig + i; // index of cell on grid
        dx = lg.zdx[gi];
	gl.area[0][k][j][i] = 1.0;
      }
    }
    else if (nd == 2){
      for (int j = 0; j < jj; ++j){
	gj = jgg*jg + j;   // index of cell on grid
	r1 = lg.zya[gj];   // inner radius of cell
	r2 = lg.zya[gj+1]; // outer radius
	for (int i = 0; i < ii; ++i){
	  gi = igg*ig + i; // index of cell on grid
	  dx = lg.zdx[gi];
	  gl.area[0][k][j][i] = mabs((r2*r2 - r1*r1)*pi);
	  gl.area[1][k][j][i] = 2.0*pi*mabs(r1)*dx; // face normal along "r" (dx is actually dz)
	}
      }
    }
    else{
      cout << "calculateAreas: check this!\n";
      quit();
      for (int k = 0; k < kk; ++k){
        gk = kgg*kg + k; // index of cell on grid
        dz = lg.zdz[gk];
        for (int j = 0; j < jj; ++j){
	  gj = jgg*jg + j;   // index of cell on grid
	  r1 = lg.zya[gj];   // inner radius of cell
	  r2 = lg.zya[gj+1]; // outer radius
	  for (int i = 0; i < ii; ++i){
	    gi = igg*ig + i; // index of cell on grid
	    dx = lg.zdx[gi];
	    gl.area[0][k][j][i] = mabs((r2*r2 - r1*r1)*pi*dz); // dz should be fraction of phi
	    gl.area[1][k][j][i] = 2.0*pi*mabs(r1)*dx*dz; // face normal along "r" (dx is actually dz) (dz should be fraction of phi)
	    gl.area[2][k][j][i] = -1.0; // Need to work out still!!!!
 	  }
	}
      }
    }
  }
  else if (gg.geom == "RTP"){
    if (nd == 1){
      for (int i = 0; i < ii; ++i){
	gi = igg*ig + i; // index of cell on grid
	r1 = lg.zxa[gi];    // inner radius of cell
	gl.area[0][k][j][i] = 4.0*pi*r1*r1;
      }
    }
    else if (nd == 2){
      for (int j = 0; j < jj; ++j){
	gj = jgg*jg + j; // index of cell on grid
	dy = lg.zdy[gj];
	for (int i = 0; i < ii; ++i){
	  gi = igg*ig + i; // index of cell on grid
	  r1 = lg.zxa[gi];    // inner radius of cell
	  r2 = lg.zxa[gi+1];  // outer radius
	  gl.area[0][k][j][i] = -1.0; // STILL TO WORK OUT! dy and dz should be fractions of pi if 2D/3D, and = 1.0 if 1D
	  gl.area[1][k][j][i] = -1.0; // STILL TO WORK OUT! dy and dz should be fractions of pi if 2D/3D, and = 1.0 if 1D
        }
      }
    }
    else{
      for (int k = 0; k < kk; ++k){
        gk = kgg*kg + k; // index of cell on grid
        dz = lg.zdz[gk];
        for (int j = 0; j < jj; ++j){
	  gj = jgg*jg + j; // index of cell on grid
	  dy = lg.zdy[gj];
	  for (int i = 0; i < ii; ++i){
	    gi = igg*ig + i; // index of cell on grid
	    r1 = lg.zxa[gi];    // inner radius of cell
	    r2 = lg.zxa[gi+1];  // outer radius
	    gl.area[0][k][j][i] = -1.0; // STILL TO WORK OUT! dy and dz should be fractions of pi if 2D/3D, and = 1.0 if 1D
	    gl.area[1][k][j][i] = -1.0; // STILL TO WORK OUT!
	    gl.area[2][k][j][i] = -1.0; // STILL TO WORK OUT!
	  }
	}
      }
    }
    if (nd > 1){
      cout << "calculateAreas: check this too!\n";
      quit();
    }
  }
  
  //for (int k = 0; k < kk; ++k){
  //  for (int j = 0; j < jj; ++j){
  //    for (int i = 0; i < ii; ++i){
  //	cout << i << "\t" << j << "\t" << k << "\t" << gl.area[0][k][j][i] << "\t" << gl.area[1][k][j][i] << "\t" << gl.area[2][k][j][i] << "\n";
  //    }
  //  }
  //}
  //quit();
  return;
}


/*!  \brief Calculate the source terms.
 *
 *   The gridlet indices igg, jgg and kgg are passed in.
 *
 *   Note that the spatial variation of geometric source term's
 *   across the cell has to be taken account of. See the Appendix in 
 *   Falle (1991, MN, 250, 581) and p85-86 of Hydro Book III for
 *   further details.
 *
 *   \author Julian Pittard (Original version 08.03.13)
 *   \version 1.0-stable (Evenstar):
 *   \date Last modified: 05.09.13 (JMP)
 */
void calculateSourceTerms(int igg, int jgg, int kgg){

  int gi,gj,gk,dicell;
  double rg,rc,dr,fac,fac2,pre,difflp,diffrp,dpdr;
    
  // Calculate the geometric source term for each real cell on the gridlet.
  // Initialize...
  for (int n = 0; n < iqmax; ++n){
    for (int k = gl.krs; k < gl.kre; ++k){
      for (int j = gl.jrs; j < gl.jre; ++j){
	for (int i = gl.irs; i < gl.ire; ++i){
	  gl.s[n][k][j][i] = 0.0;
	}
      }
    }
  }

  if (gg.geom == "ZRP"){
    for (int k = gl.krs; k < gl.kre; ++k){
      for (int j = gl.jrs; j < gl.jre; ++j){
	gj = jgg*jg + j; // j index on the local grid
	rc = lg.zyg[gj]; // Cell COM r-coordinate
	dr = lg.zdy[gj];
	dicell = gj-nghost+1 + jpro*lgrid::jmax; // j-index such that first real cell has index of 1 (to match Eq. A14 in the mg method paper) - see p85 of Hydro Book III
	fac = 2.0/(2.0*dicell - 1.0);
	fac2 = dr*(float(dicell) - 0.5) - rc;
	for (int i = gl.irs; i < gl.ire; ++i){
	  pre = gl.p[iqe][k][j][i]; // initial pressure
	  difflp = (gl.p[iqe][k][j][i] - gl.p[iqe][k][j-1][i])/(rc - lg.zyg[gj-1]);
	  diffrp = (gl.p[iqe][k][j+1][i] - gl.p[iqe][k][j][i])/(lg.zyg[gj+1] - rc);
	  dpdr = av(diffrp, difflp);
	  gl.s[iqu0+1][k][j][i] += (fac/dr)*(pre + fac2*dpdr); // r-mtm geometric source term in ZRP geometry
	}
      }
    }
  }
  else if (gg.geom == "RTP"){
    for (int k = gl.krs; k < gl.kre; ++k){
      for (int j = gl.jrs; j < gl.jre; ++j){
	for (int i = gl.irs; i < gl.ire; ++i){
	  pre = gl.p[iqe][k][j][i]; // initial pressure
	  gi = igg*ig + i; // i index on the local grid
	  rg = lg.zxg[gi]; // Cell COM r-coordinate
	  rc = lg.zxc[gi]; // Cell centre r-coordinate
	  dr = lg.zdx[gi];
	  dicell = gi-nghost+1 + ipro*lgrid::imax; // i-index such that first real cell has index of 1 (to match Eq. A14 in the mg method paper) - see p85/86 of Hydro Book III
	  fac = 3.0*(2.0*dicell - 1.0)/(3.0*dicell*dicell - 3.0*dicell + 1.0);
	  fac2 = dr*(dicell - 0.5) - rg;
	  difflp = (gl.p[iqe][k][j][i] - gl.p[iqe][k][j][i-1])/(rg - lg.zxg[gi-1]);
	  diffrp = (gl.p[iqe][k][j][i+1] - gl.p[iqe][k][j][i])/(lg.zxg[gi+1] - rg);
	  dpdr = av(diffrp, difflp);
	  gl.s[iqu0][k][j][i] += (fac/dr)*(pre + fac2*dpdr); // r-mtm geometric source term in RTP geometry 
          //if (i < 3){
          //  cout << "calcSourceTerms: i = " << i << "; gl.s    (dup) = " << gl.s[iqd][k][j][i] << " " << gl.s[iqu0][k][j][i] << " " << gl.s[iqe][k][j][i] << "\n";
	  //}
	}
      }
    }
  }
  
  //for (int k = gl.krs; k < gl.kre; ++k){
  //  for (int j = gl.jrs; j < gl.jre; ++j){
  //    for (int i = gl.irs; i < gl.ire; ++i){
  //	cout << i << "\t" << j << "\t" << k << "\t" << gl.s[iqd][k][j][i] << "\t" << gl.s[iqu0+1][k][j][i] << "\t" << gl.s[iqe][k][j][i] << "\n";
  //    }
  //  }
  //}
  //quit();


  return;
}


/*!  \brief Calculate the fluxes.
 *
 *   The gridlet indices igg, jgg and kgg are passed in.
 *
 *   The following procedure occurs:
 *        i)   Calculate the average spatial derivative (monotonicity preserved
 *             using Eq. A18 from Falle 1991, MN, 250, 581) in each direction
 *             (store in the file scope array diff)
 *        ii)  Using diff, calculate the L and R states at each face (pL and pR)
 *        iii) Calculate the fluxes from the Riemann solver (store in the
 *             file scope array flux)
 *
 *   The value stored in flux[0][kg][jg][ig] is the flux across the lower-face 
 *   of cell[kg][jg][ig] (i.e. between cells [ig-1] and [i]).
 *
 *   It makes sense to perform the flux calculations one direction at a time, since
 *   this avoids nasty conditional statements within nested loops. Because of this,
 *   it also made sense to calculate diffl and diffr directly in the loops (a function
 *   call would have reintroduced a conditional statement within the loops).
 *   [UPDATE: In fact I don't think this is the case - probably I could do something
 *   more like Sam's mg code where diffl/r are actually gradients rather than just
 *   differences].
 *
 *   To avoid mass building up on symmetry axes (e.g the r=0 axis in ZRP geometry),
 *   two things are required:
 *     1) The initial conditions/winds must be set up using the cell centre-of-mass coordinates.
 *     2) 2nd order spatial derivatives must be calculated using the cell COM coords.
 *   Note that problems still occur if only one of the above conditions is met. 
 * 
 *   NOTE: See p187 of Hydro Book II.
 *
 *   \author Julian Pittard (Original version 08.03.13)
 *   \version 1.0-stable (Evenstar):
 *   \date Last modified: 05.09.13 (JMP)
 */
void calculateFluxes(int igg, int jgg, int kgg){

  int id;
  int igmin,igmax,igmin2,jgmin,kgmin,jgmin2,kgmin2,jgmax,kgmax;
  int gi,gj,gk;
  double diffl,diffr;
  double xint,yint,zint,fracL,fracR;
    
  // Initialize diff and flux
  for (int id = 0; id < nd; ++id){
    for (int k = 0; k < kk; ++k){
      for (int j = 0; j < jj; ++j){
	for (int i = 0; i < ii; ++i){
          for (int n = 0; n < iqmax; ++n){
	    gl.diff[id][k][j][i][n] = 0.0;
	    gl.flux[id][k][j][i][n] = 0.0;
	  }
	}
      }
    }
  }

  // Use a piecewise linear interpolation to determine pL and pR, using the same
  // code as in mg. Do not make any distinction between the 1st and 2nd parts of the timestep.

  // Set default loop ranges 
  igmin = gl.irs-1;
  igmax = gl.ire+1;
  igmin2 = gl.irs;
  jgmin = 0;
  kgmin = 0;
  jgmin2 = 0;
  kgmin2 = 0;
  jgmax = 1;
  kgmax = 1;
  if (nd > 1){
    jgmin = gl.jrs-1;
    jgmin2 = gl.jrs;
    jgmax = gl.jre+1;
  }
  if (nd == 3){
    kgmin = gl.krs-1;
    kgmin2 = gl.krs;
    kgmax = gl.kre+1;
  }

  // For each direction, calculate the interpolated L/R state, then
  // pass to the Riemann solver to determine the fluxes.
  // The value in flux[id][k][j][i] is the net flux through the lower faces of cell[k,j,i] 

  // 1) Direction "0"

  id = 0;
  
  // Calculate the average gradient between cells. diff needs to be calculated
  // in the ghost zones adjacent to the real zones.
  for (int k = kgmin; k < kgmax; ++k){
    for (int j = jgmin; j < jgmax; ++j){
      for (int i = igmin; i < igmax; ++i){
	// Calculate the left and right differences, and their monotonicity preserving average
        for (int n = 0; n < iqmax; ++n){
	  diffl = gl.p[n][k][j][i] - gl.p[n][k][j][i-1];
	  diffr = gl.p[n][k][j][i+1] - gl.p[n][k][j][i];
	  gl.diff[id][k][j][i][n] = av(diffl,diffr);
	}
      }
    }
  }

  // Calculate the L and R states at each real face.
  // pL is the interpolated value of p at the left side
  //            of the lower face of cell k,j,i
  // pL is the interpolated value of p at the right side
  //            of the lower face of cell k,j,i
  // Fluxes need to be calculated for faces irs to ire+1

  for (int k = kgmin2; k < kgmax; ++k){
    for (int j = jgmin2; j < jgmax; ++j){
      for (int i = igmin2; i < igmax; ++i){
	gi = igg*ig + i;    // i index on the grid
	xint = lg.zxa[gi];  // x-coord of left face of cell i on gridlet
	fracL = (xint - lg.zxg[gi-1])/lg.zdx[gi-1];
	fracR = (lg.zxg[gi] - xint)/lg.zdx[gi];
        for (int n = 0; n < iqmax; ++n){
	  pL[n] = gl.p[n][k][j][i-1] + fracL*gl.diff[id][k][j][i-1][n];
	  pR[n] = gl.p[n][k][j][i] - fracR*gl.diff[id][k][j][i][n];
	}
	riemann(id,pL,pR,gl.flux[id][k][j][i]);
      }
    }
  }

  // 2) Direction "1"

  if (nd > 1){
    id = 1;
    // Calculate the average gradient between cells
    for (int k = kgmin; k < kgmax; ++k){
      for (int j = jgmin; j < jgmax; ++j){
        for (int i = igmin; i < igmax; ++i){
	  // Calculate the left and right differences, and their monotonicity preserving average
          for (int n = 0; n < iqmax; ++n){
	    diffl = gl.p[n][k][j][i] - gl.p[n][k][j-1][i];
	    diffr = gl.p[n][k][j+1][i] - gl.p[n][k][j][i];
	    gl.diff[id][k][j][i][n] = av(diffl,diffr);
	  }
        }
      }
    }

    // Calculate the L and R states at each real face.
    // pL is the interpolated value of p at the left side
    //            of the lower face of cell k,j,i
    // pL is the interpolated value of p at the right side
    //            of the lower face of cell k,j,i
    // Fluxes need to be calculated for faces irs to ire+1

    for (int k = kgmin2; k < kgmax; ++k){
      for (int j = jgmin2; j < jgmax; ++j){
	gj = jgg*jg + j;    // j index on the grid
	yint = lg.zya[gj];  // y-coord of left face of cell j on gridlet
	fracL = (yint - lg.zyg[gj-1])/lg.zdy[gj-1];
	fracR = (lg.zyg[gj] - yint)/lg.zdy[gj];
        for (int i = igmin2; i < igmax; ++i){
          for (int n = 0; n < iqmax; ++n){
	    pL[n] = gl.p[n][k][j-1][i] + fracL*gl.diff[id][k][j-1][i][n];
	    pR[n] = gl.p[n][k][j][i] - fracR*gl.diff[id][k][j][i][n];
	  }
  	  riemann(id,pL,pR,gl.flux[id][k][j][i]);
	}
      }
    }
  }

  
  // 3) Direction "2"

  if (nd == 3){
    id = 2;
    // Calculate the average gradient between cells
    for (int k = kgmin; k < kgmax; ++k){
      for (int j = jgmin; j < jgmax; ++j){
        for (int i = igmin; i < igmax; ++i){
	  // Calculate the left and right differences, and their monotonicity preserving average
          for (int n = 0; n < iqmax; ++n){
	    diffl = gl.p[n][k][j][i] - gl.p[n][k-1][j][i];
	    diffr = gl.p[n][k+1][j][i] - gl.p[n][k][j][i];
	    gl.diff[id][k][j][i][n] = av(diffl,diffr);
	  }
        }
      }
    }

    // Calculate the L and R states at each real face.
    // pL is the interpolated value of p at the left side
    //            of the lower face of cell k,j,i
    // pL is the interpolated value of p at the right side
    //            of the lower face of cell k,j,i
    // Fluxes need to be calculated for faces irs to ire+1

    for (int k = kgmin2; k < kgmax; ++k){
      gk = kgg*kg + k;    // k index on the grid
      zint = lg.zza[gk];  // z-coord of left face of cell k on gridlet
      fracL = (zint - lg.zzg[gk-1])/lg.zdz[gk-1];
      fracR = (lg.zzg[gk] - zint)/lg.zdz[gk];
      for (int j = jgmin2; j < jgmax; ++j){
        for (int i = igmin2; i < igmax; ++i){
          for (int n = 0; n < iqmax; ++n){
	    pL[n] = gl.p[n][k-1][j][i] + fracL*gl.diff[id][k-1][j][i][n];
	    pR[n] = gl.p[n][k][j][i] - fracR*gl.diff[id][k][j][i][n];
	  }
  	  riemann(id,pL,pR,gl.flux[id][k][j][i]);
	}
      }
    }
  }
  
  //for (int k = kgmin2; k < kgmax; ++k){
    //for (int j = jgmin2; j < jgmax; ++j){
  //  for (int j = 25; j < 26; ++j){
  //    for (int i = igmin2; i < igmax; ++i){
  //	cout << k << "\t" << j << "\t" << i << "\t" << gl.flux[0][k][j][i][iqd] << "\t" << gl.flux[0][k][j][i][iqu0] << "\t" << gl.flux[0][k][j][i][iqu0+1] << "\t" << gl.flux[0][k][j][i][iqe] << "\t" << gl.flux[1][k][j][i][iqd] << "\t" << gl.flux[1][k][j][i][iqu0] << "\t" << gl.flux[1][k][j][i][iqu0+1] << "\t" << gl.flux[1][k][j][i][iqe] << "\n";
  //    }
  //  }
  //}
  //quit();

  return;
}


/*!  \brief Calculate the intermediate state Pstar.
 *
 *   The gridlet indices igg, jgg and kgg are passed in.
 *   Update the real cells. flux has correct values for 
 *   i=gl.irs to i=gl.ire, inclusive.
 *   
 *   \author Julian Pittard (Original version 08.03.13)
 *   \version 1.0-stable (Evenstar):
 *   \date Last modified: 05.09.13 (JMP)
 */
void calculatePstar(int igg, int jgg, int kgg){

  int gi,gj,gk;
  int igmin,jgmin,kgmin,igmax,jgmax,kgmax;
  double dtvol,area1,area2;
  double corig[iqmax];
  double cstar[iqmax];
  double porig[iqmax];
  double pstar[iqmax];

  // Pstar is calculated at t^(n+1) = t^(n) + dt
  double delt = dt;
  //cout << "calcPstar: dt = " << dt << "\n";
  
  // Set default loop ranges. irs -> ire-1 inclusive will update only the real cells on the gridlet
    
  igmin = gl.irs;
  igmax = gl.ire;
  jgmin = 0;
  kgmin = 0;
  jgmax = 1;
  kgmax = 1;
  if (nd > 1){
    jgmin = gl.jrs;
    jgmax = gl.jre;
  }
  if (nd == 3){
    kgmin = gl.krs;
    kgmax = gl.kre;
  }

  for (int k = kgmin; k < kgmax; ++k){
    gk = kgg*kg + k; // k index on the grid
    for (int j = jgmin; j < jgmax; ++j){
      gj = jgg*jg + j; // j index on the grid
      for (int i = igmin; i < igmax; ++i){
	gi = igg*ig + i; // i index on the grid
	dtvol = delt/gl.vol[k][j][i];

	for (int n = 0; n < iqmax; ++n){
	  porig[n] = gl.p[n][k][j][i];
	}
	primToCons(porig,corig);  // calculate the conservative variables
	for (int n = 0; n < iqmax; ++n){
	  cstar[n] = corig[n];
	}
                
	// Multiply fluxes by cell areas
	// For cell[k,j,i], the net fluxes through the left and right faces
	// in direction "0" are (+flux[id][k][j][i] and -flux[id][k][j][i+1]). 
	// The left and right faces have an area of area[id][k][j][i] and area[id][k][j][i+1].

	// Direction "0"
	area1 = gl.area[0][k][j][i];   // area of left face 
	area2 = gl.area[0][k][j][i+1]; // area of right face 

	//if (i < 3){
	//  cout << "calcPstar: ijk = " << i << " " << j << " " << k << "; vol = " << gl.vol[k][j][i] << "; delt = " << delt << "\n";
	//  cout << "calcPstar: i = " << i << "; cstarorig(dup) = " << cstar[iqd] << " " << cstar[iqu0] << " " << cstar[iqe] << "\n";
	//}
	
	for (int n = 0; n < iqmax; ++n){
	  cstar[n] += (gl.flux[0][k][j][i][n]*area1 - gl.flux[0][k][j][i+1][n]*area2)*dtvol;
	}

	//if (i < 3){
        //  cout << "calcPstar: i = " << i << "; flux(dup) = " << gl.flux[0][k][j][i][iqd] << " " << gl.flux[0][k][j][i][iqu0] << " " << gl.flux[0][k][j][i][iqe] << "\n";
	//  cout << "calcPstar: i = " << i << "; flux(dup)+1 = " << gl.flux[0][k][j][i+1][iqd] << " " << gl.flux[0][k][j][i+1][iqu0] << " " << gl.flux[0][k][j][i+1][iqe] << "\n";
	//  cout << "calcPstar: i = " << i << "; area1 = " << area1 << "; area2 = " << area2 << "; dtvol = " << dtvol << "\n";
	//  cout << "calcPstar: i = " << i << "; cstar(dup) = " << cstar[iqd] << " " << cstar[iqu0] << " " << cstar[iqe] << "\n";
	//}

	if (nd > 1){
	  // Direction "1"
	  area1 = gl.area[1][k][j][i];   // area of left face 
	  area2 = gl.area[1][k][j+1][i]; // area of right face 

  	  for (int n = 0; n < iqmax; ++n){
	    cstar[n] += (gl.flux[1][k][j][i][n]*area1 - gl.flux[1][k][j+1][i][n]*area2)*dtvol;
	  }
	}

	if (nd == 3){
	  // Direction "2"
	  area1 = gl.area[2][k][j][i];   // area of left face 
	  area2 = gl.area[2][k+1][j][i]; // area of right face 

  	  for (int n = 0; n < iqmax; ++n){
	    cstar[n] += (gl.flux[2][k][j][i][n]*area1 - gl.flux[2][k+1][j][i][n]*area2)*dtvol;
	  }
	}
	
	// Add source terms
	for (int n = 0; n < iqmax; ++n){
	  cstar[n] += delt*gl.s[n][k][j][i];
	}

	//if (i < 3){
        //  cout << "calcPstar: i = " << i << "; s    (dup) = " << gl.s[iqd][k][j][i] << " " << gl.s[iqu0][k][j][i] << " " << gl.s[iqe][k][j][i] << "\n";
	//  cout << "calcPstar: i = " << i << "; cstar(dup) = " << cstar[iqd] << " " << cstar[iqu0] << " " << cstar[iqe] << "\n";
	//}

	// Now calculate Pstar on the grid
	consToPrim(cstar,pstar);
  	for (int n = 0; n < iqmax; ++n){
	  lg.Pstar[n][gk][gj][gi] = pstar[n];
	}

	//if (i < 3){
	//  cout << "calcPstar: i = " << i << "; pstar(dup) = " << pstar[iqd] << " " << pstar[iqu0] << " " << pstar[iqe] << "\n";
	//}

      }
    }
  }
  //quit();

  //for (int k = kgmin; k < kgmax; ++k){
  //  gk = kgg*kg + k; // k index on the grid
    //for (int j = jgmin; j < jgmax; ++j){
  //  for (int j = 25; j < 26; ++j){
  //    gj = jgg*jg + j; // j index on the grid
  //    for (int i = igmin; i < igmax; ++i){
  //	gi = igg*ig + i; // i index on the grid
  //	cout << i << "\t" << j << "\t" << k << "\t" << lg.Pstar[iqe][gk][gj][gi] << "\n";
  //  }
  //}
  //}
  //quit();
  return;
}


/*!  \brief Calculate the updated state P at time t + dt.
 *
 *   The gridlet indices igg, jgg and kgg are passed in.
 *   Update the real cells. flux has correct values for i=irs to i=ire, inclusive.
 *   
 *   \author Julian Pittard (Original version 08.03.13)
 *   \version 1.0-stable (Evenstar):
 *   \date Last modified: 05.09.13 (JMP)
 */
void calculateP(int igg, int jgg, int kgg){

  int gi,gj,gk;
  int igmin,jgmin,kgmin,igmax,jgmax,kgmax;
  double dtvol,area1,area2;
  double c[iqmax];
  double cstar[iqmax];
  double cnew[iqmax];
  double gridP[iqmax];
  double pstar[iqmax];
  double pnew[iqmax];

  // P is calculated by using 0.5*dt
  double delt = 0.5*dt;
        
  // Set default loop ranges. irs -> ire-1 inclusive will update only the real cells on the gridlet
    
  igmin = gl.irs;
  igmax = gl.ire;
  jgmin = 0;
  kgmin = 0;
  jgmax = 1;
  kgmax = 1;
  if (nd > 1){
    jgmin = gl.jrs;
    jgmax = gl.jre;
  }
  if (nd == 3){
    kgmin = gl.krs;
    kgmax = gl.kre;
  }

  for (int k = kgmin; k < kgmax; ++k){
    gk = kgg*kg + k; // k index on the grid
    for (int j = jgmin; j < jgmax; ++j){
      gj = jgg*jg + j; // j index on the grid
      for (int i = igmin; i < igmax; ++i){
	gi = igg*ig + i; // i index on the grid
	dtvol = delt/gl.vol[k][j][i];

	for (int n = 0; n < iqmax; ++n){
	  pstar[n] = gl.p[n][k][j][i];
	}
	primToCons(pstar,cstar);  // conservative variables from Pstar at t^(n+1)
	for (int n = 0; n < iqmax; ++n){
          gridP[n] = lg.P0[n][gk][gj][gi];
	}
	primToCons(gridP,c);      // conservative variables from P0 at t^(n)

	for (int n = 0; n < iqmax; ++n){
	  cnew[n] = 0.5*(cstar[n] + c[n]);
	}
                
	//if (i < 3){
	//  cout << "calcP: ijk = " << i << " " << j << " " << k << "; vol = " << vol[k][j][i] << "; delt = " << delt << "\n";
	//  cout << "calcP: i = " << i << "; pstar(dup) = " << pstar[iqd] << " " << pstar[iqu0] << " " << pstar[iqe] << "\n";
	//  cout << "calcP: i = " << i << "; cstar(dup) = " << cstar[iqd] << " " << cstar[iqu0] << " " << cstar[iqe] << "\n";
	//  cout << "calcP: i = " << i << "; c(dup) = " << c[iqd] << " " << c[iqu0] << " " << c[iqe] << "\n";
	//  cout << "calcP: i = " << i << "; cnew(dup) = " << cnew[iqd] << " " << cnew[iqu0] << " " << cnew[iqe] << "\n";
	//}

	// Multiply fluxes by cell areas
	// For cell[k,j,i], the net fluxes through the left and right faces
	// in direction "0" are (+flux[id][k][j][i] and -flux[id][k][j][i+1]). 
	// The left and right faces have an area of area[id][k][j][i] and area[id][k][j][i+1].

	// Direction "0"
	area1 = gl.area[0][k][j][i];   // area of left face 
	area2 = gl.area[0][k][j][i+1]; // area of right face 

	for (int n = 0; n < iqmax; ++n){
	  cnew[n] += (gl.flux[0][k][j][i][n]*area1 - gl.flux[0][k][j][i+1][n]*area2)*dtvol;
	}

	//if (i < 3){
        //  cout << "calcP: i = " << i << "; flux(dup) = " << gl.flux[0][k][j][i][iqd] << " " << gl.flux[0][k][j][i][iqu0] << " " << gl.flux[0][k][j][i][iqe] << "\n";
	//  cout << "calcP: i = " << i << "; flux(dup)+1 = " << gl.flux[0][k][j][i+1][iqd] << " " << gl.flux[0][k][j][i+1][iqu0] << " " << gl.flux[0][k][j][i+1][iqe] << "\n";
	//  cout << "calcP: i = " << i << "; area1 = " << area1 << "; area2 = " << area2 << "; dtvol = " << dtvol << "\n";
	//  cout << "calcP: i = " << i << "; cnew(dup) = " << cnew[iqd] << " " << cnew[iqu0] << " " << cnew[iqe] << "\n";
	//}

	if (nd > 1){
	  // Direction "1"
	  area1 = gl.area[1][k][j][i];   // area of left face 
	  area2 = gl.area[1][k][j+1][i]; // area of right face 

  	  for (int n = 0; n < iqmax; ++n){
	    cnew[n] += (gl.flux[1][k][j][i][n]*area1 - gl.flux[1][k][j+1][i][n]*area2)*dtvol;
	  }
	}

	if (nd == 3){
	  // Direction "2"
	  area1 = gl.area[2][k][j][i];   // area of left face 
	  area2 = gl.area[2][k+1][j][i]; // area of right face 

  	  for (int n = 0; n < iqmax; ++n){
	    cnew[n] += (gl.flux[2][k][j][i][n]*area1 - gl.flux[2][k+1][j][i][n]*area2)*dtvol;
	  }
	}
	
	// Add source terms
	for (int n = 0; n < iqmax; ++n){
	  cnew[n] += delt*gl.s[n][k][j][i];
	}

	// Now calculate the new P on the grid, at t^(n+1)
	consToPrim(cnew,pnew);
  	for (int n = 0; n < iqmax; ++n){
	  lg.P0[n][gk][gj][gi] = pnew[n];
	}

	//if (i < 3){
        //  cout << "calcP: i = " << i << "; s   (dup) = " << gl.s[iqd][k][j][i] << " " << gl.s[iqu0][k][j][i] << " " << gl.s[iqe][k][j][i] << " " << "\n";
	//  cout << "calcP: i = " << i << "; cnew(dup) = " << cnew[iqd] << " " << cnew[iqu0] << " " << cnew[iqe] << "\n";
	//  cout << "calcP: i = " << i << "; pnew(dup) = " << pnew[iqd] << " " << pnew[iqu0] << " " << pnew[iqe] << "\n";
	//}

      }
    }
  }
  //quit();
  
  //for (int k = kgmin; k < kgmax; ++k){
  //  gk = kgg*kg + k; // k index on the grid
    //for (int j = jgmin; j < jgmax; ++j){
  //  for (int j = 25; j < 26; ++j){
  //    gj = jgg*jg + j; // j index on the grid
  //    for (int i = igmin; i < igmax; ++i){
  //	gi = igg*ig + i; // i index on the grid
  //	cout << i << "\t" << j << "\t" << k << "\t" << lg.P0[iqe][gk][gj][gi] << "\n";
  //    }
  //  }
  //}
  //quit();
  return;
}
#endif


/*!  \brief Convert an array of primitives to conserved values.
 *
 *   \author Julian Pittard (Original version 08.03.13)
 *   \version 1.0-stable (Evenstar):
 *   \date Last modified: 05.09.13 (JMP)
 */
void primToCons(double p[iqmax], double c[iqmax]){

  c[iqd] = p[iqd];
  c[iqe] = 0.0;
  for (int n = 0; n < nd; ++n){
    c[iqu0+n] = c[iqd]*p[iqu0+n];
    c[iqe] += c[iqu0+n]*p[iqu0+n];
  }
  c[iqe] *= 0.5;
  c[iqe] += p[iqe]/gamm;

  for (int n = 0; n < smax; ++n){
    c[iqal0+n] = p[iqd]*p[iqal0+n];
  }
  
  return;
}

/*!  \brief Convert an array of conservatives to primitive values.
 *
 *   \author Julian Pittard (Original version 08.03.13)
 *   \version 1.0-stable (Evenstar):
 *   \date Last modified: 05.09.13 (JMP)
 */
void consToPrim(double c[iqmax], double p[iqmax]){

  double u,pre;
  double ke = 0.0;
  double rho = c[iqd];
    
  p[iqd] = rho;
  for (int n = 0; n < nd; ++n){
    u = c[iqu0+n]/p[iqd];
    p[iqu0+n] = u;
    ke += rho*u*u;
  }
  ke *= 0.5;
  pre = gamm*(c[iqe] - ke);
  p[iqe] = pre;

  for (int n = 0; n < smax; ++n){
    p[iqal0+n] = c[iqal0+n]/rho;
  }

  // Use additional criteria to stop p going too small. This is problem specific.
  p[iqe] = pminFixUp(avgmassamb,tamb,p[iqd],p[iqe]);
  return;
}


