/*!  \file  massLoad.cpp
 *   \brief Contains functions needed to compute mass-loading.
 *
 *   \author Julian Pittard (Original version 17.11.17)
 *   \version 0.1-development (Undómiel):
 *   \date Last modified: 16.06.11 (JMP)
 */

#undef MAIN

//  Header files:
#include <cstdlib>
#include <cmath>
#include <iomanip> 
#include <iostream>
#include <vector>
#include "defs.h"
#include "array.h"
#include "constants.h"
#include "funcs.h"
#include "gas.h"
#include "global.h"
#include "grid.h"
#include "physconst.h"
#include "star.h"

using namespace std;

// File scope variables

// Function declarations
void massLoad();
extern double findForwardShock();
extern void quit();

#ifdef MASSLOAD

// Mass-load the WBB/SNR. Mass-loading can follow one of several prescriptions:
//   1) Constant rate case:           rhodot = qML          
//   2) Mach-number dependent rate:   rhodot = qML * (M/(10 + 25/M^2)) - see "WBB" book p155
// Mass-loading occurs outside of the remapradius, and inside of the forward shock, for both
// the WBB and SNR problems.
void massLoad(){
  double rho,pre,vel,c,mach,fac,rc,drho;
  double rho_old,v_old,ke_old,E_old,temp_old,z_old,rho_dust_old;
  double rho_new,v_new,ke_new,pre_new,temp_new,z_new,rho_dust_new;
  double rho_wind_old,rho_amb_old,ze_old,za_old,ze_new,za_new;
  double rho_avail,rho_inj;
  
  const int cas = 1; // specify the desired case

  // Find forward shock
  double rFS = findForwardShock();
  
  // Inject mass due to mass-loading
  for (int k = lg.krs; k <= lg.kre; k++){
    for (int j = lg.jrs; j <= lg.jre; j++){
      for (int i = lg.irs; i <= lg.ire; i++){
        //cout << "i = " << i << "; j = " << j << "; k = " << k << "\n";
	rho = lg.P0[iqd][k][j][i];       
	pre = lg.P0[iqe][k][j][i];       
	vel = lg.P0[iqu0][k][j][i];       
	c = sqrt(gam*pre/rho);
	mach = vel/c;

	rc = lg.zxg[i]; //- stars[0].xpos;
        if ((rc > remapRadius)&&(rc < rFS)){
  	  if (cas == 1){
	    fac = 1.0;
	  }
	  else if (cas == 2){
	    fac = mach/(10.0 + 25.0/(mach*mach));
	  }
	  drho = fac*fML*q0; // potential mass injection rate
	  // Add mass to the gas and adjust the dust-to-gas mass fraction.
	  // The amount of mass added is subject to there being enough left in
	  // the "reservoir".
	  rho_avail = lg.rhoML[k][j][i];
	  rho_inj = drho*dt;
	  rho_inj = min(rho_inj,rho_avail);
	  rho_avail -= rho_inj;
	  rho_avail = max(rho_avail,0.0);
	  lg.rhoML[k][j][i] = rho_avail;
	  // Mass is injected with zero mtm and energy.
          // Use the procedure in Wunsch et al. (2008, ApJ, 683, 683):
    	  //    1) Add mass to cell
  	  //    2) Conserve mtm: v_new rho_new = v_old rho_old
	  //    3) Calculate KE from rho_new and v_new, and subtract from 
          //       old total energy to give new internal energy prior to
	  //       the addition of new energy (none in this case)
	  if (rho_inj > 0.0){ 
	    rho_old = rho;
	    v_old = vel;
	    ke_old = 0.5*rho_old*v_old*v_old;
	    E_old = pre/gamm + ke_old;
	    temp_old = (avgmassamb/boltzman)*pre/rho;
	    rho_new = rho_old + rho_inj;
	    v_new = v_old*rho_old/rho_new;
	    ke_new = 0.5*rho_new*v_new*v_new;
	    pre_new = gamm*(E_old - ke_new);
	    temp_new = (avgmassamb/boltzman)*pre_new/rho_new;
	    temp_new = max(tmin,temp_new);
	    temp_new = min(temp_new,tmax);
	    pre_new = rho_new*temp_new*(boltzman/avgmassamb);
	    lg.P0[iqd][k][j][i] = rho_new;
	    lg.P0[iqe][k][j][i] = pre_new;
	    lg.P0[iqu0][k][j][i] = v_new;
	    ze_old = lg.P0[iqal0][k][j][i];  // old wind    mass fraction
	    za_old = lg.P0[iqal0+1][k][j][i];// old ambient mass fraction
	    rho_wind_old = rho_old*ze_old;
	    rho_amb_old = rho_old*za_old;
	    ze_new = rho_wind_old/rho_new;
	    ze_new = min(ze_new,1.0);
	    ze_new = max(ze_new,0.0);
	    lg.P0[iqal0][k][j][i] = ze_new; // updated wind mass fraction
	    za_new = rho_amb_old/rho_new;
	    za_new = min(za_new,1.0);
	    za_new = max(za_new,0.0);
	    lg.P0[iqal0+1][k][j][i] = za_new;// updated ambient mass fraction
#ifdef DUST
	    z_old = lg.P0[iqal0][k][j][i];  // old dust-to-gas mass ratio
	    rho_dust_old = rho_old*z_old;       
	    rho_dust_new = rho_dust_old + rho_inj*dustToGasMassRatio;
	    z_new = rho_dust_new/rho_new;
	    z_new = min(z_new,1.0);
	    z_new = max(z_new,0.0);
	    lg.P0[iqal0][k][j][i] = z_new;
#endif	    
	  }
	}
      }
    }
  }
  return;
}

#endif
