/*!  \file mpiCode.cpp
 *   \brief Repository for code for running ARWEN with MPI.
 *
 *   Use C MPI bindings since the C++ bindings are being depreciated.
 *
 *   \author Julian Pittard (Original version 02.06.11)
 *   \version 1.0-stable (Evenstar):
 *   \date Last modified: 18.08.11 (JMP)
 */

#undef MAIN

//  Header files:
#include <cstdlib>
#include <cstring>
#include <iomanip>
#include "defs.h"
#include "array.h"
#include "constants.h"
#include "gas.h"
#include "global.h"
#include "globalMPI.h"
#include "grid.h"

// Setup namesspaces
using namespace std;

// Function declarations (needed even for functions within this file if the
// function definition occurs further down the file than its call)
void broadcast(bool& value);
void broadcast(double& value);
void broadcast(double value[], int count);
void broadcast(float value[], int count);
void broadcast(int& value);
void broadcast(int value[], int count);
void broadcast(string& value);
void broadcastfromproc(int send, int recv, int count, bool& value);
void calcstats();
void determineGlobalTimestep();
double global_max(double& sendbuf);
double global_sum(double& sendbuf);
float global_max(float& sendbuf);
float global_sum(float& sendbuf);
void initiate(int argc, char *argv[]);
float reduce_data_min(float& a);
double reduce_data_min(double& a);
void reduce_data_min(float& a, float& b);
void reduce_data_min(double& a, double& b);
void reduce_data_max(float& a, float& b);
void reduce_data_max(double& a, double& b);
void reduce_data_sum(double& a, double& b);
//cell* senddatatomaster(cell *send_buf, int nsend, int &totsize);
void sendrecv(int send, int recv, int count, float& value);
void sendrecv(int send, int recv, int count, double& value);
void sendrecv(int send, int recv, int count, bool& value);
void quit();

static double time0;  // compute loop timing
static double timeg0; // global timing

/*!  \brief Initialize MPI, find the size of the ring and the 
 *   procRank of each process.
 *
 *   \author Julian Pittard (Original version 02.06.11)
 *   \version 1.0-stable (Evenstar):
 *   \date Last modified: 02.06.11 (JMP)
 */
void initiate(int argc, char *argv[]){
  //cout << "initiate.c\n";
  //cout << "nd = " << nd << "\n";
  //cout << "gam = " << gam << "\n";
  
  MPI_Init(&argc, &argv);
  MPI_Comm_size(MPI_COMM_WORLD, &size);
  MPI_Comm_rank(MPI_COMM_WORLD, &procRank);
  timeg0 = MPI_Wtime( );

  mpi = true;

  // Check that the number of cells in the split dimension on each processor, 
  // times the number of processors, is consistent with the 
  // array sizes in arwen.cpp

  if (((ggrid::imax/niProc)*niProc != ggrid::imax)||((ggrid::jmax/njProc)*njProc != ggrid::jmax)
      ||((ggrid::kmax/nkProc)*nkProc != ggrid::kmax)){
      if (procRank == 0){
        cout << "Grid size cannot be evenly divided across processors!\n";
	cout << "  nd    = " << nd << "\n";
	cout << "  imax  = " << ggrid::imax << "\n";
	cout << "  jmax  = " << ggrid::jmax << "\n";
	cout << "  kmax  = " << ggrid::kmax << "\n";
	cout << "  niProc = " << niProc << "\n";
	cout << "  njProc = " << njProc << "\n";
	cout << "  nkProc = " << nkProc << "\n";
	cout << "   ABORTING!\n";
      }
      quit();              
  }
  /*
  if (lg.imaxd != imax/nipro + 2*nghosti){
      if (procRank == 0){
        cout << "imaxd incorretly calculated!\n";
	cout << "   ABORTING!\n";
      }
      quit();            
  }

  if (lg.jmaxd != jmax/njpro + 2*nghostj){
      if (procRank == 0){
        cout << "jmaxd incorretly calculated!\n";
	cout << "   ABORTING!\n";
      }
      quit();            
  }

  if (lg.kmaxd != kmax/nkpro + 2*nghostk){
      if (procRank == 0){
        cout << "kmaxd incorretly calculated!\n";
	cout << "   ABORTING!\n";
      }
      quit();            
  }
  */
  // Synchronize
  //barrier();

  // Create derived datatypes for use with MPI...
  // Specify the "MPI_SUBQ" derived datatype, which has iqmax doubles stored
  // non-contiguously but at a set stride...
  int nblocks = iqmax;
  int nelements = 1;
  int stride = lgrid::kmaxd*lgrid::jmaxd*lgrid::imaxd;
  MPI_Type_vector(nblocks,nelements,stride,MPI_DOUBLE,&MPI_SUBQ);
  MPI_Type_commit(&MPI_SUBQ);

  // Now set up the appropriate MPI_Type_Vector's for the array sub-blocks for
  // passing data between processors (based on mpi_deriveddatatype_3DcontiguousV2.cpp
  // in other_programs/mpi_test/cpp/). These vectors specify what data to pass taking
  // into account how the 3D grid is stored in memory.

  // For left and right X ghost cells we must pass a sub-block for cells.
  // This sub-block has the same number of blocks, elements/block, and stride between blocks,
  // for both real cells on the left X of the grid which are passed into the neighbouring
  // grid's right X ghost cells, and for the real cells on the right X of the grid that are
  // passed into the neighbouring grid's left X ghost cells.

  nblocks = lgrid::kmaxd * lgrid::jmaxd;
  nelements = nghost;
  stride = lgrid::imaxd;
  MPI_Type_vector(nblocks,nelements,stride,MPI_DOUBLE,&MPI_SUBI);
  MPI_Type_commit(&MPI_SUBI);

  // Repeat above for left and right Y ghost cells
  nblocks = lgrid::kmaxd;
  nelements = lgrid::imaxd * nghost; 
  stride = lgrid::imaxd * lgrid::jmaxd;
  MPI_Type_vector(nblocks,nelements,stride,MPI_DOUBLE,&MPI_SUBJ);
  MPI_Type_commit(&MPI_SUBJ);

  // Repeat above for left and right Z ghost cells
  nblocks = 1;
  nelements = lgrid::imaxd * lgrid::jmaxd * nghost; 
  stride = 0;
  MPI_Type_vector(nblocks,nelements,stride,MPI_DOUBLE,&MPI_SUBK);
  MPI_Type_commit(&MPI_SUBK);
 
  return;
}


/*!  \brief Exits out of programme and MPI.
 *
 *   \author Julian Pittard (Original version 02.06.11)
 *   \version 1.0-stable (Evenstar):
 *   \date Last modified: 02.06.11 (JMP)
 */
void quit(){
  // Deallocate memory
  delete4DArray(lg.P0);
#ifdef UNSPLIT
  delete4DArray(lg.Pstar);
#endif
  delete3DArray(lg.dei);
#ifdef MASSLOAD
  delete3DArray(lg.rhoML);
#endif
  
  MPI_Barrier(MPI_COMM_WORLD);
  MPI_Finalize();
  exit(EXIT_SUCCESS);
}


/*!  \brief Synchronise parallel processes.
 *
 *   \author Julian Pittard (Original version 02.06.11)
 *   \version 1.0-stable (Evenstar):
 *   \date Last modified: 02.06.11 (JMP)
 */
void barrier(){
  MPI_Barrier(MPI_COMM_WORLD);
  return;
}


/*!  \brief Times the compute loops.
 *
 *   \author Julian Pittard (Original version 02.06.11)
 *   \version 1.0-stable (Evenstar):
 *   \date Last modified: 02.06.11 (JMP)
 */
void timing(bool init){
  double elapsedTime = 0.0;

  // Number of steps per cycle..
#ifdef UNSPLIT
  double ndt = 1.0;
#else
  double ndt = double(nd);
#endif

  if (init) time0 = MPI_Wtime();
  else{
    elapsedTime = MPI_Wtime()-time0;
    if (procRank == 0){
      ec = ncycle;
      double cycps = double(ec-sc)/elapsedTime;
      double cellps = cycps*ggrid::ncells*ndt;
      cout << setprecision(4) << "That took " << elapsedTime << " seconds; total number of cycles = " << ncycle << "; cycles/s = " << cycps << "; steps/cycle = " << int(ndt) << "; cells/s = " << cellps << "\n";
    }
  }
  return;
}


/*!  \brief Times the entire run.
 *
 *   \author Julian Pittard (Original version 02.06.11)
 *   \version 1.0-stable (Evenstar):
 *   \date Last modified: 02.06.11 (JMP)
 */
void gtiming(bool init){
  double elapsedTime = 0.0;

  // Number of steps per cycle..
#ifdef UNSPLIT
  double ndt = 1.0;
#else
  double ndt = double(nd);
#endif

  if (init) timeg0 = MPI_Wtime();
  else{
    elapsedTime = MPI_Wtime() - timeg0;
    if (procRank == 0){
      ec = ncycle;
      double cycps = double(ec-sc)/elapsedTime;
      double cellps = cycps*ggrid::ncells*ndt;
      cout << setprecision(4) << "That took " << elapsedTime << " seconds; total number of cycles = " << ncycle << "; cycles/s = " << cycps << "; steps/cycle = " << int(ndt) << "; cells/s = " << cellps << "\n";
    }
  }
  return;
}



/*!  \brief Determine the minimum timestep that is possible over the "global" grid.
 *
 *   First gather (using MPI_Allreduce) and store in variable gdt, then broadcast  
 *   (using another call of MPI_Allreduce) and store in variable dt
 *
 *   \author Julian Pittard (Original version 02.06.11)
 *   \version 1.0-stable (Evenstar):
 *   \date Last modified: 02.06.11 (JMP)
 */
void determineGlobalTimestep(){
  double gdt;
  MPI_Barrier(MPI_COMM_WORLD);
  MPI_Allreduce(&dt,&gdt,1,MPI_DOUBLE,MPI_MIN,MPI_COMM_WORLD);
  MPI_Allreduce(&gdt,&dt,1,MPI_DOUBLE,MPI_MIN,MPI_COMM_WORLD);
  MPI_Barrier(MPI_COMM_WORLD);

#ifdef MHD
  // Determine the global maximum wavespeed
  MPI_Barrier(MPI_COMM_WORLD);
  MPI_Allreduce(&wsmax,&gwsmax,1,MPI_DOUBLE,MPI_MAX,MPI_COMM_WORLD);
  MPI_Barrier(MPI_COMM_WORLD);
#endif
  //MPI_Allreduce( in, out, count, MPI_INT, MPI_SUM, MPI_COMM_WORLD ); 
  return;
}


/*!  \brief Broadcast a string from the master process to the slave processes.
 *
 *   This function is overloaded.
 *
 *   \author Julian Pittard (Original version 15.08.11)
 *   \version 1.0-stable (Evenstar):
 *   \date Last modified: 15.08.11 (JMP)
 */
void broadcast(string& key){
  //Broadcast key to the other processes
  int sizekey = key.size();
  //cout << "broadcast_cmd.cpp: procRank = " << procRank << "; sizekey = " << sizekey << "\n";

  //First broadcast the size of the string from procRank=0 to the other processes
  MPI_Bcast(&sizekey,1,MPI_INT,0,MPI_COMM_WORLD);
  //Now broadcast the string
  char* ckey = new char[sizekey];   // create a pointer to a char array of size sizekey
  if (procRank == 0) strcpy(ckey,key.c_str());  //copy string into a character array
  MPI_Bcast(&ckey[0],sizekey,MPI_CHAR,0,MPI_COMM_WORLD); //broadcast character array (same as ckey)
  //cout << "procRank = " << procRank << "; ckey = " << ckey << "\n";
  // convert character array back to a C string
  key = ckey; //
  key.resize(sizekey); // resize key to correct size (this is a necessary step)
  //MPI_Barrier(MPI_COMM_WORLD);
  // delete pointer
  delete []ckey;
  // check all processes have the key
  //cout << "Got key, procRank = " << procRank << "; key = " << key << "\n";
  barrier();
  return;
}

/*!  \brief Broadcast an integer from the master process to the slave processes.
 *
 *   This function is overloaded.
 *
 *   \author Julian Pittard (Original version 15.08.11)
 *   \version 1.0-stable (Evenstar):
 *   \date Last modified: 15.08.11 (JMP)
 */
void broadcast(int& value){
  MPI_Bcast(&value,1,MPI_INT,0,MPI_COMM_WORLD); //broadcast value
  // check all processes have the value
  //cout << "Got value, procRank = " << procRank << "; value = " << value << "\n";
  //barrier();
  return;
}


/*!  \brief Broadcast a double from the master process to the slave processes.
 *
 *   This function is overloaded.
 *
 *   \author Julian Pittard (Original version 15.08.11)
 *   \version 1.0-stable (Evenstar):
 *   \date Last modified: 15.08.11 (JMP)
 */
void broadcast(double& value){  
  MPI_Bcast(&value,1,MPI_DOUBLE,0,MPI_COMM_WORLD); //broadcast value
  // check all processes have the value
  //cout << "Got value, procRank = " << procRank << "; value = " << value << "\n";
  //barrier();
  return;
}


/*!  \brief Broadcast a boolean from the master process to the slave processes.
 *
 *   This function is overloaded.
 *
 *   \author Julian Pittard (Original version 16.08.11)
 *   \version 1.0-stable (Evenstar):
 *   \date Last modified: 16.08.11 (JMP)
 */
void broadcast(bool& value){
  MPI_Bcast(&value,1,MPI::BOOL,0,MPI_COMM_WORLD); //buffer,count,datatype,root,comm
  return;
}


/*!  \brief Broadcast an integer array from the master process to the slave processes.
 *
 *   This function is overloaded.
 *
 *   \author Julian Pittard (Original version 26.09.11)
 *   \version 1.0-stable (Evenstar):
 *   \date Last modified: 26.09.11 (JMP)
 */
void broadcast(int value[], int count){
  MPI_Bcast(value,count,MPI_INT,0,MPI_COMM_WORLD); //broadcast value
  return;
}


/*!  \brief Broadcast a float array from the master process to the slave processes.
 *
 *   This function is overloaded.
 *
 *   \author Julian Pittard (Original version 26.09.11)
 *   \version 1.0-stable (Evenstar):
 *   \date Last modified: 26.09.11 (JMP)
 */
void broadcast(float value[], int count){
  MPI_Bcast(value,count,MPI_FLOAT,0,MPI_COMM_WORLD); //broadcast value
  return;
}


/*!  \brief Broadcast a double array from the master process to the slave processes.
 *
 *   This function is overloaded.
 *
 *   \author Julian Pittard (Original version 26.09.11)
 *   \version 1.0-stable (Evenstar):
 *   \date Last modified: 26.09.11 (JMP)
 */
void broadcast(double value[], int count){
  MPI_Bcast(value,count,MPI_DOUBLE,0,MPI_COMM_WORLD); //broadcast value
  return;
}


/*!  \brief Broadcast a boolean from one process (not necessarily the 
 *            master) to the other processes.
 *
 *   First we use MPISENDRECV to send the bool from the desired processor
 *   to the master, and then MPI_Bcast to broadcast to every processor
 *
 *   \author Julian Pittard (Original version 06.12.11)
 *   \version 1.0-stable (Evenstar):
 *   \date Last modified: 06.12.11 (JMP)
 */
void broadcastfromproc(int send, int recv, int count, bool& value){
  
  // Use MPI_Ssend and MPI_Recv to send data from one process to another 
  // First check that the send and recv processors are different. If they are
  // not, then the master processor already has the desired value.
  if (send != recv){
    if ((procRank == send)||(procRank == recv)){      
      if (debug_mpi) cout << "presendrecv: procRank = " << procRank << "; send  = " << send << "; recv = " << recv << "\n";
      sendrecv(send, recv, count, value);
      if (debug_mpi) cout << "pstsenrecv: procRank = " << procRank << "; send  = " << send << "; recv = " << recv << "\n";
    }
  }
  barrier();
  // Now broadcast to every processor
  if (debug_mpi) cout << "prebroadcast: procRank = " << procRank << "; value = " << value << "\n"; 
  broadcast(value);
  if (debug_mpi) cout << "pstbroadcast: procRank = " << procRank << "; value = " << value << "\n"; 
  return;
}


/*!  \brief Send a float from one process to another process.
 *
 *   This function is overloaded.
 *
 *   \author Julian Pittard (Original version 19.08.11)
 *   \version 1.0-stable (Evenstar):
 *   \date Last modified: 19.08.11 (JMP)
 */
void sendrecv(int send, int recv, int count, float& value){
  MPI_Status status;
  // Use MPI_Ssend and MPI_Recv to send data from one process to another 
  // First check that the send and recv processors are different
  if (send == recv){
    cout << "sendrecv error: send and recv processors are identical!\n";
    quit();
  }

  // Otherwise perform the data exchange
  int tag = 1;
  if (procRank == recv){
    if (debug_mpi) cout << "Processor " << recv << " receiving from " << send << "\n";
    MPI_Recv(&value, count, MPI_FLOAT, send, tag, MPI_COMM_WORLD, &status);//receive from processor send
  }
  else if (procRank == send){ // send data
    tag = procRank;
    if (debug_mpi) cout << "Processor " << send << " sending to " << recv << "\n";
    MPI_Ssend(&value, count, MPI_FLOAT, recv, tag, MPI_COMM_WORLD); //synchronous send to processor recv
  }
  //barrier();

  return;
}


/*!  \brief Send a double from one process to another process.
 *
 *   This function is overloaded.
 *
 *   \author Julian Pittard (Original version 19.08.11)
 *   \version 1.0-stable (Evenstar):
 *   \date Last modified: 19.08.11 (JMP)
 */
void sendrecv(int send, int recv, int count, double& value){
  MPI_Status status;
  // Use MPI_Ssend and MPI_Recv to send data from one process to another 
  // First check that the send and recv processors are different
  if (send == recv){
    cout << "sendrecv error: send and recv processors are identical!\n";
    quit();
  }

  // Otherwise perform the data exchange
  int tag = 1;
  if (procRank == recv){
    if (debug_mpi) cout << "Processor " << recv << " receiving from " << send << "\n";
    MPI_Recv(&value, count, MPI_DOUBLE, send, tag, MPI_COMM_WORLD, &status);//receive from processor send
  }
  else if (procRank == send){ // send data
    tag = procRank;
    if (debug_mpi) cout << "Processor " << send << " sending to " << recv << "\n";
    MPI_Ssend(&value, count, MPI_DOUBLE, recv, tag, MPI_COMM_WORLD); //synchronous send to processor recv
  }
  //barrier();
  return;
}


/*!  \brief Send a boolean from one process to another process.
 *
 *   The boolean is passed as an integer. This function is overloaded.
 *
 *   \author Julian Pittard (Original version 06.12.11)
 *   \version 1.0-stable (Evenstar):
 *   \date Last modified: 06.12.11 (JMP)
 */
void sendrecv(int send, int recv, int count, bool& value){
  MPI_Status status;
  // Use MPI_Ssend and MPI_Recv to send data from one process to another 
  // First check that the send and recv processors are different
  if (send == recv){
    cout << "sendrecv error: send and recv processors are identical!\n";
    quit();
  }

  // Otherwise perform the data exchange
  int tag = 1;
  int ivalue = (int)value;
  if (procRank == recv){
    if (debug_mpi) cout << "Processor " << recv << " receiving from " << send << "\n";
    MPI_Recv(&ivalue, count, MPI_INT, send, tag, MPI_COMM_WORLD, &status);//receive from processor send
  }
  else if (procRank == send){ // send data
    tag = procRank;
    if (debug_mpi) cout << "Processor " << send << " sending to " << recv << "\n";
    MPI_Ssend(&ivalue, count, MPI_INT, recv, tag, MPI_COMM_WORLD); //synchronous send to processor recv
    if (debug_mpi) cout << "Processor " << send << " sent to " << recv << "\n";
  }
  //barrier();
  value = (bool)ivalue;
  if (debug_mpi) cout << "PROCRANK = " << procRank << "; value = " << value << "\n"; 
  return;
}


/*!  \brief Data held in send_buf by each slave processor is sent to the
 *   master processor's recv_buf.
 *
 *   Each processor (including the master) sends nsend cells to the master.
 *   The total number of cells in recv_buf is totsize, which is returned to the
 *   calling routine. Operator overloaded.
 *
 *   Based on the example code in mpi_gatherv_jmp.
 *
 *   NOTE: When returning a dynamically allocated array to the calling function 
 *   I have 3 possible options:
 *     1) use a pointer, here it is a pointer to pointer
 *     2) use a reference, here it is a reference to pointer
 *     3) use the return value of the function, so return a cell* instead of void
 *   I have implemented option 3. (Only C++ can do option 2).
 *   See http://stackoverflow.com/questions/2625869/why-pointer-to-pointer-is-needed-to-allocate-memory-in-function
 *
 *   \author Julian Pittard (Original version 04.09.13)
 *   \version 1.0-stable (Evenstar):
 *   \date Last modified: 05.09.13 (JMP)
 */
/*
cell* senddatatomaster(cell *send_buf, int nsend, int &totsize){
  int mpi_err;

  // All processors must allocate these pointers (which must have function scope), but only 
  // the root processor allocates memory for them
  int *counts, *displacements;  
  cell *recv_buf;

  // counts and displacement arrays are only required on the root
  if (procRank == 0){
    counts = new int[size]; // DO NOT set this up as int* counts, because then a new version of the "counts" pointer is
                            // created which ONLY has scope within this conditional statement
    displacements = new int[size];
  }

  if (size > 1){
    // Gather the counts to the root (all processors (slaves+master) send "nsend", master stores into counts array)
    mpi_err = MPI_Gather(&nsend,1,MPI_INT,counts,1,MPI_INT,0,MPI_COMM_WORLD); // nsend goes into counts array on master

    // calculate displacements and the size of the recv buffer
    if (procRank == 0){
      displacements[0]=0;
      for (int i = 1; i < size; i++){
	displacements[i] = counts[i-1] + displacements[i-1];
      }
      for (int i = 0; i < size; i++){
	totsize += counts[i];
      }
      recv_buf = new cell[totsize]; // create recv buffer needed for the total number of cells passed to the master
    }

    // different amounts of data from each processor are gathered to the root 
    mpi_err = MPI_Gatherv(send_buf, nsend, MPI_CELL, recv_buf, counts, displacements, MPI_CELL, 0, MPI_COMM_WORLD);

    if (debug_mpi){
      if (procRank == 0){
        for (int i = 0; i < totsize; i++){
          cout << "senddatatomaster: cell = " << i << "; recv_buf[i].i = " << recv_buf[i].i << "; recv_buf[i].j = " << recv_buf[i].j << "; procnum = " << recv_buf[i].procnum << "; col = " << recv_buf[i].pvars.p[iqal0+1] << "\n";
        }
      }
    }

  } // size > 1
  else{
    // Code is being run serially. All plot data must be on the master and therefore in the send_buf.
    // Simply copy data from send_buf into recv_buf
    totsize = nsend;
    recv_buf = new cell[totsize]; // create recv buffer needed for the total number of cells passed to the master
    for (int i = 0; i < nsend; i++){
      recv_buf[i] = send_buf[i];
      if (debug_lev > 3){
        cout << "senddatatomaster, serial: i = " << i << "; recv_buf[i].i = " << recv_buf[i].i << "; recv_buf[i].j = " << recv_buf[i].j << "; rho = " << recv_buf[i].pvars.p[iqd] << "; pre = " << recv_buf[i].pvars.p[iqe] << "\n";
      }
    }
  }
  //if (procRank == 0) cout << "senddatatomaster: recv_buf[10].i = " << recv_buf[10].i << "\n";

  // At this point all of the cells to be plotted are in recv_buf, which is 
  // a 1D array of size totsize. The counts and displacement arrays can be
  // deleted to free up memory
  if (procRank == 0){
    delete[] counts;
    delete[] displacements;
  }
  return recv_buf;
}
*/

/*!  \brief Perform an MPI reduction operation (min).
 *
 *   Only the master processor gets the result (use MPI_ALLREDUCE to broadcast the result
 *   to all processors).
 *
 *   \author Julian Pittard (Original version 18.08.11)
 *   \version 1.0-stable (Evenstar):
 *   \date Last modified: 18.08.11 (JMP)
 */
float reduce_data_min(float& a){
  float b = a;
  MPI_Reduce(&a,&b,1,MPI_FLOAT,MPI_MIN,0,MPI_COMM_WORLD);
  return b;
}


/*!  \brief Perform an MPI reduction operation (min).
 *
 *   Only the master processor gets the result (use MPI_ALLREDUCE to broadcast the result
 *   to all processors).
 *
 *   \author Julian Pittard (Original version 18.08.11)
 *   \version 1.0-stable (Evenstar):
 *   \date Last modified: 18.08.11 (JMP)
 */
double reduce_data_min(double& a){
  double b = a;
  MPI_Reduce(&a,&b,1,MPI_DOUBLE,MPI_MIN,0,MPI_COMM_WORLD);
  return b;
}


/*!  \brief Perform an MPI reduction operation (min).
 *
 *   Overloaded.
 *
 *   \author Julian Pittard (Original version 18.08.11)
 *   \version 1.0-stable (Evenstar):
 *   \date Last modified: 18.08.11 (JMP)
 */
void reduce_data_min(float& a, float& b){
  MPI_Reduce(&a,&b,1,MPI_FLOAT,MPI_MIN,0,MPI_COMM_WORLD);
  return;
}


/*!  \brief Perform an MPI reduction operation (min).
 *
 *   Overloaded.
 *
 *   \author Julian Pittard (Original version 18.08.11)
 *   \version 1.0-stable (Evenstar):
 *   \date Last modified: 18.08.11 (JMP)
 */
void reduce_data_min(double& a, double& b){
  MPI_Reduce(&a,&b,1,MPI_DOUBLE,MPI_MIN,0,MPI_COMM_WORLD);
  return;
}


/*!  \brief Perform an MPI reduction operation (max).
 *
 *   Overloaded.
 *
 *   \author Julian Pittard (Original version 18.08.11)
 *   \version 1.0-stable (Evenstar):
 *   \date Last modified: 18.08.11 (JMP)
 */
void reduce_data_max(float& a, float& b){
  MPI_Reduce(&a,&b,1,MPI_FLOAT,MPI_MAX,0,MPI_COMM_WORLD);
  return;
}


/*!  \brief Perform an MPI reduction operation (max).
 *
 *   Overloaded.
 *
 *   \author Julian Pittard (Original version 18.08.11)
 *   \version 1.0-stable (Evenstar):
 *   \date Last modified: 18.08.11 (JMP)
 */
void reduce_data_max(double& sendbuf, double& recvbuf){
  //barrier();
  //cout << "reduce_data_max: before MPI_Reduce, procRank = " << procRank << "; sendbuf = " << sendbuf << "\n";
  MPI_Reduce(&sendbuf,&recvbuf,1,MPI_DOUBLE,MPI_MAX,0,MPI_COMM_WORLD);
  //barrier();
  //cout << "reduce_data_max: after MPI_Reduce, procRank = " << procRank << "; recvbuf = " << recvbuf << "\n";
  return;
}


/*!  \brief Perform an MPI reduction operation (sum).
 *
 *   Overloaded.
 *
 *   \author Julian Pittard (Original version 30.09.11)
 *   \version 1.0-stable (Evenstar):
 *   \date Last modified: 30.09.11 (JMP)
 */
void reduce_data_sum(double& sendbuf, double& recvbuf){
  //barrier();
  //cout << "reduce_data_sum: before MPI_Reduce, procRank = " << procRank << "; sendbuf = " << sendbuf << "\n";
  MPI_Reduce(&sendbuf,&recvbuf,1,MPI_DOUBLE,MPI_SUM,0,MPI_COMM_WORLD);
  //MPI_Allreduce(&sendbuf,&recvbuf,1,MPI_DOUBLE,MPI_SUM,MPI_COMM_WORLD);
  //barrier();
  //cout << "reduce_data_sum: after MPI_Reduce, procRank = " << procRank << "; recvbuf = " << recvbuf << "\n";
  return;
}


/*!  \brief Perform an MPI reduction operation (sum).
 *
 *   Overloaded.
 *
 *   \author Julian Pittard (Original version 03.12.13)
 *   \version 1.1-development (Hadafang):
 *   \date Last modified: 03.12.13 (JMP)
 */
double global_sum(double& sendbuf){
  double sum;
  MPI_Reduce(&sendbuf,&sum,1,MPI_DOUBLE,MPI_SUM,0,MPI_COMM_WORLD);
  return sum;
}


/*!  \brief Perform an MPI reduction operation (sum).
 *
 *   Overloaded.
 *
 *   \author Julian Pittard (Original version 03.12.13)
 *   \version 1.1-development (Hadafang):
 *   \date Last modified: 03.12.13 (JMP)
 */
float global_sum(float& sendbuf){
  float sum;
  MPI_Reduce(&sendbuf,&sum,1,MPI_FLOAT,MPI_SUM,0,MPI_COMM_WORLD);
  return sum;
}


/*!  \brief Perform an MPI reduction operation (max).
 *
 *   Overloaded.
 *
 *   \author Julian Pittard (Original version 03.12.13)
 *   \version 1.1-development (Hadafang):
 *   \date Last modified: 03.12.13 (JMP)
 */
double global_max(double& sendbuf){
  double max;
  MPI_Reduce(&sendbuf,&max,1,MPI_DOUBLE,MPI_MAX,0,MPI_COMM_WORLD);
  return max;
}


/*!  \brief Perform an MPI reduction operation (max).
 *
 *   Overloaded.
 *
 *   \author Julian Pittard (Original version 03.12.13)
 *   \version 1.1-development (Hadafang):
 *   \date Last modified: 03.12.13 (JMP)
 */
float global_max(float& sendbuf){
  float max;
  MPI_Reduce(&sendbuf,&max,1,MPI_FLOAT,MPI_MAX,0,MPI_COMM_WORLD);
  return max;
}

