/*!  \file pdcell.cpp
 *   \brief Repository for functions which use the cell value of pressure/rho.
 *
 *   Copied from ~/work/stellarcluster/arwen161211/code (JMP 28/06/21)
 *
 *   \author Julian Pittard (Original version 15.09.11)
 *   \version 0.1-development (Undómiel):
 *   \date Last modified: 15.09.11 (JMP)
 */

#undef MAIN

//  Header files:
#include <fstream>
#include <iostream>
#include <math.h>
#include <cmath>              //needed to check for infinities and NaNs
#include "constants.h"
#include "global.h"
#include "pdcell.h"

// Function declarations
double pdtmp(double pdcell);
double pdavgmu(double pdcell);
double pdfrach2(double pdcell);
double pdfrachi(double pdcell);
double pdfrachii(double pdcell);
void readtempgammamu();
double tmpmu(double temp);
double tmpg(double temp);
extern void quit();

using namespace std;


/*!  \brief Read in a data table of p/rho to calculate the monotonic temperature dependent gamma and mu.
 *  
 *   The pre-tabulated table is based on the allen1973.txt data in Fyris Alpha.
 *
 *   \author Julian Pittard (Original version 23.09.11)
 *   \version 0.1-development (Undómiel):
 *   \date Last modified: 23.09.11 (JMP)
 */
void readtempgammamu(){

  // Open the file to read in the desired data table
  ifstream file("tempdependent_mu_and_gamma.dat"); //C++ method (trackfile must be a string type)
  if (!file) {
    if (procRank == 0) cerr << "Failed to open temp dependent mu and gamma file\n";
    quit();
  } 
  if (procRank == 0) cout << "Reading in temp dependent mu and gamma file.\n"; 

  // Read in data
  for (int n=0; n < npd; n++){
    file >> pdlogt[n] >> pdmu[n] >> pdg[n] >> pdh2[n] >> pdhi[n] >> pdhii[n] >> pd[n] >> ied[n];
  }

  return;
}


 
/*!  \brief Return the cell temperature (K) given p/rho. 
 *  
 *   It uses a pre-tabulated table based on the allen1973.txt data  in Fyris Alpha.
 *
 *   \author Julian Pittard (Original version 15.09.11)
 *   \version 0.1-development (Undómiel):
 *   \date Last modified: 15.09.11 (JMP)
 */
double pdtmp(double pdcell){

  double grad;
  double temp = 0.0;

  // Set temp to lower or upper limits, or interpolate from the table

  if (pdcell < pd[0])          temp = pow(10,pdlogt[0]);
  else if (pdcell > pd[npd-1]) temp = pow(10,pdlogt[npd-1]);
  else{
    for (int n=0; n < npd-1; n++){
      if ((pdcell >= pd[n]) && (pdcell < pd[n+1])){
	grad = (pdlogt[n+1]-pdlogt[n])/(log10(pd[n+1])-log10(pd[n])); 
	temp = pow(10,(pdlogt[n] + grad*(log10(pdcell) - log10(pd[n]))));
	goto fin;
      }
    }
  }

 fin:

  if (debug_lev >= 5){
    // Check whether temp is sensible
    if ( std::isinf( temp )){
      // value is infinity
      cerr << "temp is infinity\n";
      cerr << "Aborting from pdtmp.c\n";
      quit();
    }
    if ( std::isnan( temp )){
      // value is not a number
      cerr << "temp is a NAN\n";
      cerr << "Aborting from pdtmp.c\n";
      quit();
    }
    if (temp <= 0.0){
      cerr << "temp = " << temp << "\n";
      cerr << "pdcell = " << pdcell << "\n";
      cerr << "Aborting from pdtmp.c\n";
      quit();
    }
  } // debug

  return temp;
}



/*!  \brief Return the cell average particle mass given p/rho. 
 *  
 *   It uses a pre-tabulated table based on the allen1973.txt data in Fyris Alpha.
 *
 *   \author Julian Pittard (Original version 15.09.11)
 *   \version 0.1-development (Undómiel):
 *   \date Last modified: 15.09.11 (JMP)
 */
double pdavgmu(double pdcell){

  double grad;
  double pdavgmu = 0.0;

  // Set avgmass to lower or upper limits, or interpolate from the table

  if (pdcell < pd[0]) pdavgmu = pdmu[0];
  else if (pdcell > pd[npd-1]) pdavgmu = pdmu[npd-1];
  else{
    for (int n=0; n < npd-1; n++){
      if ((pdcell >= pd[n]) && (pdcell < pd[n+1])){
	grad = (pdmu[n+1]-pdmu[n])/(pd[n+1]-pd[n]); 
	pdavgmu = pdmu[n] + grad*(pdcell - pd[n]);
	goto fin;
      }
    }
  }

 fin: return pdavgmu;
}



/*!  \brief Return the number fraction of H in H2 given p/rho. 
 *  
 *   It uses an ad-hoc table noted on p175-176 of "Notes on Constructing a Hydro Code".
 *
 *   \author Julian Pittard (Original version 15.09.11)
 *   \version 0.1-development (Undómiel):
 *   \date Last modified: 15.09.11 (JMP)
 */
double pdfrach2(double pdcell){

  int n;
  double grad,frach2;
  bool loop;
         
  frach2 = 0.0;
  //set frach2 to lower or upper limits, or interpolate from the table
  if (pdcell < pd[0]) frach2 = pdh2[0];
  else if (pdcell > pd[npd-1]) frach2 = pdh2[npd-1];
  else{
    n = 0;
    loop = true;
    while (loop){
      if ((pdcell >= pd[n])&&(pdcell < pd[n+1])){
        grad = (pdh2[n+1]-pdh2[n])/(pd[n+1]-pd[n]); 
        frach2 = pdh2[n] + grad*(pdcell - pd[n]);
        loop = false;
      }
      n++;
    }    
  }

  return frach2;
}


/*!  \brief Return the number fraction of H in HI given p/rho. 
 *  
 *   It uses an ad-hoc table noted on p175-176 of "Notes on Constructing a Hydro Code".
 *
 *   \author Julian Pittard (Original version 15.09.11)
 *   \version 0.1-development (Undómiel):
 *   \date Last modified: 15.09.11 (JMP)
 */
double pdfrachi(double pdcell){

  int n;
  double grad,frachi;
  bool loop;
         
  frachi = 0.0;
  //set frachi to lower or upper limits, or interpolate from the table
  if (pdcell < pd[0]) frachi = pdhi[0];
  else if (pdcell > pd[npd-1]) frachi = pdhi[npd-1];
  else{
    n = 0;
    loop = true;
    while (loop){
    if ((pdcell >= pd[n])&&(pdcell < pd[n+1])){
      grad = (pdhi[n+1]-pdhi[n])/(pd[n+1]-pd[n]); 
      frachi = pdhi[n] + grad*(pdcell - pd[n]);
      loop = false;
    }
    n++;
    }    
  }

  return frachi;
}



/*!  \brief Return the number fraction of H in HII given p/rho. 
 *  
 *   It uses a pre-tabulated table based on the allen1973.txt data in Fyris Alpha.
 *
 *   \author Julian Pittard (Original version 15.09.11)
 *   \version 0.1-development (Undómiel):
 *   \date Last modified: 15.09.11 (JMP)
 */
double pdfrachii(double pdcell){

  int n;
  double grad,frachii;
  bool loop;
         
  frachii = 0.0;
  //set frachii to lower or upper limits, or interpolate from the table
  if (pdcell < pd[0]) frachii = pdhii[0];
  else if (pdcell > pd[npd-1]) frachii = pdhii[npd-1];
  else{
    n = 0;
    loop = true;
    while (loop){
    if ((pdcell >= pd[n])&&(pdcell < pd[n+1])){
      grad = (pdhii[n+1]-pdhii[n])/(pd[n+1]-pd[n]); 
      frachii = pdhii[n] + grad*(pdcell - pd[n]);
      loop = false;
    }
    n++;
    }    
  }

  return frachii;
}


/*!  \brief Return the cell average particle mass given the temperature.
 *  
 *   It uses a pre-tabulated table based on the allen1973.txt data in Fyris Alpha.
 *
 *   \author Julian Pittard (Original version 15.09.11)
 *   \version 0.1-development (Undómiel):
 *   \date Last modified: 15.09.11 (JMP)
 */
double tmpmu(double temp){

  int n;
  double grad,mu,logtemp;
  bool loop;
         
  mu = 0.0;
  logtemp = log10(temp);

  //set mu to lower or upper limits, or interpolate from the table
  if (logtemp < pdlogt[0]) mu = pdmu[0];
  else if (logtemp > pdlogt[npd-1]) mu = pdmu[npd-1];
  else{
    n = 0;
    loop = true;
    while (loop){
      if ((logtemp >= pdlogt[n])&&(logtemp < pdlogt[n+1])){
        grad = (log10(pdmu[n+1])-log10(pdmu[n]))/(pdlogt[n+1]-pdlogt[n]); 
        mu = pow(10,log10(pdmu[n]) + grad*(logtemp - pdlogt[n]));
        loop = false;
      }
    n++;
    }    
  }
  return mu;
}



/*!  \brief Return the cell ratio of specific heats (gamma) given the temperature (K).
 *  
 *   It uses a pre-tabulated table based on the allen1973.txt data in Fyris Alpha.
 *
 *   \author Julian Pittard (Original version 15.09.11)
 *   \version 0.1-development (Undómiel):
 *   \date Last modified: 15.09.11 (JMP)
 */
double tmpg(double temp){

  int n;
  double grad,gam,logtemp;
  bool loop;
         
  gam = 0.0;
  logtemp = log10(temp);

  //set gam to lower or upper limits, or interpolate from the table
  if (logtemp < pdlogt[0]) gam = pdg[0];
  else if (logtemp > pdlogt[npd-1]) gam = pdg[npd-1];
  else{
    n = 0;
    loop = true;
    while (loop){
      if ((logtemp >= pdlogt[n])&&(logtemp < pdlogt[n+1])){
        grad = (log10(pdg[n+1])-log10(pdg[n]))/(pdlogt[n+1]-pdlogt[n]); 
        gam = pow(10,log10(pdg[n]) + grad*(logtemp - pdlogt[n]));
        loop = false;
      }
    n++;
    }    
  }

  return gam;
}

