/*!  \file ppmCode.cpp
 *   \brief Repository for code for Piecewise Parabolic Method (PPM).
 *
 *   \author Julian Pittard (Original version 02.06.11)
 *   \version 0.1-development (Undómiel):
 *   \date Last modified: 29.07.11 (JMP)
 */

#undef MAIN

//  Header files:
#include <algorithm>
#include <cmath>
#include <math.h>
#include <iomanip>
#include <iostream>
#include "defs.h"
#include "constants.h"
#include "gas.h"
#include "global.h"           
#include "grid.h"
#include "physconst.h"

using namespace std;

// Function declarations
void parabola(int nmin, int nmax, double para[][lgrid::maxsweep], double a[], double deltaa[], double a6[], double al[], double flat[], double jump[], int iflat, int iskip, int icd);
void paraset(int nmin, int nmax, double zpara[][lgrid::maxsweep], double dx[], double xa[]);


/*!  \brief Function to setup the PPM parabolic coefficients of the grid.
 *
 *   Sets up constants which are re-used each time we want to
 *   interpolate a parabola on a quantity. First pull out constants
 *   A, B, and C, and then compute all the equations in terms of those
 *   quantities. The quantities calculated here are stored in an array zpara,
 *   to be read by parabola()
 *
 *   nmin/nmax are the index range for which one will calculate parabolae
 * 
 *   Based on Colella and Woodward, JCompPhys 54, 174-201 (1984) eq 1.6, 1.7
 *
 *   04.08.11 (JMP) - changed zpara to a built-in array, which necessitates 
 *                     defining it to be of size [][maxsweep] - see grid.h.
 *
 *   \author Julian Pittard (Original version 31.05.11)
 *   \version 0.1-development (Undómiel):
 *   \date Last modified: 10.08.11 (JMP)
 */
void paraset(int nmin, int nmax, double zpara[][lgrid::maxsweep], double dx[], double xa[])
{
  //Locals
  double a[lgrid::maxsweep], b[lgrid::maxsweep], c[lgrid::maxsweep], d[lgrid::maxsweep], e[lgrid::maxsweep], f[lgrid::maxsweep], g[lgrid::maxsweep];
  double ai[lgrid::maxsweep], bi[lgrid::maxsweep], ci[lgrid::maxsweep];

  //--------------------------------------------------------------------------
  //                                      A =  dX_j +  dX_j+1
  //                                      B = 2dX_j +  dX_j+1
  //                                      C =  dX_j + 2dX_j+1
  //
  //                                      ai, bi, and ci are inverse quantities
  for (int n = nmin-2; n <= nmax+1; n++){
    a[n] = dx[n] + dx[n+1];
    ai[n] = 1./a[n];
    b[n] = a[n] + dx[n];
    bi[n] = 1./b[n];
    c[n] = a[n] + dx[n+1];
    ci[n] = 1./c[n];
  }
  //                                        constants for equation 1.6
  //     a(j+.5) = a(j) + C1 * (a(j+1)-a(j)) + C2 * da(j+1) + C3 * da(j)

  for (int n = nmin-1; n <= nmax; n++){
    d[n]      = 1. / (a[n-1] + a[n+1]);
    zpara[1][n] = dx[n] * ai[n] + 2. * dx[n+1] * dx[n] *
	           d[n] * ai[n] * ( a[n-1] * bi[n] - a[n+1] * ci[n] );
    zpara[2][n] = - d[n] * dx[n]   * a[n-1] * bi[n];
    zpara[3][n] =   d[n] * dx[n+1] * a[n+1] * ci[n];
    //cout << n << "\t" << zpara[1][n] << "\t" << zpara[2][n] << "\t" << zpara[3][n] << "\n";
  }
  //quit();

  //                                        constants for equation 1.7
  //     da(j) = D1 * (a(j+1) - a(j)) + D2 * (a(j) - a(j-1))

  for (int n = nmin-1; n <= nmax+1; n++){
    d[n] = dx[n] / ( a[n-1] + dx[n+1] );
    zpara[4][n] = d[n] * b[n-1] * ai[n];
    zpara[5][n] = d[n] * c[n]   * ai[n-1];
  }

  // Calculate coefficients for CD steepening. para(6,n) and para(7,n) are
  //     for calculating Eq. 35 in Fryxell et al., para(8,n) for Eq. 36

  for (int n = nmin-1; n <= nmax+2; n++){
    e[n] = a[n] + dx[n-1];         // dX_j-1 + dX_j + dX_j+1
    f[n] = pow(xa[n] - xa[n-1],3); // Note: xa, not dx, is used in Eq. 36
    g[n] = xa[n+1] - xa[n-1];
  }
  for (int n = nmin-1; n <= nmax+1; n++){
    zpara[6][n] = 1.0/(e[n]*a[n]);
    zpara[7][n] = 1.0/(e[n]*a[n-1]);
    zpara[8][n] = (f[n]+f[n+1])/g[n];
    //printf("%3d %15.6f %15.6f %15.6f %15.6f\n", n,xa[n],f[n],g[n],para[8][n]); 
  }
  //exit(EXIT_FAILURE);
  

  return;
}





/*!  \brief Calculates the parabolic coeffs for the passed fluid variable.
 *
 *   This function calculates the parabolas themselves. Call paraset() first
 *   for a given grid-spacing to set the constants, which can be reused
 *   each time parabola() is called.
 * 
 *   Based on Colella and Woodward, JCompPhys 54, 174-201 (1984), 
 *   eqs 1.5-1.8,1.10.
 *
 *   Flatening coefficients are calculated externally in flatten(). 
 *   Contact discontinuity steepening is applied if icd = 1. jump is
 *   passed into this routine to help evaluate the criteria (JMP 22/10/06).
 *
 *   nmin/nmax are indicies over which the parabolae are calculated.
 *
 *   Replaced Matrices and Vectors with built-in arrays for speed.
 *
 *   10.08.11 (JMP) - modified the effective range of the first 2 loops
 *                    so that only 3 ghost cells are required.
 *  
 *   \author Julian Pittard (Original version 31.05.11)
 *   \version 0.1-development (Undómiel):
 *   \date Last modified: 10.08.11 (JMP)
 */
void parabola(int nmin, int nmax, double para[][lgrid::maxsweep], double a[], double deltaa[], double a6[], double al[], double flat[], double jump[], int iflat, int iskip, int icd){
  
  //Locals
  //double eta, onemeta, jumpa;
  double onemfl, sgn, scrch3;
  double da[lgrid::maxsweep],     ar[lgrid::maxsweep],  diffa[lgrid::maxsweep];
  //double ali[lgrid::maxsweep], ari[lgrid::maxsweep], d2a[lgrid::maxsweep];
  double scrch1[lgrid::maxsweep], scrch2[lgrid::maxsweep];

  //JMP 10/08/11: changed loop from nmin-2 to nmax+1 to nmin-1 to nmax 
  //                and added line below for diffa[nmax+1]
  //for (int n = nmin-2; n <= nmax+1; n++){ // Original
  for (int n = nmin-1; n <= nmax; n++){ // Modified
    diffa[n] = a[n+1] - a[n];
  }
  diffa[nmax+1] = diffa[nmax]; //Modified
 
  // Skip to monotonicity when interpolating total Energy for lagrangian remap
  // since deltaa is already set up for us.

  if (iskip == 1){
    for (int n = nmin; n<= nmax; n++){
      ar[n] = deltaa[n] + al[n];
    }
  }
  else { // all other calls to parabola execute this code...
    //                                                     Equation 1.7 of C&W
    //     da(j) = D1 * (a(j+1) - a(j)) + D2 * (a(j) - a(j-1))
    //     zero out da(n) if a(n) is a local max/min
    //JMP 10/08/11: Code to approximate the code below, but which allows the loop to be over
    //               n = nmin; n <= nmax+1, which allows the use of 3 instead of 4 ghost cells.
    
    int n = nmin - 1; // This and next 5 lines is modified (added) code.
    da[n] = para[4][n] *diffa[n] + para[5][n] *diffa[n];
    if (da[n] >= 0.0) sgn = 1.0;
    else sgn = -1.0;
    scrch3 = min( abs(da[n]),double(2.*abs(diffa[n])));
    da[n] = sgn*scrch3;                                  //Eq26 in Fryxell
    //end of modified code

    //for (int n = nmin-1; n<= nmax+1; n++){ // Original
    for (int n = nmin; n<= nmax+1; n++){ // Modified
      if(diffa[n-1]*diffa[n] <= 0.0){
	da[n] = 0.0;
      }
      else{
	da[n] = para[4][n] *diffa[n] + para[5][n] *diffa[n-1];//Eq24 in Fryxell
        if (da[n] >= 0.0) sgn = 1.0;
        else sgn = -1.0;
        scrch3 = min( double(2.*abs(diffa[n-1])), double(2.*abs(diffa[n])) );
        scrch3 = min( abs(da[n]), scrch3 );
        da[n] = sgn*scrch3;                                  //Eq26 in Fryxell
      }
    }

    //                                                     Equation 1.6 of C&W
    //     a(j+.5) = a(j) + C1 * (a(j+1)-a(j)) + C2 * dma(j+1) + C3 * dma(j)
    // MONOT: Limit ar(n) to the range defined by a(n) and a(n+1)

    for (int n = nmin-1; n <= nmax; n++){
      ar[n] = a[n] + para[1][n]*diffa[n] + para[2][n]*da[n+1] +
	      para[3][n]*da[n];
      al[n+1] = ar[n];
    }

    //  eqn. 4.1 - flaten interpolation in zones with a shock ( flat(n)->1. )
    // (Eqs. 49 and 50 in Fryxell et al.)
    if( iflat == 1 ){
      for (int n = nmin; n <= nmax; n++){
	onemfl= 1.0 - flat[n];
	ar[n] = flat[n] * a[n] + onemfl * ar[n];
	al[n] = flat[n] * a[n] + onemfl * al[n];
      }
    }
  }

  // Contact discontinuity steepener for Eulerian code. The check for a CD is
  // made separately on the density and any composition variables (eg there may
  // be a discontinuity in composition in regions where the density is smooth)

  /*
  if (icd == 1){
    for (int n = nmin; n <= nmax; n++){
      // Construct new left- and right-interface densities using the  
      // montonized slopes (Eqs 33 and 34 in Fryxell et al)
      ali[n] = a[n-1] + 0.5*da[n-1]; // da is given in Eq.26 of Fryxell et al
      ari[n] = a[n+1] - 0.5*da[n+1];
    }
    for (int n = nmin; n <= nmax; n++){
      d2a[n] = para[6][n]*diffa[n] - para[7][n]*diffa[n-1]; // Eq 35
    }
    for (int n = nmin; n <= nmax; n++){
      jumpa = (fabs(a[n+1]-a[n-1]))/min(a[n+1],a[n-1]);
      if ((jumpa < 0.01) || (d2a[n-1]*d2a[n+1] > 0.0) ||
	  (jump[n] > 0.1*gam*jumpa)){                     //Eqs 37,38,39
	eta = 0.0;
      }
      else{
	eta = -(d2a[n+1]-d2a[n-1]);
	eta = eta*para[8][n]/(diffa[n]+diffa[n-1]);      // Eq 36
      }
      eta = max(0.0, min(double(20.0*(eta-0.05)),double(1.0))); // Eq 40
      onemeta = 1.0 - eta;
      al[n] = al[n]*onemeta + ali[n]*eta;                // Eq 41
      ar[n] = ar[n]*onemeta + ari[n]*eta;                // Eq 42
    }
  }
  */
  
  // MONOTONICITY constraints:

  // compute delta_a, a_6
  // MONOT: if a is a local max/min, flaten zone structure ar,al -> a.
  // MONOT: compute monotonized values using eq. 1.10 of C&W
  //        if parabola exceeds al/ar, reset ar/al so that slope -> 0.
  // Recalculate delta_a and a_6

  for (int n = nmin; n <= nmax; n++){
    deltaa[n]= ar[n] - al[n];                                 // Eq 29
    a6[n]    = 6. * (a[n] - .5 * (al[n] + ar[n]));            // Eq 30

    // Check for local maxima/minima
    scrch1[n] = (ar[n] - a[n]) * (a[n]-al[n]);                // Eq 52
    if (scrch1[n] < 0.0){
      ar[n] = a[n];
      al[n] = a[n];
    }

    // Check for non-monotonicity near the top or bottom of a steep gradient.
    // This set of equations is equivalent to Eqs. 53 in Fryxell (see CWB 2006
    // book p65)
    scrch1[n] = deltaa[n] * deltaa[n];                        
    scrch2[n] = deltaa[n] * a6[n];
    if(scrch1[n] < +scrch2[n]) al[n] = 3. * a[n] - 2. * ar[n];       
    if(scrch1[n] < -scrch2[n]) ar[n] = 3. * a[n] - 2. * al[n];

    deltaa[n]= ar[n] - al[n];                                 // Eq 29
    a6[n]    = 6. * (a[n] - .5 * (al[n] + ar[n]));            // Eq 30
  }


  return;
}


