/*!  \file readData.cpp
 *   \brief Read in a former solution. Only the fluid variables are updated.
 *   
 *   \author Julian Pittard (Original version 09.08.11)
 *   \version 0.1-development (Undómiel):
 *   \date Last modified: 09.08.11 (JMP)
 */

#undef MAIN

//  Header files:
#include <algorithm>
#include <cstdlib>
#include <cstring>
#include <fstream>
#include <iomanip>
#include <iostream>
#include <string>
#include "defs.h"
#include "constants.h"
#include "gas.h"
#include "global.h"
#include "grid.h"
#include "physconst.h"
#include "star.h"

#include "fitsio.h"      //  For FITSIO (from cfitsio directory)
#include "longnam.h"     //  For FITSIO (from cfitsio directory)

void read1dFitsImage(fitsfile *fptr, int mvfwdext, string datatype, long npixels, double array1d[]);
void readarray3d(fitsfile *fptr, int mvfwdext, string datatype, long npixels, int iq);
void read1Dfits(string restartfile);
void read2Dfits(string restartfile);
void read3Dfits(string restartfile);
void readFitsKeywords(fitsfile *fptr, long& npixels, long naxes[3]);
extern void FitsPrintError( int status );
extern void quit();

using namespace std;


void read1Dfits(string restartfile){
  // Read in a 1D FITS file for restarts
  cout << "read1Dfits not coded yet!";
  quit();
  return;
}

void read2Dfits(string restartfile){
  // Read in a 2D FITS file for restarts

  //Locals
  string stype;  
  fitsfile *fptr;
  int status = 0; 
  long npixels;
  long naxes[3];

  const char *filename;
  filename = restartfile.c_str();

  if (procRank == 0) cout << "Opening file " << filename << "\n";

  if ( fits_open_file(&fptr, filename, READONLY, &status) )
       FitsPrintError( status );

  // Read keywords and perform relevant checks (e.g. making sure the grid has the correct
  // dimensionality and size).
  readFitsKeywords(fptr,npixels,naxes);
  cout << "npixels = " << npixels << "\n";

  //Now read the desired arrays, specifying their indices
  readarray3d(fptr,0,"density",npixels,iqd);
  readarray3d(fptr,1,"pressure",npixels,iqe);
  readarray3d(fptr,1,"vx",npixels,iqu0);
  readarray3d(fptr,1,"vy",npixels,iqu0+1);
  //readarray3d(fptr,1,"vz",npixels,iqu0+2);
  readarray3d(fptr,1,"colour",npixels,iqal0);
  /*
  for (int m = 0; m < smax; m++){
    sprintf(cnum, "%4d",m);
    for (int i = 0; i < 4; i++){
      if (cnum[i] == ' ') cnum[i] = '0';
    }
    strcpy(type,"as");
    strcat(type,cnum);
    readarray3d(fptr,1,type,npixels,iqal0+m);
  }
  */
  
  /*
  //Read in the axes data
  // TODO: Commented out since coordinates for ghost zones must be constructed...
  read1dFitsImage(fptr,1,"zxa",naxes[0]+1,lg.zxa[nghost]);
  read1dFitsImage(fptr,1,"zya",naxes[1]+1,lg.zya[nghost]);
  read1dFitsImage(fptr,1,"zdx",naxes[0]+1,lg.zdx[nghost]);
  read1dFitsImage(fptr,1,"zdy",naxes[1]+1,lg.zdy[nghost]);
  */
  //if (procRank == 0){
  //  cout << "readin_2dfits: zxa[lg.irs]      = " << lg.zxa[lg.irs] << "\n";
  //  cout << "readin_2dfits: zxa[lg.ire]   = " << lg.zxa[lg.ire] << "\n";
  //}
  //cout << "readin_2dfits: procRank = " << procRank << "; zya[lg.jrs]   = " << lg.zya[lg.jrs] << "\n";
  //cout << "readin_2dfits: procRank = " << procRank << "; zya[lg.jre]   = " << lg.zya[lg.jre] << "\n";
  //if (procRank == 0){
  //  for (int n = 0; n < lgrid::jmaxd; n++) cout << lg.zya[n] << "\t";
  //}
  //cout << "naxes[1] = " << naxes[1] << "; lgrid::jmaxd = " << lgrid::jmaxd << "\n";
  //barrier();
  //quit;

  //  Close the output fits file.
  if ( fits_close_file(fptr, &status) ){   //  Close the file.
    FitsPrintError( status ); 
  }

  // Check for any error, and if so print out error messages
  if (status > 0){
    FitsPrintError(status);
    quit();
  }
  
  ++nfile2d;
  return;
}
  

void read3Dfits(string restartfile){
  // Read in a 3D FITS file for restarts
  
  //Locals
  char type[20];
  char cnum[5];             // needs an extra space for end of string character

  fitsfile *fptr;

  int status = 0; 
  long npixels;
  long naxes[3];

  const char *filename;
  filename = restartfile.c_str();

  if (procRank == 0) cout << "Opening file " << filename << "\n";

  if ( fits_open_file(&fptr, filename, READONLY, &status) )
       FitsPrintError( status );

  // Read keywords and perform relevant checks (e.g. making sure the grid has the correct
  // dimensionality and size).
  readFitsKeywords(fptr,npixels,naxes);
  //cout << "npixels = " << npixels << "\n";

  //Now read the desired arrays, specifying their indices
  readarray3d(fptr,0,"density",npixels,iqd);
  readarray3d(fptr,1,"pressure",npixels,iqe);
  readarray3d(fptr,1,"vx",npixels,iqu0);
  readarray3d(fptr,1,"vy",npixels,iqu0+1);
  readarray3d(fptr,1,"vz",npixels,iqu0+2);
  for (int m = 0; m < smax; m++){
    sprintf(cnum, "%4d",m);
    for (int i = 0; i < 4; i++){
      if (cnum[i] == ' ') cnum[i] = '0';
    }
    strcpy(type,"as");
    strcat(type,cnum);
    readarray3d(fptr,1,type,npixels,iqal0+m);
  }

  /*
  //Read in the axes data
  read1dFitsImage(fptr,1,"zxa",naxes[0]+1,lg.zxa[lg.irs]);
  read1dFitsImage(fptr,1,"zya",naxes[1]+1,lg.zya[lg.jrs]);
  read1dFitsImage(fptr,1,"zza",naxes[1]+1,lg.zza[lg.krs]);
  read1dFitsImage(fptr,1,"zdx",naxes[0]+1,lg.zdx[lg.irs]);
  read1dFitsImage(fptr,1,"zdy",naxes[1]+1,lg.zdy[lg.jrs]);
  read1dFitsImage(fptr,1,"zdz",naxes[1]+1,lg.zdz[lg.krs]);
  */

  // Close the output fits file
  if ( fits_close_file(fptr, &status) ){   //  Close the file
    FitsPrintError( status ); 
  }

  // Check for any error, and if so print out error messages
  if (status > 0){
    FitsPrintError(status);
    quit();
  }
  
  ++nfile3d;
  return;
}


void readFitsKeywords(fitsfile *fptr, long& npixels, long naxes[3]){

  // Locals
  char comment[FLEN_COMMENT];
  int nfound;
  int status = 0;

  //char pre_fix[30];
  //char pre_fixstats[70];
  //char chydrosolver[70];
  //int ndim;
  //double gamma,courant,avg_mass;

  // initialize naxes
  naxes[0] = 1; naxes[1] = 1; naxes[2] = 1;

  // read the NAXIS1 and NAXIS2 keyword to get image size 
  if ( fits_read_keys_lng(fptr, "NAXIS", 1, 3, naxes, &nfound, &status) )
       FitsPrintError( status );

  if ((nd == 1)&&(naxes[0] != lgrid::imax)){
    if (procRank == 0){
      cout << "Dimensions of file do not match lgrid::imax! \n";
      cout << "lgrid::imax = " << lgrid::imax << "; naxes[0] = " << naxes[0] << "\n";
    }
  }
  else if ((nd == 1)&&((naxes[0] != lgrid::imax) || (naxes[1] != lgrid::jmax) )){
    if (procRank == 0){
      cout << "Dimensions of file do not match lgrid::imax and lgrid::jmax! \n";
      cout << "lgrid::imax = " << lgrid::imax << "; naxes[0] = " << naxes[0] << "\n";
      cout << "lgrid::jmax = " << lgrid::jmax << "; naxes[1] = " << naxes[1] << "\n";
    }
  }
  else if ((nd == 3)&&((naxes[0] != lgrid::imax) || (naxes[1] != lgrid::jmax) || (naxes[2] != lgrid::kmax))){
    if (procRank == 0){
      cout << "Dimensions of file do not match lgrid::imax, lgrid::jmax and lgrid::kmax! \n";
      cout << "lgrid::imax = " << lgrid::imax << "; naxes[0] = " << naxes[0] << "\n";
      cout << "lgrid::jmax = " << lgrid::jmax << "; naxes[1] = " << naxes[1] << "\n";
      cout << "lgrid::kmax = " << lgrid::kmax << "; naxes[2] = " << naxes[2] << "\n";
    }
    quit();
  }
  /*
  if (procRank == 0){
    cout << "naxes[0] = " << naxes[0] << "\n";
    if (nd > 1) cout << "naxes[1] = " << naxes[1] << "\n";
    if (nd == 3) cout << "naxes[2] = " << naxes[2] << "\n";
  }
  */
  // Number of pixels in the image
  if      (nd == 1) npixels = naxes[0];
  else if (nd == 2) npixels = naxes[0] * naxes[1];
  else if (nd == 3) npixels  = naxes[0] * naxes[1] *naxes[2];

  if ( fits_read_key(fptr, TINT, "NCYCLE", &ncycle, comment, &status) ){
    cout << "  Error:  Can not read NCYCLE keyword.\n";
    quit();
  }
  if ( fits_read_key(fptr, TINT, "NFILE1D", &nfile1d, comment, &status) ){
    cout << "  Error:  Can not read NFILE1D keyword.\n";
    quit();
  }
  if ( fits_read_key(fptr, TINT, "NFILE2D", &nfile2d, comment, &status) ){
    cout << "  Error:  Can not read NFILE2D keyword.\n";
    quit();
  }
  if ( fits_read_key(fptr, TINT, "NFILE3D", &nfile3d, comment, &status) ){
    cout << "  Error:  Can not read NFILE3D keyword.\n";
    quit();
  }
  if ( fits_read_key(fptr, TINT, "NPRIN1D", &nprin1d, comment, &status) ){
    cout << "  Error:  Can not read NPRIN1D keyword.\n";
    quit();
  }
  if ( fits_read_key(fptr, TINT, "NPRIN2D", &nprin2d, comment, &status) ){
    cout << "  Error:  Can not read NPRIN2D keyword.\n";
    quit();
  }
  if ( fits_read_key(fptr, TINT, "NPRIN3D", &nprin3d, comment, &status) ){
    cout << "  Error:  Can not read NPRIN3D keyword.\n";
    quit();
  }
  if ( fits_read_key(fptr, TINT, "NCYCP1D", &ncycp1d, comment, &status) ){
    cout << "  Error:  Can not read NCYCP1D keyword.\n";
    quit();
  }
  if ( fits_read_key(fptr, TINT, "NCYCP2D", &ncycp2d, comment, &status) ){
    cout << "  Error:  Can not read NCYCP2D keyword.\n";
    quit();
  }
  if ( fits_read_key(fptr, TINT, "NCYCP3D", &ncycp3d, comment, &status) ){
    cout << "  Error:  Can not read NCYCP3D keyword.\n";
    quit();
  }
  // Now the doubles:  
  if ( fits_read_key(fptr, TDOUBLE, "TIME", &t, comment, &status) ){
    cout << "  Error:  Can not read TIME keyword.\n";
    quit();
  }
  if ( fits_read_key(fptr, TDOUBLE, "DT", &dt, comment, &status) ){
    cout << "  Error:  Can not read DT keyword.\n";
    quit();
  }
  if ( fits_read_key(fptr, TDOUBLE, "TIMEP1D", &timep1d, comment, &status) ){
    cout << "  Error:  Can not read TIMEP1D keyword.\n";
    quit();
  }
  if ( fits_read_key(fptr, TDOUBLE, "TIMEP2D", &timep2d, comment, &status) ){
    cout << "  Error:  Can not read TIMEP2D keyword.\n";
    quit();
  }
  if ( fits_read_key(fptr, TDOUBLE, "TIMEP3D", &timep3d, comment, &status) ){
    cout << "  Error:  Can not read TIMEP3D keyword.\n";
    quit();
  }
  if ( fits_read_key(fptr, TDOUBLE, "TPRIN1D", &tprin1d, comment, &status) ){
    cout << "  Error:  Can not read TPRIN1D keyword.\n";
    quit();
  }
  if ( fits_read_key(fptr, TDOUBLE, "TPRIN2D", &tprin2d, comment, &status) ){
    cout << "  Error:  Can not read TPRIN2D keyword.\n";
    quit();
  }
  if ( fits_read_key(fptr, TDOUBLE, "TPRIN3D", &tprin3d, comment, &status) ){
    cout << "  Error:  Can not read TPRIN3D keyword.\n";
    quit();
  }
  
  //cout << "Restarting from ncycle = " << ncycle << "; t = " << t << "; dt = " << dt << "\n";
  return;
}



void readarray3d(fitsfile *fptr, int mvfwdext, string datatype, long npixels, int iq){

  int status = 0; 
  int anynull = 0;
  long fpixel = 1;
  float nullval = 0.0;
  float datamin = 1.0E30;
  float datamax = -1.0E30;
  float *ptr_flt = new float[lgrid::imax*lgrid::jmax*lgrid::kmax];
  char rdgrid_data[FLEN_KEYWORD];
  char comment[FLEN_COMMENT];
  int hdutype;

  const char *cdatatype;
  cdatatype = datatype.c_str();  //convert string to char

  //Move to the desired extension, read the necessary keywords.

  if ( fits_movrel_hdu(fptr, mvfwdext, &hdutype, &status) ){
    cout << "  Error: Can not move to next HDU.\n";
    return;
  }
  if (hdutype != 0){
    cout << "  Error: HDU is not an image extension.\n";
    return;
  }
  //First read the data keyword
  if ( fits_read_key_str(fptr, "VH_DATA", rdgrid_data, 
			 comment, &status) ){
    cout << " Error: Can not read VH_DATA keyword.\n";
    return;
  } else {
    if (debug_lev >= 3) cout << " VH_DATA = " << rdgrid_data << endl;
  }

  //Check that the desired array is to be read
  if ( strcmp(rdgrid_data, cdatatype) != 0){
    cout << "  Error: Expected VH_DATA to be density.\n";
    quit();
    //return;
  }
  //read the data into a 1D array
  if ( fits_read_img(fptr, TFLOAT, fpixel, npixels, &nullval,
                ptr_flt, &anynull, &status) )
         FitsPrintError( status );
  int n = 0;
  //copy the data into the desired fluid Matrix
  for (int k=lg.krs;k<=lg.kre;k++){
    for (int j=lg.jrs;j<=lg.jre;j++){
      for (int i=lg.irs;i<=lg.ire;i++){
        lg.P0[iq][k][j][i] = ptr_flt[n];
        if (ptr_flt[n] < datamin) datamin = ptr_flt[n];
        if (ptr_flt[n] > datamax) datamax = ptr_flt[n];
        n++;
      }
    }
  }
  if (procRank == 0){
    if (debug_lev >= 3){
      cout << "Read in " << datatype << " array: (irs,jrs,krs) = " << lg.P0[iq][lg.krs][lg.jrs][lg.irs] << "\n";
    }
    cout << "Min and max " << datatype << " = " << datamin << ", " << datamax << "\n"; 
  }

  return;
}


void read1dFitsImage(fitsfile *fptr, int mvfwdext, string datatype, long npixels, double array1d[]){

  int status = 0; 
  int anynull = 0;
  long fpixel = 1;
  float nullval = 0.0;
  float datamin = 1.0E30;
  float datamax = -1.0E30;
  float *ptr_flt = new float[npixels];
  char rdgrid_data[FLEN_KEYWORD];
  char comment[FLEN_COMMENT];
  int hdutype;

  const char *cdatatype;
  cdatatype = datatype.c_str();  //convert string to char

  //Move to the desired extension, read the necessary keywords.

  if ( fits_movrel_hdu(fptr, mvfwdext, &hdutype, &status) ){
    cout << "  Error: Can not move to next HDU.\n";
    return;
  }
  if (hdutype != 0){
    cout << "  Error: HDU is not an image extension.\n";
    return;
  }
  //First read the data keyword
  if ( fits_read_key_str(fptr, "VH_DATA", rdgrid_data, 
			 comment, &status) ){
    cout << " Error: Can not read VH_DATA keyword.\n";
    return;
  } else {
    if (debug_lev >= 3) cout << " VH_DATA = " << rdgrid_data << endl;
  }

  //Check that the desired array is to be read
  if ( strcmp(rdgrid_data, cdatatype) != 0){
    cout << "  Error: Expected VH_DATA to be density.\n";
    quit();
    //return;
  }
  //read the data into a 1D array
  if ( fits_read_img(fptr, TFLOAT, fpixel, npixels, &nullval,
                ptr_flt, &anynull, &status) )
         FitsPrintError( status );

  //copy the data into the desired array
  for (int n = 0; n < npixels; n++){
    array1d[n] = ptr_flt[n];
    if (ptr_flt[n] < datamin) datamin = ptr_flt[n];
    if (ptr_flt[n] > datamax) datamax = ptr_flt[n];
  }
  //if (rank == 0){
  //if (debug_lev >= 3){
  //cout << "readin1dFitsImage: rank = " << rank << "; read in " << datatype << " array[0] = " << array1d[0] << "\n";
      //}
  //cout << "Min and max " << datatype << " = " << datamin << ", " << datamax << "\n"; 
    //}
    //cout << "readin1dFitsImage: rank = " << rank << "; read in " << datatype << " array[0] = " << array1d[0] << "; min = " << datamin << "; max = " << datamax << "\n";

  return;
}
