/*!  \file setUp.cpp
 *   \brief Sets up variables specifying the initial conditions.
 *
 *   Fixed parameters, in contrast, are specified in constants.h.
 *   Definitions are specified in defs.h.
 *   Users should mostly only need to setUp.cpp, defs.h and constants.h
 *
 *   \author Julian Pittard (Original version 02.06.11)
 *   \version 0.1-development (Undómiel):
 *   \date Last modified: 16.06.11 (JMP)
 */

#undef MAIN

//  Header files:
#include <cstdlib>
#include <cmath>
#include <iostream>
#include "defs.h"
#include "array.h"
#include "constants.h"
#include "funcs.h"
#include "gas.h"
#include "global.h"
#include "grid.h"
#include "pdcell.h"
#include "physconst.h"
#include "star.h"

using namespace std;

// Function declarations
void cellCentreOfMass();
void checkSetup();
double findForwardShock();
double findReverseShock();
void grid(int nzones, double xmin, double xmax, double xa[], double xc[], double dx[]);
void gridSetup();
void localGridPos();
void localGridSetup();
void mapWind(bool init);
void paraSetup();
void setupCWB();
void setupSOD();
void setupSNR();
void setupWBB();
void setupRestart(string fileName);
extern void barrier();
extern void computeTimestep();
extern void calculateEquilibriumTemperatureCurve(string coolCurveFile);
extern double equilibriumTemperature(double rho);
extern void paraset(int nmin, int nmax, double zpara[][lgrid::maxsweep], double dx[], double xa[]);
extern double pdtmp(double pdcell);
extern void quit();
extern void read1Dfits(string restartfile);
extern void read2Dfits(string restartfile);
extern void read3Dfits(string restartfile);
extern void readtempgammamu();
extern void setupHeatingAndCooling(string coolCurveFile);
extern double tmpmu(double temp);

/*!  \brief Sets up variables specifying the initial conditions.
 *
 *   Fixed parameters, in contrast, are specified in constants.h.
 *   Users should mostly only need to alter setup.cpp and constants.h
 *
 *   \author Julian Pittard (Original version 02.06.11)
 *   \version 0.1-development (Undómiel):
 *   \date Last modified: 02.06.11 (JMP)
 */
void setup(){
  //////////////////////////////////////////////////////////////////////
  //                                                                  //
  //                   SET UP CODE VARIABLES                          //
  // ...students can (carefully) alter the values specified below...  //
  //                                                                  //
  //       NOTE: Remember to recompile after making changes!          // 
  //                                                                  //
  //////////////////////////////////////////////////////////////////////  
  
  //problem = "SOD";
  //problem = "SNR";
  problem = "WBB";
  //problem = "CWB";
  reStart = false;
  //datadir = "/nobackup/phyjmp/arwenSimpleTileData/CWB3Dv3_chi0.125/";
  datadir = "";

  // Read the temperature dependent avgmass file
  readtempgammamu();
  
  if (reStart == false){
    if (problem == "SOD"){
      setupSOD();
    }
    else if (problem == "SNR"){
      setupSNR();
    }
    else if (problem == "WBB"){
      setupWBB();
    }
    else if (problem == "CWB"){
      setupCWB();
    }
    else{
      cout << "Unknown problem in setup!\n";
      quit();
    }
    computeTimestep();
  }
  else{
    // For restarts the philosophy is to: 1) set everything up as normal, as if
    // running the code from t=0.0. Then after everything is initialized,
    // 2) read in a FITS file to reset the fluid variables to their new values.
    // The time, and file counters are also updated. Finally, 3) one can modify
    // any other parameter as desired...
    // 1) Set the initial state up...
    if (problem == "SOD"){
      setupSOD();
    }
    else if (problem == "SNR"){
      setupSNR();
    }
    else if (problem == "WBB"){
      setupWBB();
    }
    else if (problem == "CWB"){
      setupCWB();
    }
    else{
      cout << "Unknown problem in setup!\n";
      quit();
    }
    // 2) Read in a FITS file to reset the fluid variables
    setupRestart("cwb_0000_2d_0010.fits");
    // 3) Modify any other variables as desired...
  }
    
  if (procRank == 0) cout << "Initial time = " << t << "; dt = " << dt << "; endtime = " << endtime << "\n";
  //for (int i = 0; i < lgrid::maxsweep; ++i){
  //  cout << i << " " << zparax[1][i] << " " << zparax[2][i] << " " << zparax[3][i] << " " << zparax[4][i] << " " << zparax[5][i] << "\n";
  //}
  //quit();
  return;
}


/*!  \brief Check that the setup is consistent.
 *
 *   \author Julian Pittard (Original version 05.06.11)
 *   \version 0.1-development (Undómiel):
 *   \date Last modified: 10.08.11 (JMP)
 */
void checkSetup(){

  if ((nd == 1) && ((lgrid::jmax != 1)||(lgrid::kmax != 1))){
    cout << "nd = " << nd << ", lgrid::jmax = " << lgrid::jmax << ", lgrid::kmax = " << lgrid::kmax << "\n";
    cout << "Aborting...\n";
    exit(EXIT_FAILURE);
  }
  if ((nd != 3) && (lgrid::kmax != 1)){
    cout << "nd = " << nd << ", lgrid::kmax = " << lgrid::kmax << "\n";
    cout << "Aborting...\n";
    exit(EXIT_FAILURE);
  }

  if ((gg.ngeomx == 0) && (gg.ngeomy == 0) && (gg.ngeomz == 0)) gg.geom = "XYZ";
  else if ((gg.ngeomx == 0) && (gg.ngeomy == 1) && (gg.ngeomz == 3)) gg.geom = "ZRP";
  else if ((gg.ngeomx == 2) && (gg.ngeomy == 4) && (gg.ngeomz == 5)) gg.geom = "RTP";
  else{
    if (procRank == 0){
      cout << "Error: geometry unrecognized!\n";
      cout << " gg.ngeomx = " << gg.ngeomx << "\n";
      cout << " gg.ngeomy = " << gg.ngeomy << "\n";
      cout << " gg.ngeomz = " << gg.ngeomz << "\n";
    }
    quit();
  }
  
  if ((nd == 3) && (gg.geom != "XYZ")){
    if (procRank == 0) cout << "Only XYZ geometry is currently supported in 3D. Aborting!\n";
    quit();
  }

  if (lgrid::mmax < max(max(lgrid::imax,lgrid::jmax),lgrid::kmax)){
    if (procRank == 0){
      cout << "Error: lgrid::mmax is too small!\n";
      cout << " lgrid::imax = " << lgrid::imax << "\n";
      cout << " lgrid::jmax = " << lgrid::jmax << "\n";
      cout << " lgrid::kmax = " << lgrid::kmax << "\n";
      cout << " lgrid::mmax = " << lgrid::mmax << "\n";
    }
    quit();
  }
  
#ifdef UNSPLIT
  if ((nd == 1 && cn > 0.4)||(nd == 2 && cn > 0.25)||(nd == 3 && cn > 0.2)){
    if (procRank == 0){
      cout << "courant number is too high (cn = " << cn << ")!\n";
      cout << " Maximum permissible values for directionally UNSPLIT solver are:\n";
      cout << "   Eulerian computation - 1D: cn = 0.4; 2D: cn = 0.25; 3D: cn = 0.2\n";
    }
    quit();
  }
#else
  if (cn > 0.6){
    if (procRank == 0){
      cout << "courant number is too high (cn = " << cn << ")!\n";
      cout << " Maximum permissible value for directionally SPLIT solver is cn = 0.6\n";
    }
    quit();
  }    
#endif

  return;
}


/*!  \brief Setup the grid.
 *
 *   \author Julian Pittard (Original version 02.06.11)
 *   \version 0.1-development (Undómiel):
 *   \date Last modified: 23.08.11 (JMP)
 */
void gridSetup(){
  gg.xmax0 = gg.xmax;
  lg.dr0 = (gg.xmax - gg.xmin)/double(gg.imax);
  
  // Create the P0 (and if necessary the Pstar) grid array(s)
  lg.P0 = create4DArray<double>(iqmax,lgrid::kmaxd,lgrid::jmaxd,lgrid::imaxd);
  if (lg.P0 == NULL){
    cout << "Failed to allocate memory for lg.P0 array - aborting!\n";
    quit();
  }
#ifdef UNSPLIT
  lg.Pstar = create4DArray<double>(iqmax,lgrid::kmaxd,lgrid::jmaxd,lgrid::imaxd);
  if (lg.Pstar == NULL){
    cout << "Failed to allocate memory for lg.Pstar array - aborting!\n";
    quit();
  }
#endif

  // Create the dei array for the cooling in radiateExact()
  lg.dei = create3DArray<double>(lgrid::kmaxd,lgrid::jmaxd,lgrid::imaxd);
  if (lg.dei == NULL){
    cout << "Failed to allocate memory for lg.dei array - aborting!\n";
    quit();
  }

#ifdef MASSLOAD  
  // Create the rhoML array for the gas mass-loading reservoir
  lg.rhoML = create3DArray<double>(lgrid::kmaxd,lgrid::jmaxd,lgrid::imaxd);
  if (lg.rhoML == NULL){
    cout << "Failed to allocate memory for lg.rhoML array - aborting!\n";
    quit();
  }
#endif
  
  // Determine position of local grid and the processor ID's of its neighbours
  localGridPos();
  
  // Define grid extent of local grid lg (xmin,xmax,ymin,ymax,zmin,zmax),
  // index range of real cells, cell coords and centre of mass, and if necessary
  // cell parabolae
  localGridSetup();

  return;
}


/*!  \brief Returns the ijk position of the local grid on the global domain and the processor ranks
 *   of neighbouring grids.
 *
 *   Calculates a map of the processor ranks in the global domain (stored in proMap).
 *   The latter is used to determine the ranks of processors to the L and R of
 *   the local processor. 
 *   
 *   \author Julian Pittard (Original version 02.08.13)
 *   \version 1.0-stable (Evenstar):
 *   \date Last modified: 02.08.13 (JMP)
 */
void localGridPos(){

  int cnt = 0;

  // Given the rank of the processor, calculate its ijk position (starting at (0,0,0))
  // (See p156 of "Hydro Book II")
  kpro = procRank/(niProc*njProc);
  jpro = (procRank - kpro*niProc*njProc)/niProc;
  ipro = procRank - kpro*niProc*njProc - jpro*niProc;

  // Calculate the processor map (this is similar to the pD->GData[n][m][l].ID_Comm_Domain
  // structure created in Athena - see init_mesh.c line 697)
  // Create 3D contiguous dynamically allocated array of ints, which stores
  // the ranks of processors in 3D space
  
  int*** proMap = create3DArray<int>(nkProc,njProc,niProc);
  if (proMap == NULL) goto on_error1;

  for (int k = 0; k < nkProc; k++){
  for (int j = 0; j < njProc; j++){
  for (int i = 0; i < niProc; i++){
    proMap[k][j][i] = cnt;
    cnt++;
  }}}

  // Calculate the rank of the processor to the left and right in each direction

  if (ipro != 0) liProID = proMap[kpro][jpro][ipro-1];
  else liProID = proMap[kpro][jpro][niProc-1];           // in case of periodic boundary
  if (ipro != niProc-1) riProID = proMap[kpro][jpro][ipro+1];
  else riProID = proMap[kpro][jpro][0];                 // in case of periodic boundary
  
  if (jpro != 0) ljProID = proMap[kpro][jpro-1][ipro];
  else ljProID = proMap[kpro][njProc-1][ipro];           // in case of periodic boundary
  if (jpro != njProc-1) rjProID = proMap[kpro][jpro+1][ipro];
  else rjProID = proMap[kpro][0][ipro];                 // in case of periodic boundary
  
  if (kpro != 0) lkProID = proMap[kpro-1][jpro][ipro];
  else lkProID = proMap[nkProc-1][jpro][ipro];           // in case of periodic boundary
  if (kpro != nkProc-1) rkProID = proMap[kpro+1][jpro][ipro];
  else rkProID = proMap[0][jpro][ipro];           // in case of periodic boundary
  
  // Deallocate memory
  delete3DArray(proMap);

  //if (debugMPI){
    //if (procRank == 1){
  //cout << "procRank = " << procRank << '\t' << ipro << '\t' << jpro << '\t' << kpro << '\t' << liProID << '\t' << riProID << '\t' << ljProID << '\t' << rjProID << '\t' << lkProID << '\t' << rkProID << "\n";
  //}
  //barrier();
  //quit();

  return;

  on_error1:
    cout << "localGridPos: Error allocating memory\n";
    quit();
}


/*!  \brief Setup grid coordinates. Calculate parabolic coefficients if necessary.
 *
 *   \author Julian Pittard (Original version 02.06.11)
 *   \version 0.1-development (Undómiel):
 *   \date Last modified: 23.08.11 (JMP)
 */
void localGridSetup(){
  
  lg.radius = 0.0;
  lg.theta = 0.0;
  lg.stheta = 0.0;

  // Set extent of local grid
  double dxgrid = (gg.xmax - gg.xmin)/double(niProc);     
  lg.xmin = gg.xmin + double(ipro)*dxgrid;
  lg.xmax = lg.xmin + dxgrid;

  double dygrid = (gg.ymax - gg.ymin)/double(njProc);     
  lg.ymin = gg.ymin + double(jpro)*dygrid;
  lg.ymax = lg.ymin + dygrid;
  
  double dzgrid = (gg.zmax - gg.zmin)/double(nkProc);     
  lg.zmin = gg.zmin + double(kpro)*dzgrid;
  lg.zmax = lg.zmin + dzgrid;

  // Set range for real cells on the grid (irs->ire inclusive)
  lg.irs = nghost;
  lg.ire = lgrid::imax+nghost-1;
  if (nd > 1){
    lg.jrs = nghost;
    lg.jre = lgrid::jmax+nghost-1;
  }
  else{
    lg.jrs = 0;
    lg.jre = 0;
  }
  if (nd == 3){
    lg.krs = nghost;
    lg.kre = lgrid::kmax+nghost-1;
  }
  else{
    lg.krs = 0;
    lg.kre = 0;
  }
    
  grid(lgrid::imax,lg.xmin,lg.xmax,lg.zxa,lg.zxc,lg.zdx);
  grid(lgrid::jmax,lg.ymin,lg.ymax,lg.zya,lg.zyc,lg.zdy);
  grid(lgrid::kmax,lg.zmin,lg.zmax,lg.zza,lg.zzc,lg.zdz);

  cellCentreOfMass();
#ifndef UNSPLIT
  paraSetup();
#endif
  return;
}


/*!  \brief Function to create coords for the grid in a given direction.
 *
 *   \param xmin (double) is the left edge of the grid.
 *   \param xmax (double) is the right edge of the grid.
 *   \param xa (double) returns the left edge coord of each grid cell.
 *   \param xa (double) returns the coord of the centre of each grid cell.
 *   \param dx (double) returns the width of each grid cell.
 *
 *   xa(girs) is left boundary location (at xmin). xa(gire+1) is right boundary 
 *   location (at xmax).
 *
 *   \author Julian Pittard (Original version 31.05.11)
 *   \version 0.1-development (Undómiel):
 *   \date Last modified: 04.08.11 (JMP)
 */
void grid(int nzones, double xmin, double xmax, double xa[], double xc[], double dx[])
{
  double dxfac = (xmax - xmin) / double(nzones);
  int ntot = nzones + 2*nghost;

  for (int n = 0; n <= ntot; n++){
    xa[n] = xmin + double(n-nghost)*dxfac;
    dx[n] = dxfac;
    xc[n] = xa[n] + 0.5*dx[n];
  }
  return;
}


/*!  \brief Setup the parabolic coefficients.
 *
 *   \author Julian Pittard (Original version 02.06.11)
 *   \version 0.1-development (Undómiel):
 *   \date Last modified: 10.08.11 (JMP)
 */
void paraSetup(){
  paraset(nghost,lgrid::imax-1+nghost,lg.zparax,lg.zdx,lg.zxa); 
  paraset(nghost,lgrid::jmax-1+nghost,lg.zparay,lg.zdy,lg.zya); 
  paraset(nghost,lgrid::kmax-1+nghost,lg.zparaz,lg.zdz,lg.zza); 
  return;
}


/*!  \brief Calculate the cell centre of mass arrays. Based on centis in mg.
 *
 *   \author Julian Pittard (Original version 02.06.11)
 *   \version 0.1-development (Undomiel):
 *   \date Last modified: 21.09.11 (JMP)
 */
void cellCentreOfMass(){

  double x1,x2,y1,y2,y1abs,y2abs;
  
  if (gg.geom == "XYZ"){
    for (int i = 0; i < lgrid::imaxd; ++i){ 
      lg.zxg[i] = 0.5*(lg.zxa[i] + lg.zxa[i+1]);
    }
    if (nd > 1){
      for (int j = 0; j < lgrid::jmaxd; ++j){ 
	lg.zyg[j] = 0.5*(lg.zya[j] + lg.zya[j+1]);
      }
    }
    if (nd == 3){
      for (int k = 0; k < lgrid::kmaxd; ++k){ 
	lg.zzg[k] = 0.5*(lg.zza[k] + lg.zza[k+1]);
      }
    }
  }
  else if (gg.geom == "ZRP"){
    for (int i = 0; i < lgrid::imaxd; ++i){ 
      lg.zxg[i] = 0.5*(lg.zxa[i] + lg.zxa[i+1]);
    }
    if (nd > 1){
      for (int j = 0; j < lgrid::jmaxd; ++j){ 
	y1 = lg.zya[j];
	y2 = lg.zya[j+1];
	if (y1 >= 0.0){
	  lg.zyg[j] = (2.0/3.0)*(y2*y2 + y1*y2 + y1*y1)/(y1+y2);
	}
	else{
	  y1abs = abs(y2);
	  y2abs = abs(y1);
	  lg.zyg[j] = -(2.0/3.0)*(y2abs*y2abs + y1abs*y2abs + y1abs*y1abs)/(y1abs+y2abs);
	}
      }
    }
    if (nd == 3){
      for (int k = 0; k < lgrid::kmaxd; ++k){ 
	lg.zzg[k] = 0.5*(lg.zza[k] + lg.zza[k+1]);
      }
    }
  }
  else if (gg.geom == "RTP"){
    for (int i = 0; i < lgrid::imaxd; ++i){ 
      x1 = lg.zxa[i];
      x2 = lg.zxa[i+1];
      lg.zxg[i] = 0.75*(x2*x2*x2 + x1*x2*(x2 + x1) + x1*x1*x1)/(x2*x2 + x1*x2 + x1*x1);
    }
    if (nd > 1){
      for (int j = 0; j < lgrid::jmaxd; ++j){ 
	y1 = lg.zya[j];
	y2 = lg.zya[j+1];
	lg.zyg[j] = (y1*cos(y1) - y2*cos(y2) + sin(y2) - sin(y1))/(cos(y1) - cos(y2));
      }
    }
    if (nd == 3){
      for (int k = 0; k < lgrid::kmaxd; ++k){ 
	lg.zzg[k] = 0.5*(lg.zza[k] + lg.zza[k+1]);
      }
    }
  }
  else{
    if (procRank == 0) cout << "cellCentreOfMass: error - not coded for geometry " << gg.geom << ". Aborting!\n";
    quit();
  }

  return;
}



/*!  \brief Sets up a SOD shock tube problem for the initial conditions.
 *
 *   \author Julian Pittard (Original version 02.06.11)
 *   \version 0.1-development (Undómiel):
 *   \date Last modified: 21.09.11 (JMP)
 */
void setupSOD(){

  gg.geom = "XYZ";
  gg.ngeomx = 0;    
  gg.ngeomy = 0;
  gg.ngeomz = 0;

  gg.nleftx = 0; // reflecting
  gg.nlefty = 0;
  gg.nleftz = 0;

  gg.nrightx = 0;
  gg.nrighty = 0;
  gg.nrightz = 0;

  gg.xmin = 0.0;
  gg.xmax = 1.0;
  gg.ymin = 0.0;
  gg.ymax = 1.0;
  gg.zmin = 0.0;
  gg.zmax = 1.0;

  ncycle = 0;
  ncycend = 100;
  nprin1d = 1000000;
  nprin2d = 10;
  endtime = 1.0e99;
  tprin1d = 1.0e99;
  tprin2d = 1.0e99;
  tprin3d = 1.0e99;
  
  // Other values (do not change without good reason)
  t = 0.0;
  svel = 1.0;
  prefx = "sod";
  
  //tmin = 1.0e4;
  //tmax = 1.0e9;
  
  // Check inconsistencies and setup grid...
  checkSetup();
  gridSetup();

  // Set up ambient medium and ejecta
  for (int k = lg.krs; k <= lg.kre; ++k){
    for (int j = lg.jrs; j <= lg.jre; ++j){
      for (int i = lg.irs; i <= lg.ire; ++i){
	lg.P0[iqd][k][j][i] = 0.125;
        lg.P0[iqe][k][j][i] = 0.1;
	for (int n = 0; n < nd; ++n){
	  lg.P0[iqu0+n][k][j][i] = 0.0;
	}
	for (int n = 0; n < smax; ++n){
	  lg.P0[iqal0+n][k][j][i] = 0.0;
	}      
	if (nd == 1){
	  float x = lg.zxc[i];
	  if (x < 0.5*(gg.xmin + gg.xmax)){
  	    lg.P0[iqd][k][j][i] = 1.0;
  	    lg.P0[iqe][k][j][i] = 1.0;
	    lg.P0[iqal0][k][j][i] = 1.0;
	  }
	}
	else if (nd == 2){
	  float r1 = 0.5*gg.xmax*gg.ymax;
	  float x = lg.zxc[i];
	  float y = lg.zyc[j];
	  float r2 = x*x + y*y;
	  if (r2 <= r1){
  	    lg.P0[iqd][k][j][i] = 1.0;
  	    lg.P0[iqe][k][j][i] = 1.0;
	    lg.P0[iqal0][k][j][i] = 1.0;
	  }
	}
      }
    }
  }
  
  return;
}


/*!  \brief Sets up a supernova remnant problem for the initial conditions.
 *
 *   \author Julian Pittard (Original version 02.06.11)
 *   \version 0.1-development (Undómiel):
 *   \date Last modified: 21.09.11 (JMP)
 */
void setupSNR(){

  // 1D spherically symmetric
  gg.geom = "RTP";
  gg.ngeomx = 2;    
  gg.ngeomy = 4;
  gg.ngeomz = 5;

  /*
  // 2D ZRP
  gg.geom = "ZRP";
  gg.ngeomx = 0;    
  gg.ngeomy = 1;
  gg.ngeomz = 3;
  */
  
  gg.nleftx = 0;
  gg.nlefty = 0;
  gg.nleftz = 0;

  gg.nrightx = 1;
  gg.nrighty = 1;
  gg.nrightz = 1;

  // Grid extent
  gg.xmin = 0.0;
  gg.xmax = 100.0*pc;
  gg.ymin = 0.0;
  gg.ymax = 0.0;
  //gg.ymax = 15.0*pc;
  gg.zmin = 0.0;
  gg.zmax = 0.0;

  ncycle = 0;
  ncycshow = 1000;
  ncycend = 1e9;
  nprin1d = 1e9;
  nprin2d = 1e9;
  endtime = 1.0e5*yr;
  tprin1d = 1.0e4*yr;
  tprin2d = 1.0e99;
  tprin3d = 1.0e99;
  t = 0.0;
  svel = 1.0;
    
  tmin = 1.0e4;
  tmax = 1.0e9;
  
  esn = 1.0e51;      // ergs
  msn = 10.0;        // Msol
  ramb = 1.0e-24;    // g/cm^-3
  tamb = 1.0e4;      // K
  uamb = 0.0;
  avgmassamb = 1.0e-24; // g
  pamb = (ramb/avgmassamb)*boltzman*tamb;
    
  // Set mass-loading parameters
  clumpToInterClumpMassRatio = 10.0;
  fML = 1.0;           // Set the mass-loading "strength"

  // Set dust parameters
  //double grainRadiusMicrons = 0.1;
  //grainBulkDensity = 3.0;                       // (g/cm^3)
  //dustToGasMassRatio = 0.0075;

  // Check inconsistencies and setup grid...
  checkSetup();
  gridSetup();

  remapRadius = 6.0*lg.zdx[0];
  prefx = "snr";

  double vol = (4.0/3.0)*pi*pow(remapRadius,3);
  double drho = msn*msol/vol;
  double dp = gamm*esn/vol;
  double n0 = ramb/massh; // ambient number density of H atoms (assume 100% H)
  double tSF = 2.87e4*yr*std::pow(esn/1.0e51,3.0/14.0)*std::pow(n0,-4.0/7.0);
  q0 = 6.0*ramb/(5.0*tSF);

  cout << "eSN = " << esn << " (erg); rhoamb = " << ramb << " (g cm^-3); tSF = " << tSF/yr << " (yr); fML = " << fML << "; clumpToInterclumpMassRatio = " << clumpToInterClumpMassRatio << "\n";
  
  // Set up ambient medium and ejecta. For ZRP geometry, j is R and i is Z.
  for (int k = lg.krs; k <= lg.kre; ++k){
    double zc = lg.zzc[k];
    double zc2 = zc*zc;
    for (int j = lg.jrs; j <= lg.jre; ++j){
      double yc = lg.zyc[j];
      double yc2 = yc*yc;
      for (int i = lg.irs; i <= lg.ire; ++i){
        double xc = lg.zxc[i];
        double xc2 = xc*xc;
	double r = sqrt(xc2 + yc2 + zc2);
	lg.P0[iqd][k][j][i] = ramb;
        lg.P0[iqe][k][j][i] = pamb;
	for (int n = 0; n < nd; ++n){
	  lg.P0[iqu0+n][k][j][i] = 0.0;
	}
	for (int n = 0; n < smax; ++n){
	  lg.P0[iqal0+n][k][j][i] = 0.0;
	}
	if (r < remapRadius){
	  lg.P0[iqd][k][j][i] += drho;
	  lg.P0[iqe][k][j][i] += dp;
  	  lg.P0[iqal0][k][j][i] = 1.0;
	}
      }
    }
  }

#ifdef MASSLOAD
  // Setup the mass-loading reservoir
  for (int k = lg.krs; k <= lg.kre; ++k){
    for (int j = lg.jrs; j <= lg.jre; ++j){
      for (int i = lg.irs; i <= lg.ire; ++i){
	lg.rhoML[k][j][i] = clumpToInterClumpMassRatio*ramb;
      }
    }
  }
#endif
  
  return;
}


/*!  \brief Sets up a wind-blown-bubble problem for the initial conditions.
 *
 *   \author Julian Pittard (Original version 02.06.11)
 *   \version 0.1-development (Undómiel):
 *   \date Last modified: 21.09.11 (JMP)
 */
void setupWBB(){

  tmin = 1.0e0; // use 1.0e4 when tamb=1e4
  tmax = 1.0e9; // 1e10 for cooling_curves_lambdaJMP_fromWang2014.txt, 1e9 for cooling_KI02_4.0_CLOUDY_7.6_MEKAL_correct_low.txt
  double Xsol = 0.7381;  // Grevesse et al. (2010) mass fractions
  double Ysol = 0.2485;
  double Zsol = 0.0134;
      
  double nH_over_nHe = (Xsol/1.0)/(Ysol/4.0);
  double nH_over_nC = (Xsol/1.0)/(Zsol/12.0);

  stars[0].mdot = 1.0e-7*msol/yr;
  stars[0].vinf = 2.0e8;
  stars[0].xpos = 0.0;
  stars[0].ypos = 0.0;
  stars[0].zpos = 0.0;

  // Set the ambient density. Given the cooling curve and heating rate, calculate the ambient temperature.
  // Then calculate the temperature-dependent mu, and finally pamb.
  ramb = 2.0e-21;        // g/cm^3
  GammaHeat = 2.0e-26; ; // DO NOT CHANGE THIS - THE KOYAMA & INUTSUKA PART OF THE COOLING CURVE DEPENDS ON IT!!!

  string coolCurveFile = "cooling_KI02_4.0_CLOUDY_7.6_MEKAL_correct_low.txt";
  setupHeatingAndCooling(coolCurveFile);

  // For a given value of GammaHeat, calculate the equilibrium temperature as a function of rho
  //calculateEquilibriumTemperatureCurve(coolCurveFile);
  //quit();

  // Calculate the other properties of the ambient medium:
  // equilibrium temperature, average particle mass and thermal pressure.
  tamb = equilibriumTemperature(ramb);
  //tamb = 1.0e8;
  muamb = tmpmu(tamb);
  avgmassamb = muamb*massh;
  pamb = (ramb/avgmassamb)*boltzman*tamb;
  uamb = 0.0;

  // The avgerage particle mass of the ambient medium massamb should be set appropriate for the temperature
  //if (tamb < 50.0){ // assume molecular H and atomic He
  //  avgmassamb = massh/(0.5*Xsol + 0.25*Ysol); 
  //}
  //else if (tamb < 8000.0){ // assume neutral
  //  avgmassamb = massh/(Xsol + 0.25*Ysol + Zsol/6.0);     // neutral H and He, singly ionized metals (C)
  //}
  //else{ // assume fully ionized
  //  avgmassamb = massh/(2.0*Xsol + 0.75*Ysol + 0.5*Zsol); 
  //}
  
  // Comment in only 1 of options A or B... 
  // Option A: set nH
  //double nH = 100.0;    // H nuclei number density (cm^-3)
  //double nHe = nH/nH_over_nHe;
  //double nC = nH/nH_over_nC;
  //ramb = (nH + 4.0*nHe + 12.0*nC)*massh; 
  
  // Option B: set ramb = (1.0 + 4.0/nH_over_nHe + 12.0/nH_over_nC)*nH*massh
  //ramb = 1.0e-22; // g/cm^3
  double nH = (ramb/massh)/(1.0 + 4.0/nH_over_nHe + 12.0/nH_over_nC);
  //double nHe = nH/nH_over_nHe;
  //double nC = nH/nH_over_nC;
  
  //std::cout << "nH = " << nH << "; nHe = " << nHe << "; nC = " << nC << "; avgmassamb = " << avgmassamb << "; ramb = " << ramb << "\n";
  //quit();

  
  //quit();


  // Set dust parameters
  double grainRadiusMicrons = 0.1;
  //grainBulkDensity = 3.0;                       // (g/cm^3)
  //dustToGasMassRatio = 0.0075;

  // Set mass-loading parameters
  clumpToInterClumpMassRatio = 10.0;
  double tMS = 5.0e6*yr;  // Main Sequence lifetime of star
  fML = 0.0e10;           // Set the mass-loading "strength"

  // Determine the size of the bubble and other useful quantities
  double edot = 0.5*stars[0].mdot*stars[0].vinf*stars[0].vinf;    // erg/s
  double rbub = pow(edot*125.0/(154.0*pi*ramb),0.2)*pow(tMS,0.6); // radius of adiabatic WBB at end of MS stage
  double volbub = 4.0*pi*pow(rbub,3)/3.0;
  q0 = stars[0].mdot/volbub;

  if (procRank == 0){
    cout << "nH = " << nH << " (cm^-3); rho_amb = " << ramb << " (g/cm^3); T_amb = " << tamb << " (K); P_amb = " << pamb << " (dyn/cm^2); P_amb/k = " << pamb/boltzman << " (K/cm^3); muamb = " << muamb << "; average particle mass of ambient gas = " << avgmassamb << " (g)\n";
    cout << "T_min = " << tmin << " (K); T_max = " << tmax << " (K)\n";
    cout << "Mdot = " << stars[0].mdot/(msol/yr) << " (Msol/yr); vinf = " << stars[0].vinf/1.0e5 << " (km/s); Edot = " << edot << " (erg/s); tMS = " << tMS/(1.0e6*yr) << " (Myr)\n";
    cout << "Adiabatic bubble radius at t=tMS is " << rbub/pc << " (pc); q0 = " << q0 << " (g cm^-3 s^-1)\n";    
    cout << "Using solar abundances...\n";
    double tSF = 1.3e4*pow(edot/1.0e38,3.0/11.0)*pow(nH,-8.0/11.0); // yr (Eq 4.3 Koo & McKee 1992 Paper I)
    cout << "Shell formation time (tSF) = " << tSF << " (yr)\n";
    double tcool = 0.031*pow(stars[0].vinf/1.0e8,25)*pow(edot,-2)*pow(ramb,-3); // yr (see p13 in "HII/WBB/SNR 2019" book)
    double lambda0 = 1.0e-23;
    double tcool_JMP = 3.2e-4*pow(stars[0].vinf*stars[0].vinf*avgmassamb,10)*pow(edot,-2)*pow(ramb,-3)*pow(lambda0,-5)/yr; // my estimate at t_cool
    //cout << "Cooling time of shocked stellar wind (tcool) = " << tcool << " (yr)\n";
    //cout << "Cooling time of shocked stellar wind (tcoolJMP) = " << tcool_JMP << " (yr)\n";
    double cisoamb = sqrt(boltzman*tamb/avgmassamb); // isothermal sound speed in ambient medium
    double tP_adia = 2.9e7*sqrt(edot/(1.0e38*nH*pow(cisoamb/1.0e6,5))); // yr (Eq 4.5 Koo & McKee 1992 Paper I)
    double Min = stars[0].vinf/cisoamb; // injection Mach number
    double tP_PRB = 2.2e8*pow(edot*pow(Min/100.0,7)/(1.0e38*nH*nH*pow(stars[0].vinf/1.0e8,2)),1.0/3.0); // yr (Eq. 3.8a Koo & McKee Paper II)
    cout << "Pressure confinement time for an adiabatic bubble          (tP_adia) = " << tP_adia << " (yr)\n";
    cout << "Pressure confinement time for a partially radiative bubble (tP_PRB)  = " << tP_PRB << " (yr)\n";
    cout << "\n Expected evolution of bubble:\n";
    if (tcool > tMS/yr){
      cout << "   Bubble stays adiabatic during its lifetime (tcool > tMS)\n";
      if (tP_adia < tMS) cout << "   but becomes pressure confined (tP_adia < tMS)\n";
      else               cout << "   and is not pressure confined (tP_adia > tMS)\n";
    }
    else{
      cout << "   Bubble becomes partially radiative during its lifetime (tcool < tMS)\n";
      if (tP_PRB < tMS/yr) cout << "   and becomes pressure confined (tP_PRB < tMS)\n";
      else               cout << "   and is not pressure confined (tP_PRB > tMS)\n";      
    }
    //quit();
  }
  
  gg.geom = "RTP";
  gg.ngeomx = 2;    
  gg.ngeomy = 4;
  gg.ngeomz = 5;

  gg.nleftx = 0;
  gg.nlefty = 0;
  gg.nleftz = 0;

  gg.nrightx = 1;
  gg.nrighty = 1;
  gg.nrightz = 1;

  // Grid extent
  gg.xmin = 0.0;
  //gg.xmax = 1.2*rbub; // 10.0*pc
  double dr = 1.0e-5*pc;
  gg.xmax = gg.imax*dr;
  gg.ymin = 0.0;
  gg.ymax = 1.0;
  gg.zmin = 0.0;
  gg.zmax = 1.0;
  

  /*
  gg.geom = "ZRP";
  gg.ngeomx = 0;    
  gg.ngeomy = 1;
  gg.ngeomz = 3;
  
  gg.nleftx = 0; // reflecting
  gg.nlefty = 0;
  gg.nleftz = 0;

  gg.nrightx = 1; // outflow
  gg.nrighty = 1;
  gg.nrightz = 0;
  
  gg.xmin = 0.0;
  gg.xmax = 10.0*pc;
  gg.ymin = 0.0;
  gg.ymax = 10.0*pc;
  gg.zmin = 0.0;
  gg.zmax = 1.0;
  */
  
  /*
  gg.geom = "XYZ";
  gg.ngeomx = 0;    
  gg.ngeomy = 0;
  gg.ngeomz = 0;
  
  gg.nleftx = 1; 
  gg.nlefty = 1;
  gg.nleftz = 1;

  gg.nrightx = 1;
  gg.nrighty = 1;
  gg.nrightz = 1;
  
  gg.xmin = -10.0*pc;
  gg.xmax = 10.0*pc;
  gg.ymin = -10.0*pc;
  gg.ymax = 10.0*pc;
  gg.zmin = -10.0*pc;
  gg.zmax = 10.0*pc;
  */
  
  ncycend = 1e9;// 1e9
  ncycshow = 1000; // 1000
  nprin1d = 1e9;
  nprin2d = 1e9;
  nprin3d = 1e9;
  endtime = tMS; // 1.0e5*yr
  tprin1d = 5.0e0*yr; // 1.0e4*yr
  tprin2d = 1.0e99;
  tprin3d = 1.0e99;

  // Check inconsistencies and setup grid...
  checkSetup();
  gridSetup();

  if (smax != 2){
    cout << "Error: smax must be equal to 2!\n";
    quit();
  }
  
  remapi = 10;//10,6
  remapRadius = float(remapi)*lg.zdx[0];

  double rRS_stall = sqrt(stars[0].mdot*stars[0].vinf/(4.0*pi*pamb)); 
  cout << "dr = " << lg.zdx[0]/pc << " (pc); remapi = " << remapi << "; remapRadius = " << remapRadius/pc << " (pc); r_RSstall = " << rRS_stall/pc << " (pc)\n";
  //quit();
  
  // Set up ambient medium. For ZRP geometry, j is R and i is Z.
  for (int k = lg.krs; k <= lg.kre; ++k){
    for (int j = lg.jrs; j <= lg.jre; ++j){
      for (int i = lg.irs; i <= lg.ire; ++i){
	lg.P0[iqd][k][j][i] = ramb;
        lg.P0[iqe][k][j][i] = pamb;
	for (int n = 0; n < nd; ++n){
	  lg.P0[iqu0+n][k][j][i] = 0.0;
	}
//#ifdef DUST
//	lg.P0[iqal0][k][j][i] = dustToGasMassRatio;
//#else
//	for (int n = 0; n < smax; ++n){
//	  lg.P0[iqal0+n][k][j][i] = 0.0;
//	}
//#endif
	lg.P0[iqal0]  [k][j][i] = 0.0; // wind    "colour"
	lg.P0[iqal0+1][k][j][i] = 1.0; // ambient "colour"
      }
    }
  }

#ifdef MASSLOAD
  // Setup the mass-loading reservoir
  for (int k = lg.krs; k <= lg.kre; ++k){
    for (int j = lg.jrs; j <= lg.jre; ++j){
      for (int i = lg.irs; i <= lg.ire; ++i){
	lg.rhoML[k][j][i] = clumpToInterClumpMassRatio*ramb;
      }
    }
  }
#endif
  
  // Set up wind
  mapWind(false);

  //remapi = 10;//6
  //remapRadius = float(remapi)*lg.zdx[0];

  // Other values (do not change without good reason)
  t = 0.0;
  svel = 1.0;
  prefx = "wbb";
  grainRadius = grainRadiusMicrons*1.0e-4; // cm
  
  return;
}


/*!  \brief Set up the CWB problem.
 *
 *   \author Julian Pittard (Original version 02.06.11)
 *   \version 0.1-development (Undómiel):
 *   \date Last modified: 21.09.11 (JMP)
 */
void setupCWB(){

  /*
  // 2D ZRP
  gg.geom = "ZRP";
  gg.ngeomx = 0;    
  gg.ngeomy = 1;
  gg.ngeomz = 3;
  
  gg.nleftx = 1; 
  gg.nlefty = 0;
  gg.nleftz = 0;

  gg.nrightx = 1;
  gg.nrighty = 1;
  gg.nrightz = 0;
  
  gg.xmin = -1.0e12;
  gg.xmax = 1.0e12;
  gg.ymin = 0.0;
  gg.ymax = 2.0e12;
  gg.zmin = 0.0;
  gg.zmax = 1.0;
  */
  
  
  // 3D XYZ
  gg.geom = "XYZ";
  gg.ngeomx = 0;    
  gg.ngeomy = 0;
  gg.ngeomz = 0;

  gg.nleftx = 1; // outflow
  gg.nlefty = 1;
  gg.nleftz = 1;

  gg.nrightx = 1;
  gg.nrighty = 1;
  gg.nrightz = 1;

  gg.xmin = -2.0e12;
  gg.xmax = 2.0e12;
  gg.ymin = -2.0e12;
  gg.ymax = 2.0e12;
  gg.zmin = -2.0e12;
  gg.zmax = 2.0e12;
  
  
  ncycshow = 10;
  ncycend = 100;
  nprin1d = 1e9;
  nprin2d = 1e9;
  nprin3d = 20;
  endtime = 1.0e99;
  tprin1d = 1.0e99;
  tprin2d = 1.0e99;
  tprin3d = 1.0e99;
    
  tmin = 1.0e4;
  tmax = 1.0e9;

  stars[0].mdot = 0.625e-7*msol/yr;
  stars[0].vinf = 2.00e8;
  stars[0].xpos = -9.0e11;
  stars[0].ypos = 0.0;
  stars[0].zpos = 0.0;

  stars[1].mdot = 0.625e-7*msol/yr;
  stars[1].vinf = 2.00e8;
  stars[1].xpos = 9.0e11;
  stars[1].ypos = 0.0;
  stars[1].zpos = 0.0;

  tamb = 1.0e4;         // K
  avgmassamb = 1.0e-24; // g
  remapi = 10;
    
  // Check inconsistencies and setup grid...
  checkSetup();
  if (nstars != 2){
    if (procRank == 0){
      cout << "setupCWB: nstars not equal to 2. Aborting!";
    }
    quit();
  }
  gridSetup();

  remapRadius = remapi*lg.zdx[0];

  dsep = sqrt(pow(stars[1].xpos - stars[0].xpos,2) + pow(stars[1].ypos - stars[0].ypos,2) + pow(stars[1].zpos - stars[0].zpos,2));
  eta = stars[1].mdot*stars[1].vinf/(stars[0].mdot*stars[0].vinf);        // wind mtm ratio
  rob = (sqrt(eta)/(1.0 + sqrt(eta)))*dsep;//distance of stagnation point from star 1 (distance from star 0 is rwr)

  // Check that the WCR has enough room not to interfere with the remap regions
  if (int(rob/lg.zdx[0]) < remapi + 10){
    if (procRank == 0){
      cout << "setupCWB: WCR too close to star 1. Aborting!";
      cout << "rob = " << rob << "; rob/lg.zdx[0] = " << rob/lg.zdx[0] << "\n";
    }
    quit();
  }
  if (int((dsep - rob)/lg.zdx[0]) < remapi + 10){
    if (procRank == 0){
      cout << "setupCWB: WCR too close to star 0. Aborting!";
      cout << "dsep = " << dsep << "; rob = " << rob << "; (dsep - rob)/lg.zdx[0] = " << (dsep - rob)/lg.zdx[0] << "\n";
    }
    quit();
  }
  
  // Other values (do not change without good reason)
  t = 0.0;
  svel = 1.0;
  prefx = "cwb";

  // Initialize winds
  mapWind(true);
    
  return;
}


/*!  \brief Map a wind solution onto the grid.
 *
 *   \author Julian Pittard (Original version 05.06.11)
 *   \version 0.1-development (Undómiel):
 *   \date Last modified: 10.08.11 (JMP)
 */
void mapWind(bool init){

  double xc, yc, zc, xc2, yc2, zc2, r2, r, xy;
  double sinphi, cosphi, costhta, sinthta;
  double xmaxst1;
  double mdot1 = stars[0].mdot;
  double vinf1 = stars[0].vinf;
  double xpos1 = stars[0].xpos;
  double ypos1 = stars[0].ypos;
  double zpos1 = stars[0].zpos;
  double mdot2,vinf2,xpos2,ypos2,zpos2;
  double mdot,vinf,xpos,ypos,zpos;
  double edot,drho,rold,eold,uxold,etotold,colold,rnew,uxnew,uynew,uznew,kenew,dp;
  double Minjold,Minj,vr,rho,ke,ie;
  int istar,jstar,kstar,istl,istu,jstl,jstu,kstl,kstu;

  // Wind injection options. At most one of these should be true.
  // Apart from wunschInject and directRemap, all others are set up for 1D only.
  // directRemap, wunschInject and geenInject are the 3 key methods (see "WBB Book 2019" p65).
  // See Wunsch et al. (2008), Gatto et al. (2017), Haid et al. (2018) and Geen et al. (2021).
  const bool directRemap = true;
  const bool wunschInject = false;
  const bool gattoInject = false;
  const bool haidInject = false;
  const bool geenInject = false;

  // Modify the algorithm in Gatto to account for the energy already within the injection region
  const bool gattoInjectJMPmod = false; 
  
  // Calculate volume of energy injection
  const double volinj = 4.0*pi*remapRadius*remapRadius*remapRadius/3.0;

  if (nstars > 1){
    mdot2 = stars[1].mdot;
    vinf2 = stars[1].vinf;
    xpos2 = stars[1].xpos;
    ypos2 = stars[1].ypos;
    zpos2 = stars[1].zpos;
  }

  if (!init){
    for (int n = 0; n < nstars; ++n){
      mdot = stars[n].mdot;
      vinf = stars[n].vinf;
      xpos = stars[n].xpos;
      ypos = stars[n].ypos;
      zpos = stars[n].zpos;
      // Determine extent of remap region.
      // lg.irs is the index of the first real cell, lg.ire is the last real cell
      istar = int((xpos - lg.zxa[0])/lg.zdx[0]);
      jstar = 0;
      kstar = 0;
      istl = max(lg.irs,istar - remapi - 2);
      istu = min(lg.ire,istar + remapi + 2);
      jstl = 0;
      jstu = 0;
      kstl = 0;
      kstu = 0;
      if (nd > 1){
	jstar = int((ypos - lg.zya[0])/lg.zdy[0]);
	jstl = max(lg.jrs,jstar - remapi - 2);
	jstu = min(lg.jre,jstar + remapi + 2);
      }
      if (nd == 3){
	kstar = int((zpos - lg.zza[0])/lg.zdz[0]); 
	kstl = max(lg.krs,kstar - remapi - 2);
	kstu = min(lg.kre,kstar + remapi + 2);
      }

      // If gattoInject we need to determine the sum of the mass inside the injection region
      if (gattoInject){
	Minjold = 0.0;
	etotold = 0.0;
        for (int k = kstl; k <= kstu; ++k){
          zc = lg.zzg[k] - zpos;
          zc2 = zc*zc;
          for (int j = jstl; j <= jstu; ++j){
            yc = lg.zyg[j] - ypos;
            yc2 = yc*yc;
            for (int i = istl; i <= istu; ++i){
	      xc = lg.zxg[i] - xpos;
	      xc2 = xc*xc;
	      r2 = xc2 + yc2 + zc2;
	      r = sqrt(r2);
	      if (r <= remapRadius){
		double ri = lg.zxa[i] - xpos;
		double ro = ri + lg.zdx[i];
		double vol = 4.0*pi*(std::pow(ro,3) - std::pow(ri,3))/3.0;
		rho = lg.P0[iqd][k][j][i];
	        Minjold += rho*vol;
		vr = lg.P0[iqu0][k][j][i];
		ke = 0.5*rho*vr*vr;
		ie = lg.P0[iqe][k][j][i]/gamm;
		etotold += (ke*ie)*vol;
	      }
	    }
	  }
	}
  	edot = 0.5*mdot*vinf*vinf;
	Minj = Minjold + mdot*dt; // total mass in the injection region
	
	if (gattoInjectJMPmod){
	  double mu = tmpmu(tmin);
	  ie = (Minj/mu*massh)*boltzman*tmin/gamm;    // total internal energy of the mass if set to tmin
	  vr = sqrt(2.0*(etotold + edot*dt - ie)/Minj); // accounts for the energy already in the injection region and setting the mass in the injection region to tmin
	  //cout << "vr = " << vr << " (cm/s)\n";
	}
	else{
  	  vr = sqrt(2.0*edot*dt/Minj); // Eq7 of Gatto et al. (2017)
	}
      }
      
      
      // for (int k = lg.krs; k <= lg.kre; ++k){
      for (int k = kstl; k <= kstu; ++k){
        zc = lg.zzg[k] - zpos;
        zc2 = zc*zc;
        //for (int j = lg.jrs; j <= lg.jre; ++j){
        for (int j = jstl; j <= jstu; ++j){
          yc = lg.zyg[j] - ypos;
          yc2 = yc*yc;
          //for (int i = lg.irs; i <= lg.ire; ++i){
          for (int i = istl; i <= istu; ++i){
	    xc = lg.zxg[i] - xpos;
	    xc2 = xc*xc;
	    r2 = xc2 + yc2 + zc2;
	    r = sqrt(r2);
	    xy = sqrt(xc2 + yc2);
	    sinphi = xy/r;
	    cosphi = zc/r;
	    costhta = xc/xy;
	    sinthta = yc/xy;
	    if (r <= remapRadius){
	      if (wunschInject){
                // If injecting energy (i.e. a Chevalier & Clegg (1985) wind), use the procedure
                // in Wunsch et al. (2008, ApJ, 683, 683):
    	        //    1) Add mass to cell
  	        //    2) Conserve mtm: v_new rho_new = v_old rho_old
	        //    3) Calculate KE from rho_new and v_new, and subtract from old total energy
                //       to give new internal energy prior to the addition of new energy
	        //    4) Add new energy to new internal energy
  	        edot = 0.5*mdot*vinf*vinf;
	        drho = mdot*dt/volinj;
	        rold = lg.P0[iqd][k][j][i];
	        eold = lg.P0[iqe][k][j][i]/gamm;
                uxold = lg.P0[iqu0][k][j][i];
		double uyold = 0.0;
		double uzold = 0.0;
                if (nd > 1)  uyold = lg.P0[iqu0+1][k][j][i];
                if (nd == 3) uzold = lg.P0[iqu0+2][k][j][i];
                etotold = eold + 0.5*rold*(uxold*uxold + uyold*uyold + uzold*uzold);
	        rnew = rold + drho;

                lg.P0[iqd][k][j][i] = rnew;
   	        lg.P0[iqu0][k][j][i]                *= rold/rnew; // conservation of mtm
	        if (nd > 1)  lg.P0[iqu0+1][k][j][i] *= rold/rnew; // conservation of mtm
	        if (nd == 3) lg.P0[iqu0+2][k][j][i] *= rold/rnew; // conservation of mtm
                uxnew = lg.P0[iqu0][k][j][i];
                uynew = lg.P0[iqu0+1][k][j][i];
                uznew = lg.P0[iqu0+2][k][j][i];

                kenew = 0.5*rnew*(uxnew*uxnew + uynew*uynew + uznew*uznew);
	        lg.P0[iqe][k][j][i] = gamm*(etotold - kenew);       // conservation of energy
                dp = gamm*edot*dt/volinj;
	        lg.P0[iqe][k][j][i] += dp;
		colold = lg.P0[iqal0][k][j][i];
		lg.P0[iqal0][k][j][i] = (colold*rold + drho)/rnew;
		colold = lg.P0[iqal0+1][k][j][i];
		lg.P0[iqal0+1][k][j][i] = (colold*rold)/rnew;
	      }
	      else if (haidInject){ // (ONLY CODED FOR 1D)
                // Use the procedure in Haid et al. (2018, MNRAS, 478, 4799):
    	        //    1) Add mass to cell
  	        //    2) Add mtm of new mass to cell
	        //    3) Calculate the new internal energy of the cell
  	        edot = 0.5*mdot*vinf*vinf;
	        drho = mdot*dt/volinj;
	        rold = lg.P0[iqd][k][j][i];
	        eold = lg.P0[iqe][k][j][i]/gamm;
                double uold = lg.P0[iqu0][k][j][i];
                etotold = eold + 0.5*rold*uold*uold;
	        rnew = rold + drho;

                lg.P0[iqd][k][j][i] = rnew;
   	        lg.P0[iqu0][k][j][i] = (rold*uold + drho*vinf)/rnew;
		double dv = uold - vinf;
		double enew = eold + 0.5*(rold*drho)*dv*dv/rnew; // new internal energy
		lg.P0[iqe][k][j][i] = gamm*enew;

		colold = lg.P0[iqal0][k][j][i];
		lg.P0[iqal0][k][j][i] = (colold*rold + drho)/rnew;
		colold = lg.P0[iqal0+1][k][j][i];
		lg.P0[iqal0+1][k][j][i] = (colold*rold)/rnew;
		
	      }
	      else if (geenInject){ // (ONLY CODED FOR 1D)
                // Use the procedure in Geen et al. (2021, MNRAS):
    	        //    1) Add mass to cell
  	        //    2) Add mtm of new mass to cell
	        //    3) Add ke of new mass to cell
		// (Calculate new internal energy to make it properly conserving)
  	        edot = 0.5*mdot*vinf*vinf;
	        drho = mdot*dt/volinj;
	        rold = lg.P0[iqd][k][j][i];
	        eold = lg.P0[iqe][k][j][i]/gamm;
                double uold = lg.P0[iqu0][k][j][i];
                etotold = eold + 0.5*rold*uold*uold;
	        rnew = rold + drho;

                lg.P0[iqd][k][j][i] = rnew;
		double unew = (rold*uold + drho*vinf)/rnew;
		lg.P0[iqu0][k][j][i] = unew;
		double etotnew = etotold + 0.5*drho*vinf*vinf;
		kenew = 0.5*rnew*unew*unew;
		double enew = etotnew - kenew;
		lg.P0[iqe][k][j][i] = gamm*enew;

		colold = lg.P0[iqal0][k][j][i];
		lg.P0[iqal0][k][j][i] = (colold*rold + drho)/rnew;
		colold = lg.P0[iqal0+1][k][j][i];
		lg.P0[iqal0+1][k][j][i] = (colold*rold)/rnew;
	      }		
	      else if (gattoInject){
		// It is not clear what should happen to the pressure (internal energy of the cell).
		// Gatto say in step iii) of their algorithm that "The mass that is added to the
		// injection region carries a certain amount of internal energy, which we take into account.
		// However, all of the new energy is added as kinetic energy so I will not adjust the
		// pressure of the cell
	        drho = mdot*dt/volinj;
	        rold = lg.P0[iqd][k][j][i];
  	        //eold = lg.P0[iqe][k][j][i]/gamm;
		//de = drho*

	        rnew = rold + drho;

                lg.P0[iqd][k][j][i] = rnew;
		lg.P0[iqu0][k][j][i] = vr;

		// All of the energy in the injection region has now been turned into kinetic. Set the
		// pressure to correspond to the temperature at the floor value (tmin).
		// Because we do not set it to zero, we overestimate the inject
		double mu = tmpmu(tmin);
		double avgmass = mu*massh;
	        if (gattoInjectJMPmod) lg.P0[iqe][k][j][i] = (lg.P0[iqd][k][j][i]/avgmass)*boltzman*tmin;
		  
		colold = lg.P0[iqal0][k][j][i];
		lg.P0[iqal0][k][j][i] = (colold*rold + drho)/rnew;
		colold = lg.P0[iqal0+1][k][j][i];
		lg.P0[iqal0+1][k][j][i] = (colold*rold)/rnew;
		  
	      }
	      else if (directRemap){
		bool subSampleCell = true;
		if (subSampleCell){
                  double dr = lg.zxa[istl+1] - lg.zxa[istl];
		  double ri = lg.zxa[i];
		  double ro = ri + dr;
		  double ri3 = ri*ri*ri;
		  double ro3 = ro*ro*ro;
		  double vol = 4.0*pi*(ro3 - ri3)/3.0;
		  double dmass = mdot*(ro - ri)/vinf;
		  double rho = dmass/vol;
	          lg.P0[iqd][k][j][i] = rho;

		  /*
		  const int ns = 10;
                  double dr = lg.zxa[istl+1] - lg.zxa[istl];
		  double ddr = dr/double(ns);		  
		  for (int n = 0; n < ns; ++n){
		    double ri = lg.zxa[i] + double(n)*ddr;
		    double rc = ri + 0.5*ddr;
		    double ro = ri + ddr;
		    double r2 = rc*rc;
		    double vol = 
		  }
		  */
		}
		else{
	          lg.P0[iqd][k][j][i] = mdot/(4.0*pi*r2*vinf);
		}
		double Twind = 1.0e4;
		double mu = tmpmu(Twind);
		double avgmass = mu*massh;
	        lg.P0[iqe][k][j][i] = (lg.P0[iqd][k][j][i]/avgmass)*boltzman*Twind;
	        lg.P0[iqu0][k][j][i] = vinf*sinphi*costhta;
	        if (nd > 1)  lg.P0[iqu0+1][k][j][i] = vinf*sinphi*sinthta;
	        if (nd == 3) lg.P0[iqu0+2][k][j][i] = vinf*cosphi;
//#ifdef DUST
//	        lg.P0[iqal0][k][j][i] = 0.0; // zero dust in the wind
//#else
//	        for (int m = 0; m < smax; ++m) lg.P0[iqal0+m][k][j][i] = 1.0 - float(n);
//#endif
	        lg.P0[iqal0]  [k][j][i] = 1.0; // wind    "colour" (mass fraction)
	        lg.P0[iqal0+1][k][j][i] = 0.0; // ambient "colour" (mass fraction)		
	      }
	      else{
		cout << "[mapWind] - option not recognised. Aborting!\n";
		quit();
	      }
	    }
          }
        }
      }
    } // nstars
  }   // init T/F?
  else{
    // Initial setup of gas on the grid (used for directRemap case only)
    if (directRemap){
    if (problem == "CWB") xmaxst1 = xpos1 + dsep - rob; // maximum x-extent of wind
    else xmaxst1 = lg.zxa[lgrid::imax+1];
    //cout << "xmaxst1 = " << xmaxst1 << "\n";

    double Twind = 1.0e4;
    double mu = tmpmu(Twind);
    double avgmass = mu*massh;

    // First map the wind of star 0
    for (int k = lg.krs; k <= lg.kre; ++k){
      zc = lg.zzg[k] - zpos1;
      zc2 = zc*zc;
      for (int j = lg.jrs; j <= lg.jre; ++j){
        yc = lg.zyg[j] - ypos1;
        yc2 = yc*yc;
        for (int i = lg.irs; i <= lg.ire; ++i){
	  if (lg.zxc[i] < xmaxst1){
	    xc = lg.zxg[i] - xpos1;
	    xc2 = xc*xc;
	    r2 = xc2 + yc2 + zc2;
	    r = sqrt(r2);
	    xy = sqrt(xc2 + yc2);
	    sinphi = xy/r;
	    cosphi = zc/r;
	    costhta = xc/xy;
	    sinthta = yc/xy;
	    lg.P0[iqd][k][j][i] = mdot1/(4.0*pi*r2*vinf1);
	    //cout << mdot1 << " " << r << " " << vinf1 << " " << lg.P0[iqd][k][j][i] << "\n";
	    //quit();
	    lg.P0[iqe][k][j][i] = (lg.P0[iqd][k][j][i]/avgmass)*boltzman*Twind;
	    lg.P0[iqu0][k][j][i] = vinf1*sinphi*costhta;
            if (nd > 1)  lg.P0[iqu0+1][k][j][i] = vinf1*sinphi*sinthta;
	    if (nd == 3) lg.P0[iqu0+2][k][j][i] = vinf1*cosphi;
//#ifdef DUST
//	    lg.P0[iqal0][k][j][i] = 0.0; // zero dust in the wind
//#else
//	    for (int m = 0; m < smax; ++m) lg.P0[iqal0+m][k][j][i] = 1.0;
//#endif
	      lg.P0[iqal0]  [k][j][i] = 1.0; // wind    "colour" (mass fraction)
	      lg.P0[iqal0+1][k][j][i] = 0.0; // ambient "colour" (mass fraction)
          }
        }
      }
    }
    // Now map the wind of star 1
    for (int k = lg.krs; k <= lg.kre; ++k){
      zc = lg.zzg[k] - zpos2;
      zc2 = zc*zc;
      for (int j = lg.jrs; j <= lg.jre; ++j){
        yc = lg.zyg[j] - ypos2;
        yc2 = yc*yc;
        for (int i = lg.irs; i <= lg.ire; ++i){
	  if (lg.zxc[i] >= xmaxst1){
	    xc = lg.zxg[i] - xpos2;
	    xc2 = xc*xc;
	    r2 = xc2 + yc2 + zc2;
	    r = sqrt(r2);
	    xy = sqrt(xc2 + yc2);
	    sinphi = xy/r;
	    cosphi = zc/r;
	    costhta = xc/xy;
	    sinthta = yc/xy;
	    lg.P0[iqd][k][j][i] = mdot2/(4.0*pi*r2*vinf2);
	    lg.P0[iqe][k][j][i] = (lg.P0[iqd][k][j][i]/avgmass)*boltzman*Twind;
	    lg.P0[iqu0][k][j][i] = vinf2*sinphi*costhta;
	    if (nd > 1)  lg.P0[iqu0+1][k][j][i] = vinf2*sinphi*sinthta;
	    if (nd == 3) lg.P0[iqu0+2][k][j][i] = vinf2*cosphi;
#ifdef DUST
	    lg.P0[iqal0][k][j][i] = 0.0; // zero dust in the wind
#else
	    for (int m = 0; m < smax; ++m) lg.P0[iqal0+m][k][j][i] = 0.0;
#endif
          }
        }
      }
    }
    }// directRemap?
  }
      
  return;
}



void setupRestart(string fileName){
  if      (nd == 1) read1Dfits(fileName);
  else if (nd == 2) read2Dfits(fileName);
  else              read3Dfits(fileName);
  return;
}


/*!  \brief Find the forward shock position.
 *
 *   \author Julian Pittard (Original version 05.06.11)
 *   \version 0.1-development (Undómiel):
 *   \date Last modified: 10.08.11 (JMP)
 */
double findForwardShock(){
  int j = 0;
  int k = 0;
  double rshk = 0.0;
  for (int i = lg.ire; i >= lg.irs; --i){
    if (lg.P0[iqd][k][j][i] > 1.1*ramb){
      rshk = lg.zxc[i];
      break;
    }
  }
  return rshk;
}


/*!  \brief Find the reverse shock position.
 *
 *   \author Julian Pittard (Original version 05.06.11)
 *   \version 0.1-development (Undómiel):
 *   \date Last modified: 10.08.11 (JMP)
 */
double findReverseShock(){
  int j = 0;
  int k = 0;
  double rshk = 0.0;
  double shockedTemp;
  if (problem == "WBB"){
    shockedTemp = 3.0*avgmassamb*stars[0].vinf*stars[0].vinf/(16.0*boltzman); // see Stevens & Pollock (1994)
  }
  else if (problem == "SNR"){
    double n = msn*msol/1.0e-24; // total number of particles in ejecta
    //shockedTemp = esn/(1.5*n*boltzman);
    shockedTemp = 1.0e6; // to find older SNRs
  }
  else{
    cout << "[findReverseShock] - Not coded for other problems. Aborting!\n"; quit();
  }

  // Search inwards from right boundary. Find last transition from hot to cold.
  
  double rho = lg.P0[iqd][k][j][lg.ire];
  double pre = lg.P0[iqe][k][j][lg.ire];
  //double tmpold = pdtmp(pre/rho);
  //for (int i = lg.ire; i >= lg.irs; --i){
  for (int i = lg.irs; i <= lg.ire; ++i){
    rho = lg.P0[iqd][k][j][i];
    pre = lg.P0[iqe][k][j][i];
    double temp = pdtmp(pre/rho);

    //if (tmpold > 0.1*shockedTemp && temp < max(0.1*tmpold,1.1*tamb)){
    //rshk = lg.zxc[i+1];
    //}
    if (temp > 0.1*shockedTemp){
      if (problem == "WBB"){
        if (lg.zxc[i] > float(remapi+3)*lg.zdx[0]){
	  rshk = lg.zxc[i];
	  break;
	}
	else{
          cout << "ABORTING - reverse shock is too close to remap region!\n";
	  quit();
	}
      }
    }
    //tmpold = temp;
  }
  return rshk;
}
