/*!  \file stitch.cpp
 *   \brief Read in 2D or 3D FITS files, stitch them together, and write out a single file
 *
 *   For stitch to compile, it needs mpicxx as its compiler...
 *
 *   Compile: make stitch
 *
 *   \author Julian Pittard (Original version 25.05.11)
 *   \version 0.1-development (Undómiel):
 *   \date Last modified: 18.11.11 (JMP)
 */

#define MAIN

//  Header files:
#include <cstdlib>
#include <cstring>
#include <iomanip>
#include <iostream>
#include <string>
// For stitch to compile using the readData.cpp and writeData.cpp files,
// ALL of the .h files that these files include must also be included here...
#include "array.h"
#include "defs.h"
#include "constants.h"
#include "gas.h"
#include "global.h"
#include "grid.h"
#include "physconst.h"
#include "star.h"

#include "fitsio.h"      //  For FITSIO (from cfitsio directory)
#include "longnam.h"     //  For FITSIO (from cfitsio directory)

using namespace std;

// Function declarations
void createAxisData(double xmin, double xmax, double ymin, double ymax, double zmin, double zmax);
string generateBigFilename(int nfile, int ndim, string ext);
void prin2DBigFits(string outfile);
void prin3DBigFits(string outfile);
void putIntoGlobalGrid(int i, int j, int k);
void quit();
void setLocalGridRanges();
void write2dBigFitsImage(fitsfile *fptr, const char* extname, long naxis, long naxes[2], int iq, int k);
void write3dBigFitsImage(fitsfile *fitsptr, const char* extname, long naxis, long naxes[3], int iq);
extern void FitsPrintError( int status );
extern string generateFilename(int nfile, int ndim, string ext);
extern void read1Dfits(string infile);
extern void read2Dfits(string infile);
extern void read3Dfits(string infile);
extern void writeFitsKeywords(fitsfile *fptr);
extern void write2dFitsImage(fitsfile *fptr, const char* extname, long nzones, float array1d[]);

// File scope variables
double**** bigP;
float zxa[IMAXX+1],zdx[IMAXX+1];
float zya[JMAXX+1],zdy[JMAXX+1];
float zza[KMAXX+1],zdz[KMAXX+1];

int main(int argc, char *argv[]){

  int nfile_min = 3;
  int nfile_max = 5;
  
  /*
  // SOD
  prefx = "sod";
  double xmin = 0.0;
  double xmax = 1.0;
  double ymin = 0.0;
  double ymax = 1.0;
  double zmin = 0.0;
  double zmax = 1.0;
  */

  /*
  // WBB
  prefx = "wbb";
  double xmin = 0.0;
  double xmax = 10.0*pc;
  double ymin = 0.0;
  double ymax = 10.0*pc;
  double zmin = 0.0;
  double zmax = 1.0;
  */
  
  // CWB
  problem = "CWB";
  prefx = "cwb";
  // 3D XYZ
  gg.geom = "XYZ";
  gg.ngeomx = 0;    
  gg.ngeomy = 0;
  gg.ngeomz = 0;

  gg.nleftx = 1; // outflow
  gg.nlefty = 1;
  gg.nleftz = 1;

  gg.nrightx = 1;
  gg.nrighty = 1;
  gg.nrightz = 1;

  gg.xmin = -6.0e12;
  gg.xmax = 6.0e12;
  gg.ymin = -6.0e12;
  gg.ymax = 6.0e12;
  gg.zmin = -6.0e12;
  gg.zmax = 6.0e12;
  
  ncycshow = 50;
  ncycend = 1e9;
  nprin1d = 1e9;
  nprin2d = 1e9;
  nprin3d = 1e9;
  endtime = 2.0e5;
  tprin1d = 1.0e99;
  tprin2d = 1.0e99;
  tprin3d = 2.0e4;
    
  tmin = 3.0e4;
  tmax = 1.0e9;

  stars[0].mdot = 2.5e-8*msol/yr;
  stars[0].vinf = 1.00e8;
  stars[0].xpos = 0.0e12;
  stars[0].ypos = 0.0;
  stars[0].zpos = 0.0;

  stars[1].mdot = 2.5e-9*msol/yr;
  stars[1].vinf = 1.00e8;
  stars[1].xpos = 3.0e12;
  stars[1].ypos = 0.0;
  stars[1].zpos = 0.0;

  tamb = 3.0e4;         // K
  avgmassamb = 1.0e-24; // g
  remapi = 10;

  double xmin = gg.xmin;
  double xmax = gg.xmax;
  double ymin = gg.ymin;
  double ymax = gg.ymax;
  double zmin = gg.zmin;
  double zmax = gg.zmax;

  //double xmin = -4.0e12;
  //double xmax = 4.0e12;
  //double ymin = -8.0e12;
  //double ymax = 8.0e12;
  //double zmin = -8.0e12;
  //double zmax = 8.0e12;

  // Create grids
  lg.P0 = create4DArray<double>(iqmax,lgrid::kmaxd,lgrid::jmaxd,lgrid::imaxd);
  bigP = create4DArray<double>(iqmax,KMAXX,JMAXX,IMAXX);

  setLocalGridRanges();

  // Loop over snapshots
  for (int nfile = nfile_min; nfile <= nfile_max; ++nfile){ 
  // Read in tiled data files, and store in bigP
  for (int k = 0; k < NKPRO; ++k){
    for (int j = 0; j < NJPRO; ++j){
      for (int i = 0; i < NIPRO; ++i){
	procRank = k*(NJPRO*NIPRO) + j*NIPRO + i;
        string fileName = generateFilename(nfile,nd,"fits");
        if      (nd == 1){
	  read1Dfits(fileName);
	  nfile1d--;
	}
        else if (nd == 2){
	  read2Dfits(fileName);
	  nfile2d--;
	}
        else{
	  read3Dfits(fileName);
	  nfile3d--;
	}
        //cout << "Finished reading in file " << fileName << "\n";
	putIntoGlobalGrid(i,j,k);
      }
    }
  }

  // Generate axis data
  createAxisData(xmin,xmax,ymin,ymax,zmin,zmax);

  // Write out big (stitched) file
  string bigFileName = generateBigFilename(nfile,nd,"fits");
  cout << "Writing out file " << bigFileName << "\n";
  if      (nd == 2) prin2DBigFits(bigFileName);
  else if (nd == 3) prin3DBigFits(bigFileName);
  else{
    cout << "Dimensionality not recognised. Aborting!\n";
  }
  }
  
  delete4DArray(lg.P0);
  delete4DArray(bigP);
  quit();  
}


void createAxisData(double xmin, double xmax, double ymin, double ymax, double zmin, double zmax){
  double dx = (xmax - xmin)/double(IMAXX);
  double dy = (ymax - ymin)/double(JMAXX);
  double dz = (zmax - zmin)/double(KMAXX);
  for (int i = 0; i <= IMAXX; ++i){
    zxa[i] = xmin + double(i)*dx;
    zdx[i] = dx;
  }
  for (int j = 0; j <= JMAXX; ++j){
    zya[j] = ymin + double(j)*dy;
    zdy[j] = dy;
  }
  for (int k = 0; k <= KMAXX; ++k){
    zza[k] = zmin + double(k)*dz;
    zdz[k] = dz;
  }
  
  return;
}

void setLocalGridRanges(){
  // Set range for real cells on the grid (irs->ire inclusive)
  lg.irs = nghost;
  lg.ire = lgrid::imax+nghost-1;
  if (nd > 1){
    lg.jrs = nghost;
    lg.jre = lgrid::jmax+nghost-1;
  }
  else{
    lg.jrs = 0;
    lg.jre = 0;
  }
  if (nd == 3){
    lg.krs = nghost;
    lg.kre = lgrid::kmax+nghost-1;
  }
  else{
    lg.krs = 0;
    lg.kre = 0;
  }

}


string generateBigFilename(int nfile, int ndim, string ext)
//Generate the name for creating the big stitched file.
{
  char cnum[5];             // needs an extra space for end of string character
  char ndnum[2];            // needs an extra space for end of string character
  string srank,snfile,sdim;
  string filename;

  // Create strings for procRank and nfile
  sprintf(cnum, "%4d",procRank);
  for (int i = 0; i < 4; i++){
    if (cnum[i] == ' ') cnum[i] = '0';
  }
  srank = cnum;

  sprintf(cnum, "%4d",nfile);
  for (int i = 0; i < 4; i++){
    if (cnum[i] == ' ') cnum[i] = '0';
  }
  snfile = cnum;

  sprintf(ndnum, "%1d",ndim);
  sdim = ndnum;
  sdim += 'd';
 
  // The filename takes the form: prefx_BB_CCCC.ext
  //         BB = dimensionality of data   (e.g. 1d)
  //       CCCC = file number in sequence  (e.g. 0010)
  
  filename = prefx + '_' + sdim + '_' + snfile + "." + ext;

  return filename;
}


void putIntoGlobalGrid(int ip, int jp, int kp){
  for (int iq = 0; iq < iqmax; ++iq){
    for (int k = lg.krs, k2 = 0; k <= lg.kre; ++k, ++k2){
      int kbig = kp*lgrid::kmax + k2;
      for (int j = lg.jrs, j2 = 0; j <= lg.jre; ++j, ++j2){
        int jbig = jp*lgrid::jmax + j2;
        for (int i = lg.irs, i2 = 0; i <= lg.ire; ++i, ++i2){
          int ibig = ip*lgrid::imax + i2;
	  bigP[iq][kbig][jbig][ibig] = lg.P0[iq][k][j][i];
	  //if ((iq == 0)&&(j < 3)){
	  //  cout << kbig << "\t" << jbig << "\t" << ibig << "\t" << bigP[iq][kbig][jbig][ibig] << "\n";
	  //}
	}
      }
    }
  }
  return;
}


void quit(){
  exit(EXIT_SUCCESS);
}


void prin2DBigFits(string outfile){

  // Write out a 2D FITS file compatible with the X-ray code
  char type[20];
  char cnum[5];             // needs an extra space for end of string character

  float scrh;
  FILE *fptr;
  
  //  FITSIO variables.
  int simple = 1;
  int extend = 1;
  int status = 0;                         //  FITSIO status.
  fitsfile *fitsptr;                      //  Pointer to fits file.
  int bitpix = FLOAT_IMG;                 //  Single precision.
  long naxis = 2;
  long naxes[2];
  long group = 1;
  long fpixel[2];
  long nelements;
  char extname[FLEN_VALUE];
  
  const char *filename;     //  Character to store filename

  const char* ptrbuf1;      //  To call FITS routines expecting a const char
  char buf1[30];            //  without an error

  int k = 0;
  if (nd == 3) k = lgrid::kmax/2;

  filename = outfile.c_str();                   // convert to character array

  // Delete any existing file of that name and write out to file
  remove( filename );
  if ( fits_create_file(&fitsptr, filename, &status) ){ 
    FitsPrintError( status );
  }

  naxes[0] = ggrid::imax;
  naxes[1] = ggrid::jmax;

  //  Create the density image.
  if ( fits_create_img(fitsptr, bitpix, naxis, naxes, &status) ){
    cerr << "  Error creating rho data image" << endl;
    FitsPrintError( status );
  }

  //  Write keywords
  writeFitsKeywords(fitsptr);

  // Write the first (density) image
  //  Put the data into a 1-D array  
  float *ptr_flt = new float[ggrid::imax*ggrid::jmax];
  int n = 0;
  for (int j=0; j < ggrid::jmax; j++){
    for (int i=0; i < ggrid::imax; i++){
      ptr_flt[n] = bigP[iqd][0][j][i];
      //cout << ptr_flt[i] << "\t";
      n++;
    }
  }
    
  //  Write the 1-D array, cunningly stored in the same way as a 2-D
  //  image, into the FITS image.
  if ( fits_write_2d_flt(fitsptr, group, naxes[0], 
                           naxes[0], naxes[1], ptr_flt, &status) ){ 
    FitsPrintError( status );
  } 
  
  delete []ptr_flt; // clean up array

  // Create and write the other data as (2d) images. We pass in
  // the variable index in the primitive array, and in write2dFitsImage()
  // floats are written to file (to reduce the file size).
  strcpy(type,"pressure");
  write2dBigFitsImage(fitsptr,type,naxis,naxes,iqe,k);
  strcpy(type,"vx");
  write2dBigFitsImage(fitsptr,type,naxis,naxes,iqu0,k);
  strcpy(type,"vy");
  write2dBigFitsImage(fitsptr,type,naxis,naxes,iqu0+1,k);
  //strcpy(type,"vz");
  //write2dFitsImage(fitsptr,"vz",naxis,naxes,iqu0+2,k);
  strcpy(type,"colour");
  write2dBigFitsImage(fitsptr,type,naxis,naxes,iqal0,k);
  /*
  for (int m = 0; m < smax; m++){
    sprintf(cnum, "%4d",m);
    for (int i = 0; i < 4; i++){
      if (cnum[i] == ' ') cnum[i] = '0';
    }
    strcpy(type,"as");
    strcat(type,cnum);
    write2dFitsImage(fitsptr,type,naxis,naxes,iqal0+m,k); //advected scalars
  }
  */

  //Write axis data as 2D images
  write2dFitsImage(fitsptr,"zxa",naxes[0]+1,zxa);
  write2dFitsImage(fitsptr,"zya",naxes[1]+1,zya);
  write2dFitsImage(fitsptr,"zdx",naxes[0]+1,zdx);
  write2dFitsImage(fitsptr,"zdy",naxes[1]+1,zdy);
  
  //  Close the output fits file.
  if ( fits_close_file(fitsptr, &status) ){   //  Close the file.
    FitsPrintError( status ); 
  }
  cout << " Finished writing " << outfile << endl;

  // Check for any error, and if so print out error messages
  if (status > 0){
    FitsPrintError(status);
    quit();
  }

  return;
}



void write2dBigFitsImage(fitsfile *fitsptr, const char* extname, long naxis, long naxes[2], int iq, int k){

  int status = 0;
  long group = 1;
  int bitpix = FLOAT_IMG;                 //  Single precision.

  //  Create the new image.
  if ( fits_create_img(fitsptr, bitpix, naxis, naxes, &status) ){
    cerr << "  Error creating " << extname << " data image" << endl;
    FitsPrintError( status );
  }
  if ( fits_update_key(fitsptr, TSTRING,"VH_DATA", (char *)extname, 
			 "Extension name", &status ) ){
    cerr << "  Error writing EXTNAME version keyword."  << endl;
    FitsPrintError( status );
  }
     
  // Put the desired data into a 1-D array of floats and write floats to
  // disk to reduce the file size (by not writing doubles)
  float *ptr_flt = new float[ggrid::imax*ggrid::jmax];
  int n = 0;
  for (int j=0; j < ggrid::jmax; j++){
    for (int i=0; i < ggrid::imax; i++){
      ptr_flt[n] = bigP[iq][k][j][i];
      n++;
    }
  }
    
  //  Write the 1-D array, cunningly stored in the same way as a 2-D
  //  image, into the FITS image.
  if ( fits_write_2d_flt(fitsptr, group, naxes[0], 
			   naxes[0], naxes[1], ptr_flt, &status) ){ 
    FitsPrintError( status );
  } 

  // Write pixel keywords
  //writeFitsKeywordsPixels(fitsptr);

  delete []ptr_flt; // clean up array

  return;
}


void prin3DBigFits(string outfile){

  // Write out a 3D FITS file compatible with the X-ray code
  // Locals

  int ndim;
  char type[20];
  char cnum[5];             // needs an extra space for end of string character
  
  //  FITSIO variables.
  int status = 0;                         //  FITSIO status.
  fitsfile *fitsptr;                      //  Pointer to fits file.
  int bitpix = FLOAT_IMG;                 //  Single precision.
  long naxis = 3;
  long naxes[3];
  long group = 1;
  
  const char *filename;     //  Character to store filename

  const char* ptrbuf1;      //  To call FITS routines expecting a const char
  char buf1[30];            //  without an error

  filename = outfile.c_str();                   // convert to character array

  // Delete any existing file of that name and write out to file
  remove( filename );
  if ( fits_create_file(&fitsptr, filename, &status) ){ 
    FitsPrintError( status );
  }

  naxes[0] = ggrid::imax;
  naxes[1] = ggrid::jmax;
  naxes[2] = ggrid::kmax;

  //  Create the density image.
  if ( fits_create_img(fitsptr, bitpix, naxis, naxes, &status) ){
    cerr << "  Error creating rho data image" << endl;
    FitsPrintError( status );
  }

  //  Write keywords
  writeFitsKeywords(fitsptr);

  // Write the first (density) image
  //  Put the data into a 1-D array  
  float *ptr_flt = new float[ggrid::imax*ggrid::jmax*ggrid::kmax];
  int n = 0;
  for (int k=0; k < ggrid::kmax; k++){
    for (int j=0; j < ggrid::jmax; j++){
      for (int i=0; i < ggrid::imax; i++){
        ptr_flt[n] = bigP[iqd][k][j][i];
        //cout << ptr_flt[i] << "\t";
        n++;
      }
    }
  }

  //  Write the 1-D array, cunningly stored in the same way as a 3-D
  //  image, into the FITS image.
  if ( fits_write_3d_flt(fitsptr, group, naxes[0], naxes[1],
			 naxes[0], naxes[1], naxes[2], ptr_flt, &status) ){ 
    FitsPrintError( status );
  } 

  delete []ptr_flt; // clean up array

  // Create and write the other data as (3d) images. We pass in
  // the variable index in the primitive array, and in write3dFitsImage()
  // floats are written to file (to reduce the file size).
  strcpy(type,"pressure");
  write3dBigFitsImage(fitsptr,type,naxis,naxes,iqe);
  strcpy(type,"vx");
  write3dBigFitsImage(fitsptr,type,naxis,naxes,iqu0);
  strcpy(type,"vy");
  write3dBigFitsImage(fitsptr,type,naxis,naxes,iqu0+1);
  strcpy(type,"vz");
  write3dBigFitsImage(fitsptr,type,naxis,naxes,iqu0+2);
  //strcpy(type,"colour");
  //write3dBigFitsImage(fitsptr,type,naxis,naxes,iqal0);
  for (int m = 0; m < smax; m++){
    sprintf(cnum, "%4d",m);
    for (int i = 0; i < 4; i++){
      if (cnum[i] == ' ') cnum[i] = '0';
    }
    strcpy(type,"as");
    strcat(type,cnum);
    write3dBigFitsImage(fitsptr,type,naxis,naxes,iqal0+m); //advected scalars
  }
 
  //Write axis data as 2D images
  write2dFitsImage(fitsptr,"zxa",naxes[0]+1,zxa);
  write2dFitsImage(fitsptr,"zya",naxes[1]+1,zya);
  write2dFitsImage(fitsptr,"zza",naxes[2]+1,zza);
  write2dFitsImage(fitsptr,"zdx",naxes[0]+1,zdx);
  write2dFitsImage(fitsptr,"zdy",naxes[1]+1,zdy);
  write2dFitsImage(fitsptr,"zdz",naxes[2]+1,zdz);

  //  Close the output fits file.
  if ( fits_close_file(fitsptr, &status) ){   //  Close the file.
    FitsPrintError( status ); 
  }
  cout << " Finished writing " << outfile << endl;

  // Check for any error, and if so print out error messages
  if (status > 0){
    FitsPrintError(status);
    quit();
  }

  return;
}


void write3dBigFitsImage(fitsfile *fitsptr, const char* extname, long naxis, long naxes[3], int iq){

  int status = 0;
  long group = 1;
  int bitpix = FLOAT_IMG;                 //  Single precision.

  //  Create the new image.
  if ( fits_create_img(fitsptr, bitpix, naxis, naxes, &status) ){
    cerr << "  Error creating rho data image" << endl;
    FitsPrintError( status );
  }
  if ( fits_update_key(fitsptr, TSTRING,"VH_DATA", (char *)extname, 
			 "Extension name", &status ) ){
    cerr << "  Error writing EXTNAME version keyword."  << endl;
    FitsPrintError( status );
  }
     
  // Put the desired data into a 1-D array of floats and write floats to
  // disk to reduce the file size (by not writing doubles)
  float *ptr_flt = new float[ggrid::imax*ggrid::jmax*ggrid::kmax];
  int n = 0;
  for (int k=0; k < ggrid::kmax; k++){
    for (int j=0; j < ggrid::jmax; j++){
      for (int i=0; i < ggrid::imax; i++){
        ptr_flt[n] = bigP[iq][k][j][i];
        n++;
      }
    }
  }
    
  //  Write the 1-D array, cunningly stored in the same way as a 3-D
  //  image, into the FITS image.
  if ( fits_write_3d_flt(fitsptr, group, naxes[0], naxes[1], 
			 naxes[0], naxes[1], naxes[2], ptr_flt, &status) ){ 
    FitsPrintError( status );
  } 
  
  // Write pixel keywords (required for every image extension in a FITS file)
  //writeFitsKeywordsPixels(fitsptr);

  delete []ptr_flt; // clean up array

  return;
}



