/*!  \file writeData.cpp
 *   \brief Write out the solution
 *
 *   \author Julian Pittard (Original version 09.08.11)
 *   \version 0.1-development (Undómiel):
 *   \date Last modified: 09.08.11 (JMP)
 */

#undef MAIN

//  Header files:
#include <algorithm>
#include <cstdlib>
#include <cstring>
#include <fstream>
#include <iomanip>
#include <iostream>
#include <string>
#include "defs.h"
#include "constants.h"
#include "gas.h"
#include "global.h"
#include "globalMPI.h"
#include "grid.h"
#include "physconst.h"
#include "star.h"

#include "fitsio.h"      //  For FITSIO (from cfitsio directory)
#include "longnam.h"     //  For FITSIO (from cfitsio directory)
// File scope
int bitpix = FLOAT_IMG;                 //  Single precision.

// Function declarations
string generateFilename(int nfile, int ndim, string ext);
void FitsPrintError( int status );
void writeFitsKeywords(fitsfile *fptr);
void writeFitsKeywordsCWB(fitsfile *fitsptr);
void writeFitsKeywordsPixels(fitsfile *fitsptr);
void write2dFitsImage(fitsfile *fptr, const char* extname, long naxis, long naxes[2], int iq, int k);
void write2dFitsImage(fitsfile *fptr, const char* extname, long nzones, float array1d[]);
void write3dFitsImage(fitsfile *fptr, const char* extname, long naxis, long naxes[3], int iq);
extern double findForwardShock();
extern double findReverseShock();
extern double global_sum(double& sendbuf);
extern double pdtmp(double pdcell);
extern void quit();

using namespace std;

void prinShkRadiusAndEnergyAndMassFractions(){

  // Note: It is best to calculate the velocity and deceleration parameter (eta) from
  //       post-processing the shock radius output (trying to do this within the code
  //       does not produce satisfactory results). Use the code calculateVelocity.cpp
  //       to do this. For this reason, we do not calculate the shock velocity and eta
  //       here.

  string outfile;
  int j=0;
  int k=0;
  static bool first = true;
  static double old_fshk_radius = 0.0;
  static double old_fshk_time = 0.0;

  double fshk_radius, fshk_vel, rshk_radius;
  double dr = lg.zdx[0];
  
  if (ncycle == 0){ 
    fshk_radius = remapRadius;
    //fshk_vel = 0.0;
    rshk_radius = remapRadius;
  }
  else {
    fshk_radius = findForwardShock();
    //fshk_vel = (fshk_radius - old_fshk_radius)/(t - old_fshk_time);
    rshk_radius = findReverseShock();
  }

  double KEtot = 0.0;
  double IEtot = 0.0;
  double mtmtot = 0.0; // radial momentum
  double windMass = 0.0;
  double sweptUpMass = 0.0;
  double injectedMass = 0.0;
  double Mhot = 0.0;   // "hot" gas is anything with T > 2e4 K (cf. Kim & Ostriker 2015, ApJ, 802,99)
  double Mshell = 0.0; // "shell" gas has T < 2e4 K and v_r > 1 km/s (cf. KO15)

  double gridmtmtot = 0.0; // total radial momentum on grid (including outside forward shock)
  
  // Calculate energy and mass fractions
  for (int i = lg.irs; i <= lg.ire; ++i){
    double r = lg.zxc[i];

    double ri = lg.zxa[i];
    double ro = lg.zxa[i] + lg.zdx[i];
    double vol = 4.0*pi*(std::pow(ro,3) - std::pow(ri,3))/3.0;
    double rho = lg.P0[iqd][k][j][i];
    double v = lg.P0[iqu0][k][j][i];
    double pre = lg.P0[iqe][k][j][i];
    if (r == fshk_radius){
      // for the cell that contains the shock, only add the excess mass above ramb and pressure above pamb - this should keep things smoother...
      rho -= ramb;
      pre -= pamb;
    }
    double mass = rho*vol;
    double ke = 0.5*mass*v*v;
    double ie = 1.5*pre*vol;
    double mtm = mass*v;

    // Add up quantities INSIDE the bubble
    if (r <= fshk_radius){// + 4.0*dr){ // add a small offset to the forward shock, since the shock will be broadened by the code
      KEtot += ke;
      IEtot += ie;
      mtmtot += mtm;
      double windMassFrac = lg.P0[iqal0][k][j][i];
      double ambientMassFrac = lg.P0[iqal0+1][k][j][i];
      double injectedMassFrac = 1.0 - windMassFrac - ambientMassFrac;
      injectedMassFrac = max(0.0,min(1.0,injectedMassFrac));
      windMass     += mass*windMassFrac;
      sweptUpMass  += mass*ambientMassFrac;
      injectedMass += mass*injectedMassFrac;
      double temp = pdtmp(pre/rho);
      if      (temp > 2.0e4) Mhot += mass;
      else if (v > 1.0e5) Mshell += mass;
    }

    // Add up quantities on the ENTIRE grid
    gridmtmtot += mtm;
  }

  // Get global sums
  MPI_Barrier(MPI_COMM_WORLD);
  KEtot = global_sum(KEtot);
  IEtot = global_sum(IEtot);
  windMass = global_sum(windMass);
  sweptUpMass = global_sum(sweptUpMass);
  injectedMass = global_sum(injectedMass);
  Mhot = global_sum(Mhot);
  Mshell = global_sum(Mshell);
  mtmtot = global_sum(mtmtot);
  gridmtmtot = global_sum(gridmtmtot);
  MPI_Barrier(MPI_COMM_WORLD);

  if (procRank == 0){
  // Calculate other quantities and scale to appropriate units  
  double Etot = KEtot + IEtot;
  double KEfrac = KEtot/Etot;
  double IEfrac = IEtot/Etot;
  double masstot = windMass + sweptUpMass + injectedMass;
  double wndfrac= windMass/masstot;
  double swfrac = sweptUpMass/masstot;
  double injfrac= injectedMass/masstot;
  fshk_radius /= pc;
  rshk_radius /= pc;
  masstot /= msol;
  windMass /= msol;
  sweptUpMass /= msol;
  injectedMass /= msol;
  Mhot /= msol;
  Mshell /= msol;
  mtmtot     /= (msol*1.0e5); // units of Msol km/s
  gridmtmtot /= (msol*1.0e5);
  
  // Write out data 
  outfile = "wbb_shock_and_energy_and_massdata.ascii";
  //cout << outfile << "\n";

  if (first){
    ofstream file(outfile.c_str()); //C++ method (outfile must be a string type)
    if (!file) {
      cerr << "Failed to open output file: " << outfile;
      quit();
    } 
    file << "# Problem = " << problem << "\n";
#ifdef MASSLOAD    
    file << "# Mass loading included. fML = " << fML << "; clumpToInterClumpMassRatio = " << clumpToInterClumpMassRatio << "\n";
#endif    
    file << "# ncycle          time            rFS             KEtot          IEtot           Etot            KEfrac          IEfrac         wndMtot          ambMtot         injMtot          Mtot          wndMfrac         ambMfrac        injMfrac        Mhot           Mshell         mtmtot\n";
    file << scientific << setprecision(4);
    file << ncycle << "\t\t" << t << "\t" << fshk_radius << "\t" << KEtot << "\t" << IEtot << "\t" << Etot << "\t" << KEfrac << "\t" << IEfrac << "\t" << windMass << "\t" << sweptUpMass << "\t" << injectedMass << "\t" << masstot << "\t" << wndfrac << "\t" << swfrac << "\t" << injfrac << "\t" << Mhot << "\t" << Mshell << "\t" << mtmtot << "\n";
    file.close();
    first = false;

    ofstream file2("wbb_data2.ascii");
    if (!file2) {
      cerr << "Failed to open output file wbb_shock_data.ascii...\n";
      quit();
    } 
    file2 << "# Problem = " << problem << "\n";    
    file2 << "# ncycle          time            rRS             rFS           gridmtmtot\n";
    file2 << scientific << setprecision(4);
    file2 << ncycle << "\t\t" << t << "\t" << rshk_radius << "\t" << fshk_radius << "\t" << gridmtmtot << "\n";
    file2.close();
  }
  else{
    // Open with access=append
    ofstream file(outfile.c_str(),std::ofstream::app); //C++ method (outfile must be a string type)
    file << scientific << setprecision(4);
    file << ncycle << "\t\t" << t << "\t" << fshk_radius << "\t" << KEtot << "\t" << IEtot << "\t" << Etot << "\t" << KEfrac << "\t" << IEfrac << "\t" << windMass << "\t" << sweptUpMass << "\t" << injectedMass << "\t" << masstot << "\t" << wndfrac << "\t" << swfrac << "\t" << injfrac << "\t" << Mhot << "\t" << Mshell << "\t" << mtmtot << "\n";
    file.close();

    ofstream file2("wbb_data2.ascii",std::ofstream::app);
    file2 << scientific << setprecision(4);
    file2 << ncycle << "\t\t" << t << "\t" << rshk_radius << "\t" << fshk_radius << "\t" << gridmtmtot << "\n";
    file2.close();
  }

  }
  //old_fshk_radius = fshk_radius;
  //old_fshk_time = t;

  return;
}


void prin1D(){

  string outfile;
  int j=0;
  int k=0;

  // Write out data 
  outfile = generateFilename(nfile1d,1,"ascii");
  //cout << outfile << "\n";
  ofstream file(outfile.c_str()); //C++ method (outfile must be a string type)
  if (!file) {
    cerr << "Failed to open output file: " << outfile;
    quit();
  } 
  file << "# Problem = " << problem << "; ncycle = " << ncycle << "; t = " << t << "; dt = " << dt << "\n";
#ifdef MASSLOAD
  file << "# Mass loading included. fML = " << fML << "; clumpToInterClumpMassRatio = " << clumpToInterClumpMassRatio << "\n";
#ifdef DUST
  file << "# zxa             rho             pre              vx            temp       dust mass frac       rhoML\n";
#else
  file << "# zxa             rho             pre              vx            temp       wnd mass frac  amb. mass frac  ML mass frac       rhoML\n";
#endif  
#elif DUST  
  file << "# zxa             rho             pre              vx            temp       dust mass frac\n";
#else
  file << "# zxa             rho             pre              vx            temp           col\n";
#endif
  file << scientific << setprecision(4);
  for (int i = lg.irs; i <= lg.ire; ++i){
    double temp = pdtmp(lg.P0[iqe][k][j][i]/lg.P0[iqd][k][j][i]);
    double wndMF = lg.P0[iqal0][k][j][i];
    double ambMF = lg.P0[iqal0+1][k][j][i];
    double massloadMF = (1.0 - wndMF - ambMF);
    massloadMF = min(massloadMF,1.0);
    massloadMF = max(massloadMF,0.0);
#ifdef MASSLOAD 
    file << lg.zxc[i] << "\t" << lg.P0[iqd][k][j][i] << "\t" << lg.P0[iqe][k][j][i] << "\t" << lg.P0[iqu0][k][j][i] << "\t" << temp << "\t" << wndMF << "\t" << ambMF << "\t" << massloadMF << "\t" << lg.rhoML[k][j][i] << "\n";
#else
    file << lg.zxc[i] << "\t" << lg.P0[iqd][k][j][i] << "\t" << lg.P0[iqe][k][j][i] << "\t" << lg.P0[iqu0][k][j][i] << "\t" << temp << "\t" << lg.P0[iqal0][k][j][i] << "\n";
#endif
  }
  file.close();
  ++nfile1d;
  tprin1d *= 1.02;
  return;
}


void prin2D(){

  // Write out a file containing x,y,rho,pressure,vx,vy,col
  // To plot in gnuplot use:
  // set term png
  // set output "sod_0000_2d_0009.png"
  // set size square
  // set logscale cb
  // set cbrange [0.1:15.0]
  // plot 'sod_0000_2d_0009.ascii' us 1:2:4 with image pixels
  // set term x11
  
  string outfile;
  int k=0;
  if (nd == 3) k = lgrid::kmax/2;

  // Write out data 
  outfile = generateFilename(nfile2d,2,"ascii");
  //cout << outfile << "\n";
  ofstream file(outfile.c_str()); //C++ method (outfile must be a string type)
  if (!file) {
    cerr << "Failed to open output file: " << outfile;
    quit();
  } 
  file << "# Problem = " << problem << "; ncycle = " << ncycle << "; t = " << t << "\n";
  file << "# zxa             zya              rho             pre              vx             vy             col\n";
  file << scientific << setprecision(4);
  for (int j = lg.jrs; j <= lg.jre; j++){
    for (int i = lg.irs; i <= lg.ire; ++i){
      file << lg.zxc[i] << "\t" << lg.zyc[j] << "\t" << lg.P0[iqd][k][j][i] << "\t" << lg.P0[iqe][k][j][i] << "\t" << lg.P0[iqu0][k][j][i] << "\t" << lg.P0[iqu0+1][k][j][i] << "\t" << lg.P0[iqal0][k][j][i] << "\n";
    }
    file << "\n";
  }
  file.close();
  return;
}


void prin2Dfits(){

  // Write out a 2D FITS file compatible with the X-ray code
  char type[20];
  
  //  FITSIO variables.
  int status = 0;                         //  FITSIO status.
  fitsfile *fitsptr;                      //  Pointer to fits file.
  int bitpix = FLOAT_IMG;                 //  Single precision.
  long naxis = 2;
  long naxes[2];
  long group = 1;
  
  string outfile;    
  const char *filename;     //  Character to store filename

  int k = 0;
  if (nd == 3) k = lgrid::kmax/2;

  outfile = generateFilename(nfile2d,2,"fits"); 
  filename = outfile.c_str();                   // convert to character array

  // Delete any existing file of that name and write out to file
  remove( filename );
  if ( fits_create_file(&fitsptr, filename, &status) ){ 
    FitsPrintError( status );
  }

  naxes[0] = lgrid::imax;
  naxes[1] = lgrid::jmax;

  //  Create the density image.
  if ( fits_create_img(fitsptr, bitpix, naxis, naxes, &status) ){
    cerr << "  Error creating rho data image" << endl;
    FitsPrintError( status );
  }

  //  Write keywords
  writeFitsKeywords(fitsptr);

  // Write the first (density) image
  //  Put the data into a 1-D array  
  float *ptr_flt = new float[lgrid::imax*lgrid::jmax];
  int n = 0;
  for (int j=lg.jrs; j <= lg.jre; j++){
    for (int i=lg.irs; i <= lg.ire; i++){
      ptr_flt[n] = lg.P0[iqd][0][j][i];
      //cout << ptr_flt[i] << "\t";
      n++;
    }
  }
    
  //  Write the 1-D array, cunningly stored in the same way as a 2-D
  //  image, into the FITS image.
  if ( fits_write_2d_flt(fitsptr, group, naxes[0], 
                           naxes[0], naxes[1], ptr_flt, &status) ){ 
    FitsPrintError( status );
  } 
  
  delete []ptr_flt; // clean up array

  // Create and write the other data as (2d) images. We pass in
  // the variable index in the primitive array, and in write2dFitsImage()
  // floats are written to file (to reduce the file size).
  strcpy(type,"pressure");
  write2dFitsImage(fitsptr,type,naxis,naxes,iqe,k);
  strcpy(type,"vx");
  write2dFitsImage(fitsptr,type,naxis,naxes,iqu0,k);
  strcpy(type,"vy");
  write2dFitsImage(fitsptr,type,naxis,naxes,iqu0+1,k);
  //strcpy(type,"vz");
  //write2dFitsImage(fitsptr,"vz",naxis,naxes,iqu0+2,k);
  strcpy(type,"colour");
  write2dFitsImage(fitsptr,type,naxis,naxes,iqal0,k);
  /*
  for (int m = 0; m < smax; m++){
    sprintf(cnum, "%4d",m);
    for (int i = 0; i < 4; i++){
      if (cnum[i] == ' ') cnum[i] = '0';
    }
    strcpy(type,"as");
    strcat(type,cnum);
    write2dFitsImage(fitsptr,type,naxis,naxes,iqal0+m,k); //advected scalars
  }
  */
  
  //Write axis data as 2D images
  float data1d[lgrid::mmax+1];      // array to store 1D data
  for (int i=0; i <= lgrid::imax; i++) data1d[i] = lg.zxa[i+nghost];
  write2dFitsImage(fitsptr,"zxa",naxes[0]+1,data1d);
  for (int j=0; j <= lgrid::jmax; j++) data1d[j] = lg.zya[j+nghost];
  write2dFitsImage(fitsptr,"zya",naxes[1]+1,data1d);
  for (int i=0; i <= lgrid::imax; i++) data1d[i] = lg.zdx[i+nghost];
  write2dFitsImage(fitsptr,"zdx",naxes[0]+1,data1d);
  for (int j=0; j <= lgrid::jmax; j++) data1d[j] = lg.zdy[j+nghost];
  write2dFitsImage(fitsptr,"zdy",naxes[1]+1,data1d);

  //  Close the output fits file.
  if ( fits_close_file(fitsptr, &status) ){   //  Close the file.
    FitsPrintError( status ); 
  }
  if (procRank == 0) cout << " Finished writing " << outfile << endl;

  // Check for any error, and if so print out error messages
  if (status > 0){
    FitsPrintError(status);
    quit();
  }

  return;
}


void prin3Dfits(){

  // Locals

  int ndim;
  char type[20];
  char cnum[5];             // needs an extra space for end of string character
  
  //  FITSIO variables.
  int status = 0;                         //  FITSIO status.
  fitsfile *fitsptr;                      //  Pointer to fits file.
  int bitpix = FLOAT_IMG;                 //  Single precision.
  long naxis = 3;
  long naxes[3];
  long group = 1;
  
  string strfilename;       //  String to store filename
  const char *filename;     //  Character to store filename

  ndim = 3;
  strfilename = generateFilename(nfile3d,3,"fits"); 
  filename = strfilename.c_str();               // convert to character array

  // Delete any existing file of that name and write out to file
  remove( filename );
  if ( fits_create_file(&fitsptr, filename, &status) ){ 
    FitsPrintError( status );
  }

  naxes[0] = lgrid::imax;
  naxes[1] = lgrid::jmax;
  naxes[2] = lgrid::kmax;

  //  Create the density image.
  if ( fits_create_img(fitsptr, bitpix, naxis, naxes, &status) ){
    cerr << "  Error creating rho data image" << endl;
    FitsPrintError( status );
  }

  //  Write keywords
  writeFitsKeywords(fitsptr);

  // Write the first (density) image
  //  Put the data into a 1-D array  
  float *ptr_flt = new float[lgrid::imax*lgrid::jmax*lgrid::kmax];
  int n = 0;
  for (int k=lg.krs; k <= lg.kre; k++){
    for (int j=lg.jrs; j <= lg.jre; j++){
      for (int i=lg.irs; i <= lg.ire; i++){
        ptr_flt[n] = lg.P0[iqd][k][j][i];
        //cout << ptr_flt[i] << "\t";
        n++;
      }
    }
  }

  //  Write the 1-D array, cunningly stored in the same way as a 3-D
  //  image, into the FITS image.
  if ( fits_write_3d_flt(fitsptr, group, naxes[0], naxes[1],
			 naxes[0], naxes[1], naxes[2], ptr_flt, &status) ){ 
    FitsPrintError( status );
  } 

  delete []ptr_flt; // clean up array

  // Create and write the other data as (3d) images. We pass in
  // the variable index in the primitive array, and in write3dFitsImage()
  // floats are written to file (to reduce the file size).
  strcpy(type,"pressure");
  write3dFitsImage(fitsptr,type,naxis,naxes,iqe);
  strcpy(type,"vx");
  write3dFitsImage(fitsptr,type,naxis,naxes,iqu0);
  strcpy(type,"vy");
  write3dFitsImage(fitsptr,type,naxis,naxes,iqu0+1);
  strcpy(type,"vz");
  write3dFitsImage(fitsptr,type,naxis,naxes,iqu0+2);
  //strcpy(type,"colour");
  //write3dFitsImage(fitsptr,type,naxis,naxes,iqal0);
  for (int m = 0; m < smax; m++){
    sprintf(cnum, "%4d",m);
    for (int i = 0; i < 4; i++){
      if (cnum[i] == ' ') cnum[i] = '0';
    }
    strcpy(type,"as");
    strcat(type,cnum);
    write3dFitsImage(fitsptr,type,naxis,naxes,iqal0+m); //advected scalars
  }
 
  //Write axis data as 2D images
  float data1d[lgrid::mmax+1];      // array to store 1D data
  for (int i=0; i <= lgrid::imax; i++) data1d[i] = lg.zxa[i+nghost];
  write2dFitsImage(fitsptr,"zxa",naxes[0]+1,data1d);
  for (int j=0; j <= lgrid::jmax; j++) data1d[j] = lg.zya[j+nghost];
  write2dFitsImage(fitsptr,"zya",naxes[1]+1,data1d);
  for (int k=0; k <= lgrid::kmax; k++) data1d[k] = lg.zza[k+nghost];
  write2dFitsImage(fitsptr,"zza",naxes[2]+1,data1d);
  for (int i=0; i <= lgrid::imax; i++) data1d[i] = lg.zdx[i+nghost];
  write2dFitsImage(fitsptr,"zdx",naxes[0]+1,data1d);
  for (int j=0; j <= lgrid::jmax; j++) data1d[j] = lg.zdy[j+nghost];
  write2dFitsImage(fitsptr,"zdy",naxes[1]+1,data1d);
  for (int k=0; k <= lgrid::kmax; k++) data1d[k] = lg.zdz[k+nghost];
  write2dFitsImage(fitsptr,"zdz",naxes[2]+1,data1d);

  // Write out problem specific data
  //if (problem == "STELLAR_CLUSTER") writeoutStellarClusterData(fitsptr);

  //  Close the output fits file.
  if ( fits_close_file(fitsptr, &status) ){   //  Close the file.
    FitsPrintError( status ); 
  }
  if (procRank == 0) cout << " Finished writing " << strfilename << endl;

  // Check for any error, and if so print out error messages
  if (status > 0){
    FitsPrintError(status);
    quit();
  }

  return;
}


void FitsPrintError( int status ){
  if (status)
  {
    fits_report_error(stderr, status);    //  Print error report.
    //exit( status );    //  Terminate the program, returning error status.
    quit();    //  Terminate the program
  }
  return;

}


void writeFitsKeywords(fitsfile *fitsptr){

  //Locals
  int status = 0;                         //  FITSIO status.
  char extname[FLEN_VALUE];
  char cproblem[70];

  //  Write keywords, in particular the coordinate keywords.
  //  First write current date and time (UTC).
  if ( fits_write_date(fitsptr, &status ) ){
    cerr << "  Error writing DATE keyword."  << endl;
    FitsPrintError( status );
  }

  //  Now write the density extension name
  strcpy(extname,"density");
  if ( fits_update_key(fitsptr, TSTRING, "VH_DATA", extname, 
			 "Extension name", &status ) ){
    cerr << "  Error writing EXTNAME version keyword."  << endl;
    FitsPrintError( status );
  }

  //Write introductory (mostly logical) keywords
  strcpy(cproblem,problem.c_str());   // copy string into char array
  //fits_update_key(fitsptr,TSTRING,"problem",cproblem,"problem",&status);
  fits_update_key(fitsptr,TSTRING,"problem",(char*) problem.c_str(),"problem",&status);

  //Write integers...
  fits_update_key(fitsptr,TINT,"ncycle",&ncycle,"ncycle",&status);
  fits_update_key(fitsptr,TINT,"nfile1d",&nfile1d,"nfile1d",&status);
  fits_update_key(fitsptr,TINT,"nfile2d",&nfile2d,"nfile2d",&status);
  fits_update_key(fitsptr,TINT,"nfile3d",&nfile3d,"nfile3d",&status);
  fits_update_key(fitsptr,TINT,"nprin1d",&nprin1d,"nprin1d",&status);
  fits_update_key(fitsptr,TINT,"nprin2d",&nprin2d,"nprin2d",&status);
  fits_update_key(fitsptr,TINT,"nprin3d",&nprin3d,"nprin3d",&status);
  fits_update_key(fitsptr,TINT,"ncycp1d",&ncycp1d,"ncycp1d",&status);
  fits_update_key(fitsptr,TINT,"ncycp2d",&ncycp2d,"ncycp2d",&status);
  fits_update_key(fitsptr,TINT,"ncycp3d",&ncycp3d,"ncycp3d",&status);

  // Now the doubles. const parameters must first be copied into non-const variables
  // before they can be output
  fits_update_key(fitsptr,TDOUBLE,"time",&t,"time",&status);
  fits_update_key(fitsptr,TDOUBLE,"dt",&dt,"dt",&status);
  fits_update_key(fitsptr,TDOUBLE,"timep1d",&timep1d,"timep1d",&status);
  fits_update_key(fitsptr,TDOUBLE,"timep2d",&timep2d,"timep2d",&status);
  fits_update_key(fitsptr,TDOUBLE,"timep3d",&timep3d,"timep3d",&status);
  fits_update_key(fitsptr,TDOUBLE,"tprin1d",&tprin1d,"tprin1d",&status);
  fits_update_key(fitsptr,TDOUBLE,"tprin2d",&tprin2d,"tprin2d",&status);
  fits_update_key(fitsptr,TDOUBLE,"tprin3d",&tprin3d,"tprin3d",&status);

  float avgmass1 = 1.0e-24;
  float avgmass2 = 1.0e-24;  
  fits_update_key(fitsptr,TFLOAT,"avgmass1",&avgmass1,"average mass 1",&status);
  fits_update_key(fitsptr,TFLOAT,"avgmass2",&avgmass2,"average mass 2",&status);

  // Problem specific code to come here...
  //

  if (problem == "CWB"){
    writeFitsKeywordsCWB(fitsptr);
  }

  // Check status before exiting

  if (status != 0){
    cerr << "  Error writing CTYPE keywords."  << endl;
    FitsPrintError( status );
  }

  //cout << "writeFitsKeywords: Keywords written successfully\n";
  return;
}


void writeFitsKeywordsCWB(fitsfile *fitsptr){

  int status = 0;
 
  double mdot1 = stars[0].mdot/(msol/yr);
  double mdot2 = stars[1].mdot/(msol/yr);
  double vinf1 = stars[0].vinf/1.0e5;
  double vinf2 = stars[1].vinf/1.0e5;
  
  fits_update_key(fitsptr,TDOUBLE,"Mdot_1",&mdot1,"(Msol/yr)",&status);
  fits_update_key(fitsptr,TDOUBLE,"Mdot_2",&mdot2,"(Msol/yr)",&status);
  fits_update_key(fitsptr,TDOUBLE,"vinf_1",&vinf1,"(km/s)",&status);
  fits_update_key(fitsptr,TDOUBLE,"vinf_2",&vinf2,"(km/s)",&status);
  fits_update_key(fitsptr,TDOUBLE,"eta",&eta,"(no units)",&status);
  fits_update_key(fitsptr,TDOUBLE,"dsep",&dsep,"(cm)",&status);

  fits_update_key(fitsptr,TDOUBLE,"xpos_1",&stars[0].xpos,"(cm)",&status);
  fits_update_key(fitsptr,TDOUBLE,"ypos_1",&stars[0].ypos,"(cm)",&status);
  //fits_update_key(fitsptr,TDOUBLE,"zpos1",&stzpos[0],"(cm)",&status);
  fits_update_key(fitsptr,TDOUBLE,"xpos_2",&stars[1].xpos,"(cm)",&status);
  fits_update_key(fitsptr,TDOUBLE,"ypos_2",&stars[1].ypos,"(cm)",&status);

  double X1 = 0.705;
  double X2 = 0.705;
  double Y1 = 0.275;
  double Y2 = 0.275;
  
  fits_update_key(fitsptr,TDOUBLE,"X1",&X1,"X1",&status);
  fits_update_key(fitsptr,TDOUBLE,"X2",&X2,"X2",&status);
  fits_update_key(fitsptr,TDOUBLE,"Y1",&Y1,"Y1",&status);
  fits_update_key(fitsptr,TDOUBLE,"Y2",&Y2,"Y2",&status);

  // Check status before exiting

  if (status != 0){
    cerr << "  Error writing CWB keywords."  << endl;
    FitsPrintError( status );
  }

  //cout << "writeFitsKeywordsCWB: Keywords written successfully\n";
  return;
}


void write2dFitsImage(fitsfile *fitsptr, const char* extname, long naxis, long naxes[2], int iq, int k){

  int status = 0;
  long group = 1;

  //  Create the new image.
  if ( fits_create_img(fitsptr, bitpix, naxis, naxes, &status) ){
    cerr << "  Error creating " << extname << " data image" << endl;
    FitsPrintError( status );
  }
  if ( fits_update_key(fitsptr, TSTRING,"VH_DATA", (char *)extname, 
			 "Extension name", &status ) ){
    cerr << "  Error writing EXTNAME version keyword."  << endl;
    FitsPrintError( status );
  }
     
  // Put the desired data into a 1-D array of floats and write floats to
  // disk to reduce the file size (by not writing doubles)
  float *ptr_flt = new float[lgrid::imax*lgrid::jmax];
  int n = 0;
  for (int j=lg.jrs; j <= lg.jre; j++){
    for (int i=lg.irs; i <= lg.ire; i++){
      ptr_flt[n] = lg.P0[iq][k][j][i];
      n++;
    }
  }
    
  //  Write the 1-D array, cunningly stored in the same way as a 2-D
  //  image, into the FITS image.
  if ( fits_write_2d_flt(fitsptr, group, naxes[0], 
			   naxes[0], naxes[1], ptr_flt, &status) ){ 
    FitsPrintError( status );
  } 

  // Write pixel keywords
  //writeFitsKeywordsPixels(fitsptr);

  delete []ptr_flt; // clean up array

  return;
}


void write2dFitsImage(fitsfile *fitsptr, const char* extname, long nzones, float array1d[]){

  int status = 0;
  long group = 1;
  long naxis;
  long naxes[2];

  naxis = 2;
  naxes[0] = nzones;
  naxes[1] = 1;

  //  Create the new image.
  if ( fits_create_img(fitsptr, bitpix, naxis, naxes, &status) ){
    cerr << "  Error creating " << extname << " data image" << endl;
    FitsPrintError( status );
  }
  if ( fits_update_key(fitsptr, TSTRING,"VH_DATA", (char *)extname, 
			 "Extension name", &status ) ){
    cerr << "  Error writing EXTNAME version keyword."  << endl;
    FitsPrintError( status );
  }
     
  //  Get the passed data into a 1-D array  
  float *ptr_flt = new float[nzones];
  for (int i=0; i < nzones; i++){
    ptr_flt[i] = array1d[i];
  }
    
  //  Write the 1-D array, cunningly stored in the same way as a 2-D
  //  image, into the FITS image.
  if ( fits_write_2d_flt(fitsptr, group, naxes[0], 
			 //naxes[0], naxes[1], ptr_flt, &status) ){ 
			   naxes[0], naxes[1], array1d, &status) ){ 
    FitsPrintError( status );
  } 

  delete []ptr_flt; // clean up array

  return;
}


void write3dFitsImage(fitsfile *fitsptr, const char* extname, long naxis, long naxes[3], int iq){

  int status = 0;
  long group = 1;

  //  Create the new image.
  if ( fits_create_img(fitsptr, bitpix, naxis, naxes, &status) ){
    cerr << "  Error creating rho data image" << endl;
    FitsPrintError( status );
  }
  if ( fits_update_key(fitsptr, TSTRING,"VH_DATA", (char *)extname, 
			 "Extension name", &status ) ){
    cerr << "  Error writing EXTNAME version keyword."  << endl;
    FitsPrintError( status );
  }
     
  //  Get the passed data into a 1-D array 
  float *ptr_flt = new float[lgrid::imax*lgrid::jmax*lgrid::kmax];
  int n = 0;
  for (int k=lg.krs; k <= lg.kre; k++){
    for (int j=lg.jrs; j <= lg.jre; j++){
      for (int i=lg.irs; i <= lg.ire; i++){
        ptr_flt[n] = lg.P0[iq][k][j][i];
        n++;
      }
    }
  }
    
  //  Write the 1-D array, cunningly stored in the same way as a 3-D
  //  image, into the FITS image.
  if ( fits_write_3d_flt(fitsptr, group, naxes[0], naxes[1], 
			 naxes[0], naxes[1], naxes[2], ptr_flt, &status) ){ 
    FitsPrintError( status );
  } 
  
  // Write pixel keywords (required for every image extension in a FITS file)
  writeFitsKeywordsPixels(fitsptr);

  delete []ptr_flt; // clean up array

  return;
}


string generateFilename(int nfile, int ndim, string ext)
//Generate the name for file dumps. Always include the processor number
//(for serial code this is 0000).
{
  char cnum[5];             // needs an extra space for end of string character
  char ndnum[2];            // needs an extra space for end of string character
  string srank,snfile,sdim;
  string filename;

  // Create strings for procRank and nfile
  sprintf(cnum, "%4d",procRank);
  for (int i = 0; i < 4; i++){
    if (cnum[i] == ' ') cnum[i] = '0';
  }
  srank = cnum;

  sprintf(cnum, "%4d",nfile);
  for (int i = 0; i < 4; i++){
    if (cnum[i] == ' ') cnum[i] = '0';
  }
  snfile = cnum;

  sprintf(ndnum, "%1d",ndim);
  sdim = ndnum;
  sdim += 'd';
 
  // The filename takes the form: prefx_AAAA_BB_CCCC.ext
  // where AAAA = rank number of processor (e.g. 0000)
  //         BB = dimensionality of data   (e.g. 1d)
  //       CCCC = file number in sequence  (e.g. 0010)
  
  filename = datadir + prefx + '_' + srank + '_' + sdim + '_' + snfile + "." + ext;

  return filename;
}


void writeFitsKeywordsPixels(fitsfile *fitsptr){

  int status = 0;

  //  CDELT, CRPIX etc axis information.
  char ctype1[FLEN_VALUE];  //  X-axis label
  char cunit1[FLEN_VALUE];  //  X-axis units
  float crpix1;             //  Reference pixel coordinate. The first pixel 
                            //  goes from 0.5-1.5 in FITSIO's units, cent=1.0.
  float crval1;             //  Reference pixel value.
  float cdelt1;             //  Increment per pixel of physical x-axis unit.

  char ctype2[FLEN_VALUE];  //  Y-axis label
  char cunit2[FLEN_VALUE];  //  Y-axis units

  float crpix2;             //  Reference pixel coordinate. The first pixel 
                            //  goes from 0.5-1.5 in FITSIO's units, cent=1.0.
  float crval2;             //  Reference pixel value.
  float cdelt2;             //  Increment per pixel of physical y-axis unit.

  char ctype3[FLEN_VALUE];  //  Z-axis label
  char cunit3[FLEN_VALUE];  //  Z-axis units
  float crpix3;             //  Reference pixel coordinate. The first pixel 
                            //  goes from 0.5-1.5 in FITSIO's units, cent=1.0.
  float crval3;             //  Reference pixel value.
  float cdelt3;             //  Increment per pixel of physical z-axis unit.

  const char *geomc;        //  Character to store geometry
  geomc = gg.geom.c_str();  //  convert to character array

  float rscale = 1.0;       //  local value overrides that in global.h 
  bool to2d = false;
  string splane = "2dxy";

  crpix1 = 0.5;
  crpix2 = 0.5;
  crpix3 = 0.5;

  if (!to2d){ // output normal grid geometry
  cdelt1 = lg.zdx[0]*rscale;
  cdelt2 = lg.zdy[0]*rscale;
  crval1 = lg.zxa[lg.irs]*rscale;
  crval2 = lg.zya[lg.jrs]*rscale;
  if (strcmp(geomc,"XYZ") == 0){
    strcpy( ctype1, "x" );
    strcpy( ctype2, "y" );
  }
  else if (strcmp(geomc,"RZP") == 0){
    strcpy( ctype1, "r" );
    strcpy( ctype2, "z" );
  }
  else if (strcmp(geomc,"ZRP") == 0){
    strcpy( ctype1, "z" );
    strcpy( ctype2, "r" );
  }
  else if (strcmp(geomc,"RTP") == 0){
    strcpy( ctype1, "r" );
    strcpy( ctype2, "thta" );
  }
  else{
    cout << "Further coding needed in writeFitsKeywordsPixels for current geometry\n";
    quit();
  }
  if (nd == 3){
    cdelt3 = lg.zdz[0]*rscale;
    crval3 = lg.zza[lg.krs]*rscale;
    if (strcmp(geomc,"XYZ") == 0){
      strcpy( ctype3, "z" );
    }
    else if (strcmp(geomc,"RZP") == 0){
      strcpy( ctype3, "p" );
    }
    else if (strcmp(geomc,"ZRP") == 0){
      strcpy( ctype3, "p" );
    }
    else{
      cout << "Further coding needed in writeFitsKeywordsPixels for current geometry\n";
      quit();
    }
  }
  }
  else{ // image data is a 2D plane through a 3D cube
    if (splane == "2dyz"){
      strcpy( ctype1, "y" );
      strcpy( ctype2, "z" );
      cdelt1 = lg.zdy[0];
      cdelt2 = lg.zdz[0];
      crval1 = lg.zya[lg.jrs];
      crval2 = lg.zza[lg.krs];
   }
    else if (splane == "2dxz"){
      strcpy( ctype1, "x" );
      strcpy( ctype2, "z" );
      cdelt1 = lg.zdx[0];
      cdelt2 = lg.zdz[0];
      crval1 = lg.zxa[lg.irs];
      crval2 = lg.zza[lg.krs];
    }
    else if (splane == "2dxy"){
      strcpy( ctype1, "x" );
      strcpy( ctype2, "y" );
      cdelt1 = lg.zdx[0];
      cdelt2 = lg.zdy[0];
      crval1 = lg.zxa[lg.irs];
      crval2 = lg.zya[lg.jrs];
    }
    else{
      cout << "writeFitsKeywordsPixels: unrecognized splane value. Aborting!\n";
      quit();    
    }
  }

  strcpy( cunit1, "cm" );
  strcpy( cunit2, "cm" );
  strcpy( cunit3, "cm" );

  fits_update_key(fitsptr, TSTRING, "CTYPE1", ctype1, 0, &status );
  fits_update_key(fitsptr, TSTRING, "CUNIT1", cunit1, 0, &status );
  fits_update_key(fitsptr, TFLOAT, "CRPIX1", &crpix1, 0, &status );
  fits_update_key(fitsptr, TFLOAT, "CRVAL1", &crval1, 0, &status );
  fits_update_key(fitsptr, TFLOAT, "CDELT1", &cdelt1, 0, &status );
  fits_update_key(fitsptr, TSTRING, "CTYPE2", ctype2, 0, &status );
  fits_update_key(fitsptr, TSTRING, "CUNIT2", cunit2, 0, &status );
  fits_update_key(fitsptr, TFLOAT, "CRPIX2", &crpix2, 0, &status );
  fits_update_key(fitsptr, TFLOAT, "CRVAL2", &crval2, 0, &status );
  fits_update_key(fitsptr, TFLOAT, "CDELT2", &cdelt2, 0, &status );
  if ((nd == 3)&&(!to2d)){
    fits_update_key(fitsptr, TSTRING, "CTYPE3", ctype3, 0, &status );
    fits_update_key(fitsptr, TSTRING, "CUNIT3", cunit3, 0, &status );
    fits_update_key(fitsptr, TFLOAT, "CRPIX3", &crpix3, 0, &status );
    fits_update_key(fitsptr, TFLOAT, "CRVAL3", &crval3, 0, &status );
    fits_update_key(fitsptr, TFLOAT, "CDELT3", &cdelt3, 0, &status );
  }
  //cout << "writeFitsKeywordsPixels: Keywords written successfully\n";

  return;
}
